@@ -12,12 +12,15 @@
  *
  *      By Shawn Hargreaves.
  *
+ *      Multiple bullets and extra lives added by David Cullen.
+ *
  *      See readme.txt for copyright information.
  */
 
 
 #include <string.h>
 #include <stdio.h>
+#include <time.h>
 
 #include "allegro.h"
 #include "demo.h"
@@ -64,10 +67,12 @@ int pos;
 int xspeed, yspeed, ycounter;
 int ship_state;
 int ship_burn;
+int ship_count;
 int shot;
 int skip_speed, skip_count;
 volatile int frame_count, fps;
-volatile int game_time;
+volatile unsigned int game_time;
+unsigned int prev_bullet_time;
 
 #define MAX_SPEED       32
 #define SPEED_SHIFT     3
@@ -106,11 +111,17 @@ volatile int new_alien_count;
 RLE_SPRITE *explosion[EXPLODE_FRAMES];
 
 
-/* position of the bullet */
+/* position of the bullets */
 #define BULLET_SPEED    6
+#define BULLET_DELAY   20
 
-volatile int bullet_flag;
-volatile int bullet_x, bullet_y; 
+typedef struct BULLET {
+   int x;
+   int y;
+   struct BULLET *next;
+} BULLET;
+
+BULLET *bullet_list;
 
 
 /* dirty rectangle list */
@@ -161,16 +172,70 @@ END_OF_FUNCTION(fps_proc);
 
 
 
+/* add a bullet to the list */
+BULLET *add_bullet(int x, int y)
+{
+   BULLET *iter, *bullet;
+
+   bullet = (BULLET *)malloc(sizeof(BULLET));
+   bullet->x = x;
+   bullet->y = y;
+   bullet->next = NULL;
+
+   /* special treatment for head */
+   if (!bullet_list) {
+      bullet_list = bullet;
+   }
+   else {
+      for (iter = bullet_list; iter->next; iter = iter->next)
+	 ;
+
+      iter->next = bullet;
+   }
+
+   return bullet;
+}
+
+
+
+/* delete a bullet and return the next in the list */
+BULLET *delete_bullet(BULLET *bullet)
+{
+   BULLET *iter;
+
+   /* special treatment for head */
+   if (bullet == bullet_list) {
+      bullet_list = bullet->next;
+      free(bullet);
+      return bullet_list;
+   }
+   else {
+      for (iter = bullet_list; iter->next != bullet; iter = iter->next)
+	 ;
+
+      iter->next = bullet->next;
+      free(bullet);
+      return iter->next;
+   }
+}
+
+
+
 /* the main game update function */
-void move_everyone()
+void move_everyone(void)
 {
    int c;
+   BULLET *bullet;
 
    if (shot) {
       /* player dead */
       if (skip_count <= 0) {
-	 if (ship_state >= EXPLODE_FLAG+EXPLODE_FRAMES-1)
-	    dead = TRUE;
+	 if (ship_state >= EXPLODE_FLAG+EXPLODE_FRAMES-1) {
+	    if (--ship_count <= 0)
+	       dead = TRUE;
+	    else
+	       shot = FALSE;
+	 }
 	 else
 	    ship_state++;
 
@@ -266,8 +331,8 @@ void move_everyone()
       ycounter += yspeed;
 
       while (ycounter >= (1 << SPEED_SHIFT)) {
-	 if (bullet_flag)
-	    bullet_y++;
+	 for (bullet = bullet_list; bullet; bullet = bullet->next)
+	    bullet->y++;
 
 	 for (c=0; c<MAX_STARS; c++) {
 	    if (++star[c].oy >= SCREEN_H)
@@ -282,25 +347,34 @@ void move_everyone()
    }
 
    /* move bullet */
-   if (bullet_flag) {
-      bullet_y -= BULLET_SPEED;
+   bullet = bullet_list;
+   while (bullet) {
+      bullet->y -= BULLET_SPEED;
 
-      if (bullet_y < 8) {
-	 bullet_flag = FALSE;
+      /* if the bullet is at the top of the screen, delete it */
+      if (bullet->y < 8) {
+	 bullet = delete_bullet(bullet);
+	 goto bullet_updated;
       }
       else {
 	 /* shot an alien? */
 	 for (c=0; c<alien_count; c++) {
-	    if ((ABS(bullet_y - alien[c].y) < 20) && (ABS(bullet_x - alien[c].x) < 20) && (!alien[c].shot)) {
+	    if ((ABS(bullet->y - alien[c].y) < 20) && (ABS(bullet->x - alien[c].x) < 20) && (!alien[c].shot)) {
 	       alien[c].shot = TRUE;
 	       alien[c].state = EXPLODE_FLAG;
-	       bullet_flag = FALSE;
 	       score += 10;
-	       play_sample(data[BOOM_SPL].dat, 255, PAN(bullet_x), 1000, FALSE);
-	       break; 
+	       play_sample(data[BOOM_SPL].dat, 255, PAN(bullet->x), 1000, FALSE);
+	       /* delete the bullet that killed the alien */
+	       bullet = delete_bullet(bullet);
+	       goto bullet_updated;
 	    }
 	 }
       }
+
+      bullet = bullet->next;
+
+    bullet_updated:
+      ; /* prevent GCC 3.3.1 from complaining */
    }
 
    /* move stars */
@@ -311,12 +385,14 @@ void move_everyone()
 
    /* fire bullet? */
    if (!shot) {
-      if ((key[KEY_SPACE]) || (joy[0].button[0].b) || (joy[0].button[1].b)) {
-	 if (!bullet_flag) {
-	    bullet_x = (pos>>SPEED_SHIFT)-2;
-	    bullet_y = SCREEN_H-64;
-	    bullet_flag = TRUE;
-	    play_sample(data[SHOOT_SPL].dat, 100, PAN(bullet_x), 1000, FALSE);
+      if ((key[KEY_SPACE] || key[KEY_LCONTROL] || key[KEY_RCONTROL]) ||
+	 (joy[0].button[0].b) || (joy[0].button[1].b)) {
+	 if (prev_bullet_time + BULLET_DELAY < game_time) {
+	    bullet = add_bullet((pos >> SPEED_SHIFT) - 2, SCREEN_H - 64);
+	    if (bullet) {
+	       play_sample(data[SHOOT_SPL].dat, 100, PAN(bullet->x), 1000, FALSE);
+	       prev_bullet_time = game_time;
+	    }
 	 } 
       }
    }
@@ -332,9 +408,9 @@ void move_everyone()
 		  alien[c].x += alien[c].d;
 	    }
 	    else {
-	       alien[c].x = 16 + rand() % (SCREEN_W-32);
-	       alien[c].y = -60 - (rand() & 0x3f);
-	       alien[c].d = (rand()&1) ? 1 : -1;
+	       alien[c].x = 16 + AL_RAND() % (SCREEN_W-32);
+	       alien[c].y = -60 - (AL_RAND() & 0x3f);
+	       alien[c].d = (AL_RAND()&1) ? 1 : -1;
 	       alien[c].shot = FALSE;
 	       alien[c].state = -1;
 	    }
@@ -354,15 +430,15 @@ void move_everyone()
 
       if (alien[c].y > SCREEN_H+30) {
 	 if (!alien[c].shot) {
-	    alien[c].x = rand() % (SCREEN_W-32);
-	    alien[c].y = -32 - (rand() & 0x3f);
-	    alien[c].d = (rand()&1) ? 1 : -1;
+	    alien[c].x = AL_RAND() % (SCREEN_W-32);
+	    alien[c].y = -32 - (AL_RAND() & 0x3f);
+	    alien[c].d = (AL_RAND()&1) ? 1 : -1;
 	 }
       }
       else {
 	 /* alien collided with player? */
 	 if ((ABS(alien[c].x - (pos>>SPEED_SHIFT)) < 48) && (ABS(alien[c].y - (SCREEN_H-42)) < 32)) {
-	    if (!shot) {
+            if ((!shot) && (!alien[c].shot)) {
 	       if (!cheat) {
 		  ship_state = EXPLODE_FLAG;
 		  shot = TRUE;
@@ -422,7 +498,7 @@ int rect_cmp(const void *p1, const void *p2)
 
 
 /* main screen update function */
-void draw_screen()
+void draw_screen(void)
 {
    int c;
    int i, j;
@@ -430,6 +506,7 @@ void draw_screen()
    BITMAP *bmp;
    RLE_SPRITE *spr;
    char *animation_type_str;
+   BULLET *bullet;
 
    if (animation_type == DOUBLE_BUFFER) {
       /* for double buffering, draw onto the memory bitmap. The first step 
@@ -437,7 +514,7 @@ void draw_screen()
        */
       animation_type_str = "double buffered";
       bmp = s;
-      clear(bmp);
+      clear_bitmap(bmp);
    }
    else if ((animation_type == PAGE_FLIP) || (animation_type == RETRACE_FLIP)) {
       /* for page flipping we draw onto one of the sub-bitmaps which
@@ -456,7 +533,7 @@ void draw_screen()
 	 bmp = page1;
 	 current_page = 0;
       }
-      clear(bmp);
+      clear_bitmap(bmp);
    }
    else if (animation_type == TRIPLE_BUFFER) {
       /* triple buffering works kind of like page flipping, but with three
@@ -479,7 +556,7 @@ void draw_screen()
 	 bmp = page1; 
 	 current_page = 0;
       }
-      clear(bmp);
+      clear_bitmap(bmp);
    }
    else {
       /* for dirty rectangle animation we draw onto the memory bitmap, but 
@@ -566,10 +643,10 @@ void draw_screen()
 	 add_to_list(&dirty, x-spr->h/2, y-spr->h/2, spr->w, spr->h);
    }
 
-   /* draw the bullet */
-   if (bullet_flag) {
-      x = bullet_x;
-      y = bullet_y;
+   /* draw the bullets */
+   for (bullet = bullet_list; bullet; bullet = bullet->next) {
+      x = bullet->x;
+      y = bullet->y;
 
       spr = (RLE_SPRITE *)data[ROCKET].dat;
       draw_rle_sprite(bmp, spr, x-spr->w/2, y-spr->h/2);
@@ -580,11 +657,11 @@ void draw_screen()
 
    /* draw the score and fps information */
    if (fps)
-      sprintf(score_buf, "Score: %ld - (%s, %ld fps)", (long)score, animation_type_str, (long)fps);
+      sprintf(score_buf, "Lives: %d - Score: %d - (%s, %d fps)", ship_count, score, animation_type_str, fps);
    else
-      sprintf(score_buf, "Score: %ld", (long)score);
+      sprintf(score_buf, "Lives: %d - Score: %d", ship_count, score);
 
-   textout(bmp, font, score_buf, 0, 0, 7);
+   textout_ex(bmp, font, score_buf, 0, 0, 7, 0);
 
    if (animation_type == DIRTY_RECTANGLE)
       add_to_list(&dirty, 0, 0, text_length(font, score_buf), 8);
@@ -655,7 +732,7 @@ void draw_intro_item(int item, int size)
    int w = MIN(SCREEN_W, (SCREEN_W * 2 / size));
    int h = SCREEN_H / size;
 
-   clear(screen);
+   clear_bitmap(screen);
    stretch_blit(b, screen, 0, 0, b->w, b->h, (SCREEN_W-w)/2, (SCREEN_H-h)/2, w, h);
 }
 
@@ -683,11 +760,12 @@ int fade_intro_item(int music_pos, int fade_speed)
 
 
 /* sets up and plays the game */
-void play_game()
+void play_game(void)
 {
    int c;
    BITMAP *b, *b2;
    int esc = FALSE;
+   unsigned int prev_update_time;
 
    stop_midi();
 
@@ -697,8 +775,9 @@ void play_game()
    xspeed = yspeed = ycounter = 0;
    ship_state = SHIP3;
    ship_burn = FALSE;
+   ship_count = 3;   /* 3 lives instead of one... */
    frame_count = fps = 0;
-   bullet_flag = FALSE;
+   bullet_list = NULL;
    score = 0;
    old_dirty.count = dirty.count = 0;
 
@@ -713,15 +792,15 @@ void play_game()
       skip_speed = 3;
 
    for (c=0; c<MAX_STARS; c++) {
-      star[c].ox = rand() % SCREEN_W;
-      star[c].oy = rand() % SCREEN_H;
-      star[c].z = rand() & 7;
+      star[c].ox = AL_RAND() % SCREEN_W;
+      star[c].oy = AL_RAND() % SCREEN_H;
+      star[c].z = AL_RAND() & 7;
    }
 
    for (c=0; c<MAX_ALIENS; c++) {
-      alien[c].x = 16 + rand() % (SCREEN_W-32);
-      alien[c].y = -60 - (rand() & 0x3f);
-      alien[c].d = (rand()&1) ? 1 : -1;
+      alien[c].x = 16 + AL_RAND() % (SCREEN_W-32);
+      alien[c].y = -60 - (AL_RAND() & 0x3f);
+      alien[c].d = (AL_RAND()&1) ? 1 : -1;
       alien[c].state = -1;
       alien[c].shot = FALSE;
    }
@@ -764,9 +843,8 @@ void play_game()
 
    skip:
 
-   text_mode(0);
-   clear(screen);
-   clear(s);
+   clear_bitmap(screen);
+   clear_bitmap(s);
    draw_screen();
 
    if ((!keypressed()) && (!joy[0].button[0].b) && (!joy[0].button[1].b)) {
@@ -788,15 +866,17 @@ void play_game()
       install_int(game_timer, 6400/SCREEN_W);
 
    game_time = 0;
+   prev_bullet_time = 0;
+   prev_update_time = 0;
 
    /* main game loop */
    while (!dead) {
       poll_keyboard();
       poll_joystick();
 
-      while (game_time > 0) {
+      while (prev_update_time < game_time) {
 	 move_everyone();
-	 game_time--;
+	 prev_update_time++;
       }
 
       draw_screen();
@@ -816,6 +896,9 @@ void play_game()
 
    stop_sample(data[ENGINE_SPL].dat);
 
+   while (bullet_list)
+      delete_bullet(bullet_list);
+
    if (esc) {
       while (current_page != 0)
 	 draw_screen();
@@ -845,9 +928,10 @@ void play_game()
    /* display the score */
    b = create_bitmap(160, 160);
    b2 = create_bitmap(160, 160);
-   clear(b);
-   textout_centre(b, data[END_FONT].dat, "GAME OVER", 80, 50, 2);
-   textout_centre(b, data[END_FONT].dat, score_buf, 80, 82, 2);
+   clear_bitmap(b);
+   textout_centre_ex(b, data[END_FONT].dat, "GAME OVER", 80, 50, 2, 0);
+   sprintf(score_buf, "Score: %d", score);
+   textout_centre_ex(b, data[END_FONT].dat, score_buf, 80, 82, 2, 0);
    clear_keybuf();
    scroll_count = -160;
    c = 0;
@@ -879,7 +963,7 @@ void play_game()
 
 
 /* the explosion graphics are pregenerated, using a simple particle system */
-void generate_explosions()
+void generate_explosions(void)
 {
    BITMAP *bmp = create_bitmap(EXPLODE_SIZE, EXPLODE_SIZE);
    unsigned char *p;
@@ -898,12 +982,12 @@ void generate_explosions()
 
    for (c=0; c<HOTSPOTS; c++) {
       hot[c].x = hot[c].y = (EXPLODE_SIZE/2)<<16;
-      hot[c].xc = (rand()&0xFFFF)-0x7FFF;
-      hot[c].yc = (rand()&0xFFFF)-0x7FFF;
+      hot[c].xc = (AL_RAND()&0xFFFF)-0x7FFF;
+      hot[c].yc = (AL_RAND()&0xFFFF)-0x7FFF;
    }
 
    for (c=0; c<EXPLODE_FRAMES; c++) {
-      clear(bmp);
+      clear_bitmap(bmp);
 
       color = ((c<16) ? c*4 : (80-c)) >> 2;
 
@@ -943,7 +1027,7 @@ void generate_explosions()
 
 
 /* modifies the palette to give us nice colors for the GUI dialogs */
-void set_gui_colors()
+void set_gui_colors(void)
 {
    static RGB black = { 0,  0,  0,  0 };
    static RGB grey  = { 48, 48, 48, 0 };
@@ -954,8 +1038,8 @@ void set_gui_colors()
    set_color(1, &grey); 
    set_color(255, &white); 
 
-   gui_fg_color = 0;
-   gui_bg_color = 1;
+   gui_fg_color = palette_color[0];
+   gui_bg_color = palette_color[1];
 }
 
 
@@ -1072,13 +1156,12 @@ int anim_desc_proc(int msg, DIALOG *d, int c)
 
    if (msg == MSG_DRAW) {
       rectfill(screen, d->x, d->y, d->x+d->w, d->y+d->h, d->bg);
-      text_mode(d->bg);
 
       p = descs[anim_type_dlg[2].d1];
       y = d->y;
 
       while (*p) {
-	 textout(screen, font, *p, d->x, y, d->fg);
+	 textout_ex(screen, font, *p, d->x, y, d->fg, d->bg);
 	 y += 8;
 	 p++;
       } 
@@ -1092,12 +1175,13 @@ int anim_desc_proc(int msg, DIALOG *d, int c)
 DIALOG anim_type_dlg[] =
 {
    /* (dialog proc)     (x)   (y)   (w)   (h)   (fg)  (bg)  (key) (flags)     (d1)  (d2)  (dp)                 (dp2) (dp3) */
-   { d_shadow_box_proc, 0,    0,    280,  150,  0,    1,    0,    0,          0,    0,    NULL,                NULL, NULL  },
+   { d_shadow_box_proc, 0,    0,    281,  151,  0,    1,    0,    0,          0,    0,    NULL,                NULL, NULL  },
    { d_ctext_proc,      140,  8,    1,    1,    0,    1,    0,    0,          0,    0,    "Animation Method",  NULL, NULL  },
-   { anim_list_proc,    16,   28,   152,  43,   0,    1,    0,    D_EXIT,     4,    0,    anim_list_getter,    NULL, NULL  },
+   { anim_list_proc,    16,   28,   153,  44,   0,    1,    0,    D_EXIT,     4,    0,    anim_list_getter,    NULL, NULL  },
    { anim_desc_proc,    16,   90,   248,  48,   0,    1,    0,    0,          0,    0,    0,                   NULL, NULL  },
    { d_button_proc,     184,  28,   80,   16,   0,    1,    13,   D_EXIT,     0,    0,    "OK",                NULL, NULL  },
    { d_button_proc,     184,  50,   80,   16,   0,    1,    27,   D_EXIT,     0,    0,    "Cancel",            NULL, NULL  },
+   { d_yield_proc,      0,    0,    0,    0,    0,    0,    0,    0,          0,    0,    NULL,                NULL, NULL  },
    { NULL,              0,    0,    0,    0,    0,    0,    0,    0,          0,    0,    NULL,                NULL, NULL  }
 };
 
@@ -1110,7 +1194,13 @@ int pick_animation_type(int *type)
 
    centre_dialog(anim_type_dlg);
 
-   clear(screen);
+   clear_bitmap(screen);
+
+   /* we set up colors to match screen color depth */
+   for (ret = 0; anim_type_dlg[ret].proc; ret++) {
+      anim_type_dlg[ret].fg = palette_color[0];
+      anim_type_dlg[ret].bg = palette_color[1];
+   }
 
    ret = do_dialog(anim_type_dlg, 2);
 
@@ -1174,7 +1264,7 @@ char *format_text(TEXT_LIST *head, char *eol, char *gap)
 
 
 /* loads the scroller message from readme.txt */
-void load_text()
+void load_text(void)
 {
    README_SECTION sect[] =
    {
@@ -1342,7 +1432,7 @@ CREDIT_NAME *credits = NULL;
 
 
 /* reads credit info from a source file */
-void parse_source(char *filename, int attrib, int param)
+int parse_source(AL_CONST char *filename, int attrib, void *param)
 {
    char buf[256];
    PACKFILE *f;
@@ -1361,7 +1451,7 @@ void parse_source(char *filename, int attrib, int param)
 	 put_backslash(buf);
 	 strcat(buf, "*.*");
 
-	 for_each_file(buf, FA_ARCH | FA_RDONLY | FA_DIREC, parse_source, param);
+	 for_each_file_ex(buf, 0, ~(FA_ARCH | FA_RDONLY | FA_DIREC), parse_source, param);
       }
    }
    else {
@@ -1374,9 +1464,9 @@ void parse_source(char *filename, int attrib, int param)
 	 /* parse a source file */
 	 f = pack_fopen(filename, F_READ);
 	 if (!f)
-	    return;
+	    return -1;
 
-	 textprintf_centre(screen, font, SCREEN_W/2, SCREEN_H/2+8, makecol(160, 160, 160), "                %s                ", filename+param);
+	 textprintf_centre_ex(screen, font, SCREEN_W/2, SCREEN_H/2+8, makecol(160, 160, 160), 0, "                %s                ", filename+(int)(unsigned long)param);
 
 	 while (pack_fgets(buf, sizeof(buf)-1, f) != 0) {
 	    if (strstr(buf, "*/"))
@@ -1387,14 +1477,14 @@ void parse_source(char *filename, int attrib, int param)
 	    while (c) {
 	       if (strstr(buf, c->name)) {
 		  for (d=c->files; d; d=d->next) {
-		     if (strcmp(d->text, filename+param) == 0)
+		     if (strcmp(d->text, filename+(int)(unsigned long)param) == 0)
 			break;
 		  }
 
 		  if (!d) {
 		     d = malloc(sizeof(TEXT_LIST));
-		     d->text = malloc(strlen(filename+param)+1);
-		     strcpy(d->text, filename+param);
+		     d->text = malloc(strlen(filename+(int)(unsigned long)param)+1);
+		     strcpy(d->text, filename+(int)(unsigned long)param);
 		     d->next = c->files;
 		     c->files = d;
 		  }
@@ -1407,6 +1497,8 @@ void parse_source(char *filename, int attrib, int param)
 	 pack_fclose(f);
       }
    }
+
+   return 0;
 }
 
 
@@ -1443,7 +1535,7 @@ void sort_text_list(TEXT_LIST **head)
 
 
 /* sorts a list of credit strings */
-void sort_credit_list()
+void sort_credit_list(void)
 {
    CREDIT_NAME **prev, *p;
    TEXT_LIST *t;
@@ -1486,7 +1578,7 @@ void sort_credit_list()
 
 
 /* reads credit info from various places */
-void load_credits()
+void load_credits(void)
 {
    char buf[256], buf2[256], *p, *p2;
    CREDIT_NAME *c = NULL;
@@ -1494,7 +1586,7 @@ void load_credits()
 
    /* parse thanks._tx */
    get_executable_name(buf, sizeof(buf));
-   replace_filename(buf2, buf, "../docs/thanks._tx", sizeof(buf2));
+   replace_filename(buf2, buf, "../docs/src/thanks._tx", sizeof(buf2));
 
    f = pack_fopen(buf2, F_READ);
    if (!f)
@@ -1565,7 +1657,7 @@ void load_credits()
    get_executable_name(buf, sizeof(buf));
    replace_filename(buf2, buf, "../*.*", sizeof(buf2));
 
-   for_each_file(buf2, FA_ARCH | FA_RDONLY | FA_DIREC, parse_source, strlen(buf2)-3);
+   for_each_file_ex(buf2, 0, ~(FA_ARCH | FA_RDONLY | FA_DIREC), parse_source, (void *)(unsigned long)(strlen(buf2)-3));
 
    /* sort the lists */
    sort_credit_list();
@@ -1578,7 +1670,7 @@ void load_credits()
 
 
 /* displays the title screen */
-int title_screen()
+int title_screen(void)
 {
    static int color = 0;
    int c, c2, c3, n, n2;
@@ -1648,7 +1740,7 @@ int title_screen()
    if (color > 3)
       color = 0;
 
-   clear(screen);
+   clear_bitmap(screen);
    set_palette(title_palette);
 
    scroll_count = 1;
@@ -1662,10 +1754,10 @@ int title_screen()
    blit(data[TITLE_BMP].dat, screen, 0, 0, SCREEN_W/2-160, SCREEN_H/2-96, 320, 128);
 
    buffer = create_bitmap(SCREEN_W, SCREEN_H);
-   clear(buffer);
+   clear_bitmap(buffer);
 
    text_bmp = create_bitmap(SCREEN_W, 24);
-   clear(text_bmp);
+   clear_bitmap(text_bmp);
 
    old_dirty.count = dirty.count = 0;
 
@@ -1686,16 +1778,16 @@ int title_screen()
 	 /* animate the starfield */
 	 for (c=0; c<star_count; c++) {
 	    if (star[c].z <= itofix(1)) {
-	       x = itofix(rand()&0xff);
-	       y = itofix(((rand()&3)+1)*SCREEN_W);
+	       x = itofix(AL_RAND()&0xff);
+	       y = itofix(((AL_RAND()&3)+1)*SCREEN_W);
 
-	       star[c].x = fmul(fcos(x), y);
-	       star[c].y = fmul(fsin(x), y);
-	       star[c].z = itofix((rand() & 0x1f) + 0x20);
+	       star[c].x = fixmul(fixcos(x), y);
+	       star[c].y = fixmul(fixsin(x), y);
+	       star[c].z = itofix((AL_RAND() & 0x1f) + 0x20);
 	    }
 
-	    x = fdiv(star[c].x, star[c].z);
-	    y = fdiv(star[c].y, star[c].z);
+	    x = fixdiv(star[c].x, star[c].z);
+	    y = fixdiv(star[c].y, star[c].z);
 
 	    ix = (int)(x>>16) + SCREEN_W/2;
 	    iy = (int)(y>>16) + SCREEN_H/2;
@@ -1751,7 +1843,7 @@ int title_screen()
 
 	       if (n) {
 		  credit_skip = 1 + n/50;
-		  credit_speed = 8 + fixtoi(fdiv(itofix(512), itofix(n)));
+		  credit_speed = 8 + fixtoi(fixdiv(itofix(512), itofix(n)));
 		  if (credit_speed > 200)
 		     credit_speed = 200;
 		  c = 1024 + (n-1)*credit_speed;
@@ -1793,7 +1885,7 @@ int title_screen()
 
       while (text_scroll > 0) {
 	 text_pix += text_scroll;
-	 textout(text_bmp, data[TITLE_FONT].dat, buf, SCREEN_W-text_pix, 0, -1);
+	 textout_ex(text_bmp, data[TITLE_FONT].dat, buf, SCREEN_W-text_pix, 0, -1, 0);
 	 if (text_pix >= text_width) {
 	    text_scroll = text_pix - text_width;
 	    text_char++;
@@ -1829,26 +1921,26 @@ int title_screen()
 
 	       c2 -= 96;
 
-	       c3 = (32 + fixtoi(ABS(fsin(itofix(c/(15+n%42)+n)))*128)) * SCREEN_W / 640;
+	       c3 = (32 + fixtoi(ABS(fixsin(itofix(c/(15+n%42)+n)))*128)) * SCREEN_W / 640;
 
-	       x += fsin(itofix(c2)) * c3;
-	       y += fcos(itofix(c2)) * c3;
+	       x += fixsin(itofix(c2)) * c3;
+	       y += fixcos(itofix(c2)) * c3;
 
 	       if (c < 512) {
-		  z = fsqrt(itofix(c)/512);
+		  z = fixsqrt(itofix(c)/512);
 
-		  x = fmul(itofix(32), itofix(1)-z) + fmul(x, z);
-		  y = fmul(itofix(16), itofix(1)-z) + fmul(y, z);
+		  x = fixmul(itofix(32), itofix(1)-z) + fixmul(x, z);
+		  y = fixmul(itofix(16), itofix(1)-z) + fixmul(y, z);
 	       }
 	       else if (c > 768) {
-		  z = fsqrt(itofix(1024-c)/256);
+		  z = fixsqrt(itofix(1024-c)/256);
 
 		  if (n&2)
-		     x = fmul(itofix(128), itofix(1)-z) + fmul(x, z);
+		     x = fixmul(itofix(128), itofix(1)-z) + fixmul(x, z);
 		  else
-		     x = fmul(itofix(SCREEN_W-128), itofix(1)-z) + fmul(x, z);
+		     x = fixmul(itofix(SCREEN_W-128), itofix(1)-z) + fixmul(x, z);
 
-		  y = fmul(itofix(SCREEN_H-128), itofix(1)-z) + fmul(y, z);
+		  y = fixmul(itofix(SCREEN_H-128), itofix(1)-z) + fixmul(y, z);
 	       }
 
 	       c = 128 + (512 - ABS(512-c)) / 24;
@@ -1859,7 +1951,7 @@ int title_screen()
 	       c2 = strlen(tl->text);
 	       ix -= c2*4;
 
-	       textout(buffer, font, tl->text, ix, iy, c);
+	       textout_ex(buffer, font, tl->text, ix, iy, c, 0);
 	       add_to_list(&dirty, ix, iy, c2*8, 8);
 	    }
 
@@ -1904,7 +1996,7 @@ int title_screen()
 
 		  if ((c2 > 128) && (c2 <= 255)) {
 		     cbuf[0] = *p;
-		     textout(buffer, font, cbuf, c, 16, c2);
+		     textout_ex(buffer, font, cbuf, c, 16, c2, 0);
 		  }
 
 		  p++;
@@ -1921,8 +2013,8 @@ int title_screen()
 	 if (credit_scroll < 150)
 	    c += (150-credit_scroll) * (150-credit_scroll) * SCREEN_W / 22500;
 
-	 textprintf(buffer, data[TITLE_FONT].dat, c, 4, -1, "%s:", cred->name);
-	 add_to_list(&dirty, 0, 4, SCREEN_W, 24);
+	 textprintf_ex(buffer, data[TITLE_FONT].dat, c, 4, -1, 0, "%s:", cred->name);
+	 add_to_list(&dirty, 0, 4, SCREEN_W, text_height(data[TITLE_FONT].dat));
       }
 
       /* draw the Allegro logo over the top */
@@ -1976,7 +2068,7 @@ int title_screen()
 
 int main(int argc, char *argv[])
 {
-   int c, w, h;
+   int c, w, h, num_pages;
    char buf[256], buf2[256];
    BITMAP *bmp;
    CREDIT_NAME *cred;
@@ -1993,19 +2085,21 @@ int main(int argc, char *argv[])
 	 jumpstart = TRUE;
    }
 
-   allegro_init();
+   srand(time(NULL));
+   if (allegro_init() != 0)
+      return 1;
    install_keyboard();
    install_timer();
    install_mouse();
 
    if (install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, argv[0]) != 0) {
       allegro_message("Error initialising sound\n%s\n", allegro_error);
-      exit(1);
+      install_sound (DIGI_NONE, MIDI_NONE, NULL);
    }
 
    if (install_joystick(JOY_TYPE_AUTODETECT) != 0) {
       allegro_message("Error initialising joystick\n%s\n", allegro_error);
-      exit(1);
+      install_joystick (JOY_TYPE_NONE);
    }
 
    #ifdef ALLEGRO_CONSOLE_OK
@@ -2013,10 +2107,17 @@ int main(int argc, char *argv[])
 	 fade_out(4);
    #endif
 
-   set_gfx_mode(GFX_SAFE, 320, 200, 0, 0);
+   if (set_gfx_mode(GFX_AUTODETECT, 320, 200, 0, 0) != 0) {
+      if (set_gfx_mode(GFX_SAFE, 320, 200, 0, 0) != 0) {
+         set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
+         allegro_message("Unable to set any graphic mode\n%s\n", allegro_error);
+         return 1;
+      }
+   }
 
    get_executable_name(buf, sizeof(buf));
    replace_filename(buf2, buf, "demo.dat", sizeof(buf2));
+   set_color_conversion(COLORCONV_NONE);
    data = load_datafile(buf2);
    if (!data) {
       set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
@@ -2033,8 +2134,11 @@ int main(int argc, char *argv[])
        * ashamed to admit that I actually wrote this :-)
        */
       #ifdef ALLEGRO_WINDOWS
-	 if (1)
+
+ 	 play_midi(data[INTRO_MUSIC].dat, FALSE);
+
       #else
+
 	 if ((0) ||
 	    #ifdef MIDI_DIGMID
 	       (midi_driver->id == MIDI_DIGMID) ||
@@ -2049,42 +2153,61 @@ int main(int argc, char *argv[])
 	       (midi_driver->id == MIDI_SB_OUT) ||
 	    #endif
 	    (0))
-      #endif
 	 play_midi(data[INTRO_MUSIC].dat, FALSE);
       else 
 	 play_sample(data[INTRO_SPL].dat, 255, 128, 1000, FALSE);
 
+      #endif
+
       bmp = create_sub_bitmap(screen, SCREEN_W/2-160, SCREEN_H/2-100, 320, 200);
       play_memory_fli(data[INTRO_ANIM].dat, bmp, FALSE, NULL);
       destroy_bitmap(bmp);
    }
 
-   c = retrace_count;
-
-   generate_explosions();
-
    if (!jumpstart) {
-      do {
-      } while (retrace_count-c < 120);
-
+      rest (1000);
       fade_out(1);
    }
 
-   clear(screen);
+   clear_bitmap(screen);
    set_gui_colors();
+   set_mouse_sprite(NULL);
 
    if (!gfx_mode_select(&c, &w, &h))
       exit(1);
 
    if (pick_animation_type(&animation_type) < 0)
       exit(1);
+      
+   switch (animation_type) {
 
+      case PAGE_FLIP:
+      case RETRACE_FLIP: 
+	 num_pages = 2;
+	 break;
+
+      case TRIPLE_BUFFER: 
+	 num_pages = 3;
+	 break;
+
+      default: 
+	 num_pages = 1;
+	 break;
+   }
+
+   set_color_depth(8);
+#ifdef ALLEGRO_VRAM_SINGLE_SURFACE
+   if (set_gfx_mode(c, w, h, w, h * num_pages) != 0) {
+#else
    if (set_gfx_mode(c, w, h, 0, 0) != 0) {
+#endif
       set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
-      allegro_message("Error setting graphics mode\n%s\n", allegro_error);
+      allegro_message("Error setting 8bpp graphics mode\n%s\n", allegro_error);
       exit(1);
    }
 
+   generate_explosions();
+
    switch (animation_type) {
 
       case PAGE_FLIP:
@@ -2166,9 +2289,8 @@ int main(int argc, char *argv[])
 
    stop_sample(data[INTRO_SPL].dat);
 
-   text_mode(0);
-   clear(screen);
-   textout_centre(screen, font, "Scanning for author credits...", SCREEN_W/2, SCREEN_H/2-16, makecol(160, 160, 160));
+   clear_bitmap(screen);
+   textout_centre_ex(screen, font, "Scanning for author credits...", SCREEN_W/2, SCREEN_H/2-16, makecol(160, 160, 160), 0);
 
    if (SCREEN_W >= 640)
       load_credits();
@@ -2196,7 +2318,7 @@ int main(int argc, char *argv[])
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
 
    if (end_text) {
-      allegro_message("%shttp://www.talula.demon.co.uk/allegro/\n\n", end_text);
+      allegro_message("%shttp://alleg.sourceforge.net/\n\n", end_text);
       free(end_text);
    }
 
