@@ -14,15 +14,22 @@ int main(int argc, char *argv[])
    BITMAP *the_image;
    PALETTE the_palette;
 
-   allegro_init();
-
    if (argc != 2) {
       allegro_message("Usage: 'exbitmap filename.[bmp|lbm|pcx|tga]'\n");
       return 1;
    }
 
-   install_keyboard(); 
-   set_gfx_mode(GFX_SAFE, 320, 200, 0, 0);
+   if (allegro_init() != 0)
+      return 1;
+   install_keyboard();
+
+   if (set_gfx_mode(GFX_AUTODETECT, 320, 200, 0, 0) != 0) {
+      if (set_gfx_mode(GFX_SAFE, 320, 200, 0, 0) != 0) {
+	 set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
+	 allegro_message("Unable to set any graphic mode\n%s\n", allegro_error);
+	 return 1;
+      }
+   }
 
    /* read in the bitmap file */
    the_image = load_bitmap(argv[1], the_palette);
