@@ -11,23 +11,32 @@
 
 
 
-int main()
+int main(void)
 {
    BITMAP *memory_bitmap;
    int x, y;
 
-   allegro_init();
-   install_keyboard(); 
-   set_gfx_mode(GFX_SAFE, 320, 200, 0, 0);
+   if (allegro_init() != 0)
+      return 1;
+   install_keyboard();
+
+   if (set_gfx_mode(GFX_AUTODETECT, 320, 200, 0, 0) != 0) {
+      if (set_gfx_mode(GFX_SAFE, 320, 200, 0, 0) != 0) {
+	 set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
+	 allegro_message("Unable to set any graphic mode\n%s\n", allegro_error);
+	 return 1;
+      }
+   }
+
    set_palette(desktop_palette);
 
    /* make a memory bitmap sized 20x20 */
    memory_bitmap = create_bitmap(20, 20);
 
    /* draw some circles onto it */
-   clear(memory_bitmap);
+   clear_bitmap(memory_bitmap);
    for (x=0; x<16; x++)
-      circle(memory_bitmap, 10, 10, x, x);
+      circle(memory_bitmap, 10, 10, x, palette_color[x]);
 
    /* blit lots of copies of it onto the screen */
    acquire_screen();
