@@ -13,7 +13,8 @@ int main(int argc, char *argv[])
 {
    MIDI *the_music;
 
-   allegro_init();
+   if (allegro_init() != 0)
+      return 1;
 
    if (argc != 2) {
       allegro_message("Usage: 'exmidi filename.mid'\n");
@@ -36,10 +37,21 @@ int main(int argc, char *argv[])
       return 1;
    }
 
-   set_gfx_mode(GFX_SAFE, 320, 200, 0, 0);
+   if (set_gfx_mode(GFX_AUTODETECT, 320, 200, 0, 0) != 0) {
+      if (set_gfx_mode(GFX_SAFE, 320, 200, 0, 0) != 0) {
+	 set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
+	 allegro_message("Unable to set any graphic mode\n%s\n", allegro_error);
+	 return 1;
+      }
+   }
 
-   textprintf_centre(screen, font, SCREEN_W/2, SCREEN_H/3, 255, "Driver: %s", midi_driver->name);
-   textprintf_centre(screen, font, SCREEN_W/2, SCREEN_H/2, 255, "Playing %s", argv[1]);
+   set_palette(desktop_palette);
+   clear_to_color(screen, makecol(255, 255, 255));
+
+   textprintf_centre_ex(screen, font, SCREEN_W/2, SCREEN_H/3, makecol(0, 0, 0), -1,
+			"Driver: %s", midi_driver->name);
+   textprintf_centre_ex(screen, font, SCREEN_W/2, SCREEN_H/2, makecol(0, 0, 0), -1,
+			"Playing %s", argv[1]);
 
    /* start up the MIDI file */
    play_midi(the_music, TRUE);
