@@ -11,15 +11,21 @@
 
 
 
-int main()
+int main(void)
 {
    PALETTE palette;
    RGB temp;
    int c;
 
-   allegro_init();
+   if (allegro_init() != 0)
+      return 1;
    install_keyboard(); 
-   set_gfx_mode(GFX_SAFE, 320, 200, 0, 0);
+   if (set_gfx_mode(GFX_AUTODETECT, 320, 200, 0, 0) != 0) {
+      if (set_gfx_mode(GFX_AUTODETECT, 640, 480, 0, 0) != 0) {
+	 allegro_message("Error setting graphics mode\n%s\n", allegro_error);
+	 return 1;
+      }
+   }
 
    /* first set the palette to black to hide what we are doing */
    set_palette(black_palette);
