@@ -17,9 +17,17 @@ int main(int argc, char *argv[])
    BITMAP *scr_buffer;
    char pcx_name[256];
 
-   allegro_init();
+   if (allegro_init() != 0)
+      return 1;
    install_keyboard();
-   set_gfx_mode(GFX_SAFE, 320, 200, 0, 0);
+
+   if (set_gfx_mode(GFX_AUTODETECT, 320, 200, 0, 0) != 0) {
+      if (set_gfx_mode(GFX_SAFE, 320, 200, 0, 0) != 0) {
+	 set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
+	 allegro_message("Unable to set any graphic mode\n%s\n", allegro_error);
+	 return 1;
+      }
+   }
 
    replace_filename(pcx_name, argv[0], "mysha.pcx", sizeof(pcx_name));
    scr_buffer = load_pcx(pcx_name, my_palette);
@@ -30,12 +38,12 @@ int main(int argc, char *argv[])
    }
 
    set_palette(my_palette);
-   blit(scr_buffer, screen, 0, 0, 0, 0, 320, 200);
+   blit(scr_buffer, screen, 0, 0, 0, 0, scr_buffer->w, scr_buffer->h);
 
    while (!keypressed()) {
-      stretch_blit(scr_buffer, screen, 0, 0, 
-		   rand()%scr_buffer->w, rand()%scr_buffer->h,
-		   rand()%320, rand()%200, rand()%320, rand()%200);
+      stretch_blit(scr_buffer, screen, 0, 0, AL_RAND()%scr_buffer->w,
+		   AL_RAND()%scr_buffer->h, AL_RAND()%SCREEN_W, AL_RAND()%SCREEN_H,
+		   AL_RAND()%SCREEN_W, AL_RAND()%SCREEN_H);
       vsync();
    }
 
