@@ -8,7 +8,9 @@
  *                                           /\____/
  *                                           \_/__/
  *
- *      BMP reader by Seymour Shlien.
+ *      BMP reader.
+ *
+ *      By Seymour Shlien.
  *
  *      OS/2 BMP support and BMP save function by Jonas Petersen.
  *
@@ -16,8 +18,10 @@
  */
 
 
+#include <string.h>
+
 #include "allegro.h"
-#include "allegro/aintern.h"
+#include "allegro/internal/aintern.h"
 
 
 #define BI_RGB          0
@@ -260,9 +264,7 @@ static void read_24bit_line(int length, PACKFILE *f, BITMAP *bmp, int line)
       c.b = pack_getc(f);
       c.g = pack_getc(f);
       c.r = pack_getc(f);
-      bmp->line[line][i*3+_rgb_r_shift_24/8] = c.r;
-      bmp->line[line][i*3+_rgb_g_shift_24/8] = c.g;
-      bmp->line[line][i*3+_rgb_b_shift_24/8] = c.b;
+      WRITE3BYTES(bmp->line[line]+i*3, makecol24(c.r, c.g, c.b));
       nbytes += 3;
    }
 
@@ -274,10 +276,60 @@ static void read_24bit_line(int length, PACKFILE *f, BITMAP *bmp, int line)
 
 
 
+/* read_bitfields_image:
+ *  For reading the bitfield compressed BMP image format.
+ */
+static void read_bitfields_image(PACKFILE *f, BITMAP *bmp, AL_CONST BITMAPINFOHEADER *infoheader)
+{
+   int k, i;
+   int bpp;
+   int bytes_per_pixel;
+   int red, grn, blu;
+   unsigned long buffer;
+
+   bpp = bitmap_color_depth(bmp);
+   bytes_per_pixel = BYTES_PER_PIXEL(bpp);
+
+   for (i=0; i<(int)infoheader->biHeight; i++) {
+      for (k=0; k<(int)infoheader->biWidth; k++) {
+
+	 pack_fread(&buffer, bytes_per_pixel, f);
+
+	 if (bpp == 15) {
+	    red = (buffer >> 10) & 0x1f;
+	    grn = (buffer >> 5) & 0x1f;
+	    blu = (buffer) & 0x1f;
+	    buffer = (red << _rgb_r_shift_15) |
+		     (grn << _rgb_g_shift_15) |
+		     (blu << _rgb_b_shift_15);
+	 }
+	 else if (bpp == 16) {
+	    red = (buffer >> 11) & 0x1f;
+	    grn = (buffer >> 5) & 0x3f;
+	    blu = (buffer) & 0x1f;
+	    buffer = (red << _rgb_r_shift_16) |
+		     (grn << _rgb_g_shift_16) |
+		     (blu << _rgb_b_shift_16);
+	 }
+	 else {
+	    red = (buffer >> 16) & 0xff;
+	    grn = (buffer >> 8) & 0xff;
+	    blu = (buffer) & 0xff;
+	    buffer = (red << _rgb_r_shift_32) |
+		     (grn << _rgb_g_shift_32) |
+		     (blu << _rgb_b_shift_32);
+	 }
+
+	 memcpy(&bmp->line[(infoheader->biHeight - i) - 1][k * bytes_per_pixel], &buffer, bytes_per_pixel);
+      }
+   }
+}
+
+
 /* read_image:
  *  For reading the noncompressed BMP image format.
  */
-static void read_image(PACKFILE *f, BITMAP *bmp, BITMAPINFOHEADER *infoheader)
+static void read_image(PACKFILE *f, BITMAP *bmp, AL_CONST BITMAPINFOHEADER *infoheader)
 {
    int i, line;
 
@@ -310,7 +362,7 @@ static void read_image(PACKFILE *f, BITMAP *bmp, BITMAPINFOHEADER *infoheader)
 /* read_RLE8_compressed_image:
  *  For reading the 8 bit RLE compressed BMP image format.
  */
-static void read_RLE8_compressed_image(PACKFILE *f, BITMAP *bmp, BITMAPINFOHEADER *infoheader)
+static void read_RLE8_compressed_image(PACKFILE *f, BITMAP *bmp, AL_CONST BITMAPINFOHEADER *infoheader)
 {
    unsigned char count, val, val0;
    int j, pos, line;
@@ -380,7 +432,7 @@ static void read_RLE8_compressed_image(PACKFILE *f, BITMAP *bmp, BITMAPINFOHEADE
 /* read_RLE4_compressed_image:
  *  For reading the 4 bit RLE compressed BMP image format.
  */
-static void read_RLE4_compressed_image(PACKFILE *f, BITMAP *bmp, BITMAPINFOHEADER *infoheader)
+static void read_RLE4_compressed_image(PACKFILE *f, BITMAP *bmp, AL_CONST BITMAPINFOHEADER *infoheader)
 {
    unsigned char b[8];
    unsigned char count;
@@ -462,19 +514,24 @@ static void read_RLE4_compressed_image(PACKFILE *f, BITMAP *bmp, BITMAPINFOHEADE
  *
  *  Thanks to Seymour Shlien for contributing this function.
  */
-BITMAP *load_bmp(char *filename, RGB *pal)
+BITMAP *load_bmp(AL_CONST char *filename, RGB *pal)
 {
    BITMAPFILEHEADER fileheader;
    BITMAPINFOHEADER infoheader;
    PACKFILE *f;
    BITMAP *bmp;
    PALETTE tmppal;
+   int want_palette = TRUE;
    int ncol;
    unsigned long biSize;
    int bpp, dest_depth;
+   ASSERT(filename);
 
-   if (!pal)
+   /* we really need a palette */
+   if (!pal) {
+      want_palette = FALSE;
       pal = tmppal;
+   }
 
    f = pack_fopen(filename, F_READ);
    if (!f)
@@ -494,7 +551,9 @@ BITMAP *load_bmp(char *filename, RGB *pal)
       }
       /* compute number of colors recorded */
       ncol = (fileheader.bfOffBits - 54) / 4;
-      read_bmicolors(ncol, pal, f, 1);
+
+      if (infoheader.biCompression != BI_BITFIELDS)
+	 read_bmicolors(ncol, pal, f, 1);
    }
    else if (biSize == OS2INFOHEADERSIZE) {
       if (read_os2_bminfoheader(f, &infoheader) != 0) {
@@ -503,20 +562,41 @@ BITMAP *load_bmp(char *filename, RGB *pal)
       }
       /* compute number of colors recorded */
       ncol = (fileheader.bfOffBits - 26) / 3;
-      read_bmicolors(ncol, pal, f, 0);
+
+      if (infoheader.biCompression != BI_BITFIELDS)
+	 read_bmicolors(ncol, pal, f, 0);
    }
    else {
       pack_fclose(f);
       return NULL;
    }
 
-   if (infoheader.biBitCount == 24) {
+   if (infoheader.biBitCount == 24)
       bpp = 24;
-      generate_332_palette(pal);
-   }
+   else if (infoheader.biBitCount == 16)
+      bpp = 16;
    else
       bpp = 8;
 
+   if (infoheader.biCompression == BI_BITFIELDS) {
+      unsigned long redMask = pack_igetl(f);
+      unsigned long grnMask = pack_igetl(f);
+      unsigned long bluMask = pack_igetl(f);
+
+      if ((bluMask == 0x001f) && (redMask == 0x7C00))
+	 bpp = 15;
+      else if ((bluMask == 0x001f) && (redMask == 0xF800))
+	 bpp = 16;
+      else if ((bluMask == 0x0000FF) && (redMask == 0xFF0000))
+	 bpp = 32;
+      else {
+	 /* Unrecognised bit masks/depth, refuse to load. */
+	 pack_fclose(f);
+	 return NULL;
+      }
+   }
+
+
    dest_depth = _color_load_depth(bpp, FALSE);
 
    bmp = create_bitmap_ex(bpp, infoheader.biWidth, infoheader.biHeight);
@@ -525,7 +605,7 @@ BITMAP *load_bmp(char *filename, RGB *pal)
       return NULL;
    }
 
-   clear(bmp);
+   clear_bitmap(bmp);
 
    switch (infoheader.biCompression) {
 
@@ -541,13 +621,26 @@ BITMAP *load_bmp(char *filename, RGB *pal)
 	 read_RLE4_compressed_image(f, bmp, &infoheader);
 	 break;
 
+      case BI_BITFIELDS:
+	 read_bitfields_image(f, bmp, &infoheader);
+	 break;
+
       default:
 	 destroy_bitmap(bmp);
 	 bmp = NULL;
    }
 
-   if (dest_depth != bpp)
+   if (dest_depth != bpp) {
+      /* restore original palette except if it comes from the bitmap */
+      if ((bpp != 8) && (!want_palette))
+	 pal = NULL;
+
       bmp = _fixup_loaded_bitmap(bmp, pal, dest_depth);
+   }
+   
+   /* construct a fake palette if 8-bit mode is not involved */
+   if ((bpp != 8) && (dest_depth != 8) && want_palette)
+      generate_332_palette(pal);
 
    pack_fclose(f);
    return bmp;
@@ -559,15 +652,22 @@ BITMAP *load_bmp(char *filename, RGB *pal)
  *  Writes a bitmap into a BMP file, using the specified palette (this
  *  should be an array of at least 256 RGB structures).
  */
-int save_bmp(char *filename, BITMAP *bmp, RGB *pal) 
+int save_bmp(AL_CONST char *filename, BITMAP *bmp, AL_CONST RGB *pal) 
 {
    PACKFILE *f;
    PALETTE tmppal;
    int bfSize;
-   int depth = bitmap_color_depth(bmp);
-   int bpp = (depth == 8) ? 8 : 24;
-   int filler = 3 - ((bmp->w*(bpp/8)-1) & 3);
+   int biSizeImage;
+   int depth;
+   int bpp;
+   int filler;
    int c, i, j;
+   ASSERT(filename);
+   ASSERT(bmp);
+
+   depth = bitmap_color_depth(bmp);
+   bpp = (depth == 8) ? 8 : 24;
+   filler = 3 - ((bmp->w*(bpp/8)-1) & 3);
 
    if (!pal) {
       get_palette(tmppal);
@@ -575,18 +675,21 @@ int save_bmp(char *filename, BITMAP *bmp, RGB *pal)
    }
 
    if (bpp == 8) {
-      bfSize = 54                      /* header */
-	       +256*4                  /* palette */
-	       +bmp->w*bmp->h;         /* image data */
+      biSizeImage = (bmp->w + filler) * bmp->h;
+      bfSize = (54		       /* header */
+		+ 256*4		       /* palette */
+		+ biSizeImage);	       /* image data */
    }
    else {
-      bfSize = 54                      /* header */
-	       +bmp->w*bmp->h*3;       /* image data */
+      biSizeImage = (bmp->w*3 + filler) * bmp->h;
+      bfSize = 54 + biSizeImage;       /* header + image data */
    }
 
    f = pack_fopen(filename, F_WRITE);
    if (!f) 
-      return *allegro_errno;
+      return -1;
+
+   *allegro_errno = 0;
 
    /* file_header */
    pack_iputw(0x4D42, f);              /* bfType ("BM") */
@@ -600,17 +703,15 @@ int save_bmp(char *filename, BITMAP *bmp, RGB *pal)
       pack_iputl(54, f); 
 
    /* info_header */
-   bfSize = bmp->w * bmp->h * bpp/8;
-
    pack_iputl(40, f);                  /* biSize */
    pack_iputl(bmp->w, f);              /* biWidth */
    pack_iputl(bmp->h, f);              /* biHeight */
    pack_iputw(1, f);                   /* biPlanes */
    pack_iputw(bpp, f);                 /* biBitCount */
    pack_iputl(0, f);                   /* biCompression */
-   pack_iputl(bfSize, f);              /* biSizeImage */
-   pack_iputl(0, f);                   /* biXPelsPerMeter */
-   pack_iputl(0, f);                   /* biYPelsPerMeter */
+   pack_iputl(biSizeImage, f);         /* biSizeImage */
+   pack_iputl(0xB12, f);               /* biXPelsPerMeter (0xB12 = 72 dpi) */
+   pack_iputl(0xB12, f);               /* biYPelsPerMeter */
 
    if (bpp == 8) {
       pack_iputl(256, f);              /* biClrUsed */
@@ -648,6 +749,10 @@ int save_bmp(char *filename, BITMAP *bmp, RGB *pal)
    }
 
    pack_fclose(f);
-   return *allegro_errno;
+
+   if (*allegro_errno)
+      return -1;
+   else
+      return 0;
 }
 
