@@ -10,7 +10,7 @@
  *
  *      Interpolation routines for hicolor and truecolor pixels.
  *
- *      By Burton Radons.
+ *      By Cloud Wu and Burton Radons.
  *
  *      Alpha blending optimised by Peter Cech.
  *
@@ -19,7 +19,7 @@
 
 
 #include "allegro.h"
-#include "allegro/aintern.h"
+#include "allegro/internal/aintern.h"
 
 
 
@@ -157,9 +157,9 @@ unsigned long _blender_alpha24_bgr(unsigned long x, unsigned long y, unsigned lo
  */
 unsigned long _blender_add24(unsigned long x, unsigned long y, unsigned long n)
 {
-   int r = getr24(x) + getr24(y) * n / 256;
-   int g = getg24(x) + getg24(y) * n / 256;
-   int b = getb24(x) + getb24(y) * n / 256;
+   int r = getr24(y) + getr24(x) * n / 256;
+   int g = getg24(y) + getg24(x) * n / 256;
+   int b = getb24(y) + getb24(x) * n / 256;
 
    r = MIN(r, 255);
    g = MIN(g, 255);
@@ -209,9 +209,9 @@ unsigned long _blender_color24(unsigned long x, unsigned long y, unsigned long n
  */
 unsigned long _blender_difference24(unsigned long x, unsigned long y, unsigned long n)
 {
-   return BLEND(24, ABS(getr24(x) - getr24(y)),
-		    ABS(getg24(x) - getg24(y)),
-		    ABS(getb24(x) - getb24(y)));
+   return BLEND(24, ABS(getr24(y) - getr24(x)),
+		    ABS(getg24(y) - getg24(x)),
+		    ABS(getb24(y) - getb24(x)));
 }
 
 
@@ -221,7 +221,10 @@ unsigned long _blender_difference24(unsigned long x, unsigned long y, unsigned l
  */
 unsigned long _blender_dissolve24(unsigned long x, unsigned long y, unsigned long n)
 {
-   return ((rand() & 255) > (int)n) ? x : y;
+   if (n == 255)
+      return x;
+
+   return ((_al_rand() & 255) < (int)n) ? x : y;
 }
 
 
@@ -441,9 +444,9 @@ unsigned long _blender_alpha16_bgr(unsigned long x, unsigned long y, unsigned lo
  */
 unsigned long _blender_add16(unsigned long x, unsigned long y, unsigned long n)
 {
-   int r = getr16(x) + getr16(y) * n / 256;
-   int g = getg16(x) + getg16(y) * n / 256;
-   int b = getb16(x) + getb16(y) * n / 256;
+   int r = getr16(y) + getr16(x) * n / 256;
+   int g = getg16(y) + getg16(x) * n / 256;
+   int b = getb16(y) + getb16(x) * n / 256;
 
    r = MIN(r, 255);
    g = MIN(g, 255);
@@ -493,9 +496,9 @@ unsigned long _blender_color16(unsigned long x, unsigned long y, unsigned long n
  */
 unsigned long _blender_difference16(unsigned long x, unsigned long y, unsigned long n)
 {
-   return BLEND(16, ABS(getr16(x) - getr16(y)),
-		    ABS(getg16(x) - getg16(y)),
-		    ABS(getb16(x) - getb16(y)));
+   return BLEND(16, ABS(getr16(y) - getr16(x)),
+		    ABS(getg16(y) - getg16(x)),
+		    ABS(getb16(y) - getb16(x)));
 }
 
 
@@ -505,7 +508,10 @@ unsigned long _blender_difference16(unsigned long x, unsigned long y, unsigned l
  */
 unsigned long _blender_dissolve16(unsigned long x, unsigned long y, unsigned long n)
 {
-   return ((rand() & 255) > (int)n) ? x : y;
+   if (n == 255)
+      return x;
+
+   return ((_al_rand() & 255) < (int)n) ? x : y;
 }
 
 
@@ -718,9 +724,9 @@ unsigned long _blender_alpha15_bgr(unsigned long x, unsigned long y, unsigned lo
  */
 unsigned long _blender_add15(unsigned long x, unsigned long y, unsigned long n)
 {
-   int r = getr15(x) + getr15(y) * n / 256;
-   int g = getg15(x) + getg15(y) * n / 256;
-   int b = getb15(x) + getb15(y) * n / 256;
+   int r = getr15(y) + getr15(x) * n / 256;
+   int g = getg15(y) + getg15(x) * n / 256;
+   int b = getb15(y) + getb15(x) * n / 256;
 
    r = MIN(r, 255);
    g = MIN(g, 255);
@@ -770,9 +776,9 @@ unsigned long _blender_color15(unsigned long x, unsigned long y, unsigned long n
  */
 unsigned long _blender_difference15(unsigned long x, unsigned long y, unsigned long n)
 {
-   return BLEND(15, ABS(getr15(x) - getr15(y)),
-		    ABS(getg15(x) - getg15(y)),
-		    ABS(getb15(x) - getb15(y)));
+   return BLEND(15, ABS(getr15(y) - getr15(x)),
+		    ABS(getg15(y) - getg15(x)),
+		    ABS(getb15(y) - getb15(x)));
 }
 
 
@@ -782,7 +788,10 @@ unsigned long _blender_difference15(unsigned long x, unsigned long y, unsigned l
  */
 unsigned long _blender_dissolve15(unsigned long x, unsigned long y, unsigned long n)
 {
-   return ((rand() & 255) > (int)n) ? x : y;
+   if (n == 255)
+      return x;
+
+   return ((_al_rand() & 255) < (int)n) ? x : y;
 }
 
 
