@@ -17,7 +17,8 @@
  *
  *      Jan Hubicka wrote the super-fast version of create_rgb_table().
  *
- *      Michael Bevin optimised the create_trans_table() function.
+ *      Michael Bevin and Sven Sandberg optimised the create_trans_table()
+ *      function.
  *
  *      Sven Sandberg optimised the create_light_table() function.
  *
@@ -30,6 +31,7 @@
 #include <math.h>
 
 #include "allegro.h"
+#include "allegro/internal/aintern.h"
 
 
 
@@ -267,7 +269,7 @@ static unsigned int col_diff[3*128];
  *  Color matching is done with weighted squares, which are much faster
  *  if we pregenerate a little lookup table...
  */
-static void bestfit_init()
+static void bestfit_init(void)
 {
    int i;
 
@@ -284,7 +286,7 @@ static void bestfit_init()
 /* bestfit_color:
  *  Searches a palette for the color closest to the requested R, G, B value.
  */
-int bestfit_color(PALETTE pal, int r, int g, int b)
+int bestfit_color(AL_CONST PALETTE pal, int r, int g, int b)
 {
    int i, coldiff, lowest, bestfit;
 
@@ -294,13 +296,14 @@ int bestfit_color(PALETTE pal, int r, int g, int b)
    bestfit = 0;
    lowest = INT_MAX;
 
+   /* only the transparent (pink) color can be mapped to index 0 */
    if ((r == 63) && (g == 0) && (b == 63))
       i = 0;
    else
       i = 1;
 
    while (i<PAL_SIZE) {
-      RGB *rgb = &pal[i];
+      AL_CONST RGB *rgb = &pal[i];
       coldiff = (col_diff + 0) [ (rgb->g - g) & 0x7F ];
       if (coldiff < lowest) {
 	 coldiff += (col_diff + 128) [ (rgb->r - r) & 0x7F ];
@@ -345,28 +348,61 @@ void hsv_to_rgb(float h, float s, float v, int *r, int *g, int *b)
    float f, x, y, z;
    int i;
 
-   v *= 255.0;
+   v *= 255.0f;
 
-   if (s == 0.0) {
-      *r = *g = *b = (int)v;
+   if (s == 0.0f) { /* ok since we don't divide by s, and faster */
+      *r = *g = *b = v + 0.5f;
    }
    else {
-      while (h < 0)
-	 h += 360;
-      h = fmod(h, 360) / 60.0;
+      h = fmod(h, 360.0f) / 60.0f;
+      if (h < 0.0f)
+	 h += 6.0f;
+
       i = (int)h;
       f = h - i;
-      x = v * (1.0 - s);
-      y = v * (1.0 - (s * f));
-      z = v * (1.0 - (s * (1.0 - f)));
+      x = v * s;
+      y = x * f;
+      v += 0.5f; /* round to the nearest integer below */
+      z = v - x;
 
       switch (i) {
-	 case 0: *r = v; *g = z; *b = x; break;
-	 case 1: *r = y; *g = v; *b = x; break;
-	 case 2: *r = x; *g = v; *b = z; break;
-	 case 3: *r = x; *g = y; *b = v; break;
-	 case 4: *r = z; *g = x; *b = v; break;
-	 case 5: *r = v; *g = x; *b = y; break;
+
+	 case 6:
+	 case 0:
+	    *r = v;
+	    *g = z + y;
+	    *b = z;
+	    break;
+
+	 case 1:
+	    *r = v - y;
+	    *g = v;
+	    *b = z;
+	    break;
+
+	 case 2:
+	    *r = z;
+	    *g = v;
+	    *b = z + y;
+	    break;
+
+	 case 3:
+	    *r = z;
+	    *g = v - y;
+	    *b = v;
+	    break;
+
+	 case 4:
+	    *r = z + y;
+	    *g = z;
+	    *b = v;
+	    break;
+
+	 case 5:
+	    *r = v;
+	    *g = z;
+	    *b = v - y;
+	    break;
       }
    }
 }
@@ -378,36 +414,53 @@ void hsv_to_rgb(float h, float s, float v, int *r, int *g, int *b)
  */
 void rgb_to_hsv(int r, int g, int b, float *h, float *s, float *v)
 {
-   float min, max, delta, rc, gc, bc;
+   int delta;
 
-   rc = (float)r / 255.0;
-   gc = (float)g / 255.0;
-   bc = (float)b / 255.0;
-   max = MAX(rc, MAX(gc, bc));
-   min = MIN(rc, MIN(gc, bc));
-   delta = max - min;
-   *v = max;
-
-   if (max != 0.0)
-      *s = delta / max;
-   else
-      *s = 0.0;
-
-   if (*s == 0.0) {
-      *h = 0.0; 
+   if (r > g) {
+      if (b > r) {
+	 /* b>r>g */
+	 delta = b-g;
+	 *h = 240.0f + ((r-g) * 60) / (float)delta;
+	 *s = (float)delta / (float)b;
+	 *v = (float)b * (1.0f/255.0f);
+      }
+      else {
+	 /* r>g and r>b */
+	 delta = r - MIN(g, b);
+	 *h = ((g-b) * 60) / (float)delta;
+	 if (*h < 0.0f)
+	    *h += 360.0f;
+	 *s = (float)delta / (float)r;
+	 *v = (float)r * (1.0f/255.0f);
+      }
    }
    else {
-      if (rc == max)
-	 *h = (gc - bc) / delta;
-      else if (gc == max)
-	 *h = 2 + (bc - rc) / delta;
-      else if (bc == max)
-	 *h = 4 + (rc - gc) / delta;
-
-      *h *= 60.0;
-      if (*h < 0)
-	 *h += 360.0;
-    }
+      if (b > g) {
+	 /* b>g>=r */
+	 delta = b-r;
+	 *h = 240.0f + ((r-g) * 60) / (float)delta;
+	 *s = (float)delta / (float)b;
+	 *v = (float)b * (1.0f/255.0f);
+      }
+      else {
+	 /* g>=b and g>=r */
+	 delta = g - MIN(r, b);
+	 if (delta == 0) {
+	    *h = 0.0f;
+	    if (g == 0)
+	       *s = *v = 0.0f;
+	    else {
+	       *s = (float)delta / (float)g;
+	       *v = (float)g * (1.0f/255.0f);
+	    }
+	 }
+	 else {
+	    *h = 120.0f + ((b-r) * 60) / (float)delta;
+	    *s = (float)delta / (float)g;
+	    *v = (float)g * (1.0f/255.0f);
+	 }
+      }
+   }
 }
 
 
@@ -416,16 +469,16 @@ void rgb_to_hsv(int r, int g, int b, float *h, float *s, float *v)
  *  Fills an RGB_MAP lookup table with conversion data for the specified
  *  palette. This is the faster version by Jan Hubicka.
  *
- *  Uses alg. similiar to foodfill - it adds one seed per every color in 
- *  palette to its best possition. Then areas around seed are filled by 
- *  same color because it is best aproximation for them, and then areas 
+ *  Uses alg. similar to floodfill - it adds one seed per every color in
+ *  palette to its best position. Then areas around seed are filled by
+ *  same color because it is best approximation for them, and then areas
  *  about them etc...
  *
  *  It does just about 80000 tests for distances and this is about 100
- *  times better than normal 256*32000 tests so the caluclation time
+ *  times better than normal 256*32000 tests so the calculation time
  *  is now less than one second at all computers I tested.
  */
-void create_rgb_table(RGB_MAP *table, PALETTE pal, void (*callback)(int pos))
+void create_rgb_table(RGB_MAP *table, AL_CONST PALETTE pal, void (*callback)(int pos))
 {
    #define UNUSED 65535
    #define LAST 65532
@@ -460,24 +513,29 @@ void create_rgb_table(RGB_MAP *table, PALETTE pal, void (*callback)(int pos))
 		     (((int)dist((r1), (g1), (b1), \
 				 (pal1).r, (pal1).g, (pal1).b)) > (int)dist2)
 
-   /* checking of possition */
+   /* checking of position */
    #define dopos(rp, gp, bp, ts) \
       if ((rp > -1 || r > 0) && (rp < 1 || r < 61) && \
 	  (gp > -1 || g > 0) && (gp < 1 || g < 61) && \
 	  (bp > -1 || b > 0) && (bp < 1 || b < 61)) { \
 	 i = first + rp * 32 * 32 + gp * 32 + bp; \
-	 if (ts ? data[i] != val : !data[i]) { \
+         if (!data[i]) { \
+	    data[i] = val; \
+	    add1(i); \
+	 } \
+	 else if ((ts) && (data[i] != val)) { \
 	    dist2 = (rp ? (col_diff+128)[(r+2*rp-pal[val].r) & 0x7F] : r2) + \
-		    (gp ? (col_diff)[(g+2*gp-pal[val].g) & 0x7F] : g2) + \
+		    (gp ? (col_diff    )[(g+2*gp-pal[val].g) & 0x7F] : g2) + \
 		    (bp ? (col_diff+256)[(b+2*bp-pal[val].b) & 0x7F] : b2); \
 	    if (better((r+2*rp), (g+2*gp), (b+2*bp), pal[data[i]])) { \
 	       data[i] = val; \
-	       add1 (i); \
+	       add1(i); \
 	    } \
 	 } \
       }
 
-   int i, curr, r, g, b, val, r2, g2, b2, dist2;
+   int i, curr, r, g, b, val, dist2;
+   unsigned int r2, g2, b2;
    unsigned short next[32*32*32];
    unsigned char *data;
    int first = LAST;
@@ -495,8 +553,8 @@ void create_rgb_table(RGB_MAP *table, PALETTE pal, void (*callback)(int pos))
 
    data = (unsigned char *)table->data;
 
-   /* add starting seeds for foodfill */
-   for (i=1; i<256; i++) { 
+   /* add starting seeds for floodfill */
+   for (i=1; i<PAL_SIZE; i++) { 
       curr = pos(pal[i].r, pal[i].g, pal[i].b);
       if (next[curr] == UNUSED) {
 	 data[curr] = i;
@@ -504,7 +562,7 @@ void create_rgb_table(RGB_MAP *table, PALETTE pal, void (*callback)(int pos))
       }
    }
 
-   /* main foodfill: two versions of loop for faster growing in blue axis */
+   /* main floodfill: two versions of loop for faster growing in blue axis */
    while (first != LAST) { 
       depos(first, r, g, b);
 
@@ -585,6 +643,10 @@ void create_rgb_table(RGB_MAP *table, PALETTE pal, void (*callback)(int pos))
       }
    }
 
+   /* only the transparent (pink) color can be mapped to index 0 */
+   if ((pal[0].r == 63) && (pal[0].g == 0) && (pal[0].b == 63))
+      table->data[31][0][31] = 0;
+
    if (callback)
       while (cbcount < 256)
 	 callback(cbcount++);
@@ -600,35 +662,55 @@ void create_rgb_table(RGB_MAP *table, PALETTE pal, void (*callback)(int pos))
  *  not NULL, it will be called 256 times during the calculation, allowing
  *  you to display a progress indicator.
  */
-void create_light_table(COLOR_MAP *table, PALETTE pal, int r, int g, int b, void (*callback)(int pos))
+void create_light_table(COLOR_MAP *table, AL_CONST PALETTE pal, int r, int g, int b, void (*callback)(int pos))
 {
-   int r1, g1, b1, x, y;
-   unsigned int tmp;
-   RGB c;
+   int r1, g1, b1, r2, g2, b2, x, y;
+   unsigned int t1, t2;
 
-   for (x=0; x<PAL_SIZE; x++) {
-      tmp = (255 - x) * 65793;
+   if (rgb_map) {
+      for (x=0; x<PAL_SIZE-1; x++) {
+	 t1 = x * 0x010101;
+	 t2 = 0xFFFFFF - t1;
 
-      r1 = r * tmp + (1 << 23);
-      g1 = g * tmp + (1 << 23);
-      b1 = b * tmp + (1 << 23);
+	 r1 = (1 << 24) + r * t2;
+	 g1 = (1 << 24) + g * t2;
+	 b1 = (1 << 24) + b * t2;
 
-      tmp = (1 << 24) - tmp;
+	 for (y=0; y<PAL_SIZE; y++) {
+	    r2 = (r1 + pal[y].r * t1) >> 25;
+	    g2 = (g1 + pal[y].g * t1) >> 25;
+	    b2 = (b1 + pal[y].b * t1) >> 25;
 
-      for (y=0; y<PAL_SIZE; y++) {
-	 c.r = (r1 + (unsigned int)pal[y].r * tmp) >> 24;
-	 c.g = (g1 + (unsigned int)pal[y].g * tmp) >> 24;
-	 c.b = (b1 + (unsigned int)pal[y].b * tmp) >> 24;
+	    table->data[x][y] = rgb_map->data[r2][g2][b2];
+	 }
+      }
+      if (callback)
+	 (*callback)(x);
+   }
+   else {
+      for (x=0; x<PAL_SIZE-1; x++) {
+	 t1 = x * 0x010101;
+	 t2 = 0xFFFFFF - t1;
 
-	 if (rgb_map)
-	    table->data[x][y] = rgb_map->data[c.r>>1][c.g>>1][c.b>>1];
-	 else
-	    table->data[x][y] = bestfit_color(pal, c.r, c.g, c.b);
+	 r1 = (1 << 23) + r * t2;
+	 g1 = (1 << 23) + g * t2;
+	 b1 = (1 << 23) + b * t2;
+
+	 for (y=0; y<PAL_SIZE; y++) {
+	    r2 = (r1 + pal[y].r * t1) >> 24;
+	    g2 = (g1 + pal[y].g * t1) >> 24;
+	    b2 = (b1 + pal[y].b * t1) >> 24;
+
+	    table->data[x][y] = bestfit_color(pal, r2, g2, b2);
+	 }
       }
 
       if (callback)
 	 (*callback)(x);
    }
+
+   for (y=0; y<PAL_SIZE; y++)
+      table->data[255][y] = y;
 }
 
 
@@ -642,53 +724,72 @@ void create_light_table(COLOR_MAP *table, PALETTE pal, int r, int g, int b, void
  *  function is not NULL, it will be called 256 times during the calculation, 
  *  allowing you to display a progress indicator.
  */
-void create_trans_table(COLOR_MAP *table, PALETTE pal, int r, int g, int b, void (*callback)(int pos))
+void create_trans_table(COLOR_MAP *table, AL_CONST PALETTE pal, int r, int g, int b, void (*callback)(int pos))
 {
    int tmp[768], *q;
    int x, y, i, j, k;
    unsigned char *p;
-   RGB c;
+   int tr, tg, tb;
+   int add;
+
+   /* This is a bit ugly, but accounts for the solidity parameters
+      being in the range 0-255 rather than 0-256. Given that the
+      precision of r,g,b components is only 6 bits it shouldn't do any
+      harm. */
+   if (r > 128)
+      r++;
+   if (g > 128)
+      g++;
+   if (b > 128)
+      b++;
+
+   if (rgb_map)
+      add = 255;
+   else
+      add = 127;
 
    for (x=0; x<256; x++) {
-      tmp[x*3]   = pal[x].r * (255-r) / 255;
-      tmp[x*3+1] = pal[x].g * (255-g) / 255;
-      tmp[x*3+2] = pal[x].b * (255-b) / 255;
+      tmp[x*3]   = pal[x].r * (256-r) + add;
+      tmp[x*3+1] = pal[x].g * (256-g) + add;
+      tmp[x*3+2] = pal[x].b * (256-b) + add;
    }
 
-   for (y=0; y<PAL_SIZE; y++)
-      table->data[0][y] = y;
-
-   if (callback)
-      (*callback)(0);
-
-   for (x=1; x<256; x++) {
-      i = pal[x].r * r / 255;
-      j = pal[x].g * g / 255;
-      k = pal[x].b * b / 255;
+   for (x=1; x<PAL_SIZE; x++) {
+      i = pal[x].r * r;
+      j = pal[x].g * g;
+      k = pal[x].b * b;
 
       p = table->data[x];
       q = tmp;
 
       if (rgb_map) {
-	 for (y=0; y<256; y++) {
-	    c.r = i + *(q++);
-	    c.g = j + *(q++);
-	    c.b = k + *(q++);
-	    p[y] = rgb_map->data[c.r>>1][c.g>>1][c.b>>1];
+	 for (y=0; y<PAL_SIZE; y++) {
+	    tr = (i + *(q++)) >> 9;
+	    tg = (j + *(q++)) >> 9;
+	    tb = (k + *(q++)) >> 9;
+	    p[y] = rgb_map->data[tr][tg][tb];
 	 }
       }
       else {
-	 for (y=0; y<256; y++) {
-	    c.r = i + *(q++); 
-	    c.g = j + *(q++); 
-	    c.b = k + *(q++);
-	    p[y] = bestfit_color(pal, c.r, c.g, c.b);
+	 for (y=0; y<PAL_SIZE; y++) {
+	    tr = (i + *(q++)) >> 8;
+	    tg = (j + *(q++)) >> 8;
+	    tb = (k + *(q++)) >> 8;
+	    p[y] = bestfit_color(pal, tr, tg, tb);
 	 }
       }
 
       if (callback)
-	 (*callback)(x);
+	 (*callback)(x-1);
    }
+
+   for (y=0; y<PAL_SIZE; y++) {
+      table->data[0][y] = y;
+      table->data[y][y] = y;
+   }
+
+   if (callback)
+      (*callback)(255);
 }
 
 
@@ -702,7 +803,7 @@ void create_trans_table(COLOR_MAP *table, PALETTE pal, int r, int g, int b, void
  *  256 times during the calculation, allowing you to display a progress 
  *  indicator.
  */
-void create_color_table(COLOR_MAP *table, PALETTE pal, void (*blend)(PALETTE pal, int x, int y, RGB *rgb), void (*callback)(int pos))
+void create_color_table(COLOR_MAP *table, AL_CONST PALETTE pal, void (*blend)(AL_CONST PALETTE pal, int x, int y, RGB *rgb), void (*callback)(int pos))
 {
    int x, y;
    RGB c;
@@ -729,7 +830,7 @@ void create_color_table(COLOR_MAP *table, PALETTE pal, void (*blend)(PALETTE pal
  *  paletted equivalent of whatever truecolor blender mode is currently 
  *  selected.
  */
-void create_blender_table(COLOR_MAP *table, PALETTE pal, void (*callback)(int pos))
+void create_blender_table(COLOR_MAP *table, AL_CONST PALETTE pal, void (*callback)(int pos))
 {
    int x, y, c;
    int r, g, b;
