@@ -14,12 +14,14 @@
  *
  *      Hook functions added by Martijn Versteegh.
  *
+ *      Annie Testes lifted several hardcoded length limitations.
+ *
  *      See readme.txt for copyright information.
  */
 
 
 #include "allegro.h"
-#include "allegro/aintern.h"
+#include "allegro/internal/aintern.h"
 
 
 
@@ -42,9 +44,9 @@ typedef struct CONFIG
 typedef struct CONFIG_HOOK
 {
    char *section;                   /* hooked config section info */
-   int (*intgetter)(char *name, int def);
-   char *(*stringgetter)(char *name, char *def);
-   void (*stringsetter)(char *name, char *value);
+   int (*intgetter)(AL_CONST char *name, int def);
+   AL_CONST char *(*stringgetter)(AL_CONST char *name, AL_CONST char *def);
+   void (*stringsetter)(AL_CONST char *name, AL_CONST char *value);
    struct CONFIG_HOOK *next; 
 } CONFIG_HOOK;
 
@@ -62,6 +64,63 @@ static int config_installed = FALSE;
 
 
 
+/* flush_config:
+ *  Writes out a config structure to disk if the contents
+ *  have changed.
+ */
+static void flush_config(CONFIG *cfg)
+{
+   CONFIG_ENTRY *pos;
+   PACKFILE *f;
+   char cr[16];
+
+   usetc(cr+usetc(cr, '\n'), 0);
+
+   if (cfg && cfg->filename && cfg->dirty) {
+      /* write changed data to disk */
+      f = pack_fopen(cfg->filename, F_WRITE);
+
+      if (f) {
+	 pos = cfg->head;
+
+	 while (pos) {
+	    if (pos->name) {
+	       pack_fputs(pos->name, f);
+
+	       if (ugetc(pos->name) != '[') {
+		  pack_putc(' ', f);
+		  pack_putc('=', f);
+		  pack_putc(' ', f);
+	       }
+	    }
+
+	    if (pos->data)
+	       pack_fputs(pos->data, f);
+
+	    pack_fputs(cr, f);
+
+	    pos = pos->next;
+         }
+
+	 pack_fclose(f);
+	 cfg->dirty = FALSE;
+      }
+   }
+}
+
+
+
+/* flush_config_file:
+ *  Writes out the config file to disk if the contents
+ *  have changed.
+ */
+void flush_config_file(void)
+{
+   flush_config(config[0]);
+}
+
+
+
 /* destroy_config:
  *  Destroys a config structure, writing it out to disk if the contents
  *  have changed.
@@ -69,44 +128,12 @@ static int config_installed = FALSE;
 static void destroy_config(CONFIG *cfg)
 {
    CONFIG_ENTRY *pos, *prev;
-   char cr[16];
-
-   usetc(cr+usetc(cr, '\n'), 0);
 
    if (cfg) {
-      if (cfg->filename) {
-	 if (cfg->dirty) {
-	    /* write changed data to disk */
-	    PACKFILE *f = pack_fopen(cfg->filename, F_WRITE);
-
-	    if (f) {
-	       pos = cfg->head;
-
-	       while (pos) {
-		  if (pos->name) {
-		     pack_fputs(pos->name, f);
-
-		     if (ugetc(pos->name) != '[') {
-			pack_putc(' ', f);
-			pack_putc('=', f);
-			pack_putc(' ', f);
-		     }
-		  }
-
-		  if (pos->data)
-		     pack_fputs(pos->data, f);
-
-		  pack_fputs(cr, f);
-
-		  pos = pos->next;
-	       }
-
-	       pack_fclose(f);
-	    }
-	 }
+      flush_config(cfg);
 
+      if (cfg->filename)
 	 free(cfg->filename);
-      }
 
       /* destroy the variable list */
       pos = cfg->head;
@@ -189,7 +216,7 @@ static void config_cleanup(void)
  */
 static void init_config(int loaddata)
 {
-   char filename[512], tmp[128], *cfg_name;
+   char filename[1024], tmp[128], *cfg_name;
 
    if (!config_installed) {
       _add_exit_func(config_cleanup);
@@ -202,7 +229,7 @@ static void init_config(int loaddata)
       if (find_allegro_resource(filename, cfg_name, NULL, NULL, NULL, NULL, NULL, sizeof(filename)) != 0) {
 	 get_executable_name(filename, sizeof(filename));
 	 usetc(get_filename(filename), 0);
-	 ustrncat(filename, cfg_name, sizeof(filename)-ustrsizez(filename));
+	 ustrzcat(filename, sizeof(filename), cfg_name);
       }
 
       set_config_file(filename);
@@ -220,12 +247,18 @@ static void init_config(int loaddata)
 
 
 
-/* get_line: 
- *  Helper for splitting files up into individual lines.
+/* get_line:
+ *  Helper for splitting files up into individual lines. Returns the length
+ *  in bytes of the sequence of characters delimited by the first EOL marker
+ *  in the array DATA of length LENGTH, and allocates NAME and VAL to record
+ *  the name and the value of the config entry respectively; otherwise set
+ *  NAME to NULL and returns a copy of the line through VAL if the line was
+ *  blank or a comment. Returns -1 and set allegro_errno on failure.
  */
-static int get_line(char *data, int length, char *name, char *val)
+static int get_line(AL_CONST char *data, int length, char **name, char **val)
 {
-   char buf[256], buf2[256];
+   char *buf;
+   int buf_size=256;
    int inpos, outpos, i, j;
    int c, c2, w0;
 
@@ -233,7 +266,14 @@ static int get_line(char *data, int length, char *name, char *val)
    outpos = 0;
    w0 = ucwidth(0);
 
-   while ((inpos<length) && (outpos<(int)sizeof(buf)-w0)) {
+   buf = malloc(buf_size);
+   if (!buf) {
+     *allegro_errno = ENOMEM;
+     return -1;
+   }
+
+   /* search for an EOL marker */
+   while (inpos<length) {
       c = ugetc(data+inpos);
       if ((c == '\r') || (c == '\n')) {
 	 inpos += uwidth(data+inpos);
@@ -245,6 +285,16 @@ static int get_line(char *data, int length, char *name, char *val)
 	 break;
       }
 
+      /* increase the buffer size if needed */
+      if (outpos>=(int)buf_size-w0) {
+	 buf_size *= 2;
+	 buf = _al_sane_realloc(buf, buf_size);
+	 if (!buf) {
+	    *allegro_errno = ENOMEM;
+	    return -1;
+	 }
+      }
+
       outpos += usetc(buf+outpos, c);
       inpos += uwidth(data+inpos);
    }
@@ -263,35 +313,53 @@ static int get_line(char *data, int length, char *name, char *val)
    /* read name string */
    j = 0;
 
+   /* compute name length */
    while ((c) && (!uisspace(c)) && (c != '=') && (c != '#')) {
-      j += usetc(buf2+j, c);
+      j += ucwidth(c);
       i += uwidth(buf+i);
       c = ugetc(buf+i);
    }
 
    if (j) {
       /* got a variable */
-      usetc(buf2+j, 0);
-      ustrcpy(name, buf2);
+      *name = malloc(j+w0);
+      if (!(*name)) {
+	 *allegro_errno = ENOMEM;
+	 free(buf);
+	 return -1;
+      }
+
+      ustrzcpy(*name, j+w0, buf+i-j);
 
       while ((c) && ((uisspace(c)) || (c == '='))) {
 	 i += uwidth(buf+i);
 	 c = ugetc(buf+i);
       }
 
-      ustrcpy(val, buf+i);
+      *val = ustrdup(buf+i);
+      if (!(*val)) {
+	 free(name);
+	 free(buf);
+	 return -1;
+      }
 
       /* strip trailing spaces */
-      i = ustrlen(val) - 1;
-      while ((i >= 0) && (uisspace(ugetat(val, i))))
-	 usetat(val, i--, 0);
+      i = ustrlen(*val) - 1;
+      while ((i >= 0) && (uisspace(ugetat(*val, i))))
+	 usetat(*val, i--, 0);
    }
    else {
       /* blank line or comment */
-      usetc(name, 0);
-      ustrcpy(val, buf);
+      *name = NULL;
+      *val = ustrdup(buf);
+      if (!(*val)) {
+	 free(buf);
+	 return -1;
+      }
    }
 
+   free(buf);
+
    return inpos;
 }
 
@@ -300,12 +368,11 @@ static int get_line(char *data, int length, char *name, char *val)
 /* set_config:
  *  Does the work of setting up a config structure.
  */
-static void set_config(CONFIG **config, char *data, int length, char *filename)
+static void set_config(CONFIG **config, AL_CONST char *data, int length, AL_CONST char *filename)
 {
-   char name[256];
-   char val[256];
    CONFIG_ENTRY **prev, *p;
-   int pos;
+   char *name, *val;
+   int ret, pos;
 
    init_config(FALSE);
 
@@ -315,16 +382,21 @@ static void set_config(CONFIG **config, char *data, int length, char *filename)
    }
 
    *config = malloc(sizeof(CONFIG));
-   if (!(*config))
+   if (!(*config)) {
+      *allegro_errno = ENOMEM;
       return;
+   }
 
    (*config)->head = NULL;
    (*config)->dirty = FALSE;
 
    if (filename) {
-      (*config)->filename = malloc(ustrsizez(filename));
-      if ((*config)->filename)
-	 ustrcpy((*config)->filename, filename); 
+      (*config)->filename = ustrdup(filename);
+      if (!(*config)->filename) {
+	 free(*config);
+	 *config = NULL;
+	 return;
+      }
    }
    else
       (*config)->filename = NULL;
@@ -333,23 +405,25 @@ static void set_config(CONFIG **config, char *data, int length, char *filename)
    pos = 0;
 
    while (pos < length) {
-      pos += get_line(data+pos, length-pos, name, val);
+      ret = get_line(data+pos, length-pos, &name, &val);
+      if (ret<0) {
+	 free(*config);
+	 *config = NULL;
+	 return;
+      }
+
+      pos += ret;
 
       p = malloc(sizeof(CONFIG_ENTRY));
-      if (!p)
+      if (!p) {
+	 *allegro_errno = ENOMEM;
+	 free(*config);
+	 *config = NULL;
 	 return;
-
-      if (ugetc(name)) {
-	 p->name = malloc(ustrsizez(name));
-	 if (p->name)
-	    ustrcpy(p->name, name);
       }
-      else
-	 p->name = NULL;
 
-      p->data = malloc(ustrsizez(val));
-      if (p->data)
-	 ustrcpy(p->data, val);
+      p->name = name;
+      p->data = val;
 
       p->next = NULL;
       *prev = p;
@@ -362,7 +436,7 @@ static void set_config(CONFIG **config, char *data, int length, char *filename)
 /* load_config_file:
  *  Does the work of loading a config file.
  */
-static void load_config_file(CONFIG **config, char *filename, char *savefile)
+static void load_config_file(CONFIG **config, AL_CONST char *filename, AL_CONST char *savefile)
 {
    char *tmp, *tmp2;
    int length;
@@ -389,7 +463,7 @@ static void load_config_file(CONFIG **config, char *filename, char *savefile)
 	       tmp2 = malloc(length);
 
 	       if (tmp2)
-		  do_uconvert(tmp, U_UTF8, tmp2, U_CURRENT, -1);
+		  do_uconvert(tmp, U_UTF8, tmp2, U_CURRENT, length);
 
 	       length -= ucwidth(0);
 	    }
@@ -422,8 +496,9 @@ static void load_config_file(CONFIG **config, char *filename, char *savefile)
 /* set_config_file:
  *  Sets the file to be used for all future configuration operations.
  */
-void set_config_file(char *filename)
+void set_config_file(AL_CONST char *filename)
 {
+   ASSERT(filename);
    load_config_file(&config[0], filename, filename);
 }
 
@@ -433,8 +508,9 @@ void set_config_file(char *filename)
  *  Sets the block of data to be used for all future configuration 
  *  operations.
  */
-void set_config_data(char *data, int length)
+void set_config_data(AL_CONST char *data, int length)
 {
+   ASSERT(data);
    set_config(&config[0], data, length, NULL);
 }
 
@@ -443,8 +519,9 @@ void set_config_data(char *data, int length)
 /* override_config_file:
  *  Sets the file that will override all future configuration operations.
  */
-void override_config_file(char *filename)
+void override_config_file(AL_CONST char *filename)
 {
+   ASSERT(filename);
    load_config_file(&config_override, filename, NULL);
 }
 
@@ -454,8 +531,9 @@ void override_config_file(char *filename)
  *  Sets the block of data that will override all future configuration 
  *  operations.
  */
-void override_config_data(char *data, int length)
+void override_config_data(AL_CONST char *data, int length)
 {
+   ASSERT(data);
    set_config(&config_override, data, length, NULL);
 }
 
@@ -464,7 +542,7 @@ void override_config_data(char *data, int length)
 /* push_config_state:
  *  Pushes the current config state onto the stack.
  */
-void push_config_state()
+void push_config_state(void)
 {
    int i;
 
@@ -482,7 +560,7 @@ void push_config_state()
 /* pop_config_state:
  *  Pops the current config state off the stack.
  */
-void pop_config_state()
+void pop_config_state(void)
 {
    int i;
 
@@ -500,11 +578,11 @@ void pop_config_state()
 /* prettify_section_name:
  *  Helper for ensuring that a section name is enclosed by [ ] braces.
  */
-static void prettify_section_name(char *in, char *out)
+static void prettify_section_name(AL_CONST char *in, char *out, int out_size)
 {
    int p;
 
-   if (in) {
+   if ((in) && (ustrlen(in))) {
       if (ugetc(in) != '[') {
 	 p = usetc(out, '[');
 	 usetc(out+p, 0);
@@ -512,7 +590,7 @@ static void prettify_section_name(char *in, char *out)
       else
 	 usetc(out, 0);
 
-      ustrcat(out, in);
+      ustrzcat(out, out_size - ucwidth(']'), in);
 
       out += uoffset(out, -1);
 
@@ -533,14 +611,14 @@ static void prettify_section_name(char *in, char *out)
  *  override the normal table of values, and give the provider of the hooks 
  *  complete control over that section.
  */
-void hook_config_section(char *section, int (*intgetter)(char *, int), char *(*stringgetter)(char *, char *), void (*stringsetter)(char *,char *))
+void hook_config_section(AL_CONST char *section, int (*intgetter)(AL_CONST char *, int), AL_CONST char *(*stringgetter)(AL_CONST char *, AL_CONST char *), void (*stringsetter)(AL_CONST char *, AL_CONST char *))
 {
    CONFIG_HOOK *hook, **prev;
    char section_name[256];
 
    init_config(FALSE);
 
-   prettify_section_name(section, section_name);
+   prettify_section_name(section, section_name, sizeof(section_name));
 
    hook = config_hook;
    prev = &config_hook;
@@ -572,12 +650,11 @@ void hook_config_section(char *section, int (*intgetter)(char *, int), char *(*s
    if (!hook)
       return;
 
-   hook->section = malloc(ustrsizez(section_name));
+   hook->section = ustrdup(section_name);
    if (!(hook->section)) {
       free(hook);
       return;
    }
-   ustrcpy(hook->section, section_name);
 
    hook->intgetter = intgetter;
    hook->stringgetter = stringgetter;
@@ -592,12 +669,12 @@ void hook_config_section(char *section, int (*intgetter)(char *, int), char *(*s
 /* is_config_hooked:
  *  Checks whether a specific section is hooked in any way.
  */
-int config_is_hooked(char *section)
+int config_is_hooked(AL_CONST char *section)
 {
    CONFIG_HOOK *hook = config_hook;
    char section_name[256];
 
-   prettify_section_name(section, section_name);
+   prettify_section_name(section, section_name, sizeof(section_name));
 
    while (hook) {
       if (ustricmp(section_name, hook->section) == 0)
@@ -614,10 +691,10 @@ int config_is_hooked(char *section)
 /* find_config_string:
  *  Helper for finding an entry in the configuration file.
  */
-static CONFIG_ENTRY *find_config_string(CONFIG *config, char *section, char *name, CONFIG_ENTRY **prev)
+static CONFIG_ENTRY *find_config_string(CONFIG *config, AL_CONST char *section, AL_CONST char *name, CONFIG_ENTRY **prev)
 {
    CONFIG_ENTRY *p;
-   int in_section = TRUE;
+   int in_section;
 
    if (config) {
       p = config->head;
@@ -625,6 +702,11 @@ static CONFIG_ENTRY *find_config_string(CONFIG *config, char *section, char *nam
       if (prev)
 	 *prev = NULL;
 
+      if (section && ugetc(section))
+         in_section = FALSE;
+      else
+         in_section = TRUE;
+
       while (p) {
 	 if (p->name) {
 	    if ((section) && (ugetc(p->name) == '[') && (ugetat(p->name, -1) == ']')) {
@@ -653,7 +735,7 @@ static CONFIG_ENTRY *find_config_string(CONFIG *config, char *section, char *nam
 /* get_config_string:
  *  Reads a string from the configuration file.
  */
-char *get_config_string(char *section, char *name, char *def)
+AL_CONST char *get_config_string(AL_CONST char *section, AL_CONST char *name, AL_CONST char *def)
 {
    char section_name[256];
    CONFIG_HOOK *hook;
@@ -661,7 +743,7 @@ char *get_config_string(char *section, char *name, char *def)
 
    init_config(TRUE);
 
-   prettify_section_name(section, section_name);
+   prettify_section_name(section, section_name, sizeof(section_name));
 
    /* check for hooked sections */
    hook = config_hook;
@@ -686,8 +768,8 @@ char *get_config_string(char *section, char *name, char *def)
 	 p = find_config_string(config[0], section_name, name, NULL);
    }
 
-   if (p)
-      return (p->data ? p->data : empty_string);
+   if (p && p->data && (ustrlen(p->data) != 0))
+      return p->data;
    else
       return def;
 }
@@ -697,13 +779,13 @@ char *get_config_string(char *section, char *name, char *def)
 /* get_config_int:
  *  Reads an integer from the configuration file.
  */
-int get_config_int(char *section, char *name, int def)
+int get_config_int(AL_CONST char *section, AL_CONST char *name, int def)
 {
    CONFIG_HOOK *hook;
    char section_name[256];
-   char *s;
+   AL_CONST char *s;
 
-   prettify_section_name(section, section_name);
+   prettify_section_name(section, section_name, sizeof(section_name));
 
    /* check for hooked sections */
    hook = config_hook;
@@ -740,14 +822,15 @@ int get_config_int(char *section, char *name, int def)
 /* get_config_hex:
  *  Reads a hexadecimal integer from the configuration file.
  */
-int get_config_hex(char *section, char *name, int def)
+int get_config_hex(AL_CONST char *section, AL_CONST char *name, int def)
 {
-   char *s = get_config_string(section, name, NULL);
+   AL_CONST char *s = get_config_string(section, name, NULL);
+   char tmp[64];
    int i;
 
    if ((s) && (ugetc(s))) {
       i = ustrtol(s, NULL, 16);
-      if ((i == 0x7FFFFFFF) && (ustricmp(s, uconvert_ascii("7FFFFFFF", NULL)) != 0))
+      if ((i == 0x7FFFFFFF) && (ustricmp(s, uconvert_ascii("7FFFFFFF", tmp)) != 0))
 	 i = -1;
       return i;
    }
@@ -760,9 +843,9 @@ int get_config_hex(char *section, char *name, int def)
 /* get_config_float:
  *  Reads a float from the configuration file.
  */
-float get_config_float(char *section, char *name, float def)
+float get_config_float(AL_CONST char *section, AL_CONST char *name, float def)
 {
-   char *s = get_config_string(section, name, NULL);
+   AL_CONST char* s = get_config_string(section, name, NULL);
 
    if ((s) && (ugetc(s)))
       return uatof(s);
@@ -775,11 +858,11 @@ float get_config_float(char *section, char *name, float def)
 /* get_config_id:
  *  Reads a driver ID number from the configuration file.
  */
-int get_config_id(char *section, char *name, int def)
+int get_config_id(AL_CONST char *section, AL_CONST char *name, int def)
 {
-   char *s = get_config_string(section, name, NULL);
+   AL_CONST char *s = get_config_string(section, name, NULL);
    char tmp[4];
-   char *endp;
+   char* endp;
    int val, i;
 
    if ((s) && (ugetc(s))) {
@@ -807,22 +890,36 @@ int get_config_id(char *section, char *name, int def)
 /* get_config_argv:
  *  Reads an argc/argv style token list from the configuration file.
  */
-char **get_config_argv(char *section, char *name, int *argc)
+char **get_config_argv(AL_CONST char *section, AL_CONST char *name, int *argc)
 {
    #define MAX_ARGV  16
 
-   static char buf[256];
+   static char *buf = NULL;
+   static int buf_size = 0;
    static char *argv[MAX_ARGV];
    int pos, ac, q, c;
+   int s_size;
 
-   char *s = get_config_string(section, name, NULL);
+   AL_CONST char *s = get_config_string(section, name, NULL);
 
    if (!s) {
       *argc = 0;
       return NULL;
    }
 
-   ustrcpy(buf, s);
+   /* increase the buffer size if needed */
+   s_size = ustrsizez(s);
+   if (s_size>buf_size) {
+      buf_size = s_size;
+      buf = _al_sane_realloc(buf, buf_size);
+      if (!buf) {
+	 *allegro_errno = ENOMEM;
+	 *argc = 0;
+	 return NULL;
+      }
+   }
+
+   ustrzcpy(buf, buf_size, s);
    pos = 0;
    ac = 0;
 
@@ -867,26 +964,20 @@ char **get_config_argv(char *section, char *name, int *argc)
 /* insert_variable:
  *  Helper for inserting a new variable into a configuration file.
  */
-static CONFIG_ENTRY *insert_variable(CONFIG *the_config, CONFIG_ENTRY *p, char *name, char *data)
+static CONFIG_ENTRY *insert_variable(CONFIG *the_config, CONFIG_ENTRY *p, AL_CONST char *name, AL_CONST char *data)
 {
    CONFIG_ENTRY *n = malloc(sizeof(CONFIG_ENTRY));
 
    if (!n)
       return NULL;
 
-   if (name) {
-      n->name = malloc(ustrsizez(name));
-      if (n->name)
-	 ustrcpy(n->name, name);
-   }
+   if (name)
+      n->name = ustrdup(name);
    else
       n->name = NULL;
 
-   if (data) {
-      n->data = malloc(ustrsizez(data));
-      if (n->data)
-	 ustrcpy(n->data, data);
-   }
+   if (data)
+      n->data = ustrdup(data);
    else
       n->data = NULL;
 
@@ -907,7 +998,7 @@ static CONFIG_ENTRY *insert_variable(CONFIG *the_config, CONFIG_ENTRY *p, char *
 /* set_config_string:
  *  Writes a string to the configuration file.
  */
-void set_config_string(char *section, char *name, char *val)
+void set_config_string(AL_CONST char *section, AL_CONST char *name, AL_CONST char *val)
 {
    CONFIG *the_config;
    CONFIG_HOOK *hook;
@@ -916,7 +1007,7 @@ void set_config_string(char *section, char *name, char *val)
 
    init_config(TRUE);
 
-   prettify_section_name(section, section_name);
+   prettify_section_name(section, section_name, sizeof(section_name));
 
    /* check for hooked sections */
    hook = config_hook;
@@ -945,9 +1036,7 @@ void set_config_string(char *section, char *name, char *val)
 	    if (p->data)
 	       free(p->data);
 
-	    p->data = malloc(ustrsizez(val));
-	    if (p->data)
-	       ustrcpy(p->data, val);
+	    p->data = ustrdup(val);
 	 }
 	 else {
 	    /* delete variable */
@@ -1009,10 +1098,10 @@ void set_config_string(char *section, char *name, char *val)
 /* set_config_int:
  *  Writes an integer to the configuration file.
  */
-void set_config_int(char *section, char *name, int val)
+void set_config_int(AL_CONST char *section, AL_CONST char *name, int val)
 {
-   char buf[32];
-   usprintf(buf, uconvert_ascii("%d", NULL), val);
+   char buf[32], tmp[32];
+   uszprintf(buf, sizeof(buf), uconvert_ascii("%d", tmp), val);
    set_config_string(section, name, buf);
 }
 
@@ -1021,12 +1110,12 @@ void set_config_int(char *section, char *name, int val)
 /* set_config_hex:
  *  Writes a hexadecimal integer to the configuration file.
  */
-void set_config_hex(char *section, char *name, int val)
+void set_config_hex(AL_CONST char *section, AL_CONST char *name, int val)
 {
-   char buf[32];
+   char buf[32], tmp[32];
 
    if (val >= 0) {
-      usprintf(buf, uconvert_ascii("%X", NULL), val);
+      uszprintf(buf, sizeof(buf), uconvert_ascii("%X", tmp), val);
       set_config_string(section, name, buf);
    }
    else
@@ -1038,10 +1127,10 @@ void set_config_hex(char *section, char *name, int val)
 /* set_config_float:
  *  Writes a float to the configuration file.
  */
-void set_config_float(char *section, char *name, float val)
+void set_config_float(AL_CONST char *section, AL_CONST char *name, float val)
 {
-   char buf[32];
-   usprintf(buf, uconvert_ascii("%f", NULL), val);
+   char buf[32], tmp[32];
+   uszprintf(buf, sizeof(buf), uconvert_ascii("%f", tmp), val);
    set_config_string(section, name, buf);
 }
 
@@ -1050,15 +1139,15 @@ void set_config_float(char *section, char *name, float val)
 /* set_config_id:
  *  Writes a driver ID to the configuration file.
  */
-void set_config_id(char *section, char *name, int val)
+void set_config_id(AL_CONST char *section, AL_CONST char *name, int val)
 {
-   char buf[32];
+   char buf[32], tmp[32];
    int v[4];
    int pos = 0;
    int i;
 
    if (val < 256) {
-      usprintf(buf, uconvert_ascii("%d", NULL), val);
+      uszprintf(buf, sizeof(buf), uconvert_ascii("%d", tmp), val);
    }
    else {
       v[0] = (val>>24)&0xFF;
@@ -1077,34 +1166,46 @@ void set_config_id(char *section, char *name, int val)
 
 
 
-/* _load_config_text:
+/* reload_config_texts:
  *  Reads in a block of translated system text, looking for either a
  *  user-specified file, a ??text.cfg file, or a language.dat#??TEXT_CFG 
- *  datafile object.
+ *  datafile object. If new_language is not NULL, the language config
+ *  variable will be set to new_language before reloading the
+ *  configuration files.
  */
-void _load_config_text()
+void reload_config_texts(AL_CONST char *new_language)
 {
-   char filename[512], tmp1[128], tmp2[128], *name, *ext, *datafile;
+   char filename[1024], tmp1[128], tmp2[128];
+   AL_CONST char *name, *ext, *datafile;
+   char *namecpy;
 
    if (config_language) {
       destroy_config(config_language);
       config_language = NULL;
    }
 
-   name = get_config_string(NULL, uconvert_ascii("language", tmp1), NULL);
+   if (new_language)
+      set_config_string("system", "language", new_language);
+
+   name = get_config_string(uconvert_ascii("system", tmp1), uconvert_ascii("language", tmp2), NULL);
 
    if ((name) && (ugetc(name))) {
-      if ((ustrlen(name)<4) || (ustricmp(name+uoffset(name, -4), uconvert_ascii("text", tmp1)) != 0))
+      namecpy = ustrdup(name);
+      ustrlwr (namecpy);
+      if ((ustrlen(namecpy)<4) || (ustricmp(namecpy+uoffset(namecpy, -4), uconvert_ascii("text", tmp1)) != 0))
 	 ext = uconvert_ascii("text.cfg", tmp1);
       else
 	 ext = uconvert_ascii(".cfg", tmp1);
 
       datafile = uconvert_ascii("language.dat", tmp2);
 
-      if (find_allegro_resource(filename, name, ext, datafile, NULL, NULL, NULL, sizeof(filename)) == 0) {
+      if (find_allegro_resource(filename, namecpy, ext, datafile, NULL, NULL, NULL, sizeof(filename)) == 0) {
+	 free(namecpy);
 	 load_config_file(&config_language, filename, NULL);
 	 return;
       }
+
+      free(namecpy);
    }
 
    config_language = malloc(sizeof(CONFIG));
@@ -1122,22 +1223,52 @@ void _load_config_text()
  *  returning a suitable message in the current language if one is
  *  available, or a copy of the parameter if no translation can be found.
  */
-char *get_config_text(char *msg)
+AL_CONST char *get_config_text(AL_CONST char *msg)
 {
-   char tmp1[256], tmp2[256], name[256];
-   char *section = uconvert_ascii("[language]", tmp1);
-   char *umsg = uconvert_ascii(msg, tmp2);
-   char *s;
+   char tmp1[256];
+   AL_CONST char *section = uconvert_ascii("[language]", tmp1);
+   AL_CONST char *umsg;
+   AL_CONST char *s;
+   AL_CONST char *ret = NULL;
+   char *name;
    CONFIG_HOOK *hook;
    CONFIG_ENTRY *p;
-   int c, pos;
+   int c, pos, size;
+   ASSERT(msg);
 
    init_config(TRUE);
 
+   /* allocate memory and convert message to current encoding format */
+   if (need_uconvert(msg, U_ASCII, U_CURRENT)) {
+      size = uconvert_size(msg, U_ASCII, U_CURRENT);
+      umsg = malloc(size);
+      if (!umsg) {
+	 *allegro_errno = ENOMEM;
+	 return empty_string;
+      }
+
+      name = malloc(size);
+      if (!name) {
+	 free((char *)umsg);  /* remove constness */
+	 *allegro_errno = ENOMEM;
+	 return empty_string;
+      }
+
+      do_uconvert(msg, U_ASCII, (char*)umsg, U_CURRENT, size);
+   }
+   else {
+      umsg = msg;
+      name = malloc(ustrsizez(msg));
+      if (!name) {
+	 *allegro_errno = ENOMEM;
+	 return empty_string;
+      }
+   }
+
    s = umsg;
    pos = 0;
 
-   while ((c = ugetx(&s)) != 0) {
+   while ((c = ugetxc(&s)) != 0) {
       if ((uisspace(c)) || (c == '=') || (c == '#'))
 	 pos += usetc(name+pos, '_');
       else
@@ -1151,29 +1282,44 @@ char *get_config_text(char *msg)
 
    while (hook) {
       if (ustricmp(section, hook->section) == 0) {
-	 if (hook->stringgetter)
-	    return hook->stringgetter(name, umsg);
+	 if (hook->stringgetter) {
+	    ret = hook->stringgetter(name, umsg);
+	    break;
+	 }
       }
+
       hook = hook->next;
    }
 
-   /* find the string */
-   p = find_config_string(config_override, section, name, NULL);
+   if (!ret) {
+      /* find the string */
+      p = find_config_string(config_override, section, name, NULL);
 
-   if (!p) {
-      p = find_config_string(config[0], section, name, NULL);
+      if (!p) {
+	 p = find_config_string(config[0], section, name, NULL);
 
-      if (!p)
-	 p = find_config_string(config_language, section, name, NULL);
+	 if (!p)
+	    p = find_config_string(config_language, section, name, NULL);
+      }
+
+      if (p) {
+	 ret = (p->data ? p->data : empty_string);
+      }
+      else {
+	 /* no translation, so store off this value in the file */
+	 p = config_language->head;
+	 insert_variable(config_language, NULL, name, umsg);
+	 config_language->head->next = p;
+	 ret = config_language->head->data;
+      }
    }
 
-   if (p)
-      return (p->data ? p->data : empty_string);
+   /* free memory */
+   if (umsg!=msg)
+      free((char*) umsg);  /* remove constness */
 
-   /* no translation, so store off this value in the file */
-   p = config_language->head;
-   insert_variable(config_language, NULL, name, umsg);
-   config_language->head->next = p;
-   return config_language->head->data;
+   free(name);
+
+   return ret;
 }
 
