@@ -8,7 +8,9 @@
  *                                           /\____/
  *                                           \_/__/
  *
- *      Ensoniq Soundscape driver, by Andreas Kluge.
+ *      Ensoniq Soundscape driver.
+ *
+ *      By Andreas Kluge.
  *
  *      Based on code by Andrew P. Weir.
  *
@@ -20,8 +22,8 @@
 #include <string.h>
 
 #include "allegro.h"
-#include "allegro/aintern.h"
-#include "allegro/aintdos.h"
+#include "allegro/internal/aintern.h"
+#include "allegro/platform/aintdos.h"
 
 #ifndef ALLEGRO_DOS
    #error something is wrong with the makefile
@@ -107,7 +109,7 @@ static int soundscape_bufnum = 0;            /* the one currently in use */
 
 static void soundscape_lock_mem(void);
 
-static char soundscape_desc[160] = EMPTY_STRING;
+static char soundscape_desc[256] = EMPTY_STRING;
 
 static int cd_cfg_save;             /* gate array register save area */
 static int dma_cfg_save;            /* gate array register save area */
@@ -203,7 +205,7 @@ DIGI_DRIVER digi_soundscape =
 /* soundscape_buffer_size:
  *  Returns the current DMA buffer size, for use by the audiostream code.
  */
-static int soundscape_buffer_size()
+static int soundscape_buffer_size(void)
 {
    return soundscape_dma_size/4; /* convert bytes to stereo 16 bit samples */
 }
@@ -265,8 +267,8 @@ static int ga_read(int rnum)
  */
 static void set_dac_vol(int lvol, int rvol)
 {
-   cd_write(CD_DACL_REG, ~(lvol >> 1) & 0x3F);
-   cd_write(CD_DACR_REG, ~(rvol >> 1) & 0x3F);
+   cd_write(CD_DACL_REG, ~(lvol >> 2) & 0x3F);
+   cd_write(CD_DACR_REG, ~(rvol >> 2) & 0x3F);
 }
 
 
@@ -345,7 +347,7 @@ static int set_format(int srate, int stereo, int size16bit, int direction)
 /* stop_codec:
  *  This function will stop the CoDec auto-restart DMA process.
  */
-static void stop_codec()
+static void stop_codec(void)
 {
    int i;
 
@@ -366,14 +368,14 @@ static void stop_codec()
  *  if found, writes its associated right-hand value to a destination buffer.
  *  This function is case-insensitive.
  */
-static int get_ini_config_entry(char *entry, char *dest, FILE *fp)
+static int get_ini_config_entry(char *entry, char *dest, unsigned int dest_size, FILE *fp)
 {
    char str[83];
    char tokstr[33];
    char *p;
 
    /* make a local copy of the entry, upper-case it */
-   strcpy(tokstr, entry);
+   _al_sane_strncpy(tokstr, entry, sizeof(tokstr));
    strupr(tokstr);
 
    /* rewind the file and try to find it... */
@@ -409,7 +411,7 @@ static int get_ini_config_entry(char *entry, char *dest, FILE *fp)
 	 continue;
 
       /* it's our string - copy the right-hand value to buffer */
-      strcpy(dest, str+strlen(str)+1);
+      _al_sane_strncpy(dest, str+strlen(str)+1, dest_size);
       break;
    }
 
@@ -421,7 +423,7 @@ static int get_ini_config_entry(char *entry, char *dest, FILE *fp)
 /* get_init_config:
  *  This function gets all parameters from a file (SNDSCAPE.INI)
  */
-static int get_init_config()
+static int get_init_config(void)
 {
    FILE *fp = NULL;
    char str[78];
@@ -431,18 +433,18 @@ static int get_init_config()
    if (!(ep = getenv("SNDSCAPE")))
       return FALSE;
 
-   strcpy(str, ep);
+   _al_sane_strncpy(str, ep, sizeof(str));
 
    if (str[strlen(str)-1] == '\\')
       str[strlen(str)-1] = 0;
 
-   strcat(str, "\\SNDSCAPE.INI");
+   strncat(str, "\\SNDSCAPE.INI", sizeof(str)-1);
 
    if (!(fp = fopen(str, "r")))
       return FALSE;
 
    /* read all of the necessary config info ... */
-   if (get_ini_config_entry("Product", str, fp)) {
+   if (get_ini_config_entry("Product", str, sizeof(str), fp)) {
       fclose(fp);
       return FALSE;
    }
@@ -454,21 +456,21 @@ static int get_init_config()
    else
       soundscape_irqset = ss_irqs;
 
-   if (get_ini_config_entry("Port", str, fp)) {
+   if (get_ini_config_entry("Port", str, sizeof(str), fp)) {
       fclose(fp);
       return FALSE;
    }
 
    soundscape_baseport = strtol(str, NULL, 16);
 
-   if (get_ini_config_entry("WavePort", str, fp)) {
+   if (get_ini_config_entry("WavePort", str, sizeof(str), fp)) {
       fclose(fp);
       return FALSE;
    }
 
    soundscape_waveport = strtol(str, NULL, 16);
 
-   if (get_ini_config_entry("IRQ", str, fp)) {
+   if (get_ini_config_entry("IRQ", str, sizeof(str), fp)) {
       fclose(fp);
       return FALSE;
    }
@@ -478,7 +480,7 @@ static int get_init_config()
    if (soundscape_midiirq == 2)
       soundscape_midiirq = 9;
 
-   if (get_ini_config_entry("SBIRQ", str, fp)) {
+   if (get_ini_config_entry("SBIRQ", str, sizeof(str), fp)) {
       fclose(fp);
       return FALSE;
    }
@@ -488,7 +490,7 @@ static int get_init_config()
    if (soundscape_waveirq == 2)
       soundscape_waveirq = 9;
 
-   if (get_ini_config_entry("DMA", str, fp)) {
+   if (get_ini_config_entry("DMA", str, sizeof(str), fp)) {
       fclose(fp);
       return FALSE;
    }
@@ -511,7 +513,7 @@ static int get_init_config()
  *  present. If this function is not explicitly called by the application, it
  *  it will be called by the OpenSoundscape function.
  */
-static int detect_soundscape()
+static int detect_soundscape(void)
 {
    int tmp;
 
@@ -568,7 +570,7 @@ static int soundscape_mixer_volume(int volume)
 /* soundscape_interrupt:
  *  The Soundscape end-of-buffer interrupt handler.
  */
-static int soundscape_interrupt()
+static int soundscape_interrupt(void)
 {
    /* if the CoDec is interrupting ... */
    if (inportb(soundscape_waveport + CD_STATUS_OFF) & 0x01) {
@@ -617,7 +619,7 @@ static int soundscape_detect(int input)
       return FALSE;
 
    if (!detect_soundscape()) {
-      ustrcpy(allegro_error, get_config_text("Soundscape not found"));
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Soundscape not found"));
       return FALSE;
    } 
 
@@ -640,9 +642,10 @@ static int soundscape_detect(int input)
       soundscape_freq = 48000;
 
    /* set up the card description */
-   usprintf(soundscape_desc, get_config_text("Soundscape %s (%d hz) on port %X, using IRQ %d, DMA channel %d and waveport %X"),
-			     uconvert_ascii(ensoniq_gate_array[soundscape_hw_ver], tmp), 
-			     soundscape_freq, soundscape_baseport, soundscape_waveirq, soundscape_dma, soundscape_waveport);
+   uszprintf(soundscape_desc, sizeof(soundscape_desc),
+             get_config_text("Soundscape %s (%d hz) on port %X, using IRQ %d, DMA channel %d and waveport %X"),
+             uconvert_ascii(ensoniq_gate_array[soundscape_hw_ver], tmp), 
+	     soundscape_freq, soundscape_baseport, soundscape_waveirq, soundscape_dma, soundscape_waveport);
 
    digi_soundscape.desc = soundscape_desc;
 
@@ -823,7 +826,7 @@ static void soundscape_exit(int input)
  *  Locks all the memory touched by parts of the Soundscape code that are
  *  executed in an interrupt context.
  */
-static void soundscape_lock_mem()
+static void soundscape_lock_mem(void)
 {
    LOCK_VARIABLE(digi_soundscape);
    LOCK_VARIABLE(soundscape_freq);
