@@ -1,5 +1,5 @@
-/*         ______   ___    ___ 
- *        /\  _  \ /\_ \  /\_ \ 
+/*         ______   ___    ___
+ *        /\  _  \ /\_ \  /\_ \
  *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
  *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
  *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
@@ -13,6 +13,8 @@
  *
  *      By Shawn Hargreaves.
  *
+ *      Refresh rate control added by S. Sakamaki.
+ *
  *      See readme.txt for copyright information.
  */
 
@@ -20,8 +22,8 @@
 #include <string.h>
 
 #include "allegro.h"
-#include "allegro/aintern.h"
-#include "allegro/aintdos.h"
+#include "allegro/internal/aintern.h"
+#include "allegro/platform/aintdos.h"
 
 #ifndef ALLEGRO_DOS
    #error something is wrong with the makefile
@@ -36,9 +38,10 @@ static BITMAP *vesa_3_init(int w, int h, int v_w, int v_h, int color_depth);
 static void vesa_exit(BITMAP *b);
 static int vesa_scroll(int x, int y);
 static void vesa_vsync(void);
-static void vesa_set_palette_range(PALETTE p, int from, int to, int vsync);
+static void vesa_set_palette_range(AL_CONST PALETTE p, int from, int to, int vsync);
 static int vesa_request_scroll(int x, int y);
 static int vesa_poll_scroll(void);
+static GFX_MODE_LIST *vesa_fetch_mode_list(void);
 
 static char vesa_desc[256] = EMPTY_STRING;
 
@@ -61,7 +64,8 @@ GFX_DRIVER gfx_vesa_1 =
    NULL, NULL, NULL, NULL,       /* no hardware cursor */
    NULL,                         /* no drawing mode hook */
    NULL, NULL,                   /* no state saving */
-   0, 0, FALSE, 0, 0, 0, 0
+   vesa_fetch_mode_list,         /* aye! */
+   0, 0, FALSE, 0, 0, 0, 0, FALSE
 };
 
 
@@ -83,7 +87,8 @@ GFX_DRIVER gfx_vesa_2b =
    NULL, NULL, NULL, NULL,       /* no hardware cursor */
    NULL,                         /* no drawing mode hook */
    NULL, NULL,                   /* no state saving */
-   0, 0, FALSE, 0, 0, 0, 0
+   vesa_fetch_mode_list,         /* aye! */
+   0, 0, FALSE, 0, 0, 0, 0, FALSE
 };
 
 
@@ -105,7 +110,8 @@ GFX_DRIVER gfx_vesa_2l =
    NULL, NULL, NULL, NULL,       /* no hardware cursor */
    NULL,                         /* no drawing mode hook */
    NULL, NULL,                   /* no state saving */
-   0, 0, FALSE, 0, 0, 0, 0
+   vesa_fetch_mode_list,         /* aye! */
+   0, 0, FALSE, 0, 0, 0, 0, FALSE
 };
 
 
@@ -129,15 +135,16 @@ GFX_DRIVER gfx_vesa_3 =
    NULL, NULL, NULL, NULL,       /* no hardware cursor */
    NULL,                         /* no drawing mode hook */
    NULL, NULL,                   /* no state saving */
-   0, 0, FALSE, 0, 0, 0, 0
+   vesa_fetch_mode_list,         /* aye! */
+   0, 0, FALSE, 0, 0, 0, 0, FALSE
 };
 
 
 
-#define MASK_LINEAR(addr)     (addr & 0x000FFFFF)
-#define RM_TO_LINEAR(addr)    (((addr & 0xFFFF0000) >> 12) + (addr & 0xFFFF))
-#define RM_OFFSET(addr)       (addr & 0xF)
-#define RM_SEGMENT(addr)      ((addr >> 4) & 0xFFFF)
+#define MASK_LINEAR(addr)     ((addr) & 0x000FFFFF)
+#define RM_TO_LINEAR(addr)    ((((addr) & 0xFFFF0000) >> 12) + ((addr) & 0xFFFF))
+#define RM_OFFSET(addr)       ((addr) & 0xF)
+#define RM_SEGMENT(addr)      (((addr) >> 4) & 0xFFFF)
 
 
 
@@ -237,6 +244,30 @@ typedef struct PM_INFO           /* VESA 2.0 protected mode interface */
 
 
 
+typedef struct CRTCInfoBlock     /* VESA 3.0 CRTC timings structure */
+{
+    unsigned short HorizontalTotal     __PACKED__;
+    unsigned short HorizontalSyncStart __PACKED__;
+    unsigned short HorizontalSyncEnd   __PACKED__;
+    unsigned short VerticalTotal       __PACKED__;
+    unsigned short VerticalSyncStart   __PACKED__;
+    unsigned short VerticalSyncEnd     __PACKED__;
+    unsigned char  Flags               __PACKED__;
+    unsigned long  PixelClock          __PACKED__;    /* units of Hz */
+    unsigned short RefreshRate         __PACKED__;    /* units of 0.01 Hz */
+    unsigned char  reserved[40]        __PACKED__;
+} CRTCInfoBlock;
+
+
+#define HPOS         0
+#define HNEG         (1 << 2)
+#define VPOS         0
+#define VNEG         (1 << 3)
+#define INTERLACED   (1 << 1)
+#define DOUBLESCAN   (1 << 0)
+
+
+
 static VESA_INFO vesa_info;               /* SVGA info block */
 static MODE_INFO mode_info;               /* info for this video mode */
 static char oem_string[256];              /* vendor name */
@@ -269,7 +300,7 @@ static int evilness_flag = 0;             /* set if we are doing dodgy things
 /* get_vesa_info:
  *  Retrieves a VESA info block structure, returning 0 for success.
  */
-static int get_vesa_info()
+static int get_vesa_info(void)
 {
    unsigned long addr;
    int c;
@@ -371,12 +402,12 @@ static int find_vesa_mode(int w, int h, int color_depth, int vbe_version)
    long mode_ptr;
 
    if (get_vesa_info() != 0) {
-      ustrcpy(allegro_error, get_config_text("VESA not available"));
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("VESA not available"));
       return 0;
    }
 
    if (vesa_info.VESAVersion < (vbe_version<<8)) {
-      usprintf(allegro_error, get_config_text("VBE %d.0 not available"), vbe_version);
+      uszprintf(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("VBE %d.0 not available"), vbe_version);
       return 0;
    }
 
@@ -455,7 +486,7 @@ static int find_vesa_mode(int w, int h, int color_depth, int vbe_version)
       #endif
 
       default:
-	 ustrcpy(allegro_error, get_config_text("Resolution not supported"));
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Resolution not supported"));
 	 return 0;
    }
 
@@ -519,14 +550,14 @@ static int find_vesa_mode(int w, int h, int color_depth, int vbe_version)
    else if ((w == 1280) && (h == 1024) && (color_depth == 8))
       c = 0x107;
    else {
-      ustrcpy(allegro_error, get_config_text("Resolution not supported"));
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Resolution not supported"));
       return 0; 
    }
 
    if (get_mode_info(c) == 0)
       return c;
 
-   ustrcpy(allegro_error, get_config_text("Resolution not supported"));
+   ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Resolution not supported"));
    return 0;
 }
 
@@ -537,31 +568,31 @@ static int find_vesa_mode(int w, int h, int color_depth, int vbe_version)
  */
 static void setup_vesa_desc(GFX_DRIVER *driver, int vbe_version, int linear)
 {
-   char tmp1[64], tmp2[64];
+   char tmp1[64], tmp2[1600];
 
    /* VESA version number */
-   usprintf(vesa_desc, uconvert_ascii("%4.4s %d.%d (", tmp1), 
-		       uconvert_ascii(vesa_info.VESASignature, tmp2),
-		       vesa_info.VESAVersion>>8, 
-		       vesa_info.VESAVersion&0xFF);
+   uszprintf(vesa_desc, sizeof(vesa_desc), uconvert_ascii("%4.4s %d.%d (", tmp1), 
+		                           uconvert_ascii(vesa_info.VESASignature, tmp2),
+		                           vesa_info.VESAVersion>>8, 
+		                           vesa_info.VESAVersion&0xFF);
 
    /* VESA description string */
-   ustrcat(vesa_desc, uconvert_ascii(oem_string, NULL));
-   ustrcat(vesa_desc, uconvert_ascii(")", NULL));
+   ustrzcat(vesa_desc, sizeof(vesa_desc), uconvert_ascii(oem_string, tmp2));
+   ustrzcat(vesa_desc, sizeof(vesa_desc), uconvert_ascii(")", tmp1));
 
    /* warn about dodgy hardware */
    if (evilness_flag & 1)
-      ustrcat(vesa_desc, uconvert_ascii(", Trio64 bodge", NULL));
+      ustrzcat(vesa_desc, sizeof(vesa_desc), uconvert_ascii(", Trio64 bodge", tmp1));
 
    if (evilness_flag & 2)
-      ustrcat(vesa_desc, uconvert_ascii(", 0x4F06 N.A.", NULL));
+      ustrzcat(vesa_desc, sizeof(vesa_desc), uconvert_ascii(", 0x4F06 N.A.", tmp1));
 
    /* banked/linear status for VBE 3.0 */
    if (vbe_version >= 3) {
       if (linear)
-	 ustrcat(vesa_desc, uconvert_ascii(", linear", NULL));
+	 ustrzcat(vesa_desc, sizeof(vesa_desc), uconvert_ascii(", linear", tmp1));
       else
-	 ustrcat(vesa_desc, uconvert_ascii(", banked", NULL));
+	 ustrzcat(vesa_desc, sizeof(vesa_desc), uconvert_ascii(", banked", tmp1));
    }
 
    driver->desc = vesa_desc;
@@ -569,6 +600,79 @@ static void setup_vesa_desc(GFX_DRIVER *driver, int vbe_version, int linear)
 
 
 
+/* vesa_fetch_mode_list:
+ *  Generates a list of valid video modes for the VESA drivers.
+ *  Returns the mode list on success or NULL on failure.
+ */
+static GFX_MODE_LIST *vesa_fetch_mode_list(void)
+{
+   GFX_MODE_LIST  *mode_list;
+   unsigned long  mode_ptr;
+   unsigned short *vesa_mode;
+   int            mode, pos, vesa_list_length;
+
+   mode_list = malloc(sizeof(GFX_MODE_LIST));
+   if (!mode_list) return NULL;
+
+   /* fetch list of VESA modes */
+   if (get_vesa_info()) return NULL;
+
+   _farsetsel(_dos_ds);
+
+   /* count number of VESA modes */
+   mode_ptr = RM_TO_LINEAR(vesa_info.VideoModePtr);
+   for (mode_list->num_modes = 0; _farnspeekw(mode_ptr) != 0xFFFF; mode_list->num_modes++)
+      mode_ptr += sizeof(unsigned short);
+
+   /* allocate and fill in temporary vesa mode list */
+   vesa_mode = malloc(sizeof(unsigned short) * mode_list->num_modes);
+   if (!vesa_mode) return NULL;
+
+   mode_ptr = RM_TO_LINEAR(vesa_info.VideoModePtr);
+   for (mode = 0; _farnspeekw(mode_ptr) != 0xFFFF; mode++) {
+      vesa_mode[mode] = _farnspeekw(mode_ptr);
+      mode_ptr += sizeof(unsigned short);
+   }
+
+   vesa_list_length = mode_list->num_modes;
+
+   /* zero out text and <8 bpp modes for later exclusion */
+   for (mode = 0; mode < mode_list->num_modes; mode++) {
+      if (get_mode_info(vesa_mode[mode])) return NULL;
+      if ((mode_info.MemoryModel == 0) || (mode_info.BitsPerPixel < 8)) {
+         vesa_mode[mode] = 0;
+         mode_list->num_modes--;
+      }
+   }
+
+   /* allocate mode list */
+   mode_list->mode = malloc(sizeof(GFX_MODE) * (mode_list->num_modes + 1));
+   if (!mode_list->mode) return NULL;
+
+   /* fill in width, height and color depth for each VESA mode */
+   mode = 0;
+   for (pos = 0; pos < vesa_list_length; pos++) {
+      if (get_mode_info(vesa_mode[pos])) return NULL;
+      if (!vesa_mode[pos]) continue;
+      mode_list->mode[mode].width  = mode_info.XResolution;
+      mode_list->mode[mode].height = mode_info.YResolution;
+      mode_list->mode[mode].bpp    = mode_info.BitsPerPixel;
+      mode++;
+   }
+
+   /* terminate list */
+   mode_list->mode[mode_list->num_modes].width  = 0;
+   mode_list->mode[mode_list->num_modes].height = 0;
+   mode_list->mode[mode_list->num_modes].bpp    = 0;
+
+   /* free up temporary vesa mode list */
+   free(vesa_mode);
+
+   return mode_list;
+}
+
+
+
 /* get_pmode_functions:
  *  Attempts to use VBE 2.0 functions to get access to protected mode bank
  *  switching, set display start, and palette routines, storing the bank
@@ -702,6 +806,110 @@ static BITMAP *make_linear_bitmap(GFX_DRIVER *driver, int width, int height, int
 
 
 
+/* calc_crtc_timing:
+ *  Calculates CRTC mode timings.
+ */
+static void calc_crtc_timing(CRTCInfoBlock *crtc, int xres, int yres, int xadjust, int yadjust)
+{
+   int HTotal, VTotal;
+   int HDisp, VDisp;
+   int HSS, VSS;
+   int HSE, VSE;
+   int HSWidth, VSWidth;
+   int SS, SE;
+   int doublescan = FALSE;
+
+   if (yres < 400){
+      doublescan = TRUE;
+      yres *= 2;
+   }
+
+   HDisp = xres;
+   HTotal = (int)(HDisp * 1.27) & ~0x7;
+   HSWidth = (int)((HTotal - HDisp) / 5) & ~0x7;
+   HSS = HDisp + 16;
+   HSE = HSS + HSWidth;
+   VDisp = yres;
+   VTotal = VDisp * 1.07;
+   VSWidth = (VTotal / 100) + 1;
+   VSS = VDisp + ((int)(VTotal - VDisp) / 5) + 1;
+   VSE = VSS + VSWidth;
+
+   SS = HSS + xadjust;
+   SE = HSE + xadjust;
+
+   if (xadjust < 0) {
+      if (SS < (HDisp + 8)) {
+	 SS = HDisp + 8;
+	 SE = SS + HSWidth;
+      }
+   }
+   else {
+      if ((HTotal - 24) < SE) {
+	 SE = HTotal - 24;
+	 SS = SE - HSWidth;
+      }
+   }
+
+   HSS = SS;
+   HSE = SE;
+
+   SS = VSS + yadjust;
+   SE = VSE + yadjust;
+
+   if (yadjust < 0) {
+      if (SS < (VDisp + 3)) {
+	 SS = VDisp + 3;
+	 SE = SS + VSWidth;
+      }
+   }
+   else {
+      if ((VTotal - 4) < SE) {
+	 SE = VTotal - 4;
+	 SS = SE - VSWidth;
+      }
+   }
+
+   VSS = SS;
+   VSE = SE;
+
+   crtc->HorizontalTotal     = HTotal;
+   crtc->HorizontalSyncStart = HSS;
+   crtc->HorizontalSyncEnd   = HSE;
+   crtc->VerticalTotal       = VTotal;
+   crtc->VerticalSyncStart   = VSS;
+   crtc->VerticalSyncEnd     = VSE;
+   crtc->Flags               = HNEG | VNEG;
+
+   if (doublescan)
+      crtc->Flags |= DOUBLESCAN;
+}
+
+
+
+/* get_closest_pixel_clock:
+ *  Uses VESA 3.0 function 0x4F0B to find the closest pixel clock to the
+ *  requested value.
+ */
+static unsigned long get_closest_pixel_clock(int mode_no, unsigned long vclk)
+{
+   __dpmi_regs r;
+
+   r.x.ax = 0x4F0B;
+   r.h.bl = 0;
+   r.d.ecx = vclk;
+   r.x.dx = mode_no;
+
+   __dpmi_int(0x10, &r);
+
+   if (r.h.ah != 0)
+      return 0;
+
+   return r.d.ecx;
+}
+
+
+
 /* vesa_init:
  *  Tries to enter the specified graphics mode, and makes a screen bitmap
  *  for it. Will use a linear framebuffer if one is available.
@@ -744,7 +952,7 @@ static BITMAP *vesa_init(GFX_DRIVER *driver, int linear, int vbe_version, int w,
 	 driver->bank_size = driver->bank_gran = 0;
       }
       else {
-	 ustrcpy(allegro_error, get_config_text("Linear framebuffer not available"));
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Linear framebuffer not available"));
 	 return NULL;
       }
    }
@@ -756,13 +964,13 @@ static BITMAP *vesa_init(GFX_DRIVER *driver, int linear, int vbe_version, int w,
 	 driver->bank_gran = mode_info.WinGranularity * 1024;
       }
       else {
-	 ustrcpy(allegro_error, get_config_text("Banked framebuffer not available"));
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Banked framebuffer not available"));
 	 return NULL;
       }
    }
 
    if (MAX(w, v_w) * MAX(h, v_h) * bpp > driver->vid_mem) {
-      ustrcpy(allegro_error, get_config_text("Insufficient video memory"));
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Insufficient video memory"));
       return NULL;
    }
 
@@ -782,10 +990,37 @@ static BITMAP *vesa_init(GFX_DRIVER *driver, int linear, int vbe_version, int w,
    if (driver->linear)
       _dpmi_reg.x.bx |= 0x4000;
 
+   if ((_refresh_rate_request > 0) && (vbe_version >= 3)) {
+      /* VESA 3.0 stuff for controlling the refresh rate */
+      CRTCInfoBlock crtc;
+      unsigned long vclk;
+      double f0;
+
+      calc_crtc_timing(&crtc, w, h, 0, 0);
+
+      vclk = (double)crtc.HorizontalTotal * crtc.VerticalTotal * _refresh_rate_request;
+      vclk = get_closest_pixel_clock(vesa_mode, vclk);
+
+      if (vclk != 0) {
+	 f0 = (double)vclk / (crtc.HorizontalTotal * crtc.VerticalTotal);
+
+	 _set_current_refresh_rate((int)(f0 + 0.5));
+
+	 crtc.PixelClock  = vclk;
+	 crtc.RefreshRate = _refresh_rate_request * 100;
+
+	 dosmemput(&crtc, sizeof(CRTCInfoBlock), MASK_LINEAR(__tb));
+
+	 _dpmi_reg.x.di = RM_OFFSET(__tb);
+	 _dpmi_reg.x.es = RM_SEGMENT(__tb);
+	 _dpmi_reg.x.bx |= 0x0800;
+      }
+   }
+
    __dpmi_int(0x10, &_dpmi_reg);
 
    if (_dpmi_reg.h.ah) {
-      ustrcpy(allegro_error, get_config_text("VESA function 0x4F02 failed"));
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("VESA function 0x4F02 failed"));
       return NULL;
    }
 
@@ -814,7 +1049,7 @@ static BITMAP *vesa_init(GFX_DRIVER *driver, int linear, int vbe_version, int w,
 	  * work on 99% of cards, if VESA lets me down.
 	  */
 	 if ((width != 1024) && (width != 2048)) {
-	    ustrcpy(allegro_error, get_config_text("VESA function 0x4F06 failed"));
+	    ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("VESA function 0x4F06 failed"));
 	    return NULL;
 	 }
 
@@ -827,7 +1062,7 @@ static BITMAP *vesa_init(GFX_DRIVER *driver, int linear, int vbe_version, int w,
 	 __dpmi_int(0x10, &_dpmi_reg);
 
 	 if (_dpmi_reg.h.ah) {
-	    ustrcpy(allegro_error, get_config_text("VESA function 0x4F02 failed"));
+	    ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("VESA function 0x4F02 failed"));
 	    return NULL;
 	 }
 
@@ -847,13 +1082,13 @@ static BITMAP *vesa_init(GFX_DRIVER *driver, int linear, int vbe_version, int w,
    }
 
    if ((width/bpp < v_w) || (width/bpp < w) || (height < v_h) || (height < h)) {
-      ustrcpy(allegro_error, get_config_text("Virtual screen size too large"));
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Virtual screen size too large"));
       return NULL;
    }
 
    if (vbe_version >= 2) {
       if (get_pmode_functions(&w1, &w2) != 0) {
-	 ustrcpy(allegro_error, get_config_text("VESA protected mode interface not available"));
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("VESA protected mode interface not available"));
 	 return NULL;
       }
    }
@@ -861,7 +1096,7 @@ static BITMAP *vesa_init(GFX_DRIVER *driver, int linear, int vbe_version, int w,
    if (driver->linear) {                        /* make linear bitmap? */
       b = make_linear_bitmap(driver, width/bpp, height, color_depth, width);
       if (!b) {
-	 ustrcpy(allegro_error, get_config_text("Can't make linear screen bitmap"));
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Can't make linear screen bitmap"));
 	 return NULL;
       }
    }
@@ -999,7 +1234,7 @@ static BITMAP *vesa_3_init(int w, int h, int v_w, int v_h, int color_depth)
  *  blanking registers. VBE doesn't provide a vsync function, but we 
  *  can emulate it by changing the display start with the vsync flag set.
  */
-static void vesa_vsync()
+static void vesa_vsync(void)
 {
    vesa_scroll(vesa_xscroll, vesa_yscroll);
 }
@@ -1091,7 +1326,7 @@ static int vesa_scroll(int x, int y)
 /* vesa_set_palette_range:
  *  Uses VESA function #9 (VBE 2.0 only) to set the palette.
  */
-static void vesa_set_palette_range(PALETTE p, int from, int to, int vsync)
+static void vesa_set_palette_range(AL_CONST PALETTE p, int from, int to, int vsync)
 {
    PALETTE tmp;
    int c, mode, seg;
@@ -1184,7 +1419,7 @@ static int vesa_request_scroll(int x, int y)
 /* vesa_poll_scroll:
  *  VBE 3.0 triple buffering test routine.
  */
-static int vesa_poll_scroll()
+static int vesa_poll_scroll(void)
 {
    _dpmi_reg.x.ax = 0x4F07;
    _dpmi_reg.x.bx = 4;
