@@ -16,12 +16,15 @@
  *
  *      Bresenham arc routine by Romano Signorelli.
  *
+ *      Cohen-Sutherland line clipping by Jon Rafkind.
+ *
  *      See readme.txt for copyright information.
  */
 
 
+#include <math.h>
 #include "allegro.h"
-#include "allegro/aintern.h"
+#include "allegro/internal/aintern.h"
 
 
 
@@ -133,9 +136,9 @@ void set_blender_mode_ex(BLENDER_FUNC b15, BLENDER_FUNC b16, BLENDER_FUNC b24, B
 /* xor_mode:
  *  Shortcut function for toggling XOR mode on and off.
  */
-void xor_mode(int xor)
+void xor_mode(int on)
 {
-   drawing_mode(xor ? DRAW_MODE_XOR : DRAW_MODE_SOLID, NULL, 0, 0);
+   drawing_mode(on ? DRAW_MODE_XOR : DRAW_MODE_SOLID, NULL, 0, 0);
 }
 
 
@@ -150,10 +153,10 @@ void solid_mode()
 
 
 
-/* clear:
+/* clear_bitmap:
  *  Clears the bitmap to color 0.
  */
-void clear(BITMAP *bitmap)
+void clear_bitmap(BITMAP *bitmap)
 {
    clear_to_color(bitmap, 0);
 }
@@ -176,9 +179,9 @@ void vsync()
 /* set_color:
  *  Sets a single palette entry.
  */
-void set_color(int index, RGB *p)
+void set_color(int index, AL_CONST RGB *p)
 {
-   set_palette_range(p-index, index, index, FALSE);
+   set_palette_range((struct RGB *)p-index, index, index, FALSE);
 }
 
 
@@ -186,7 +189,7 @@ void set_color(int index, RGB *p)
 /* set_palette:
  *  Sets the entire color palette.
  */
-void set_palette(PALETTE p)
+void set_palette(AL_CONST PALETTE p)
 {
    set_palette_range(p, 0, PAL_SIZE-1, TRUE);
 }
@@ -196,7 +199,7 @@ void set_palette(PALETTE p)
 /* set_palette_range:
  *  Sets a part of the color palette.
  */
-void set_palette_range(PALETTE p, int from, int to, int vsync)
+void set_palette_range(AL_CONST PALETTE p, int from, int to, int vsync)
 {
    int c;
 
@@ -232,24 +235,24 @@ static int prev_palette_color[256];
  *  The previous settings are copied onto a one-deep stack, from where they
  *  can be restored by calling unselect_palette().
  */
-void select_palette(PALETTE p)
+void select_palette(AL_CONST PALETTE p)
 {
    int c;
 
    for (c=0; c<256; c++) {
       _prev_current_palette[c] = _current_palette[c];
-      prev_palette_color[c] = palette_color[c];
+      _current_palette[c] = p[c];
+   }
+
+   if (_color_depth != 8) {
+      for (c=0; c<256; c++) {
+	 prev_palette_color[c] = palette_color[c];
+	 palette_color[c] = makecol(_rgb_scale_6[p[c].r], _rgb_scale_6[p[c].g], _rgb_scale_6[p[c].b]);
+      }
    }
 
    _got_prev_current_palette = TRUE;
 
-   for (c=0; c<256; c++) {
-      _current_palette[c] = p[c];
-
-      if (_color_depth != 8)
-	 palette_color[c] = makecol(_rgb_scale_6[p[c].r], _rgb_scale_6[p[c].g], _rgb_scale_6[p[c].b]);
-   }
-
    _current_palette_changed = 0xFFFFFFFF & ~(1<<(_color_depth-1));
 }
 
@@ -262,10 +265,11 @@ void unselect_palette()
 {
    int c;
 
-   for (c=0; c<256; c++) {
+   for (c=0; c<256; c++)
       _current_palette[c] = _prev_current_palette[c];
 
-      if (_color_depth != 8)
+   if (_color_depth != 8) {
+      for (c=0; c<256; c++)
 	 palette_color[c] = prev_palette_color[c];
    }
 
@@ -285,10 +289,10 @@ static int *palette_expansion_table(int bpp)
    int c;
 
    switch (bpp) {
-      case 15: table = palette_color15; break;
-      case 16: table = palette_color16; break;
-      case 24: table = palette_color24; break;
-      case 32: table = palette_color32; break;
+      case 15: table = _palette_color15; break;
+      case 16: table = _palette_color16; break;
+      case 24: table = _palette_color24; break;
+      case 32: table = _palette_color32; break;
       default: ASSERT(FALSE); return NULL;
    }
 
@@ -378,7 +382,7 @@ void get_palette_range(PALETTE p, int from, int to)
  *  be: 0 = return source, 64 = return dest, 32 = return exactly half way.
  *  Only affects colors between from and to (inclusive).
  */
-void fade_interpolate(PALETTE source, PALETTE dest, PALETTE output, int pos, int from, int to)
+void fade_interpolate(AL_CONST PALETTE source, AL_CONST PALETTE dest, PALETTE output, int pos, int from, int to)
 {
    int c;
 
@@ -396,7 +400,7 @@ void fade_interpolate(PALETTE source, PALETTE dest, PALETTE output, int pos, int
  *  is instantaneous). Only affects colors between from and to (inclusive,
  *  pass 0 and 255 to fade the entire palette).
  */
-void fade_from_range(PALETTE source, PALETTE dest, int speed, int from, int to)
+void fade_from_range(AL_CONST PALETTE source, AL_CONST PALETTE dest, int speed, int from, int to)
 {
    PALETTE temp;
    int c, start, last;
@@ -434,7 +438,7 @@ void fade_from_range(PALETTE source, PALETTE dest, int speed, int from, int to)
  *  the slowest, 64 is instantaneous). Only affects colors between from and 
  *  to (inclusive, pass 0 and 255 to fade the entire palette).
  */
-void fade_in_range(PALETTE p, int speed, int from, int to)
+void fade_in_range(AL_CONST PALETTE p, int speed, int from, int to)
 {
    fade_from_range(black_palette, p, speed, from, to);
 }
@@ -461,7 +465,7 @@ void fade_out_range(int speed, int from, int to)
  *  Fades from source to dest, at the specified speed (1 is the slowest, 64
  *  is instantaneous).
  */
-void fade_from(PALETTE source, PALETTE dest, int speed)
+void fade_from(AL_CONST PALETTE source, AL_CONST PALETTE dest, int speed)
 {
    fade_from_range(source, dest, speed, 0, PAL_SIZE-1);
 }
@@ -472,7 +476,7 @@ void fade_from(PALETTE source, PALETTE dest, int speed)
  *  Fades from a solid black palette to p, at the specified speed (1 is
  *  the slowest, 64 is instantaneous).
  */
-void fade_in(PALETTE p, int speed)
+void fade_in(AL_CONST PALETTE p, int speed)
 {
    fade_in_range(p, speed, 0, PAL_SIZE-1);
 }
@@ -529,7 +533,7 @@ void rect(BITMAP *bmp, int x1, int y1, int x2, int y2, int color)
 
 
 /* _normal_rectfill:
- *  Draws a solid filled rectangle, using hline() to do the work.
+ *  Draws a solid filled rectangle, using hfill() to do the work.
  */
 void _normal_rectfill(BITMAP *bmp, int x1, int y1, int x2, int y2, int color)
 {
@@ -575,7 +579,7 @@ void _normal_rectfill(BITMAP *bmp, int x1, int y1, int x2, int y2, int color)
    acquire_bitmap(bmp);
 
    while (y1 <= y2) {
-      hline(bmp, x1, y1, x2, color);
+      bmp->vtable->hfill(bmp, x1, y1, x2, color);
       y1++;
    };
 
@@ -619,13 +623,13 @@ void do_line(BITMAP *bmp, int x1, int y1, int x2, int y2, int d, void (*proc)(BI
 	 proc(bmp, x, y, d);                                                 \
 									     \
 	 if (dd sec_cond 0) {                                                \
-	    sec_c sec_sign##= 1;                                             \
+	    sec_c = sec_c sec_sign 1;                                        \
 	    dd += i2;                                                        \
 	 }                                                                   \
 	 else                                                                \
 	    dd += i1;                                                        \
 									     \
-	 pri_c pri_sign##= 1;                                                \
+	 pri_c = pri_c pri_sign 1;                                           \
       }                                                                      \
    }
 
@@ -673,6 +677,8 @@ void do_line(BITMAP *bmp, int x1, int y1, int x2, int y2, int d, void (*proc)(BI
 	 }
       }
    }
+
+   #undef DO_LINE
 }
 
 
@@ -735,6 +741,137 @@ void _normal_line(BITMAP *bmp, int x1, int y1, int x2, int y2, int color)
 
 
 
+/* _fast_line:
+ *  Draws a line from x1, y1 to x2, y2, using putpixel() to do the work.
+ *  This is an implementation of the Cohen-Sutherland line clipping algorithm.
+ *  Loops over the line until it can be either trivially rejected or trivially
+ *  accepted. If it is neither rejected nor accepted, subdivide it into two
+ *  segments, one of which can be rejected.
+ */
+void _fast_line(BITMAP *bmp, int x1, int y1, int x2, int y2, int color)
+{
+   int code0, code1;
+   int outcode;
+   int x, y;
+   int xmax, xmin, ymax, ymin;
+   int done = 0, accept = 0;
+   int clip_orig;
+   ASSERT(bmp);
+
+   if ((clip_orig = bmp->clip) != 0) {  /* save clipping state */
+      #define TOP     0x8
+      #define BOTTOM  0x4
+      #define LEFT    0x2
+      #define RIGHT   0x1
+
+      #define COMPCLIP(code, x, y)  \
+      {                             \
+	 code = 0;                  \
+	 if (y < ymin)              \
+	    code |= TOP;            \
+	 else if (y > ymax)         \
+	    code |= BOTTOM;         \
+	 if (x < xmin)              \
+	    code |= LEFT;           \
+	 else if (x > xmax)         \
+	    code |= RIGHT;          \
+      }
+
+      xmin = bmp->cl;
+      xmax = bmp->cr-1;
+      ymin = bmp->ct;
+      ymax = bmp->cb-1;
+
+      COMPCLIP(code0, x1, y1);
+      COMPCLIP(code1, x2, y2);
+
+      do {
+
+	 if (!(code0 | code1)) {
+	    /* Trivially accept. */
+	    accept = done = 1;
+	 }
+	 else if (code0 & code1) {
+	    /* Trivially reject. */
+	    done = 1;
+	 }
+	 else {
+	    /* Didn't reject or accept, so do some calculations. */
+	    outcode = code0 ? code0 : code1;  /* pick one endpoint */
+
+	    if (outcode & TOP) {
+	       if (y2 == y1)
+		  x = x1;
+	       else
+		  x = x1 + (x2 - x1) * (ymin - y1) / (y2 - y1);
+	       y = ymin;
+	    }
+	    else if (outcode & BOTTOM) {
+	       if (y2 == y1)
+		  x = x1;
+	       else
+		  x = x1 + (x2 - x1) * (ymax - y1) / (y2 - y1);
+	       y = ymax;
+	    }
+	    else if (outcode & LEFT) {
+	       if (x2 == x1)
+		  y = y1;
+	       else
+		  y = y1 + (y2 - y1) * (xmin - x1) / (x2 - x1);
+	       x = xmin;
+	    }
+	    else {  /* outcode & RIGHT */
+	       if (x2 == x1)
+		  y = y1;
+	       else
+		  y = y1 + (y2 - y1) * (xmax - x1) / (x2 - x1);
+	       x = xmax;
+	    }
+
+	    if (outcode == code0) {
+	       x1 = x;
+	       y1 = y;
+	       COMPCLIP(code0, x1, y1);
+	    }
+	    else {
+	       x2 = x;
+	       y2 = y;
+	       COMPCLIP(code1, x2, y2);
+	    }
+	 }
+      } while (!done);
+
+      #undef COMPCLIP
+      #undef TOP
+      #undef BOTTOM
+      #undef LEFT
+      #undef RIGHT
+
+      if (!accept)
+	 return;
+
+      /* We have already done the clipping, no need to do it again. */
+      bmp->clip = FALSE;
+   }
+
+   if (x1 == x2) {
+      bmp->vtable->vline(bmp, x1, y1, y2, color);
+   }
+   else if (y1 == y2) {
+      bmp->vtable->hline(bmp, x1, y1, x2, color);
+   }
+   else {
+      acquire_bitmap(bmp);
+      do_line(bmp, x1, y1, x2, y2, color, bmp->vtable->putpixel);
+      release_bitmap(bmp);
+   }
+
+   /* Restore original clipping state. */
+   bmp->clip = clip_orig;
+}
+
+
+
 /* do_circle:
  *  Helper function for the circle drawing routines. Calculates the points
  *  in a circle of radius r around point x, y, and calls the specified 
@@ -800,6 +937,7 @@ void do_circle(BITMAP *bmp, int x, int y, int radius, int d, void (*proc)(BITMAP
 void circle(BITMAP *bmp, int x, int y, int radius, int color)
 {
    int clip, sx, sy, dx, dy;
+   ASSERT(bmp);
 
    if (bmp->clip) {
       sx = x-radius-1;
@@ -840,6 +978,7 @@ void circlefill(BITMAP *bmp, int x, int y, int radius, int color)
    int d_e = 3;
    int d_se = -2 * radius + 5;
    int clip, sx, sy, dx, dy;
+   ASSERT(bmp);
 
    if (bmp->clip) {
       sx = x-radius-1;
@@ -861,10 +1000,10 @@ void circlefill(BITMAP *bmp, int x, int y, int radius, int color)
    acquire_bitmap(bmp);
 
    do {
-      hline(bmp, x-cy, y-cx, x+cy, color);
+      bmp->vtable->hfill(bmp, x-cy, y-cx, x+cy, color);
 
       if (cx)
-	 hline(bmp, x-cy, y+cx, x+cy, color);
+	 bmp->vtable->hfill(bmp, x-cy, y+cx, x+cy, color);
 
       if (df < 0)  {
 	 df += d_e;
@@ -873,10 +1012,10 @@ void circlefill(BITMAP *bmp, int x, int y, int radius, int color)
       }
       else { 
 	 if (cx != cy) {
-	    hline(bmp, x-cx, y-cy, x+cx, color);
+	    bmp->vtable->hfill(bmp, x-cx, y-cy, x+cx, color);
 
 	    if (cy)
-	       hline(bmp, x-cx, y+cy, x+cx, color);
+	       bmp->vtable->hfill(bmp, x-cx, y+cy, x+cx, color);
 	 }
 
 	 df += d_se;
@@ -1011,6 +1150,7 @@ void do_ellipse(BITMAP *bmp, int x, int y, int rx, int ry, int d, void (*proc)(B
 void ellipse(BITMAP *bmp, int x, int y, int rx, int ry, int color)
 {
    int clip, sx, sy, dx, dy;
+   ASSERT(bmp);
 
    if (bmp->clip) {
       sx = x-rx-1;
@@ -1050,6 +1190,7 @@ void ellipsefill(BITMAP *bmp, int x, int y, int rx, int ry, int color)
    int da, db, dc, dd;
    int na, nb, nc, nd;
    int clip, sx, sy, dx, dy;
+   ASSERT(bmp);
 
    if (bmp->clip) {
       sx = x-rx-1;
@@ -1100,15 +1241,15 @@ void ellipsefill(BITMAP *bmp, int x, int y, int rx, int ry, int color)
 	 nd = (nb * ry) / rx;
 
 	 if ((c > dc) && (c < dd)) {
-	    hline(bmp, x-b, y+c, x+b, color);
+	    bmp->vtable->hfill(bmp, x-b, y+c, x+b, color);
 	    if (c)
-	       hline(bmp, x-b, y-c, x+b, color);
+	       bmp->vtable->hfill(bmp, x-b, y-c, x+b, color);
 	    dc = c;
 	 }
 
 	 if ((d < dd) && (d > dc)) { 
-	    hline(bmp, x-a, y+d, x+a, color);
-	    hline(bmp, x-a, y-d, x+a, color);
+	    bmp->vtable->hfill(bmp, x-a, y+d, x+a, color);
+	    bmp->vtable->hfill(bmp, x-a, y-d, x+a, color);
 	    dd = d;
 	 }
 
@@ -1138,15 +1279,15 @@ void ellipsefill(BITMAP *bmp, int x, int y, int rx, int ry, int color)
 	 nd = (nb * rx) / ry;
 
 	 if ((a > da) && (a < db)) {
-	    hline(bmp, x-d, y+a, x+d, color); 
+	    bmp->vtable->hfill(bmp, x-d, y+a, x+d, color); 
 	    if (a)
-	       hline(bmp, x-d, y-a, x+d, color);
+	       bmp->vtable->hfill(bmp, x-d, y-a, x+d, color);
 	    da = a;
 	 }
 
 	 if ((b < db) && (b > da)) { 
-	    hline(bmp, x-c, y+b, x+c, color);
-	    hline(bmp, x-c, y-b, x+c, color);
+	    bmp->vtable->hfill(bmp, x-c, y+b, x+c, color);
+	    bmp->vtable->hfill(bmp, x-c, y-b, x+c, color);
 	    db = b;
 	 }
 
@@ -1160,6 +1301,24 @@ void ellipsefill(BITMAP *bmp, int x, int y, int rx, int ry, int color)
 
 
 
+/* get_point_on_arc:
+ *  Helper function for the do_arc() function, converting from (radius, angle)
+ *  to (x, y).
+ */
+static INLINE void get_point_on_arc(int r, fixed a, int *out_x, int *out_y)
+{
+   double s, c;
+   double double_a = a * (AL_PI * 2 / (1 << 24));
+   s = sin(double_a);
+   c = cos(double_a);
+   s = -s * r;
+   c = c * r;
+   *out_x = (int)((c < 0) ? (c - 0.5) : (c + 0.5));
+   *out_y = (int)((s < 0) ? (s - 0.5) : (s + 0.5));
+}
+
+
+
 /* do_arc:
  *  Helper function for the arc function. Calculates the points in an arc
  *  of radius r around point x, y, going anticlockwise from fixed point
@@ -1170,243 +1329,180 @@ void ellipsefill(BITMAP *bmp, int x, int y, int rx, int ry, int color)
  */
 void do_arc(BITMAP *bmp, int x, int y, fixed ang1, fixed ang2, int r, int d, void (*proc)(BITMAP *, int, int, int))
 {
-   unsigned long rr = r*r;
-   unsigned long rr1, rr2, rr3;
+   /* start position */
+   int sx, sy;
+   /* current position */
    int px, py;
+   /* end position */
    int ex, ey;
-   int px1, px2, px3;
-   int py1, py2, py3;
-   int d1, d2, d3;
-   int ax, ay;
-   int q, qe;
-   long tg_cur, tg_end;
-   int done = FALSE;
+   /* square of radius of circle */
+   long rr;
+   /* difference between main radius squared and radius squared of three
+      potential next points */
+   long rr1, rr2, rr3;
+   /* square of x and of y */
+   unsigned long xx, yy, xx_new, yy_new;
+   /* start quadrant, current quadrant and end quadrant */
+   int sq, q, qe;
+   /* direction of movement */
+   int dx, dy;
+   /* temporary variable for determining if we have reached end point */
+   int det;
 
-   rr1 = r;
-   rr2 = itofix(x);
-   rr3 = itofix(y);
+   /* Calculate the start point and the end point. */
+   /* We have to flip y because bitmaps count y coordinates downwards. */
+   get_point_on_arc(r, ang1, &sx, &sy);
+   px = sx;
+   py = sy;
+   get_point_on_arc(r, ang2, &ex, &ey);
 
-   /* evaluate the start point and the end point */
-   px = fixtoi(rr2 + rr1 * fcos(ang1));
-   py = fixtoi(rr3 - rr1 * fsin(ang1));
-   ex = fixtoi(rr2 + rr1 * fcos(ang2));
-   ey = fixtoi(rr3 - rr1 * fsin(ang2));
+   rr = r*r;
+   xx = px*px;
+   yy = py*py - rr;
 
-   /* start quadrant */
-   if (px >= x) {
-      if (py <= y)
-	 q = 1;                           /* quadrant 1 */
-      else
-	 q = 4;                           /* quadrant 4 */
-   }
-   else {
-      if (py < y)
-	 q = 2;                           /* quadrant 2 */
+   /* Find start quadrant. */
+   if (px >= 0) {
+      if (py <= 0)
+	 q = 0;                           /* quadrant 0 */
       else
 	 q = 3;                           /* quadrant 3 */
    }
-
-   /* end quadrant */
-   if (ex >= x) {
-      if (ey <= y)
-	 qe = 1;                          /* quadrant 1 */
-      else
-	 qe = 4;                          /* quadrant 4 */
-   }
    else {
-      if (ey < y)
-	 qe = 2;                          /* quadrant 2 */
+      if (py < 0)
+	 q = 1;                           /* quadrant 1 */
+      else
+	 q = 2;                           /* quadrant 2 */
+   }
+   sq = q;
+
+   /* Find end quadrant. */
+   if (ex >= 0) {
+      if (ey <= 0)
+	 qe = 0;                          /* quadrant 0 */
       else
 	 qe = 3;                          /* quadrant 3 */
    }
-
-   #define loc_tg(_y, _x)  (_x-x) ? itofix(_y-y)/(_x-x) : itofix(_y-y)
-
-   tg_end = loc_tg(ey, ex);
-
-   while (!done) {
-      proc(bmp, px, py, d);
-
-      /* from here, we have only 3 possible direction of movement, eg.
-       * for the first quadrant:
-       *
-       *    OOOOOOOOO
-       *    OOOOOOOOO
-       *    OOOOOO21O
-       *    OOOOOO3*O
-       */
-
-      /* evaluate the 3 possible points */
-      switch (q) {
-
-	 case 1:
-	    px1 = px;
-	    py1 = py-1;
-	    px2 = px-1;
-	    py2 = py-1;
-	    px3 = px-1;
-	    py3 = py;
-
-	    /* 2nd quadrant check */
-	    if (px != x) {
-	       break;
-	    }
-	    else {
-	       /* we were in the end quadrant, changing is illegal. Exit. */
-	       if (qe == q)
-		  done = TRUE;
-	       q++;
-	    }
-	    /* fall through */
-
-	 case 2:
-	    px1 = px-1;
-	    py1 = py;
-	    px2 = px-1;
-	    py2 = py+1;
-	    px3 = px;
-	    py3 = py+1;
-
-	    /* 3rd quadrant check */
-	    if (py != y) {
-	       break;
-	    }
-	    else {
-	       /* we were in the end quadrant, changing is illegal. Exit. */
-	       if (qe == q)
-		  done = TRUE;
-	       q++;
-	    }
-	    /* fall through */
-
-	 case 3:
-	    px1 = px;
-	    py1 = py+1;
-	    px2 = px+1;
-	    py2 = py+1;
-	    px3 = px+1;
-	    py3 = py;
-
-	    /* 4th quadrant check */
-	    if (px != x) {
-	       break;
-	    }
-	    else {
-	       /* we were in the end quadrant, changing is illegal. Exit. */
-	       if (qe == q)
-		  done = TRUE;
-	       q++;
-	    }
-	    /* fall through */
-
-	 case 4:
-	    px1 = px+1;
-	    py1 = py;
-	    px2 = px+1;
-	    py2 = py-1;
-	    px3 = px;
-	    py3 = py-1;
-
-	    /* 1st quadrant check */
-	    if (py == y) {
-	       /* we were in the end quadrant, changing is illegal. Exit. */
-	       if (qe == q)
-		  done = TRUE;
-
-	       q = 1;
-	       px1 = px;
-	       py1 = py-1;
-	       px2 = px-1;
-	       py2 = py-1;
-	       px3 = px-1;
-	       py3 = py;
-	    }
-	    break;
-
-	 default:
-	    return;
-      }
-
-      /* now, we must decide which of the 3 points is the right point.
-       * We evaluate the distance from center and, then, choose the
-       * nearest point.
-       */
-      ax = x-px1;
-      ay = y-py1;
-      rr1 = ax*ax + ay*ay;
-
-      ax = x-px2;
-      ay = y-py2;
-      rr2 = ax*ax + ay*ay;
-
-      ax = x-px3;
-      ay = y-py3;
-      rr3 = ax*ax + ay*ay;
-
-      /* difference from the main radius */
-      if (rr1 > rr)
-	 d1 = rr1-rr;
+   else {
+      if (ey < 0)
+	 qe = 1;                          /* quadrant 1 */
       else
-	 d1 = rr-rr1;
-      if (rr2 > rr)
-	 d2 = rr2-rr;
-      else
-	 d2 = rr-rr2;
-      if (rr3 > rr)
-	 d3 = rr3-rr;
-      else
-	 d3 = rr-rr3;
-
-      /* what is the minimum? */
-      if (d1 <= d2) {
-	 px = px1;
-	 py = py1;
-      }
-      else if (d2 <= d3) {
-	 px = px2;
-	 py = py2;
-      }
-      else {
-	 px = px3;
-	 py = py3;
-      }
-
-      /* are we in the final quadrant? */
-      if (qe == q) {
-	 tg_cur = loc_tg(py, px);
-
-	 /* is the arc finished? */
-	 switch (q) {
-
-	    case 1:
-	       /* end quadrant = 1? */
-	       if (tg_cur <= tg_end)
-		  done = TRUE;
-	       break;
-
-	    case 2:
-	       /* end quadrant = 2? */
-	       if (tg_cur <= tg_end)
-		  done = TRUE;
-	       break;
-
-	    case 3:
-	       /* end quadrant = 3? */
-	       if (tg_cur <= tg_end)
-		  done = TRUE;
-	       break;
-
-	    case 4:
-	       /* end quadrant = 4? */
-	       if (tg_cur <= tg_end)
-		  done = TRUE;
-	       break;
-	 }
-      }
+	 qe = 2;                          /* quadrant 2 */
    }
 
-   /* draw the last evaluated point */
-   proc(bmp, px, py, d);
+   if (q > qe) {
+      /* qe must come after q. */
+      qe += 4;
+   }
+   else if (q == qe) {
+      /* If q==qe but the beginning comes after the end, make qe be
+       * strictly after q.
+       */
+      if (((ang2&0xffffff) < (ang1&0xffffff)) ||
+	  (((ang1&0xffffff) < 0x400000) && ((ang2&0xffffff) >= 0xc00000)))
+         qe += 4;
+   }
+
+   /* initial direction of movement */
+   if (((q+1)&2) == 0)
+      dy = -1;
+   else
+      dy = 1;
+   if ((q&2) == 0)
+      dx = -1;
+   else
+      dx = 1;
+
+   while (TRUE) {
+      /* Change quadrant when needed.
+       * dx and dy determine the possible directions to go in this
+       * quadrant, so they must be updated when we change quadrant.
+       */
+      if ((q&1) == 0) {
+         if (px == 0) {
+            if (qe == q)
+	       break;
+	    q++;
+	    dy = -dy;
+	 }
+      }
+      else {
+         if (py == 0) {
+	    if (qe == q)
+	       break;
+	    q++;
+	    dx = -dx;
+	 }
+      }
+
+      /* Are we in the final quadrant? */
+      if (qe == q) {
+	 /* Have we reached (or passed) the end point both in x and y? */
+	 det = 0;
+
+	 if (dy > 0) {
+	    if (py >= ey)
+	       det++;
+	 }
+	 else {
+	    if (py <= ey)
+	       det++;
+	 }
+	 if (dx > 0) {
+	    if (px >= ex)
+	       det++;
+	 }
+	 else {
+	    if (px <= ex)
+	       det++;
+	 }
+	 
+	 if (det == 2)
+	    break;
+      }
+
+      proc(bmp, x+px, y+py, d);
+
+      /* From here, we have only 3 possible directions of movement, eg.
+       * for the first quadrant:
+       *
+       *    .........
+       *    .........
+       *    ......21.
+       *    ......3*.
+       *
+       * These are reached by adding dx to px and/or adding dy to py.
+       * We need to find which of these points gives the best
+       * approximation of the (square of the) radius.
+       */
+
+      xx_new = (px+dx) * (px+dx);
+      yy_new = (py+dy) * (py+dy) - rr;
+      rr1 = xx_new + yy;
+      rr2 = xx_new + yy_new;
+      rr3 = xx     + yy_new;
+
+      /* Set rr1, rr2, rr3 to be the difference from the main radius of the
+       * three points.
+       */
+      if (rr1 < 0)
+	 rr1 = -rr1;
+      if (rr2 < 0)
+	 rr2 = -rr2;
+      if (rr3 < 0)
+	 rr3 = -rr3;
+
+      if (rr3 >= MIN(rr1, rr2)) {
+         px += dx;
+	 xx = xx_new;
+      }
+      if (rr1 > MIN(rr2, rr3)) {
+         py += dy;
+	 yy = yy_new;
+      }
+   }
+   /* Only draw last point if it doesn't overlap with first one. */
+   if ((px != sx) || (py != sy) || (sq == qe))
+      proc(bmp, x+px, y+py, d);
 }
 
 
@@ -1416,6 +1512,7 @@ void do_arc(BITMAP *bmp, int x, int y, fixed ang1, fixed ang2, int r, int d, voi
  */
 void arc(BITMAP *bmp, int x, int y, fixed ang1, fixed ang2, int r, int color)
 {
+   ASSERT(bmp);
    acquire_bitmap(bmp);
 
    do_arc(bmp, x, y, ang1, ang2, r, color, bmp->vtable->putpixel);
