@@ -19,7 +19,7 @@
 #include <string.h>
 
 #include "allegro.h"
-#include "allegro/aintern.h"
+#include "allegro/internal/aintern.h"
 
 
 
@@ -31,23 +31,30 @@ int _gfx_mode_set_count = 0;           /* has the graphics mode changed? */
 
 int _screen_split_position = 0;        /* has the screen been split? */
 
+int _safe_gfx_mode_change = 0;         /* are we getting through GFX_SAFE? */
+
 RGB_MAP *rgb_map = NULL;               /* RGB -> palette entry conversion */
 
 COLOR_MAP *color_map = NULL;           /* translucency/lighting table */
 
 int _color_depth = 8;                  /* how many bits per pixel? */
 
+int _refresh_rate_request = 0;         /* requested refresh rate */
+static int current_refresh_rate = 0;   /* refresh rate set by the driver */
+
+int _wait_for_vsync = TRUE;            /* vsync when page-flipping? */
+
 int _color_conv = COLORCONV_TOTAL;     /* which formats to auto convert? */
 
 static int color_conv_set = FALSE;     /* has the user set conversion mode? */
 
-int palette_color8[256];               /* palette -> pixel mapping */
-int palette_color15[256];
-int palette_color16[256];
-int palette_color24[256];
-int palette_color32[256];
+int _palette_color8[256];               /* palette -> pixel mapping */
+int _palette_color15[256];
+int _palette_color16[256];
+int _palette_color24[256];
+int _palette_color32[256];
 
-int *palette_color = palette_color8; 
+int *palette_color = _palette_color8; 
 
 BLENDER_FUNC _blender_func15 = NULL;   /* truecolor pixel blender routines */
 BLENDER_FUNC _blender_func16 = NULL;
@@ -65,19 +72,19 @@ int _blender_col_32 = 0;
 
 int _blender_alpha = 0;                /* for truecolor translucent drawing */
 
-int _rgb_r_shift_15 = 0;               /* truecolor pixel format */
-int _rgb_g_shift_15 = 5;
-int _rgb_b_shift_15 = 10;
-int _rgb_r_shift_16 = 0;
-int _rgb_g_shift_16 = 5;
-int _rgb_b_shift_16 = 11;
-int _rgb_r_shift_24 = 0;
-int _rgb_g_shift_24 = 8;
-int _rgb_b_shift_24 = 16;
-int _rgb_r_shift_32 = 0;
-int _rgb_g_shift_32 = 8;
-int _rgb_b_shift_32 = 16;
-int _rgb_a_shift_32 = 24;
+int _rgb_r_shift_15 = DEFAULT_RGB_R_SHIFT_15;     /* truecolor pixel format */
+int _rgb_g_shift_15 = DEFAULT_RGB_G_SHIFT_15;
+int _rgb_b_shift_15 = DEFAULT_RGB_B_SHIFT_15;
+int _rgb_r_shift_16 = DEFAULT_RGB_R_SHIFT_16;
+int _rgb_g_shift_16 = DEFAULT_RGB_G_SHIFT_16;
+int _rgb_b_shift_16 = DEFAULT_RGB_B_SHIFT_16;
+int _rgb_r_shift_24 = DEFAULT_RGB_R_SHIFT_24;
+int _rgb_g_shift_24 = DEFAULT_RGB_G_SHIFT_24;
+int _rgb_b_shift_24 = DEFAULT_RGB_B_SHIFT_24;
+int _rgb_r_shift_32 = DEFAULT_RGB_R_SHIFT_32;
+int _rgb_g_shift_32 = DEFAULT_RGB_G_SHIFT_32;
+int _rgb_b_shift_32 = DEFAULT_RGB_B_SHIFT_32;
+int _rgb_a_shift_32 = DEFAULT_RGB_A_SHIFT_32;
 
 
 /* lookup table for scaling 5 bit colors up to 8 bits */
@@ -111,13 +118,20 @@ typedef struct VRAM_BITMAP             /* list of video memory bitmaps */
 {
    int x, y, w, h;
    BITMAP *bmp;
-   struct VRAM_BITMAP *next;
+   struct VRAM_BITMAP *next_x, *next_y;
 } VRAM_BITMAP;
 
 
 static VRAM_BITMAP *vram_bitmap_list = NULL;
 
 
+#define BMP_MAX_SIZE  46340   /* sqrt(INT_MAX) */
+
+/* the product of these must fit in an int */
+static int failed_bitmap_w = BMP_MAX_SIZE;
+static int failed_bitmap_h = BMP_MAX_SIZE;
+
+
 
 /* lock_bitmap:
  *  Locks all the memory used by a bitmap structure.
@@ -133,26 +147,171 @@ void lock_bitmap(BITMAP *bmp)
 
 
 
+/* request_refresh_rate:
+ *  Requests that the next call to set_gfx_mode() use the specified refresh
+ *  rate.
+ */
+void request_refresh_rate(int rate)
+{
+   _refresh_rate_request = rate;
+}
+
+
+
+/* get_refresh_rate:
+ *  Returns the refresh rate set by the most recent call to set_gfx_mode().
+ */
+int get_refresh_rate(void)
+{
+   return current_refresh_rate;
+}
+
+
+
+/* _set_current_refresh_rate:
+ *  Sets the current refresh rate.
+ *  (This function must be called by the gfx drivers)
+ */
+void _set_current_refresh_rate(int rate)
+{
+   /* sanity check to discard bogus values */
+   if ((rate<40) || (rate>200))
+      rate = 0;
+
+   current_refresh_rate = rate;
+
+   /* adjust retrace speed */
+   _vsync_speed = rate ? BPS_TO_TIMER(rate) : BPS_TO_TIMER(70);
+}
+
+
+
+/* sort_gfx_mode_list:
+ *  Callback for quick-sorting a mode-list.
+ */
+static int sort_gfx_mode_list(GFX_MODE *entry1, GFX_MODE *entry2)
+{
+   if (entry1->width > entry2->width) {
+      return +1;
+   }
+   else if (entry1->width < entry2->width) {
+      return -1;
+   }
+   else {
+      if (entry1->height > entry2->height) {
+         return +1;
+      }
+      else if (entry1->height < entry2->height) {
+         return -1;
+      }
+      else {
+         if (entry1->bpp > entry2->bpp) {
+            return +1;
+         }
+         else if (entry1->bpp < entry2->bpp) {
+            return -1;
+         }
+         else {
+            return 0;
+         }
+      }
+   }
+}
+
+
+
+/* get_gfx_mode_list:
+ *  Attempts to create a list of all the supported video modes for a certain
+ *  GFX driver.
+ */
+GFX_MODE_LIST *get_gfx_mode_list(int card)
+{
+   _DRIVER_INFO *list_entry;
+   GFX_DRIVER *drv = NULL;
+   GFX_MODE_LIST *gfx_mode_list = NULL;
+
+   ASSERT(system_driver);
+
+   /* ask the system driver for a list of graphics hardware drivers */
+   if (system_driver->gfx_drivers)
+      list_entry = system_driver->gfx_drivers();
+   else
+      list_entry = _gfx_driver_list;
+
+   /* find the graphics driver, and if it can fetch mode lists, do so */
+   while (list_entry->driver) {
+      if (list_entry->id == card) {
+         drv = list_entry->driver;
+         if (!drv->fetch_mode_list)
+            return NULL;
+
+         gfx_mode_list = drv->fetch_mode_list();
+         if (!gfx_mode_list)
+            return NULL;
+
+         break;
+      }
+
+      list_entry++;
+   }
+
+   if (!drv)
+      return NULL;
+
+   /* sort the list and finish */
+   qsort(gfx_mode_list->mode, gfx_mode_list->num_modes, sizeof(GFX_MODE),
+         (int (*) (AL_CONST void *, AL_CONST void *))sort_gfx_mode_list);
+
+   return gfx_mode_list;
+}
+
+
+
+/* destroy_gfx_mode_list:
+ *  Removes the mode list created by get_gfx_mode_list() from memory.
+ */
+void destroy_gfx_mode_list(GFX_MODE_LIST *gfx_mode_list)
+{
+   if (gfx_mode_list) {
+      if (gfx_mode_list->mode)
+         free(gfx_mode_list->mode);
+
+      free(gfx_mode_list);
+   }
+}
+
+
+
 /* set_color_depth:
  *  Sets the pixel size (in bits) which will be used by subsequent calls to 
- *  set_gfx_mode() and create_bitmap(). Valid depths are 8, 15, 16, and 32.
+ *  set_gfx_mode() and create_bitmap(). Valid depths are 8, 15, 16, 24 and 32.
  */
 void set_color_depth(int depth)
 {
    _color_depth = depth;
 
    switch (depth) {
-      case 8:  palette_color = palette_color8;  break;
-      case 15: palette_color = palette_color15; break;
-      case 16: palette_color = palette_color16; break;
-      case 24: palette_color = palette_color24; break;
-      case 32: palette_color = palette_color32; break;
+      case 8:  palette_color = _palette_color8;  break;
+      case 15: palette_color = _palette_color15; break;
+      case 16: palette_color = _palette_color16; break;
+      case 24: palette_color = _palette_color24; break;
+      case 32: palette_color = _palette_color32; break;
       default: ASSERT(FALSE);
    }
 }
 
 
 
+/* get_color_depth:
+ *  Returns the current color depth.
+ */
+int get_color_depth(void)
+{
+   return _color_depth;
+}
+
+
+
 /* set_color_conversion:
  *  Sets a bit mask specifying which types of color format conversions are
  *  valid when loading data from disk.
@@ -166,6 +325,17 @@ void set_color_conversion(int mode)
 
 
 
+/* get_color_conversion:
+ *  Returns the bitmask specifying which types of color format
+ *  conversion are valid when loading data from disk.
+ */
+int get_color_conversion(void)
+{
+   return _color_conv;
+}
+
+
+
 /* _color_load_depth:
  *  Works out which color depth an image should be loaded as, given the
  *  current conversion mode.
@@ -284,6 +454,123 @@ static void shutdown_gfx(void)
 }
 
 
+#define GFX_DRIVER_FULLSCREEN_FLAG    0x01
+#define GFX_DRIVER_WINDOWED_FLAG      0x02
+
+
+/* gfx_driver_is_valid:
+ *  Checks that the graphics driver 'drv' fulfills the condition
+ *  expressed by the bitmask 'flags'.
+ */
+static int gfx_driver_is_valid(GFX_DRIVER *drv, int flags)
+{
+   if ((flags & GFX_DRIVER_FULLSCREEN_FLAG) && drv->windowed)
+      return FALSE;
+
+   if ((flags & GFX_DRIVER_WINDOWED_FLAG) && !drv->windowed)
+      return FALSE;
+
+   return TRUE;
+}
+
+
+
+/* get_gfx_driver_from_id:
+ *  Retrieves a pointer to the graphics driver identified by 'card' from
+ *  the list 'driver_list' or NULL if it doesn't exist.
+ */
+static GFX_DRIVER *get_gfx_driver_from_id(int card, _DRIVER_INFO *driver_list)
+{
+   int c;
+
+   for (c=0; driver_list[c].driver; c++) {
+      if (driver_list[c].id == card)
+	 return driver_list[c].driver;
+   }
+
+   return NULL;
+}
+
+
+
+/* init_gfx_driver:
+ *  Helper function for initializing a graphics driver.
+ */
+static BITMAP *init_gfx_driver(GFX_DRIVER *drv, int w, int h, int v_w, int v_h)
+{
+   drv->name = drv->desc = get_config_text(drv->ascii_name);
+
+   /* set gfx_driver so that it is visible when initializing the driver */
+   gfx_driver = drv;
+
+   return drv->init(w, h, v_w, v_h, _color_depth);
+}
+
+
+
+/* get_config_gfx_driver:
+ *  Helper function for set_gfx_mode: it reads the gfx_card* config variables and
+ *  tries to set the graphics mode if a matching driver is found. Returns TRUE if
+ *  at least one matching driver was found, FALSE otherwise.
+ */
+static int get_config_gfx_driver(char *gfx_card, int w, int h, int v_w, int v_h, int flags, _DRIVER_INFO *driver_list)
+{
+   char buf[512], tmp[64];
+   GFX_DRIVER *drv;
+   int found = FALSE;
+   int card, n;
+
+   /* try the drivers that are listed in the config file */
+   for (n=-2; n<255; n++) {
+      switch (n) {
+
+	 case -2:
+	    /* example: gfx_card_640x480x16 = */
+	    uszprintf(buf, sizeof(buf), uconvert_ascii("%s_%dx%dx%d", tmp), gfx_card, w, h, _color_depth);
+	    break;
+
+	 case -1:
+	    /* example: gfx_card_24bpp = */
+	    uszprintf(buf, sizeof(buf), uconvert_ascii("%s_%dbpp", tmp), gfx_card, _color_depth);
+	    break;
+
+	 case 0:
+	    /* example: gfx_card = */
+	    uszprintf(buf, sizeof(buf), uconvert_ascii("%s", tmp), gfx_card);
+	    break;
+
+	 default:
+	    /* example: gfx_card1 = */
+	    uszprintf(buf, sizeof(buf), uconvert_ascii("%s%d", tmp), gfx_card, n);
+	    break;
+      }
+
+      card = get_config_id(uconvert_ascii("graphics", tmp), buf, 0);
+
+      if (card) {
+	 drv = get_gfx_driver_from_id(card, driver_list);
+
+	 if (drv && gfx_driver_is_valid(drv, flags)) {
+	    found = TRUE;
+	    screen = init_gfx_driver(drv, w, h, v_w, v_h);
+
+	    if (screen)
+	       break;
+	 }
+      }
+      else {
+	 /* Stop searching the gfx_card#n (n>0) family at the first missing member,
+	  * except gfx_card1 which could have been identified with gfx_card.
+	  */
+	 if (n > 1)
+	    break;
+      }
+   }
+
+   return found;
+}
+
+
 
 /* set_gfx_mode:
  *  Sets the graphics mode. The card should be one of the GFX_* constants
@@ -299,34 +586,57 @@ static void shutdown_gfx(void)
 int set_gfx_mode(int card, int w, int h, int v_w, int v_h)
 {
    static int allow_config = TRUE;
-   extern void blit_end();
+   extern void blit_end(void);
    _DRIVER_INFO *driver_list;
-   int tried = FALSE;
-   char buf[512], tmp[64];
-   int c, n;
+   GFX_DRIVER *drv;
+   struct GFX_MODE mode;
+   char buf[ALLEGRO_ERROR_SIZE], tmp1[64], tmp2[64];
+   AL_CONST char *dv;
+   int flags = 0;
+   int c, driver, ret;
+   ASSERT(system_driver);
 
    _gfx_mode_set_count++;
 
    /* special bodge for the GFX_SAFE driver */
    if (card == GFX_SAFE) {
-      ustrncpy(buf, allegro_error, sizeof(buf)-ucwidth(0));
-      if (set_gfx_mode(GFX_AUTODETECT, w, h, 0, 0) != 0) {
-	 set_color_depth(8);
+      if (system_driver->get_gfx_safe_mode) {
+	 ustrzcpy(buf, sizeof(buf), allegro_error);
+
+	 /* retrieve the safe graphics mode */
+	 system_driver->get_gfx_safe_mode(&driver, &mode);
+
+	 /* try using the specified depth and resolution */
+	 if (set_gfx_mode(driver, w, h, 0, 0) == 0)
+	    return 0;
+
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, buf);
+
+	 /* finally use the safe settings */
+	 set_color_depth(mode.bpp);
+	 if (set_gfx_mode(driver, mode.width, mode.height, 0, 0) == 0)
+	    return 0;
+
+	 ASSERT(FALSE);  /* the safe graphics mode must always work */
+      }
+      else {
+	 /* no safe graphics mode, try hard-coded autodetected modes with custom settings */
 	 allow_config = FALSE;
-      #ifdef GFX_SAFE_ID
-	 if (set_gfx_mode(GFX_SAFE_ID, GFX_SAFE_W, GFX_SAFE_H, 0, 0) != 0) {
-      #else
-	 if (set_gfx_mode(GFX_AUTODETECT, 0, 0, 0, 0) != 0) {
-      #endif
-	    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
-	    allegro_message("%s\n", get_config_text("Fatal error: unable to set GFX_SAFE"));
-	    exit(1);
-	 }
+	 _safe_gfx_mode_change = 1;
+
+	 ret = set_gfx_mode(GFX_AUTODETECT, w, h, 0, 0);
+
+	 _safe_gfx_mode_change = 0;
 	 allow_config = TRUE;
+
+	 if (ret == 0)
+	    return 0;
       }
 
-      ustrcpy(allegro_error, buf);
-      return 0;
+      /* failing to set GFX_SAFE is a fatal error */
+      set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
+      allegro_message(uconvert_ascii("%s\n", tmp1), get_config_text("Fatal error: unable to set GFX_SAFE"));
+      return -1;
    }
 
    /* remember the current console state */
@@ -342,6 +652,7 @@ int set_gfx_mode(int card, int w, int h, int v_w, int v_h)
       gfx_virgin = FALSE;
    }
 
+   /* lock the application in the foreground */
    if (system_driver->display_switch_lock)
       system_driver->display_switch_lock(TRUE, TRUE);
 
@@ -350,7 +661,8 @@ int set_gfx_mode(int card, int w, int h, int v_w, int v_h)
 
    /* close down any existing graphics driver */
    if (gfx_driver) {
-      show_mouse(NULL);
+      if (_al_linker_mouse)
+         _al_linker_mouse->show_mouse(NULL);
 
       while (vram_bitmap_list)
 	 destroy_bitmap(vram_bitmap_list->bmp);
@@ -372,6 +684,12 @@ int set_gfx_mode(int card, int w, int h, int v_w, int v_h)
       gfx_capabilities = 0;
    }
 
+   /* We probably don't want to do this because it makes
+    * Allegro "forget" the color layout of previously set
+    * graphics modes. But it should be retained if bitmaps
+    * created in those modes are to be used in the new mode.
+    */
+#if 0
    /* restore default truecolor pixel format */
    _rgb_r_shift_15 = 0; 
    _rgb_g_shift_15 = 5;
@@ -386,9 +704,12 @@ int set_gfx_mode(int card, int w, int h, int v_w, int v_h)
    _rgb_g_shift_32 = 8;
    _rgb_b_shift_32 = 16;
    _rgb_a_shift_32 = 24;
+#endif
 
    gfx_capabilities = 0;
 
+   _set_current_refresh_rate(0);
+
    /* return to text mode? */
    if (card == GFX_TEXT) {
       if (system_driver->restore_console_state)
@@ -405,98 +726,70 @@ int set_gfx_mode(int card, int w, int h, int v_w, int v_h)
       return 0;
    }
 
+   /* now to the interesting part: let's try to find a graphics driver */
+   usetc(allegro_error, 0);
+
    /* ask the system driver for a list of graphics hardware drivers */
    if (system_driver->gfx_drivers)
       driver_list = system_driver->gfx_drivers();
    else
       driver_list = _gfx_driver_list;
 
-   usetc(allegro_error, 0);
-
-   if ((card == GFX_AUTODETECT) && (allow_config)) {
-      /* try the drivers that are listed in the config file */
-      for (n=-2; n<255; n++) {
-	 switch (n) {
-
-	    case -2:
-	       /* example: gfx_card_640x480x16 = */
-	       usprintf(buf, uconvert_ascii("gfx_card_%dx%dx%d", tmp), w, h, _color_depth);
-	       break;
-
-	    case -1:
-	       /* example: gfx_card_24bpp = */
-	       usprintf(buf, uconvert_ascii("gfx_card_%dbpp", tmp), _color_depth);
-	       break;
-
-	    case 0:
-	       /* example: gfx_card = */
-	       ustrcpy(buf, uconvert_ascii("gfx_card", tmp));
-	       break;
-
-	    default:
-	       /* example: gfx_card1 = */
-	       usprintf(buf, uconvert_ascii("gfx_card%d", tmp), n);
-	       break;
-	 }
-
-	 card = get_config_id(NULL, buf, GFX_AUTODETECT);
-
-	 if (card != GFX_AUTODETECT) {
-	    for (c=0; driver_list[c].driver; c++) {
-	       if (driver_list[c].id == card) {
-		  gfx_driver = driver_list[c].driver;
-		  break;
-	       }
-	    }
-	    if (gfx_driver) {
-	       tried = TRUE;
-	       gfx_driver->name = gfx_driver->desc = get_config_text(gfx_driver->ascii_name);
-	       screen = gfx_driver->init(w, h, v_w, v_h, _color_depth);
-	       if (screen)
-		  break;
-	       else
-		  gfx_driver = NULL;
-	    }
-	 }
-	 else {
-	    if (n > 1)
-	       break;
-	 }
-      }
+   /* filter specific fullscreen/windowed driver requests */
+   if (card == GFX_AUTODETECT_FULLSCREEN) {
+      flags |= GFX_DRIVER_FULLSCREEN_FLAG;
+      card = GFX_AUTODETECT;
+   }
+   else if (card == GFX_AUTODETECT_WINDOWED) {
+      flags |= GFX_DRIVER_WINDOWED_FLAG;
+      card = GFX_AUTODETECT;
    }
 
-   if (!tried) {
-      /* search table for the requested driver */
-      for (c=0; driver_list[c].driver; c++) {
-	 if (driver_list[c].id == card) {
-	    gfx_driver = driver_list[c].driver;
-	    break;
-	 }
+   if (card == GFX_AUTODETECT) {
+      /* autodetect the driver */
+      int found = FALSE;
+
+      /* first try the config variables */
+      if (allow_config) {
+	 /* try the gfx_card variable if GFX_AUTODETECT or GFX_AUTODETECT_FULLSCREEN was selected */
+	 if (!(flags & GFX_DRIVER_WINDOWED_FLAG))
+	    found = get_config_gfx_driver(uconvert_ascii("gfx_card", tmp1), w, h, v_w, v_h, flags, driver_list);
+
+	 /* try the gfx_cardw variable if GFX_AUTODETECT or GFX_AUTODETECT_WINDOWED was selected */
+	 if (!(flags & GFX_DRIVER_FULLSCREEN_FLAG) && !found)
+	    found = get_config_gfx_driver(uconvert_ascii("gfx_cardw", tmp1), w, h, v_w, v_h, flags, driver_list);
       }
 
-      if (gfx_driver) {                            /* specific driver? */
-	 gfx_driver->name = gfx_driver->desc = get_config_text(gfx_driver->ascii_name);
-	 screen = gfx_driver->init(w, h, v_w, v_h, _color_depth);
-      }
-      else {                                       /* otherwise autodetect */
+      /* go through the list of autodetected drivers if none was previously found */
+      if (!found) {
 	 for (c=0; driver_list[c].driver; c++) {
 	    if (driver_list[c].autodetect) {
-	       gfx_driver = driver_list[c].driver;
-	       gfx_driver->name = gfx_driver->desc = get_config_text(gfx_driver->ascii_name);
-	       screen = gfx_driver->init(w, h, v_w, v_h, _color_depth);
-	       if (screen)
-		  break;
+	       drv = driver_list[c].driver;
+
+	       if (gfx_driver_is_valid(drv, flags)) {
+		  screen = init_gfx_driver(drv, w, h, v_w, v_h);
+
+		  if (screen)
+		     break;
+	       }
 	    }
 	 }
       }
    }
+   else {
+      /* search the list for the requested driver */
+      drv = get_gfx_driver_from_id(card, driver_list);
 
+      if (drv)
+	 screen = init_gfx_driver(drv, w, h, v_w, v_h);
+   }
+
+   /* gracefully handle failure */
    if (!screen) {
-      gfx_driver = NULL;
-      screen = NULL;
+      gfx_driver = NULL;  /* set by init_gfx_driver() */
 
       if (!ugetc(allegro_error))
-	 ustrcpy(allegro_error, get_config_text("Unable to find a suitable graphics driver"));
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Unable to find a suitable graphics driver"));
 
       if (system_driver->display_switch_lock)
 	 system_driver->display_switch_lock(FALSE, FALSE);
@@ -504,6 +797,7 @@ int set_gfx_mode(int card, int w, int h, int v_w, int v_h)
       return -1;
    }
 
+   /* set the basic capabilities of the driver */
    if ((VIRTUAL_W > SCREEN_W) || (VIRTUAL_H > SCREEN_H)) {
       if (gfx_driver->scroll)
 	 gfx_capabilities |= GFX_CAN_SCROLL;
@@ -512,8 +806,21 @@ int set_gfx_mode(int card, int w, int h, int v_w, int v_h)
 	 gfx_capabilities |= GFX_CAN_TRIPLE_BUFFER;
    }
 
+   /* check whether we are instructed to disable vsync */
+   dv = get_config_string(uconvert_ascii("graphics", tmp1),
+                          uconvert_ascii("disable_vsync", tmp2),
+                          NULL);
+
+   if ((dv) && ((c = ugetc(dv)) != 0) && ((c == 'y') || (c == 'Y') || (c == '1')))
+      _wait_for_vsync = FALSE;
+   else
+      _wait_for_vsync = TRUE;
+
+   clear_bitmap(screen);
+
+   /* set up the default colors */
    for (c=0; c<256; c++)
-      palette_color8[c] = c;
+      _palette_color8[c] = c;
 
    set_palette(default_palette);
 
@@ -528,8 +835,9 @@ int set_gfx_mode(int card, int w, int h, int v_w, int v_h)
       gui_bg_color = makecol(255, 255, 255);
    }
 
-   clear(screen);
-   _set_mouse_range();
+   if (_al_linker_mouse)
+      _al_linker_mouse->set_mouse_etc();
+
    LOCK_DATA(gfx_driver, sizeof(GFX_DRIVER));
 
    _register_switch_bitmap(screen, NULL);
@@ -624,7 +932,7 @@ BITMAP *_make_bitmap(int w, int h, unsigned long addr, GFX_DRIVER *driver, int c
 
    driver->vid_phys_base = addr;
 
-   b->line[0] = (char *)addr;
+   b->line[0] = (unsigned char *)addr;
    _gfx_bank[0] = 0;
 
    if (driver->linear) {
@@ -661,6 +969,9 @@ BITMAP *create_bitmap_ex(int color_depth, int width, int height)
    GFX_VTABLE *vtable;
    BITMAP *bitmap;
    int i;
+   ASSERT(width >= 0);
+   ASSERT(height > 0);
+   ASSERT(system_driver);
 
    if (system_driver->create_bitmap)
       return system_driver->create_bitmap(color_depth, width, height);
@@ -708,6 +1019,8 @@ BITMAP *create_bitmap_ex(int color_depth, int width, int height)
  */
 BITMAP *create_bitmap(int width, int height)
 {
+   ASSERT(width >= 0);
+   ASSERT(height > 0);
    return create_bitmap_ex(_color_depth, width, height);
 }
 
@@ -726,14 +1039,10 @@ BITMAP *create_sub_bitmap(BITMAP *parent, int x, int y, int width, int height)
    BITMAP *bitmap;
    int i;
 
-   if (!parent) 
-      return NULL;
-
-   if (x < 0) 
-      x = 0; 
-
-   if (y < 0) 
-      y = 0;
+   ASSERT(parent);
+   ASSERT((x >= 0) && (y >= 0) && (x < parent->w) && (y < parent->h));
+   ASSERT((width > 0) && (height > 0));
+   ASSERT(system_driver);
 
    if (x+width > parent->w) 
       width = parent->w-x;
@@ -781,6 +1090,7 @@ BITMAP *create_sub_bitmap(BITMAP *parent, int x, int y, int width, int height)
    }
 
    bitmap->id = parent->id | BMP_ID_SUB;
+   bitmap->id &= ~BMP_ID_LOCKED;
 
    if (is_planar_bitmap(bitmap))
       x /= 4;
@@ -810,59 +1120,83 @@ BITMAP *create_sub_bitmap(BITMAP *parent, int x, int y, int width, int height)
 
 
 
-/* try_vram_location:
- *  Attempts to create a video memory bitmap out of the specified region
- *  of the larger screen surface, returning a pointer to it if that area
- *  is free.
+/* add_vram_block:
+ *  Creates a video memory bitmap out of the specified region
+ *  of the larger screen surface, returning a pointer to it.
  */
-static BITMAP *try_vram_location(int x, int y, int w, int h)
+static BITMAP *add_vram_block(int x, int y, int w, int h)
 {
-   VRAM_BITMAP *block = vram_bitmap_list;
+   VRAM_BITMAP *b, *new_b;
+   VRAM_BITMAP **last_p;
 
-   if ((x<0) || (y<0) || (x+w > VIRTUAL_W) || (y+h > VIRTUAL_H))
+   new_b = malloc(sizeof(VRAM_BITMAP));
+   if (!new_b)
       return NULL;
 
-   while (block) {
-      if (((x < block->x+block->w) && (x+w > block->x)) &&
-	  ((y < block->y+block->h) && (y+h > block->y)))
-	 return NULL;
+   new_b->x = x;
+   new_b->y = y;
+   new_b->w = w;
+   new_b->h = h;
 
-      block = block->next;
-   }
-
-   block = malloc(sizeof(VRAM_BITMAP));
-   if (!block)
-      return NULL;
-
-   block->x = x;
-   block->y = y;
-   block->w = w;
-   block->h = h;
-
-   block->bmp = create_sub_bitmap(screen, x, y, w, h);
-   if (!(block->bmp)) {
-      free(block);
+   new_b->bmp = create_sub_bitmap(screen, x, y, w, h);
+   if (!new_b->bmp) {
+      free(new_b);
       return NULL;
    }
 
-   block->next = vram_bitmap_list;
-   vram_bitmap_list = block;
+   /* find sorted y-position */
+   last_p = &vram_bitmap_list;
+   for (b = vram_bitmap_list; b && (b->y < new_b->y); b = b->next_y)
+      last_p = &b->next_y;
 
-   return block->bmp;
+   /* insert */
+   *last_p = new_b;
+   new_b->next_y = b;
+
+   return new_b->bmp;
 }
 
 
 
 /* create_video_bitmap:
  *  Attempts to make a bitmap object for accessing offscreen video memory.
+ *
+ *  The algorithm is similar to algorithms for drawing polygons. Think of
+ *  a horizontal stripe whose height is equal to that of the new bitmap.
+ *  It is initially aligned to the top of video memory and then it moves
+ *  downwards, stopping each time its top coincides with the bottom of a
+ *  video bitmap. For each stop, create a list of video bitmaps intersecting
+ *  the stripe, sorted by the left coordinate, from left to right, in the
+ *  next_x linked list. We look through this list and stop as soon as the
+ *  gap between the rightmost right edge seen so far and the current left
+ *  edge is big enough for the new video bitmap. In that case, we are done.
+ *  If we don't find such a gap, we move the stripe further down.
+ *
+ *  To make it efficient to find new bitmaps intersecting the stripe, the
+ *  list of all bitmaps is always sorted by top, from top to bottom, in
+ *  the next_y linked list. The list of bitmaps intersecting the stripe is
+ *  merely updated, not recalculated from scratch, when we move the stripe.
+ *  So every bitmap gets bubbled to its correct sorted x-position at most
+ *  once. Bubbling a bitmap takes at most as many steps as the number of
+ *  video bitmaps. So the algorithm behaves at most quadratically with
+ *  regard to the number of video bitmaps. I think that a linear behaviour
+ *  is more typical in practical applications (this happens, for instance,
+ *  if you allocate many bitmaps of the same height).
+ *
+ *  There's one more optimization: if a call fails, we cache the size of the
+ *  requested bitmap, so that next time we can detect very quickly that this
+ *  size is too large (this needs to be maintained when destroying bitmaps).
  */
 BITMAP *create_video_bitmap(int width, int height)
 {
-   VRAM_BITMAP *blockx;
-   VRAM_BITMAP *blocky;
-   VRAM_BITMAP *block;
+   VRAM_BITMAP *active_list, *b, *vram_bitmap;
+   VRAM_BITMAP **last_p;
    BITMAP *bmp;
+   int x = 0, y = 0;
 
+   ASSERT(width >= 0);
+   ASSERT(height > 0);
+   
    if (_dispsw_status)
       return NULL;
 
@@ -872,36 +1206,111 @@ BITMAP *create_video_bitmap(int width, int height)
       if (!bmp)
 	 return NULL;
 
-      block = malloc(sizeof(VRAM_BITMAP));
-      block->x = -1;
-      block->y = -1;
-      block->w = 0;
-      block->h = 0;
-      block->bmp = bmp;
-      block->next = vram_bitmap_list;
-      vram_bitmap_list = block;
+      b = malloc(sizeof(VRAM_BITMAP));
+      b->x = -1;
+      b->y = -1;
+      b->w = 0;
+      b->h = 0;
+      b->bmp = bmp;
+      b->next_y = vram_bitmap_list;
+      vram_bitmap_list = b;
 
       return bmp;
    }
 
-   /* otherwise fall back on subdividing the normal screen surface */
-   if ((bmp = try_vram_location(0, 0, width, height)) != NULL)
-      return bmp;
+   /* check bad args */
+   if ((width > VIRTUAL_W) || (height > VIRTUAL_H) ||
+       (width < 0) || (height < 0))
+      return NULL;
 
-   for (blocky = vram_bitmap_list; blocky; blocky = blocky->next) {
-      for (blockx = vram_bitmap_list; blockx; blockx = blockx->next) {
-	 if ((bmp = try_vram_location((blockx->x+blockx->w+15)&~15, blocky->y, width, height)) != NULL)
-	    return bmp;
-	 if ((bmp = try_vram_location((blockx->x-width)&~15, blocky->y, width, height)) != NULL)
-	    return bmp;
-	 if ((bmp = try_vram_location(blockx->x, blocky->y+blocky->h, width, height)) != NULL)
-	    return bmp;
-	 if ((bmp = try_vram_location(blockx->x, blocky->y-height, width, height)) != NULL)
-	    return bmp;
+   /* check cached bitmap size */
+   if ((width >= failed_bitmap_w) && (height >= failed_bitmap_h))
+      return NULL;
+
+   vram_bitmap = vram_bitmap_list;
+   active_list = NULL;
+   y = 0;
+
+   while (TRUE) {
+      /* Move the blocks from the VRAM_BITMAP_LIST that intersect the
+       * stripe to the ACTIVE_LIST: look through the VRAM_BITMAP_LIST
+       * following the next_y link, grow the ACTIVE_LIST following
+       * the next_x link and sorting by x-position.
+       * (this loop runs in amortized quadratic time)
+       */
+      while (vram_bitmap && (vram_bitmap->y < y+height)) {
+	 /* find sorted x-position */
+	 last_p = &active_list;
+	 for (b = active_list; b && (vram_bitmap->x > b->x); b = b->next_x)
+	    last_p = &b->next_x;
+
+	 /* insert */
+	 *last_p = vram_bitmap;
+	 vram_bitmap->next_x = b;
+
+	 /* next video bitmap */
+	 vram_bitmap = vram_bitmap->next_y;
+      }
+
+      x = 0;
+
+      /* Look for holes to put our bitmap in.
+       * (this loop runs in quadratic time)
+       */
+      for (b = active_list; b; b = b->next_x) {
+	 if (x+width <= b->x)  /* hole is big enough? */
+	    return add_vram_block(x, y, width, height);
+
+         /* Move search x-position to the right edge of the
+          * skipped bitmap if we are not already past it.
+          * And check there is enough room before continuing.
+          */
+	 if (x < b->x + b->w) {
+	    x = (b->x + b->w + 15) & ~15;
+	    if (x+width > VIRTUAL_W)
+	       break;
+	 }
+      }
+
+      /* If the whole ACTIVE_LIST was scanned, there is a big
+       * enough hole to the right of the rightmost bitmap.
+       */
+      if (b == NULL)
+	 return add_vram_block(x, y, width, height);
+
+      /* Move search y-position to the topmost bottom edge
+       * of the bitmaps intersecting the stripe.
+       * (this loop runs in quadratic time)
+       */
+      y = active_list->y + active_list->h;
+      for (b = active_list->next_x; b; b = b->next_x) {
+	 if (y > b->y + b->h)
+	    y = b->y + b->h;
+      }
+
+      if (y+height > VIRTUAL_H) {  /* too close to bottom? */
+	 /* Before failing, cache this bitmap size so that later calls can
+	  * learn from it. Use area as a measure to sort the bitmap sizes.
+          */
+	 if (width * height < failed_bitmap_w * failed_bitmap_h) {
+	    failed_bitmap_w = width;
+	    failed_bitmap_h = height;
+	 }
+
+	 return NULL;
+      }
+
+      /* Remove the blocks that don't intersect the new stripe from ACTIVE_LIST.
+       * (this loop runs in quadratic time)
+       */
+      last_p = &active_list;
+      for (b = active_list; b; b = b->next_x) {
+	 if (y >= b->y + b->h)
+	    *last_p = b->next_x;
+	 else
+	    last_p = &b->next_x;
       }
    }
-
-   return NULL;
 }
 
 
@@ -913,6 +1322,8 @@ BITMAP *create_system_bitmap(int width, int height)
 {
    BITMAP *bmp;
 
+   ASSERT(width >= 0);
+   ASSERT(height > 0);
    ASSERT(gfx_driver != NULL);
 
    if (gfx_driver->create_system_bitmap)
@@ -944,9 +1355,9 @@ void destroy_bitmap(BITMAP *bitmap)
 	 while (pos) {
 	    if (pos->bmp == bitmap) {
 	       if (prev)
-		  prev->next = pos->next;
+		  prev->next_y = pos->next_y;
 	       else
-		  vram_bitmap_list = pos->next;
+		  vram_bitmap_list = pos->next_y;
 
 	       if (pos->x < 0) {
 		  /* the driver is in charge of this object */
@@ -955,12 +1366,24 @@ void destroy_bitmap(BITMAP *bitmap)
 		  return;
 	       } 
 
+	       /* Update cached bitmap size using worst case scenario:
+		* the bitmap lies between two holes whose size is the cached
+		* size on each axis respectively.
+		*/
+	       failed_bitmap_w = failed_bitmap_w * 2 + ((bitmap->w + 15) & ~15);
+	       if (failed_bitmap_w > BMP_MAX_SIZE)
+		  failed_bitmap_w = BMP_MAX_SIZE;
+
+	       failed_bitmap_h = failed_bitmap_h * 2 + bitmap->h;
+	       if (failed_bitmap_h > BMP_MAX_SIZE)
+		  failed_bitmap_h = BMP_MAX_SIZE;
+
 	       free(pos);
 	       break;
 	    }
 
 	    prev = pos;
-	    pos = pos->next;
+	    pos = pos->next_y;
 	 }
 
 	 _unregister_switch_bitmap(bitmap);
@@ -990,6 +1413,53 @@ void destroy_bitmap(BITMAP *bitmap)
 
 
 
+/* set_clip_rect:
+ *  Sets the two opposite corners of the clipping rectangle to be used when
+ *  drawing to the bitmap. Nothing will be drawn to positions outside of this 
+ *  rectangle. When a new bitmap is created the clipping rectangle will be 
+ *  set to the full area of the bitmap.
+ */
+void set_clip_rect(BITMAP *bitmap, int x1, int y1, int x2, int y2)
+{
+   ASSERT(bitmap);
+
+   /* internal clipping is inclusive-exclusive */
+   x2++;
+   y2++;
+
+   bitmap->cl = MID(0, x1, bitmap->w-1);
+   bitmap->ct = MID(0, y1, bitmap->h-1);
+   bitmap->cr = MID(0, x2, bitmap->w);
+   bitmap->cb = MID(0, y2, bitmap->h);
+
+   if (bitmap->vtable->set_clip)
+      bitmap->vtable->set_clip(bitmap);
+}
+
+
+
+/* add_clip_rect:
+ *  Makes the new clipping rectangle the intersection between the given
+ *  rectangle and the current one.
+ */
+void add_clip_rect(BITMAP *bitmap, int x1, int y1, int x2, int y2)
+{
+   int cx1, cy1, cx2, cy2;
+
+   ASSERT(bitmap);
+
+   get_clip_rect(bitmap, &cx1, &cy1, &cx2, &cy2);
+
+   x1 = MAX(x1, cx1);
+   y1 = MAX(y1, cy1);
+   x2 = MIN(x2, cx2);
+   y2 = MIN(y2, cy2);
+
+   set_clip_rect(bitmap, x1, y1, x2, y2);
+}
+
+
+
 /* set_clip:
  *  Sets the two opposite corners of the clipping rectangle to be used when
  *  drawing to the bitmap. Nothing will be drawn to positions outside of this 
@@ -1006,14 +1476,8 @@ void set_clip(BITMAP *bitmap, int x1, int y1, int x2, int y2)
    ASSERT(bitmap);
 
    if ((!x1) && (!y1) && (!x2) && (!y2)) {
-      bitmap->clip = FALSE;
-      bitmap->cl = bitmap->ct = 0;
-      bitmap->cr = bitmap->w;
-      bitmap->cb = bitmap->h;
-
-      if (bitmap->vtable->set_clip)
-	 bitmap->vtable->set_clip(bitmap);
-
+      set_clip_rect(bitmap, 0, 0, bitmap->w-1, bitmap->h-1);
+      set_clip_state(bitmap, FALSE);
       return;
    }
 
@@ -1029,17 +1493,8 @@ void set_clip(BITMAP *bitmap, int x1, int y1, int x2, int y2)
       y2 = t;
    }
 
-   x2++;
-   y2++;
-
-   bitmap->clip = TRUE;
-   bitmap->cl = MID(0, x1, bitmap->w-1);
-   bitmap->ct = MID(0, y1, bitmap->h-1);
-   bitmap->cr = MID(0, x2, bitmap->w);
-   bitmap->cb = MID(0, y2, bitmap->h);
-
-   if (bitmap->vtable->set_clip)
-      bitmap->vtable->set_clip(bitmap);
+   set_clip_rect(bitmap, x1, y1, x2, y2);
+   set_clip_state(bitmap, TRUE);
 }
 
 
