@@ -17,7 +17,7 @@
 
 
 #include "allegro.h"
-#include "allegro/aintern.h"
+#include "allegro/internal/aintern.h"
 
 
 
@@ -127,8 +127,6 @@ static INLINE void add_key(volatile KEY_BUFFER *buffer, int key, int scancode)
  */
 void clear_keybuf()
 {
-   int c;
-
    if (keyboard_polled)
       poll_keyboard();
 
@@ -138,11 +136,6 @@ void clear_keybuf()
    key_buffer.start = key_buffer.end = 0;
    _key_buffer.start = _key_buffer.end = 0;
 
-   for (c=0; c<KEY_MAX; c++) {
-      key[c] = 0;
-      _key[c] = 0;
-   }
-
    key_buffer.lock--;
    _key_buffer.lock--;
 
@@ -153,6 +146,21 @@ void clear_keybuf()
 
 
 
+/* clear_key:
+ *  Helper function to clear the key[] array.
+ */
+static void clear_key(void)
+{
+   int c;
+
+   for (c=0; c<KEY_MAX; c++) {
+      key[c] = 0;
+      _key[c] = 0;
+   }
+}
+
+
+
 /* keypressed:
  *  Returns TRUE if there are keypresses waiting in the keyboard buffer.
  */
@@ -223,7 +231,7 @@ int ureadkey(int *scancode)
       if (keyboard_polled)
 	 poll_keyboard();
 
-      yield_timeslice();
+      rest(1);
    }
 
    c = key_buffer.key[key_buffer.start];
@@ -326,6 +334,7 @@ void install_keyboard_hooks(int (*keypressed)(void), int (*readkey)(void))
    key_buffer.lock = _key_buffer.lock = 0;
 
    clear_keybuf();
+   clear_key();
 
    keypressed_hook = keypressed;
    readkey_hook = readkey;
@@ -336,7 +345,7 @@ void install_keyboard_hooks(int (*keypressed)(void), int (*readkey)(void))
 /* update_shifts:
  *  Helper function to update the key_shifts variable and LED state.
  */
-static INLINE void update_shifts()
+static INLINE void update_shifts(void)
 {
    #define LED_FLAGS  (KB_SCROLOCK_FLAG | KB_NUMLOCK_FLAG | KB_CAPSLOCK_FLAG)
 
@@ -387,7 +396,8 @@ void _handle_key_press(int keycode, int scancode)
    }
 
    /* autorepeat? */
-   if ((keyboard_driver->autorepeat) && (repeat_delay) && (keycode >= 0) && 
+   if ((keyboard_driver->autorepeat) && (repeat_delay) && 
+       (keycode >= 0) && (scancode > 0) && (scancode != KEY_PAUSE) &&
        ((keycode != repeat_key) || (scancode != repeat_scan))) {
       repeat_key = keycode;
       repeat_scan = scancode;
@@ -551,6 +561,7 @@ int install_keyboard()
    key_buffer.lock = _key_buffer.lock = 0;
 
    clear_keybuf();
+   clear_key();
 
    if (system_driver->keyboard_drivers)
       driver_list = system_driver->keyboard_drivers();
@@ -614,6 +625,7 @@ void remove_keyboard(void)
    keyboard_polled = FALSE;
 
    clear_keybuf();
+   clear_key();
 
    key_shifts = _key_shifts = 0;
 
