@@ -14,16 +14,18 @@
  *
  *      Proper mode setting support added by George Foot.
  *
+ *      Modified by Grzegorz Adam Hankiewicz.
+ *
  *      See readme.txt for copyright information.
  */
 
 
 #include "allegro.h"
-#include "allegro/aintern.h"
-#include "allegro/aintunix.h"
+#include "allegro/internal/aintern.h"
+#include "allegro/platform/aintunix.h"
 
 
-#ifdef ALLEGRO_LINUX_FBCON
+#if (defined ALLEGRO_LINUX_FBCON) && ((!defined ALLEGRO_WITH_MODULES) || (defined ALLEGRO_MODULE))
 
 #if !defined(_POSIX_MAPPED_FILES) || !defined(HAVE_MMAP)
 #error "Sorry, mapped files are required for Linux console Allegro to work!"
@@ -35,15 +37,25 @@
 #include <linux/fb.h>
 #include <sys/mman.h>
 
+/* 2.4 kernel doesn't seem to have these, so assume that they don't
+ * "have sticky," heheh. */
 
+#ifndef FB_VBLANK_HAVE_STICKY
+#define FB_VBLANK_HAVE_STICKY 0
+#endif
+
+#ifndef FB_VBLANK_STICKY
+#define FB_VBLANK_STICKY 0
+#endif
 
 static BITMAP *fb_init(int w, int h, int v_w, int v_h, int color_depth);
+static int fb_open_device(void);
 static void fb_exit(BITMAP *b);
 static void fb_save(void);
 static void fb_restore(void);
 static int  fb_scroll(int x, int y);
 static void fb_vsync(void);
-static void fb_set_palette(RGB *p, int from, int to, int vsync);
+static void fb_set_palette(AL_CONST RGB *p, int from, int to, int vsync);
 static void fb_save_cmap(void);
 static void fb_restore_cmap(void);
 
@@ -67,14 +79,14 @@ GFX_DRIVER gfx_fbcon =
    NULL,                         /* no drawing mode hook */
    fb_save,
    fb_restore,
+   NULL,                         /* no fetch mode hook */
    0, 0,
    TRUE,
-   0, 0, 0, 0
+   0, 0, 0, 0, FALSE
 };
 
 
-
-static char fb_desc[128] = EMPTY_STRING;     /* description string */
+static char fb_desc[256] = EMPTY_STRING;     /* description string */
 
 static struct fb_fix_screeninfo fix_info;    /* read-only video mode info */
 static struct fb_var_screeninfo orig_mode;   /* original video mode info */
@@ -95,62 +107,59 @@ static int fb_approx;                        /* emulate small resolution */
 static int update_timings(struct fb_var_screeninfo *mode);
 
 
+
 /* fb_init:
  *  Sets a graphics mode.
  */
 static BITMAP *fb_init(int w, int h, int v_w, int v_h, int color_depth)
 {
-   char fname[256], tmp[256], *p;
-   int stride, tries;
+   AL_CONST char *p;
+   int stride, tries, original_color_depth = _color_depth;
    BITMAP *b;
+   char tmp[16];
 
-   /* find the device filename */
-   p = get_config_string(NULL, uconvert_ascii("framebuffer", tmp), NULL);
-
-   if ((p) && (ugetc(p))) {
-      do_uconvert(p, U_CURRENT, fname, U_ASCII, sizeof(fname));
-   }
-   else {
-      p = getenv("FRAMEBUFFER");
-
-      if ((p) && (p[0]))
-	 strcpy(fname, p);
-      else
-	 strcpy(fname, "/dev/fb0");
-   }
-
-   /* open the framebuffer device */
-   if ((fbfd = open(fname, O_RDWR)) < 0) {
-      usprintf(allegro_error, get_config_text("Can't open framebuffer %s"), uconvert_ascii(fname, tmp));
+   /* open framebuffer and store info in global variables */
+   if (fb_open_device() != 0)
       return NULL;
-   }
-
-   /* read video mode information */
-   if ((ioctl(fbfd, FBIOGET_FSCREENINFO, &fix_info) != 0) ||
-       (ioctl(fbfd, FBIOGET_VSCREENINFO, &orig_mode) != 0)) {
-      usprintf(allegro_error, get_config_text("Framebuffer ioctl() failed"));
-      return NULL;
-   }
 
    /* look for a nice graphics mode in several passes */
    fb_approx = FALSE;
 
-   if ((!w) && (!h)) {
+   ASSERT (w >= 0);
+   ASSERT (h >= 0);
+
+   /* preset a resolution if the user didn't ask for one */
+   if (((!w) && (!h)) || _safe_gfx_mode_change) {
       w = orig_mode.xres;
       h = orig_mode.yres;
    }
 
-   for (tries=0; tries<3; tries++) {
+   if (_safe_gfx_mode_change) tries = -1;
+   else tries = 0;
+   
+   for (; tries<3; tries++) {
       my_mode = orig_mode;
 
       switch (tries) {
 
+	 case -1:
+	    /* let's see if we can get the actual screen mode */
+	    /* shouldn't we be keeping the previous color depth setting? */
+	    switch (orig_mode.bits_per_pixel) {
+	       case 8:
+	       case 16:
+	       case 24:
+	       case 32:
+		  color_depth = orig_mode.bits_per_pixel;
+		  set_color_depth(color_depth);
+	    }
+	    break;
 	 case 0:
 	    /* try for largest possible virtual screen size */
 	    my_mode.xres = w;
 	    my_mode.yres = h;
 	    my_mode.xres_virtual = MAX(w, v_w);
-	    my_mode.yres_virtual = MAX(MAX(h, v_h), fix_info.smem_len / (my_mode.xres_virtual * BYTES_PER_PIXEL(color_depth)));
+	    my_mode.yres_virtual = MAX(MAX(h, v_h), (int)(fix_info.smem_len / (my_mode.xres_virtual * BYTES_PER_PIXEL(color_depth))));
 	    break;
 
 	 case 1:
@@ -163,7 +172,7 @@ static BITMAP *fb_init(int w, int h, int v_w, int v_h, int color_depth)
 
 	 case 2:
 	    /* see if we can fake a smaller mode (better than nothing) */
-	    if ((my_mode.xres < w) || (my_mode.yres < h) || (v_w > w) || (v_h > h))
+	    if (((int)my_mode.xres < w) || ((int)my_mode.yres < h) || (v_w > w) || (v_h > h))
 	       continue;
 	    fb_approx = TRUE;
 	    break;
@@ -237,9 +246,12 @@ static BITMAP *fb_init(int w, int h, int v_w, int v_h, int color_depth)
    }
 
    /* oops! */
+   if (_safe_gfx_mode_change)
+      set_color_depth(original_color_depth);
+      
    ioctl(fbfd, FBIOPUT_VSCREENINFO, &orig_mode);
    close(fbfd);
-   usprintf(allegro_error, get_config_text("Framebuffer resolution not available"));
+   ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Framebuffer resolution not available"));
    return NULL;
 
    got_a_nice_mode:
@@ -249,7 +261,7 @@ static BITMAP *fb_init(int w, int h, int v_w, int v_h, int color_depth)
    if (fbaddr == MAP_FAILED) {
       ioctl(fbfd, FBIOPUT_VSCREENINFO, &orig_mode);
       close(fbfd);
-      usprintf(allegro_error, get_config_text("Can't map framebuffer %s"), uconvert_ascii(fname, tmp));
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Can't map framebuffer"));
       return NULL;
    }
 
@@ -259,7 +271,7 @@ static BITMAP *fb_init(int w, int h, int v_w, int v_h, int color_depth)
 
    gfx_fbcon.vid_mem = fix_info.smem_len;
 
-   stride = my_mode.xres_virtual * BYTES_PER_PIXEL(color_depth);
+   stride = fix_info.line_length;
    v_w = my_mode.xres_virtual;
    v_h = my_mode.yres_virtual;
    p = fbaddr;
@@ -286,8 +298,8 @@ static BITMAP *fb_init(int w, int h, int v_w, int v_h, int color_depth)
    do_uconvert(fix_info.id, U_ASCII, fb_desc, U_CURRENT, sizeof(fb_desc));
 
    if (fb_approx) {
-      ustrcat(fb_desc, uconvert_ascii(", ", NULL));
-      ustrcat(fb_desc, get_config_text("approx."));
+      ustrzcat(fb_desc, sizeof(fb_desc), uconvert_ascii(", ", tmp));
+      ustrzcat(fb_desc, sizeof(fb_desc), get_config_text("approx."));
    }
 
    gfx_fbcon.desc = fb_desc;
@@ -346,8 +358,8 @@ static BITMAP *fb_init(int w, int h, int v_w, int v_h, int color_depth)
    if (!(vblank_flags & (FB_VBLANK_HAVE_VBLANK | FB_VBLANK_HAVE_STICKY | FB_VBLANK_HAVE_VCOUNT)))
  #endif
    {
-      ustrcat(fb_desc, uconvert_ascii(", ", NULL));
-      ustrcat(fb_desc, get_config_text("no vsync"));
+      ustrzcat(fb_desc, sizeof(fb_desc), uconvert_ascii(", ", tmp));
+      ustrzcat(fb_desc, sizeof(fb_desc), get_config_text("no vsync"));
    }
 
    /* is scrolling available? */
@@ -377,6 +389,48 @@ static BITMAP *fb_init(int w, int h, int v_w, int v_h, int color_depth)
 
 
 
+/* fb_open_device:
+ *  Opens the framebuffer device, first checking config values or
+ *  environment variables. Returns 0 on success.
+ */
+static int fb_open_device(void)
+{
+   char fname[1024], tmp1[256], tmp2[256];
+   AL_CONST char *p;
+
+   /* find the device filename */
+   p = get_config_string(uconvert_ascii("graphics", tmp1), uconvert_ascii("framebuffer", tmp2), NULL);
+
+   if (p && ugetc(p))
+      do_uconvert(p, U_CURRENT, fname, U_ASCII, sizeof(fname));
+   else {
+      p = getenv("FRAMEBUFFER");
+
+      if ((p) && (p[0])) {
+	 _al_sane_strncpy(fname, p, sizeof(fname));
+      }
+      else
+	 _al_sane_strncpy(fname, "/dev/fb0", 1024);
+   }
+
+   /* open the framebuffer device */
+   if ((fbfd = open(fname, O_RDWR)) < 0) {
+      uszprintf(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Can't open framebuffer %s"), uconvert_ascii(fname, tmp1));
+      return 1;
+   }
+
+   /* read video mode information */
+   if ((ioctl(fbfd, FBIOGET_FSCREENINFO, &fix_info) != 0) ||
+       (ioctl(fbfd, FBIOGET_VSCREENINFO, &orig_mode) != 0)) {
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Framebuffer ioctl() failed"));
+      return 2;
+   }
+
+   return 0;
+}
+
+
+
 /* fb_exit:
  *  Unsets the video mode.
  */
@@ -428,13 +482,29 @@ static int fb_scroll(int x, int y)
 
    ret = ioctl(fbfd, FBIOPAN_DISPLAY, &my_mode);
 
-   fb_vsync();
-
- #ifdef FBIOGET_VBLANK
-   if (!vblank_flags)
- #endif
+   if (_wait_for_vsync) {
+      /* On a lot of graphics hardware the scroll registers don't
+       * take effect until the start of the next frame, so when you do a
+       * scroll you want to make sure you've waited at least until then.
+       * If you have reliable vsyncing, one call to vysnc is enough,
+       * but if you don't then the first call waits some time between 0
+       * and 1/60th of a second, and since it's not really synced to the
+       * retrace, this may or may not include an actual retrace. However,
+       * by making a second call, we then wait for a full 1/60th of a second
+       * after the first call, and this is sure to include a retrace,
+       * assuming the refresh rate is at least 60Hz. Sometimes overall
+       * we'll have waited past two retraces, or even more if the refresh
+       * rate is much higher than the fake vblank timer frequency, but
+       * it's still better than sometimes missing the retrace completely.
+       */
       fb_vsync();
 
+    #ifdef FBIOGET_VBLANK
+      if (!vblank_flags)
+    #endif
+         fb_vsync();
+   }
+
    return (ret) ? -1 : 0;
 }
 
@@ -443,9 +513,9 @@ static int fb_scroll(int x, int y)
 /* fb_vsync:
  *  Waits for a retrace.
  */
-static void fb_vsync()
+static void fb_vsync(void)
 {
-   int prev;
+   unsigned int prev;
 
  #ifdef FBIOGET_VBLANK
 
@@ -493,7 +563,7 @@ static void fb_vsync()
       prev = retrace_count;
 
       do {
-      } while (retrace_count == prev);
+      } while (retrace_count == (int)prev);
    }
 }
 
@@ -502,11 +572,12 @@ static void fb_vsync()
 /* fb_set_palette:
  *  Sets the palette.
  */
-static void fb_set_palette(RGB *p, int from, int to, int vsync)
+static void fb_set_palette(AL_CONST RGB *p, int from, int to, int vsync)
 {
    unsigned short r[256], g[256], b[256];
    struct fb_cmap cmap;
    int i;
+   ASSERT(p);
 
    cmap.start = from;
    cmap.len = to-from+1;
@@ -515,7 +586,7 @@ static void fb_set_palette(RGB *p, int from, int to, int vsync)
    cmap.blue = b;
    cmap.transp = NULL;
 
-   for (i=0; i<cmap.len; i++) {
+   for (i=0; i < (int)cmap.len; i++) {
       r[i] = p[from+i].r << 10;
       g[i] = p[from+i].g << 10;
       b[i] = p[from+i].b << 10;
@@ -571,7 +642,7 @@ static void fb_restore_cmap (void)
 
 
 
-struct timings {
+static struct timings {
    char config_item[1024];
    int pixclock;
    int left_margin;
@@ -598,7 +669,7 @@ static int tweak_timings (int w, int h);
  *  Returns a pointer to a struct as above containing timings for the given
  *  resolution, or NULL on error.
  */
-struct timings *_fb_get_timings (int w, int h)
+static struct timings *_fb_get_timings (int w, int h)
 {
    /* First try the config file */
    if (read_config_file (w, h)) return &temp_timings;
@@ -617,11 +688,12 @@ struct timings *_fb_get_timings (int w, int h)
  */
 static int read_config_file (int w, int h)
 {
+   char tmp[128];
    char **argv;
    int argc;
 
    /* Let the setup program know what config string we read for this mode */
-   usprintf(temp_timings.config_item, get_config_text("fb_mode_%dx%d"), w, h);
+   uszprintf(temp_timings.config_item, sizeof(temp_timings.config_item), uconvert_ascii("fb_mode_%dx%d", tmp), w, h);
 
    /* First try the config file */
    argv = get_config_argv (NULL, temp_timings.config_item, &argc);
@@ -636,11 +708,11 @@ static int read_config_file (int w, int h)
       get_info(vsync_len);
 
       if (*argv) {
-	 if (!ustrcmp (*argv, "none"))
+	 if (!ustrcmp (*argv, uconvert_ascii("none", tmp)))
 	    temp_timings.vmode = FB_VMODE_NONINTERLACED;
-	 else if (!ustrcmp (*argv, "interlaced"))
+	 else if (!ustrcmp (*argv, uconvert_ascii("interlaced", tmp)))
 	    temp_timings.vmode = FB_VMODE_INTERLACED;
-	 else if (!ustrcmp (*argv, "doublescan"))
+	 else if (!ustrcmp (*argv, uconvert_ascii("doublescan", tmp)))
 	    temp_timings.vmode = FB_VMODE_DOUBLE;
 	 argv++;
       } else
@@ -663,18 +735,24 @@ static char *get_line (FILE *file)
 {
    static char buffer[1024];
    char *ch;
-   if (feof (file)) return NULL;
-   fgets (buffer, sizeof buffer, file);
-   if (!strchr (buffer, '\n')) {
+   ASSERT (file);
+
+   if (!fgets (buffer, sizeof (buffer), file))
+      return 0;
+
+   /* if there's no eol, get one before continuing */
+   if (!strchr (buffer, '\n') && strlen (buffer) == 1 + sizeof (buffer)) {
+      char waste[128], *ret;
+
       do {
-	 fgets (buffer, sizeof buffer, file);
-      } while (!strchr (buffer, '\n'));
-      return NULL;
+	 ret = fgets (waste, sizeof (waste), file);
+      } while (ret && !strchr (waste, '\n'));
+      /* this doesn't actually exit because we still have buffer */
    }
-   ch = strchr (buffer, '#');
-   if (ch) *ch = '\0';
-   ch = strchr (buffer, '\n');
-   if (ch) *ch = '\0';
+   
+   if ((ch = strpbrk (buffer, "#\n")))
+      *ch = 0;
+      
    ch = buffer;
    while (uisspace(*ch)) ch++;
    return ch;
@@ -801,6 +879,8 @@ static int tweak_timings (int w, int h)
 
 static void set_default_timings (void)
 {
+   char tmp[128];
+
    #define cp(x) temp_timings.x = orig_mode.x
    cp(pixclock);
    cp(left_margin);
@@ -814,7 +894,8 @@ static void set_default_timings (void)
    cp(xres);
    cp(yres);
    #undef cp
-   usprintf(temp_timings.config_item, get_config_text("fb_mode_%dx%d"), orig_mode.xres, orig_mode.yres);
+   uszprintf(temp_timings.config_item, sizeof(temp_timings.config_item), uconvert_ascii("fb_mode_%dx%d", tmp),
+	     orig_mode.xres, orig_mode.yres);
 }
 
 
@@ -833,6 +914,7 @@ static void set_default_timings (void)
 static int update_timings(struct fb_var_screeninfo *mode)
 {
    struct timings *t;
+   ASSERT(mode);
 
    set_default_timings();
    t = _fb_get_timings (mode->xres, mode->yres);
@@ -859,14 +941,16 @@ static int update_timings(struct fb_var_screeninfo *mode)
 /* I'm not sure whether these work or not -- my Matrox seems capable
  * of setting whatever you ask it to. */
 
-int _fb_get_pixclock(void)
+#if 0
+
+static int _fb_get_pixclock(void)
 {
    struct fb_var_screeninfo mode;
    if (ioctl (fbfd, FBIOGET_VSCREENINFO, &mode)) return -1;
    return mode.pixclock;
 }
 
-void _fb_set_pixclock(int new_val)
+static void _fb_set_pixclock(int new_val)
 {
    struct fb_var_screeninfo mode;
    if (ioctl (fbfd, FBIOGET_VSCREENINFO, &mode)) return;
@@ -874,6 +958,23 @@ void _fb_set_pixclock(int new_val)
    if (ioctl (fbfd, FBIOPUT_VSCREENINFO, &mode)) return;
 }
 
+#endif
 
 
-#endif      /* ifdef ALLEGRO_LINUX_FBCON */
+
+#ifdef ALLEGRO_MODULE
+
+/* _module_init:
+ *  Called when loaded as a dynamically linked module.
+ */
+void _module_init(int system_driver)
+{
+   if (system_driver == SYSTEM_LINUX)
+      _unix_register_gfx_driver(GFX_FBCON, &gfx_fbcon, TRUE, TRUE);
+}
+
+#endif      /* ifdef ALLEGRO_MODULE */
+
+
+
+#endif      /* if (defined ALLEGRO_LINUX_FBCON) ... */
