@@ -22,7 +22,7 @@
 #include <string.h>
 
 #include "allegro.h"
-#include "allegro/aintern.h"
+#include "allegro/internal/aintern.h"
 
 
 
@@ -83,7 +83,7 @@ volatile long _midi_tick = 0;                   /* counter for killing notes */
 
 static void midi_player(void);                  /* core MIDI player routine */
 static void prepare_to_play(MIDI *midi);
-static void midi_lock_mem();
+static void midi_lock_mem(void);
 
 static MIDI *midifile = NULL;                   /* the file that is playing */
 
@@ -95,7 +95,7 @@ long midi_loop_end = -1;                        /* loop at this position */
 static int midi_semaphore = 0;                  /* reentrancy flag */
 static int midi_loaded_patches = FALSE;         /* loaded entire patch set? */
 
-static int midi_timer_speed;                    /* midi_player's timer speed */
+static long midi_timer_speed;                   /* midi_player's timer speed */
 static int midi_pos_speed;                      /* MIDI delta -> midi_pos */
 static int midi_speed;                          /* MIDI delta -> timer */
 static int midi_new_speed;                      /* for tempo change events */
@@ -117,8 +117,8 @@ static int midi_looping;                        /* set during loops */
 
 /* hook functions */
 void (*midi_msg_callback)(int msg, int byte1, int byte2) = NULL;
-void (*midi_meta_callback)(int type, unsigned char *data, int length) = NULL;
-void (*midi_sysex_callback)(unsigned char *data, int length) = NULL;
+void (*midi_meta_callback)(int type, AL_CONST unsigned char *data, int length) = NULL;
+void (*midi_sysex_callback)(AL_CONST unsigned char *data, int length) = NULL;
 
 
 
@@ -145,14 +145,15 @@ void lock_midi(MIDI *midi)
  *  Loads a standard MIDI file, returning a pointer to a MIDI structure,
  *  or NULL on error. 
  */
-MIDI *load_midi(char *filename)
+MIDI *load_midi(AL_CONST char *filename)
 {
    int c;
-   char buf[256];
+   char buf[4];
    long data;
    PACKFILE *fp;
    MIDI *midi;
    int num_tracks;
+   ASSERT(filename);
 
    fp = pack_fopen(filename, F_READ);        /* open the file */
    if (!fp)
@@ -247,7 +248,7 @@ void destroy_midi(MIDI *midi)
  *  number read, and alters the data pointer according to the number of
  *  bytes it used.
  */
-static unsigned long parse_var_len(unsigned char **data)
+static unsigned long parse_var_len(AL_CONST unsigned char **data)
 {
    unsigned long val = **data & 0x7F;
 
@@ -412,7 +413,7 @@ int _midi_allocate_voice(int min, int max)
    int c;
    int layer;
    int voice = -1;
-   int best_time = LONG_MAX;
+   long best_time = LONG_MAX;
 
    if (min < 0)
       min = 0;
@@ -574,6 +575,23 @@ END_OF_STATIC_FUNCTION(all_notes_off);
 
 
 
+/* all_sound_off:
+ *  Turns off sound.
+ */
+static void all_sound_off(int channel)
+{
+   if (midi_driver->raw_midi) {
+      midi_driver->raw_midi(0xB0+channel);
+      midi_driver->raw_midi(120);
+      midi_driver->raw_midi(0);
+      return;
+   }
+}
+
+END_OF_STATIC_FUNCTION(all_sound_off);
+
+
+
 /* reset_controllers:
  *  Resets volume, pan, pitch bend, etc, to default positions.
  */
@@ -602,7 +620,7 @@ END_OF_STATIC_FUNCTION(reset_controllers);
 /* update_controllers:
  *  Checks cached controller information and updates active voices.
  */
-static void update_controllers()
+static void update_controllers(void)
 {
    int c, c2, vol, bend, note;
 
@@ -673,6 +691,10 @@ static void process_controller(int channel, int ctrl, int data)
 	 }
 	 break;
 
+      case 120:                                 /* all sound off */
+	 all_sound_off(channel);
+	 break;
+
       case 121:                                 /* reset all controllers */
 	 reset_controllers(channel);
 	 break;
@@ -684,6 +706,14 @@ static void process_controller(int channel, int ctrl, int data)
       case 127:                                 /* poly mode on */
 	 all_notes_off(channel);
 	 break;
+
+      default:
+	 if (midi_driver->raw_midi) {
+	    midi_driver->raw_midi(0xB0+channel);
+	    midi_driver->raw_midi(ctrl);
+	    midi_driver->raw_midi(data);
+	 }
+	 break;
    }
 }
 
@@ -694,7 +724,7 @@ END_OF_STATIC_FUNCTION(process_controller);
 /* process_meta_event:
  *  Processes the next meta-event on the specified track.
  */
-static void process_meta_event(unsigned char **pos, long *timer)
+static void process_meta_event(AL_CONST unsigned char **pos, long *timer)
 {
    unsigned char metatype = *((*pos)++);
    long length = parse_var_len(pos);
@@ -725,7 +755,7 @@ END_OF_STATIC_FUNCTION(process_meta_event);
 /* process_midi_event:
  *  Processes the next MIDI event on the specified track.
  */
-static void process_midi_event(unsigned char **pos, unsigned char *running_status, long *timer)
+static void process_midi_event(AL_CONST unsigned char **pos, unsigned char *running_status, long *timer)
 {
    unsigned char byte1, byte2; 
    int channel;
@@ -864,13 +894,13 @@ static void midi_player(void)
 
 	 /* while events are waiting, process them */
 	 while (midi_track[c].timer <= 0) { 
-	    process_midi_event(&midi_track[c].pos, 
+	    process_midi_event((AL_CONST unsigned char**) &midi_track[c].pos, 
 			       &midi_track[c].running_status,
 			       &midi_track[c].timer); 
 
 	    /* read next time offset */
 	    if (midi_track[c].pos) { 
-	       l = parse_var_len(&midi_track[c].pos);
+	       l = parse_var_len((AL_CONST unsigned char**) &midi_track[c].pos);
 	       l *= midi_speed;
 	       midi_track[c].timer += l;
 	    }
@@ -929,8 +959,10 @@ static void midi_player(void)
 	    goto do_it_all_again;
 	 }
 	 else {
-	    for (c=0; c<16; c++)
+	    for (c=0; c<16; c++) {
 	       all_notes_off(c);
+	       all_sound_off(c);
+	    }
 	    prepare_to_play(midifile);
 	    goto do_it_all_again;
 	 }
@@ -974,8 +1006,7 @@ static int midi_init(void)
    int c, c2, c3;
    char **argv;
    int argc;
-   char buf[32];
-   char tmp[64];
+   char buf[32], tmp[64];
 
    midi_loaded_patches = FALSE;
 
@@ -996,7 +1027,7 @@ static int midi_init(void)
    }
 
    for (c=0; c<128; c++) {
-      usprintf(buf, uconvert_ascii("p%d", tmp), c+1);
+      uszprintf(buf, sizeof(buf), uconvert_ascii("p%d", tmp), c+1);
       argv = get_config_argv(uconvert_ascii("midimap", tmp), buf, &argc);
 
       if ((argv) && (argc == 4)) {
@@ -1054,6 +1085,10 @@ static int load_patches(MIDI *midi)
       running_status = 0;
 
       while (p < end) {                         /* work through data stream */
+#ifdef ALLEGRO_BEOS
+         /* Is there a bug in this routine, or in gcc under BeOS/x86? --PW */
+         { int i; for (i=1; i; i--); }
+#endif
 	 event = *p; 
 	 if (event & 0x80) {                    /* regular message */
 	    p++;
@@ -1091,7 +1126,7 @@ static int load_patches(MIDI *midi)
 	       switch (event) {
 		  case 0xF0:                    /* sysex */
 		  case 0xF7: 
-		     l = parse_var_len(&p);
+		     l = parse_var_len((AL_CONST unsigned char**) &p);
 		     p += l;
 		     break;
 
@@ -1105,7 +1140,7 @@ static int load_patches(MIDI *midi)
 
 		  case 0xFF:                    /* meta-event */
 		     p++;
-		     l = parse_var_len(&p);
+		     l = parse_var_len((AL_CONST unsigned char**) &p);
 		     p += l;
 		     break;
 
@@ -1122,7 +1157,7 @@ static int load_patches(MIDI *midi)
 	 }
 
 	 if (p < end)                           /* skip time offset */
-	    parse_var_len(&p);
+	    parse_var_len((AL_CONST unsigned char**) &p);
       }
    }
 
@@ -1163,7 +1198,7 @@ static void prepare_to_play(MIDI *midi)
    for (c=0; c<MIDI_TRACKS; c++) {
       if (midi->track[c].data) {
 	 midi_track[c].pos = midi->track[c].data;
-	 midi_track[c].timer = parse_var_len(&midi_track[c].pos);
+	 midi_track[c].timer = parse_var_len((AL_CONST unsigned char**) &midi_track[c].pos);
 	 midi_track[c].timer *= midi_speed;
       }
       else {
@@ -1193,8 +1228,10 @@ int play_midi(MIDI *midi, int loop)
 
    remove_int(midi_player);
 
-   for (c=0; c<16; c++)
+   for (c=0; c<16; c++) {
       all_notes_off(c);
+      all_sound_off(c);
+   }
 
    if (midi) {
       if (!midi_loaded_patches)
@@ -1268,8 +1305,10 @@ void midi_pause()
 
    remove_int(midi_player);
 
-   for (c=0; c<16; c++)
+   for (c=0; c<16; c++) {
       all_notes_off(c);
+      all_sound_off(c);
+   }
 }
 
 END_OF_FUNCTION(midi_pause);
@@ -1325,7 +1364,7 @@ int midi_seek(int target)
 
    /* save some variables and give temporary values */
    old_driver = midi_driver;
-   midi_driver = &midi_none;
+   midi_driver = &_midi_none;
    old_midi_loop = midi_loop;
    midi_loop = 0;
    old_midifile = midifile;
@@ -1338,7 +1377,7 @@ int midi_seek(int target)
       prepare_to_play(midifile);
 
    /* now sit back and let midi_player get to the position */
-   while ((midi_pos < target) && (midi_pos != -1)) {
+   while ((midi_pos < target) && (midi_pos >= 0)) {
       int mmpc = midi_pos_counter;
       int mmp = midi_pos;
 
@@ -1359,7 +1398,7 @@ int midi_seek(int target)
    midi_driver = old_driver;
    midi_seeking = 0;
 
-   if (midi_pos != -1) {
+   if (midi_pos >= 0) {
       /* refresh the driver with any changed parameters */
       if (midi_driver->raw_midi) {
 	 for (c=0; c<16; c++) {
@@ -1391,7 +1430,7 @@ int midi_seek(int target)
       return 0;
    }
 
-   if ((midi_loop) && (!midi_looping)) {  /* was file was looped? */
+   if ((midi_loop) && (!midi_looping)) {  /* was file looped? */
       prepare_to_play(old_midifile);
       install_int(midi_player, 20);
       return 2;                           /* seek past EOF => file restarted */
@@ -1417,7 +1456,7 @@ void midi_out(unsigned char *data, int length)
    _midi_tick++;
 
    while (pos < data+length)
-      process_midi_event(&pos, &running_status, &timer);
+      process_midi_event((AL_CONST unsigned char**) &pos, &running_status, &timer);
 
    update_controllers();
 
@@ -1452,7 +1491,7 @@ int load_midi_patches()
  *  Locks all the memory that the midi player touches inside the timer
  *  interrupt handler (which is most of it).
  */
-static void midi_lock_mem()
+static void midi_lock_mem(void)
 {
    LOCK_VARIABLE(midi_pos);
    LOCK_VARIABLE(midi_pos_counter);
@@ -1486,6 +1525,7 @@ static void midi_lock_mem()
    LOCK_FUNCTION(_midi_allocate_voice);
    LOCK_FUNCTION(midi_note_on);
    LOCK_FUNCTION(all_notes_off);
+   LOCK_FUNCTION(all_sound_off);
    LOCK_FUNCTION(reset_controllers);
    LOCK_FUNCTION(update_controllers);
    LOCK_FUNCTION(process_controller);
@@ -1505,13 +1545,17 @@ static void midi_lock_mem()
 /* midi_constructor:
  *  Register my functions with the code in sound.c.
  */
-#ifdef CONSTRUCTOR_FUNCTION
-   CONSTRUCTOR_FUNCTION(void _midi_constructor());
+#ifdef ALLEGRO_USE_CONSTRUCTOR
+   CONSTRUCTOR_FUNCTION(void _midi_constructor(void));
 #endif
 
-void _midi_constructor()
+static struct _AL_LINKER_MIDI midi_linker = {
+   midi_init,
+   midi_exit
+};
+
+void _midi_constructor(void)
 {
-   _midi_init = midi_init;
-   _midi_exit = midi_exit;
+   _al_linker_midi = &midi_linker;
 }
 
