@@ -24,6 +24,8 @@
  *
  *      Dynamic keyboard switching by Peter Cech.
  *
+ *      Peter Pavlovic improved the handling of accented keymaps.
+ *
  *      See readme.txt for copyright information.
  */
 
@@ -31,14 +33,15 @@
 #include <string.h>
 
 #include "allegro.h"
-#include "allegro/aintern.h"
+#include "allegro/internal/aintern.h"
 
 
 
-#define KB_MODIFIERS    (KB_SHIFT_FLAG | KB_CTRL_FLAG | KB_ALT_FLAG | KB_LWIN_FLAG | KB_RWIN_FLAG | KB_MENU_FLAG)
-#define KB_LED_FLAGS    (KB_SCROLOCK_FLAG | KB_NUMLOCK_FLAG | KB_CAPSLOCK_FLAG)
-#define KB_ACCENTS      (KB_ACCENT1_FLAG | KB_ACCENT1_S_FLAG | KB_ACCENT2_FLAG | KB_ACCENT2_S_FLAG)
-#define KB_CTRL_ALT     (KB_CTRL_FLAG | KB_ALT_FLAG)
+#define KB_MODIFIERS     (KB_SHIFT_FLAG | KB_CTRL_FLAG | KB_ALT_FLAG | KB_LWIN_FLAG | KB_RWIN_FLAG | KB_MENU_FLAG)
+#define KB_LED_FLAGS     (KB_SCROLOCK_FLAG | KB_NUMLOCK_FLAG | KB_CAPSLOCK_FLAG)
+#define KB_ACCENTS       (KB_ACCENT1_FLAG | KB_ACCENT2_FLAG | KB_ACCENT3_FLAG | KB_ACCENT4_FLAG)
+#define KB_SH_CTRL_ALT   (KB_SHIFT_FLAG | KB_CTRL_FLAG | KB_ALT_FLAG)
+#define KB_CTRL_ALT      (KB_CTRL_FLAG | KB_ALT_FLAG)
 
 
 static int key_extended = FALSE;
@@ -49,6 +52,13 @@ static int key_paused = FALSE;
 
 int _key_accent1 = 0;
 int _key_accent2 = 0;
+int _key_accent3 = 0;
+int _key_accent4 = 0;
+
+int _key_accent1_flag = 0;
+int _key_accent2_flag = 0;
+int _key_accent3_flag = 0;
+int _key_accent4_flag = 0;
 
 int _key_standard_kb = TRUE;
 
@@ -85,10 +95,10 @@ static unsigned char hw_to_mycode[128] =
    /* 0x64 */  0,              0,              0,              0,
    /* 0x68 */  0,              0,              0,              0, 
    /* 0x6C */  0,              0,              0,              0,
-   /* 0x70 */  0,              0,              0,              KEY_YEN, 
+   /* 0x70 */  KEY_KANA,       0,              0,              KEY_ABNT_C1, 
    /* 0x74 */  0,              0,              0,              0,
-   /* 0x78 */  0,              0,              0,              0, 
-   /* 0x7C */  0,              KEY_YEN2,       0,              0
+   /* 0x78 */  0,              KEY_CONVERT,    0,              KEY_NOCONVERT, 
+   /* 0x7C */  0,              KEY_YEN,        0,              0
 };
 
 
@@ -100,8 +110,8 @@ static unsigned char hw_to_mycode_ex[128] =
    /* 0x04 */  KEY_3,          KEY_4,          KEY_5,          KEY_6,
    /* 0x08 */  KEY_7,          KEY_8,          KEY_9,          KEY_0,
    /* 0x0C */  KEY_MINUS,      KEY_EQUALS,     KEY_BACKSPACE,  KEY_TAB,
-   /* 0x10 */  KEY_Q,          KEY_W,          KEY_E,          KEY_R,
-   /* 0x14 */  KEY_T,          KEY_Y,          KEY_U,          KEY_I,
+   /* 0x10 */  KEY_CIRCUMFLEX, KEY_AT,         KEY_COLON2,     KEY_R,
+   /* 0x14 */  KEY_KANJI,      KEY_Y,          KEY_U,          KEY_I,
    /* 0x18 */  KEY_O,          KEY_P,          KEY_OPENBRACE,  KEY_CLOSEBRACE,
    /* 0x1C */  KEY_ENTER_PAD,  KEY_RCONTROL,   KEY_A,          KEY_S,
    /* 0x20 */  KEY_D,          KEY_F,          KEY_G,          KEY_H,
@@ -163,7 +173,8 @@ static unsigned short standard_key_ascii_table[KEY_MAX] =
    /* misc chars */  27, '`', '-', '=', 8, 9, '[', ']', 13, ';', '\'', '\\', '\\', ',', '.', '/', ' ',
    /* controls */    0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
    /* numpad */      '/', '*', '-', '+', '.', 13,
-   /* modifiers */   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
+   /* others */      0, 0, 0, 0, 0, 0, 0, 0, 0, '\'', 0, 0, 0, 0, 0,
+   /* modifiers */   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
 };
 
 
@@ -179,7 +190,8 @@ static unsigned short standard_key_capslock_table[KEY_MAX] =
    /* misc chars */  27, '`', '-', '=', 8, 9, '[', ']', 13, ';', '\'', '\\', '\\', ',', '.', '/', ' ',
    /* controls */    0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
    /* numpad */      '/', '*', '-', '+', '.', 13,
-   /* modifiers */   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
+   /* others */      0, 0, 0, 0, 0, 0, 0, 0, 0, '\'', 0, 0, 0, 0, 0,
+   /* modifiers */   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
 };
 
 
@@ -192,10 +204,11 @@ static unsigned short standard_key_shift_table[KEY_MAX] =
    /* numbers */     ')', '!', '@', '#', '$', '%', '^', '&', '*', '(',
    /* numpad */      '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
    /* func keys */   0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
-   /* misc chars */  27, '~', '_', '+', 126, 126, '{', '}', 126, ':', '"', '|', '|', '<', '>', '?', 1,
+   /* misc chars */  27, '~', '_', '+', 8, 9, '{', '}', 13, ':', '"', '|', '|', '<', '>', '?', ' ',
    /* controls */    0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
-   /* numpad */      '/', '*', '-', '+', '.', 126,
-   /* modifiers */   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
+   /* numpad */      '/', '*', '-', '+', '.', 13,
+   /* others */      0, 0, 0, 0, 0, 0, 0, 0, 0, '"', 0, 0, 0, 0, 0,
+   /* modifiers */   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
 };
 
 
@@ -211,7 +224,8 @@ static unsigned short standard_key_control_table[KEY_MAX] =
    /* misc chars */  27, 2, 2, 2, 127, 127, 2, 2, 10, 2, 2, 2, 2, 2, 2, 2, 2,
    /* controls */    0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
    /* numpad */      2, 2, 2, 2, 2, 10,
-   /* modifiers */   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
+   /* others */      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+   /* modifiers */   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
 };
 
 
@@ -226,53 +240,55 @@ static unsigned short custom_key_ascii_table[KEY_MAX];
 static unsigned short custom_key_capslock_table[KEY_MAX];
 static unsigned short custom_key_shift_table[KEY_MAX];
 static unsigned short custom_key_control_table[KEY_MAX];
-static unsigned short custom_key_altgr_table[KEY_MAX];
+static unsigned short custom_key_altgr_lower_table[KEY_MAX];
+static unsigned short custom_key_altgr_upper_table[KEY_MAX];
 static unsigned short custom_key_accent1_lower_table[KEY_MAX];
 static unsigned short custom_key_accent1_upper_table[KEY_MAX];
-static unsigned short custom_key_accent1_shift_lower_table[KEY_MAX];
-static unsigned short custom_key_accent1_shift_upper_table[KEY_MAX];
 static unsigned short custom_key_accent2_lower_table[KEY_MAX];
 static unsigned short custom_key_accent2_upper_table[KEY_MAX];
-static unsigned short custom_key_accent2_shift_lower_table[KEY_MAX];
-static unsigned short custom_key_accent2_shift_upper_table[KEY_MAX];
-
+static unsigned short custom_key_accent3_lower_table[KEY_MAX];
+static unsigned short custom_key_accent3_upper_table[KEY_MAX];
+static unsigned short custom_key_accent4_lower_table[KEY_MAX];
+static unsigned short custom_key_accent4_upper_table[KEY_MAX];
 
 
 /* shortcut pointers to the current layout mappings */
-unsigned short *_key_ascii_table = standard_key_ascii_table;
-unsigned short *_key_capslock_table = standard_key_capslock_table;
-unsigned short *_key_shift_table = standard_key_shift_table;
-unsigned short *_key_control_table = standard_key_control_table;
-unsigned short *_key_altgr_table = standard_key_empty_table;
+unsigned short *_key_ascii_table         = standard_key_ascii_table;
+unsigned short *_key_capslock_table      = standard_key_capslock_table;
+unsigned short *_key_shift_table         = standard_key_shift_table;
+unsigned short *_key_control_table       = standard_key_control_table;
+unsigned short *_key_altgr_lower_table   = standard_key_empty_table;
+unsigned short *_key_altgr_upper_table   = standard_key_empty_table;
 unsigned short *_key_accent1_lower_table = standard_key_empty_table;
 unsigned short *_key_accent1_upper_table = standard_key_empty_table;
-unsigned short *_key_accent1_shift_lower_table = standard_key_empty_table;
-unsigned short *_key_accent1_shift_upper_table = standard_key_empty_table;
 unsigned short *_key_accent2_lower_table = standard_key_empty_table;
 unsigned short *_key_accent2_upper_table = standard_key_empty_table;
-unsigned short *_key_accent2_shift_lower_table = standard_key_empty_table;
-unsigned short *_key_accent2_shift_upper_table = standard_key_empty_table;
+unsigned short *_key_accent3_lower_table = standard_key_empty_table;
+unsigned short *_key_accent3_upper_table = standard_key_empty_table;
+unsigned short *_key_accent4_lower_table = standard_key_empty_table;
+unsigned short *_key_accent4_upper_table = standard_key_empty_table;
 
 
 
 /* set_standard_keyboard:
  *  Sets up pointers ready to use the standard US keyboard mapping.
  */
-static INLINE void set_standard_keyboard()
+static INLINE void set_standard_keyboard(void)
 {
-   _key_ascii_table = standard_key_ascii_table;
-   _key_capslock_table = standard_key_capslock_table;
-   _key_shift_table = standard_key_shift_table;
-   _key_control_table = standard_key_control_table;
-   _key_altgr_table = standard_key_empty_table;
+   _key_ascii_table         = standard_key_ascii_table;
+   _key_capslock_table      = standard_key_capslock_table;
+   _key_shift_table         = standard_key_shift_table;
+   _key_control_table       = standard_key_control_table;
+   _key_altgr_lower_table   = standard_key_empty_table;
+   _key_altgr_upper_table   = standard_key_empty_table;
    _key_accent1_lower_table = standard_key_empty_table;
    _key_accent1_upper_table = standard_key_empty_table;
-   _key_accent1_shift_lower_table = standard_key_empty_table;
-   _key_accent1_shift_upper_table = standard_key_empty_table;
    _key_accent2_lower_table = standard_key_empty_table;
    _key_accent2_upper_table = standard_key_empty_table;
-   _key_accent2_shift_lower_table = standard_key_empty_table;
-   _key_accent2_shift_upper_table = standard_key_empty_table;
+   _key_accent3_lower_table = standard_key_empty_table;
+   _key_accent3_upper_table = standard_key_empty_table;
+   _key_accent4_lower_table = standard_key_empty_table;
+   _key_accent4_upper_table = standard_key_empty_table;
 
    _key_standard_kb = TRUE;
 }
@@ -282,21 +298,22 @@ static INLINE void set_standard_keyboard()
 /* set_custom_keyboard:
  *  Sets up pointers ready to use the custom keyboard mapping.
  */
-static INLINE void set_custom_keyboard()
+static INLINE void set_custom_keyboard(void)
 {
-   _key_ascii_table = custom_key_ascii_table;
-   _key_capslock_table = custom_key_capslock_table;
-   _key_shift_table = custom_key_shift_table;
-   _key_control_table = custom_key_control_table;
-   _key_altgr_table = custom_key_altgr_table;
+   _key_ascii_table         = custom_key_ascii_table;
+   _key_capslock_table      = custom_key_capslock_table;
+   _key_shift_table         = custom_key_shift_table;
+   _key_control_table       = custom_key_control_table;
+   _key_altgr_lower_table   = custom_key_altgr_lower_table;
+   _key_altgr_upper_table   = custom_key_altgr_upper_table;
    _key_accent1_lower_table = custom_key_accent1_lower_table;
    _key_accent1_upper_table = custom_key_accent1_upper_table;
-   _key_accent1_shift_lower_table = custom_key_accent1_shift_lower_table;
-   _key_accent1_shift_upper_table = custom_key_accent1_shift_upper_table;
    _key_accent2_lower_table = custom_key_accent2_lower_table;
    _key_accent2_upper_table = custom_key_accent2_upper_table;
-   _key_accent2_shift_lower_table = custom_key_accent2_shift_lower_table;
-   _key_accent2_shift_upper_table = custom_key_accent2_shift_upper_table;
+   _key_accent3_lower_table = custom_key_accent3_lower_table;
+   _key_accent3_upper_table = custom_key_accent3_upper_table;
+   _key_accent4_lower_table = custom_key_accent4_lower_table;
+   _key_accent4_upper_table = custom_key_accent4_upper_table;
 
    _key_standard_kb = FALSE;
 }
@@ -310,7 +327,7 @@ int _pckey_scancode_to_ascii(int scancode)
 {
    int val;
 
-   if ((scancode < 0) && (scancode >= KEY_MAX))
+   if ((scancode < 0) || (scancode >= KEY_MAX))
       return 0;
 
    val = _key_ascii_table[scancode];
@@ -332,7 +349,7 @@ int _pckey_scancode_to_ascii(int scancode)
  */
 void _handle_pckey(int code)
 {
-   int origcode, mycode, flag, numflag, i;
+   int origcode, mycode, flag, numflag, i=0;
    unsigned short *table;
 
    if (key_pause_loop) { 
@@ -431,16 +448,25 @@ void _handle_pckey(int code)
       return;
    }
 
-   if (((mycode == _key_accent1) || (mycode == _key_accent2)) && (!_key_standard_kb)) {
-      if (key_altgr) {
-	 /* alt+gr overrides accent escapes */
-	 _handle_key_press(_key_altgr_table[mycode], origcode);
-	 return; 
+   /* new ACCENT stuff */
+   if (!_key_standard_kb) {
+      if ((mycode == _key_accent1) && ((_key_shifts & KB_SH_CTRL_ALT) == _key_accent1_flag)) {
+	 _key_shifts |= KB_ACCENT1_FLAG;
+	 _handle_key_press(-1, origcode);
+	 return;
       }
-      else {
-	 /* accent escape key */
-	 i = (mycode == _key_accent1) ? KB_ACCENT1_FLAG : KB_ACCENT2_FLAG;
-	 _key_shifts |= (_key_shifts & KB_SHIFT_FLAG) ? (i<<1) : i;
+      else if ((mycode == _key_accent2) && ((_key_shifts & KB_SH_CTRL_ALT) == _key_accent2_flag)) {
+	 _key_shifts |= KB_ACCENT2_FLAG;
+	 _handle_key_press(-1, origcode);
+	 return;
+      }
+      else if ((mycode == _key_accent3) && ((_key_shifts & KB_SH_CTRL_ALT) == _key_accent3_flag)) {
+	 _key_shifts |= KB_ACCENT3_FLAG;
+	 _handle_key_press(-1, origcode);
+	 return;
+      }
+      else if ((mycode == _key_accent4) && ((_key_shifts & KB_SH_CTRL_ALT) == _key_accent4_flag)) {
+	 _key_shifts |= KB_ACCENT4_FLAG;
 	 _handle_key_press(-1, origcode);
 	 return;
       }
@@ -451,24 +477,24 @@ void _handle_pckey(int code)
       if (((_key_shifts & KB_SHIFT_FLAG) != 0) ^ ((_key_shifts & KB_CAPSLOCK_FLAG) != 0)) {
 	 if (_key_shifts & KB_ACCENT1_FLAG)
 	    table = _key_accent1_upper_table;
-	 else if (_key_shifts & KB_ACCENT1_S_FLAG)
-	    table = _key_accent1_shift_upper_table;
 	 else if (_key_shifts & KB_ACCENT2_FLAG)
 	    table = _key_accent2_upper_table;
-	 else if (_key_shifts & KB_ACCENT2_S_FLAG)
-	    table = _key_accent2_shift_upper_table;
+	 else if (_key_shifts & KB_ACCENT3_FLAG)
+	    table = _key_accent3_upper_table;
+	 else if (_key_shifts & KB_ACCENT4_FLAG)
+	    table = _key_accent4_upper_table;
 	 else
 	    table = NULL;
       }
       else {
 	 if (_key_shifts & KB_ACCENT1_FLAG)
 	    table = _key_accent1_lower_table;
-	 else if (_key_shifts & KB_ACCENT1_S_FLAG)
-	    table = _key_accent1_shift_lower_table;
 	 else if (_key_shifts & KB_ACCENT2_FLAG)
 	    table = _key_accent2_lower_table;
-	 else if (_key_shifts & KB_ACCENT2_S_FLAG)
-	    table = _key_accent2_shift_lower_table;
+	 else if (_key_shifts & KB_ACCENT3_FLAG)
+	    table = _key_accent3_lower_table;
+	 else if (_key_shifts & KB_ACCENT4_FLAG)
+	    table = _key_accent4_lower_table;
 	 else
 	    table = NULL;
       }
@@ -481,7 +507,11 @@ void _handle_pckey(int code)
       }
       else {
 	 /* add the accent as an individual character */
-	 i = (_key_shifts & (KB_ACCENT1_FLAG | KB_ACCENT1_S_FLAG)) ? _key_accent1 : _key_accent2;
+	 if      (_key_shifts & (KB_ACCENT1_FLAG)) i = _key_accent1;
+	 else if (_key_shifts & (KB_ACCENT2_FLAG)) i = _key_accent2;
+	 else if (_key_shifts & (KB_ACCENT3_FLAG)) i = _key_accent3;
+	 else if (_key_shifts & (KB_ACCENT4_FLAG)) i = _key_accent4;
+
 	 _handle_key_press(_key_ascii_table[i], i);
 	 _key_shifts &= ~KB_ACCENTS;
       }
@@ -504,8 +534,12 @@ void _handle_pckey(int code)
 	 /* alt+key */
 	 if (_key_ascii_table[mycode] == 0xFFFF)
 	    i = 0xFFFF;
-	 else if (key_altgr)
-	    i = _key_altgr_table[mycode];
+         else if (key_altgr) {
+            if (((_key_shifts & KB_SHIFT_FLAG) != 0) ^ ((_key_shifts & KB_CAPSLOCK_FLAG) != 0))
+               i = _key_altgr_upper_table[mycode];
+            else
+               i = _key_altgr_lower_table[mycode];
+         }
 	 else
 	    i = 0;
       }
@@ -572,13 +606,13 @@ END_OF_FUNCTION(_handle_pckey);
  */
 static void read_key_table(unsigned short *out, unsigned short *in, char *section)
 {
-   char tmp1[80], tmp2[80], name[80];
+   char tmp1[64], tmp2[128], name[128];
    char *fmt = uconvert_ascii("key%d", tmp1);
    char *sec = uconvert_ascii(section, tmp2);
    int i;
 
    for (i=0; i<KEY_MAX; i++) {
-      usprintf(name, fmt, i);
+      uszprintf(name, sizeof(name), fmt, i);
       out[i] = get_config_int(sec, name, in[i]);
    }
 }
@@ -588,11 +622,12 @@ static void read_key_table(unsigned short *out, unsigned short *in, char *sectio
 /* read_keyboard_config:
  *  Reads in the keyboard config tables.
  */
-static void read_keyboard_config()
+static void read_keyboard_config(void)
 {
-   char filename[512], tmp1[128], tmp2[128], *name, *ext, *datafile;
+   char filename[1024], tmp1[128], tmp2[128], *ext, *datafile;
+   AL_CONST char* name;
 
-   name = get_config_string(NULL, uconvert_ascii("keyboard", tmp1), NULL);
+   name = get_config_string(uconvert_ascii("system", tmp1), uconvert_ascii("keyboard", tmp2), NULL);
 
    if ((!name) || (!ugetc(name)))
       return;
@@ -606,22 +641,36 @@ static void read_keyboard_config()
    push_config_state();
    set_config_file(filename);
 
-   read_key_table(custom_key_ascii_table,                standard_key_ascii_table,     "key_ascii");
-   read_key_table(custom_key_capslock_table,             standard_key_capslock_table,  "key_capslock");
-   read_key_table(custom_key_shift_table,                standard_key_shift_table,     "key_shift");
-   read_key_table(custom_key_control_table,              standard_key_control_table,   "key_control");
-   read_key_table(custom_key_altgr_table,                standard_key_empty_table,     "key_altgr");
-   read_key_table(custom_key_accent1_lower_table,        standard_key_empty_table,     "key_accent1_lower");
-   read_key_table(custom_key_accent1_upper_table,        standard_key_empty_table,     "key_accent1_upper");
-   read_key_table(custom_key_accent1_shift_lower_table,  standard_key_empty_table,     "key_accent1_shift_lower");
-   read_key_table(custom_key_accent1_shift_upper_table,  standard_key_empty_table,     "key_accent1_shift_upper");
-   read_key_table(custom_key_accent2_lower_table,        standard_key_empty_table,     "key_accent2_lower");
-   read_key_table(custom_key_accent2_upper_table,        standard_key_empty_table,     "key_accent2_upper");
-   read_key_table(custom_key_accent2_shift_lower_table,  standard_key_empty_table,     "key_accent2_shift_lower");
-   read_key_table(custom_key_accent2_shift_upper_table,  standard_key_empty_table,     "key_accent2_shift_upper");
+   read_key_table(custom_key_ascii_table,          standard_key_ascii_table,     "key_ascii");
+   read_key_table(custom_key_capslock_table,       standard_key_capslock_table,  "key_capslock");
+   read_key_table(custom_key_shift_table,          standard_key_shift_table,     "key_shift");
+   read_key_table(custom_key_control_table,        standard_key_control_table,   "key_control");
+
+   /* preserve backward compatibility with former unique key_altgr table */
+   read_key_table(custom_key_altgr_lower_table,    standard_key_empty_table,     "key_altgr");
+   read_key_table(custom_key_altgr_upper_table,    standard_key_empty_table,     "key_altgr");
+
+   read_key_table(custom_key_altgr_lower_table,    custom_key_altgr_lower_table, "key_altgr_lower");
+   read_key_table(custom_key_altgr_upper_table,    custom_key_altgr_upper_table, "key_altgr_upper");
+
+   read_key_table(custom_key_accent1_lower_table,  standard_key_empty_table,     "key_accent1_lower");
+   read_key_table(custom_key_accent1_upper_table,  standard_key_empty_table,     "key_accent1_upper");
+   read_key_table(custom_key_accent2_lower_table,  standard_key_empty_table,     "key_accent2_lower");
+   read_key_table(custom_key_accent2_upper_table,  standard_key_empty_table,     "key_accent2_upper");
+   read_key_table(custom_key_accent3_lower_table,  standard_key_empty_table,     "key_accent3_lower");
+   read_key_table(custom_key_accent3_upper_table,  standard_key_empty_table,     "key_accent3_upper");
+   read_key_table(custom_key_accent4_lower_table,  standard_key_empty_table,     "key_accent4_lower");
+   read_key_table(custom_key_accent4_upper_table,  standard_key_empty_table,     "key_accent4_upper");
 
    _key_accent1 = get_config_int(uconvert_ascii("key_escape", tmp1), uconvert_ascii("accent1", tmp2), 0);
    _key_accent2 = get_config_int(uconvert_ascii("key_escape", tmp1), uconvert_ascii("accent2", tmp2), 0);
+   _key_accent3 = get_config_int(uconvert_ascii("key_escape", tmp1), uconvert_ascii("accent3", tmp2), 0);
+   _key_accent4 = get_config_int(uconvert_ascii("key_escape", tmp1), uconvert_ascii("accent4", tmp2), 0);
+
+   _key_accent1_flag = get_config_int(uconvert_ascii("key_escape", tmp1), uconvert_ascii("accent1_flag", tmp2), 0);
+   _key_accent2_flag = get_config_int(uconvert_ascii("key_escape", tmp1), uconvert_ascii("accent2_flag", tmp2), 0);
+   _key_accent3_flag = get_config_int(uconvert_ascii("key_escape", tmp1), uconvert_ascii("accent3_flag", tmp2), 0);
+   _key_accent4_flag = get_config_int(uconvert_ascii("key_escape", tmp1), uconvert_ascii("accent4_flag", tmp2), 0);
 
    pop_config_state();
 
@@ -640,6 +689,12 @@ void _pckeys_init()
 
    _key_accent1 = 0;
    _key_accent2 = 0;
+   _key_accent3 = 0;
+   _key_accent4 = 0;
+   _key_accent1_flag = 0;
+   _key_accent2_flag = 0;
+   _key_accent3_flag = 0;
+   _key_accent4_flag = 0;
 
    key_extended = FALSE;
    key_altgr = FALSE;
@@ -650,19 +705,20 @@ void _pckeys_init()
    for (i=0; i<KEY_MAX; i++)
       standard_key_empty_table[i] = 0;
 
-   memcpy(custom_key_ascii_table,                standard_key_ascii_table,     sizeof(custom_key_ascii_table));
-   memcpy(custom_key_capslock_table,             standard_key_capslock_table,  sizeof(custom_key_capslock_table));
-   memcpy(custom_key_shift_table,                standard_key_shift_table,     sizeof(custom_key_shift_table));
-   memcpy(custom_key_control_table,              standard_key_control_table,   sizeof(custom_key_control_table));
-   memcpy(custom_key_altgr_table,                standard_key_empty_table,     sizeof(custom_key_altgr_table));
-   memcpy(custom_key_accent1_lower_table,        standard_key_empty_table,     sizeof(custom_key_accent1_lower_table));
-   memcpy(custom_key_accent1_upper_table,        standard_key_empty_table,     sizeof(custom_key_accent1_upper_table));
-   memcpy(custom_key_accent1_shift_lower_table,  standard_key_empty_table,     sizeof(custom_key_accent1_shift_lower_table));
-   memcpy(custom_key_accent1_shift_upper_table,  standard_key_empty_table,     sizeof(custom_key_accent1_shift_upper_table));
-   memcpy(custom_key_accent2_lower_table,        standard_key_empty_table,     sizeof(custom_key_accent2_lower_table));
-   memcpy(custom_key_accent2_upper_table,        standard_key_empty_table,     sizeof(custom_key_accent2_upper_table));
-   memcpy(custom_key_accent2_shift_lower_table,  standard_key_empty_table,     sizeof(custom_key_accent2_shift_lower_table));
-   memcpy(custom_key_accent2_shift_upper_table,  standard_key_empty_table,     sizeof(custom_key_accent2_shift_upper_table));
+   memcpy(custom_key_ascii_table,            standard_key_ascii_table,     sizeof(custom_key_ascii_table));
+   memcpy(custom_key_capslock_table,         standard_key_capslock_table,  sizeof(custom_key_capslock_table));
+   memcpy(custom_key_shift_table,            standard_key_shift_table,     sizeof(custom_key_shift_table));
+   memcpy(custom_key_control_table,          standard_key_control_table,   sizeof(custom_key_control_table));
+   memcpy(custom_key_altgr_lower_table,      standard_key_empty_table,     sizeof(custom_key_altgr_lower_table));
+   memcpy(custom_key_altgr_upper_table,      standard_key_empty_table,     sizeof(custom_key_altgr_upper_table));
+   memcpy(custom_key_accent1_lower_table,    standard_key_empty_table,     sizeof(custom_key_accent1_lower_table));
+   memcpy(custom_key_accent1_upper_table,    standard_key_empty_table,     sizeof(custom_key_accent1_upper_table));
+   memcpy(custom_key_accent2_lower_table,    standard_key_empty_table,     sizeof(custom_key_accent2_lower_table));
+   memcpy(custom_key_accent2_upper_table,    standard_key_empty_table,     sizeof(custom_key_accent2_upper_table));
+   memcpy(custom_key_accent3_lower_table,    standard_key_empty_table,     sizeof(custom_key_accent1_lower_table));
+   memcpy(custom_key_accent3_upper_table,    standard_key_empty_table,     sizeof(custom_key_accent1_upper_table));
+   memcpy(custom_key_accent4_lower_table,    standard_key_empty_table,     sizeof(custom_key_accent2_lower_table));
+   memcpy(custom_key_accent4_upper_table,    standard_key_empty_table,     sizeof(custom_key_accent2_upper_table));
 
    LOCK_VARIABLE(hw_to_mycode);
    LOCK_VARIABLE(hw_to_mycode_ex);
@@ -677,28 +733,30 @@ void _pckeys_init()
    LOCK_VARIABLE(custom_key_capslock_table);
    LOCK_VARIABLE(custom_key_shift_table);
    LOCK_VARIABLE(custom_key_control_table);
-   LOCK_VARIABLE(custom_key_altgr_table);
+   LOCK_VARIABLE(custom_key_altgr_lower_table);
+   LOCK_VARIABLE(custom_key_altgr_upper_table);
    LOCK_VARIABLE(custom_key_accent1_lower_table);
    LOCK_VARIABLE(custom_key_accent1_upper_table);
-   LOCK_VARIABLE(custom_key_accent1_shift_lower_table);
-   LOCK_VARIABLE(custom_key_accent1_shift_upper_table);
    LOCK_VARIABLE(custom_key_accent2_lower_table);
    LOCK_VARIABLE(custom_key_accent2_upper_table);
-   LOCK_VARIABLE(custom_key_accent2_shift_lower_table);
-   LOCK_VARIABLE(custom_key_accent2_shift_upper_table);
+   LOCK_VARIABLE(custom_key_accent3_lower_table);
+   LOCK_VARIABLE(custom_key_accent3_upper_table);
+   LOCK_VARIABLE(custom_key_accent4_lower_table);
+   LOCK_VARIABLE(custom_key_accent4_upper_table);
    LOCK_VARIABLE(_key_ascii_table);
    LOCK_VARIABLE(_key_capslock_table);
    LOCK_VARIABLE(_key_shift_table);
    LOCK_VARIABLE(_key_control_table);
-   LOCK_VARIABLE(_key_altgr_table);
+   LOCK_VARIABLE(_key_altgr_lower_table);
+   LOCK_VARIABLE(_key_altgr_upper_table);
    LOCK_VARIABLE(_key_accent1_lower_table);
    LOCK_VARIABLE(_key_accent1_upper_table);
-   LOCK_VARIABLE(_key_accent1_shift_lower_table);
-   LOCK_VARIABLE(_key_accent1_shift_upper_table);
    LOCK_VARIABLE(_key_accent2_lower_table);
    LOCK_VARIABLE(_key_accent2_upper_table);
-   LOCK_VARIABLE(_key_accent2_shift_lower_table);
-   LOCK_VARIABLE(_key_accent2_shift_upper_table);
+   LOCK_VARIABLE(_key_accent3_lower_table);
+   LOCK_VARIABLE(_key_accent3_upper_table);
+   LOCK_VARIABLE(_key_accent4_lower_table);
+   LOCK_VARIABLE(_key_accent4_upper_table);
    LOCK_VARIABLE(_key_standard_kb);
    LOCK_VARIABLE(key_extended);
    LOCK_VARIABLE(key_altgr);
@@ -707,6 +765,12 @@ void _pckeys_init()
    LOCK_VARIABLE(key_paused);
    LOCK_VARIABLE(_key_accent1);
    LOCK_VARIABLE(_key_accent2);
+   LOCK_VARIABLE(_key_accent3);
+   LOCK_VARIABLE(_key_accent4);
+   LOCK_VARIABLE(_key_accent1_flag);
+   LOCK_VARIABLE(_key_accent2_flag);
+   LOCK_VARIABLE(_key_accent3_flag);
+   LOCK_VARIABLE(_key_accent4_flag);
    LOCK_FUNCTION(_handle_pckey);
 
    _key_standard_kb = TRUE;
