@@ -67,13 +67,13 @@
 
 #if (defined ALLEGRO_DOS) || (defined ALLEGRO_LINUX_VBEAF)
 
-#include "allegro/aintern.h"
+#include "allegro/internal/aintern.h"
 
 #ifdef ALLEGRO_INTERNAL_HEADER
    #include ALLEGRO_INTERNAL_HEADER
 #endif
 
-#ifdef ALLEGRO_DJGPP
+#if (defined ALLEGRO_DJGPP) && (!defined SCAN_DEPEND)
    #include <crt0.h>
    #include <sys/nearptr.h>
    #include <sys/exceptn.h>
@@ -86,9 +86,10 @@ static BITMAP *vbeaf_init(int w, int h, int v_w, int v_h, int color_depth);
 static void vbeaf_exit(BITMAP *b);
 static void vbeaf_save(void);
 static void vbeaf_restore(void);
+static GFX_MODE_LIST *vbeaf_fetch_mode_list(void);
 static void vbeaf_vsync(void);
 static int vbeaf_scroll(int x, int y);
-static void vbeaf_set_palette_range(PALETTE p, int from, int to, int vsync);
+static void vbeaf_set_palette_range(AL_CONST PALETTE p, int from, int to, int vsync);
 static int vbeaf_request_scroll(int x, int y);
 static int vbeaf_poll_scroll(void);
 static int vbeaf_set_mouse_sprite(BITMAP *sprite, int xfocus, int yfocus);
@@ -96,6 +97,8 @@ static int vbeaf_show_mouse(BITMAP *bmp, int x, int y);
 static void vbeaf_hide_mouse(void);
 static void vbeaf_move_mouse(int x, int y);
 static void vbeaf_drawing_mode(void);
+static int vbeaf_locate_driver(void);
+static int vbeaf_lowlevel_init(void);
 
 
 /* accelerated drawing functions */
@@ -105,7 +108,7 @@ static void vbeaf_vline_b(BITMAP *bmp, int x, int y1, int y2, int color);
 static void vbeaf_line(BITMAP *bmp, int x1, int y1, int x2, int y2, int color);
 static void vbeaf_rectfill(BITMAP *bmp, int x1, int y1, int x2, int y2, int color);
 static int  vbeaf_triangle(BITMAP *bmp, int x1, int y1, int x2, int y2, int x3, int y3, int color);
-static void vbeaf_draw_glyph(BITMAP *bmp, FONT_GLYPH *glyph, int x, int y, int color);
+static void vbeaf_draw_glyph(BITMAP *bmp, AL_CONST FONT_GLYPH *glyph, int x, int y, int color, int bg);
 static void vbeaf_draw_sprite(BITMAP *bmp, BITMAP *sprite, int x, int y);
 static void vbeaf_blit_from_memory(BITMAP *source, BITMAP *dest, int source_x, int source_y, int dest_x, int dest_y, int width, int height);
 static void vbeaf_blit_to_self(BITMAP *source, BITMAP *dest, int source_x, int source_y, int dest_x, int dest_y, int width, int height);
@@ -125,12 +128,12 @@ static void (*orig_vline)(BITMAP *bmp, int x, int y1, int y2, int color);
 static void (*orig_hline)(BITMAP *bmp, int x1, int y, int x2, int color);
 static void (*orig_line)(BITMAP *bmp, int x1, int y1, int x2, int y2, int color);
 static void (*orig_rectfill)(BITMAP *bmp, int x1, int y1, int x2, int y2, int color);
-static void (*orig_draw_glyph)(BITMAP *bmp, FONT_GLYPH *glyph, int x, int y, int color);
+static void (*orig_draw_glyph)(BITMAP *bmp, AL_CONST FONT_GLYPH *glyph, int x, int y, int color, int bg);
 static void (*orig_draw_sprite)(BITMAP *bmp, BITMAP *sprite, int x, int y);
 static void (*orig_masked_blit)(BITMAP *source, BITMAP *dest, int source_x, int source_y, int dest_x, int dest_y, int width, int height);
 
 
-static char vbeaf_desc[256] = "";
+static char vbeaf_desc[256] = EMPTY_STRING;
 
 
 
@@ -155,7 +158,8 @@ GFX_DRIVER gfx_vbeaf =
    NULL,                         /* no drawing mode hook */
    vbeaf_save,
    vbeaf_restore,
-   0, 0, FALSE, 0, 0, 0, 0
+   vbeaf_fetch_mode_list,        /* fetch mode hook */
+   0, 0, FALSE, 0, 0, 0, 0, FALSE
 };
 
 
@@ -539,7 +543,7 @@ static int saved_vw;
 static int saved_vh;
 static int saved_wret;
 
-extern void _af_int86(), _af_call_rm(), _af_wrapper(), _af_wrapper_end();
+extern void _af_int86(void), _af_call_rm(void), _af_wrapper(void), _af_wrapper_end(void);
 
 
 
@@ -675,7 +679,7 @@ static int call_vbeaf_asm(void *proc)
 /* vbeaf_no_wait:
  *  Dummy wait-till-idle routine for non-accelerated drivers.
  */
-static void vbeaf_no_wait()
+static void vbeaf_no_wait(void)
 {
 }
 
@@ -687,7 +691,7 @@ END_OF_STATIC_FUNCTION(vbeaf_no_wait);
  *  Tries to load the specified VBE/AF driver file, returning TRUE on 
  *  success. Allocates memory and reads the driver into it.
  */
-static int load_vbeaf_driver(char *filename)
+static int load_vbeaf_driver(AL_CONST char *filename)
 {
    long size;
    PACKFILE *f;
@@ -702,7 +706,7 @@ static int load_vbeaf_driver(char *filename)
 	 return 0;
 
       if ((s.st_uid != 0) || (s.st_mode & (S_IWGRP | S_IWOTH))) {
-	 usprintf(allegro_error, get_config_text("%s must only be writeable by root"), filename);
+	 uszprintf(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("%s must only be writeable by root"), filename);
 	 return -1;
       }
 
@@ -743,7 +747,7 @@ static int load_vbeaf_driver(char *filename)
 /* initialise_freebeaf_extensions:
  *  Prepares the FreeBE/AF extension functions.
  */
-static void initialise_freebeaf_extensions()
+static void initialise_freebeaf_extensions(void)
 {
    typedef unsigned long (*EXT_INIT_FUNC)(AF_DRIVER *af, unsigned long id);
    EXT_INIT_FUNC ext_init;
@@ -796,7 +800,7 @@ static void initialise_freebeaf_extensions()
  *  Sets up the DPMI memory mappings required by the VBE/AF driver, 
  *  returning zero on success.
  */
-static int initialise_vbeaf_driver()
+static int initialise_vbeaf_driver(void)
 {
    int c;
 
@@ -942,7 +946,7 @@ static int find_vbeaf_mode(int w, int h, int v_w, int v_h, int color_depth, AF_M
 	    /* make sure the mode supports scrolling */
 	    if ((v_w > w) || (v_h > h)) {
 	       if (!(mode_info->Attributes & 2)) {
-		  ustrcpy(allegro_error, get_config_text("Hardware scrolling not supported"));
+		  ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Hardware scrolling not supported"));
 		  return 0;
 	       }
 	    }
@@ -952,7 +956,7 @@ static int find_vbeaf_mode(int w, int h, int v_w, int v_h, int color_depth, AF_M
 	     * call and use VBE 3.0 instead.
 	     */
 	    if ((!(mode_info->Attributes & 16)) && (!faf_id)) {
-	       ustrcpy(allegro_error, get_config_text("Hardware acceleration not available"));
+	       ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Hardware acceleration not available"));
 	       return 0;
 	    }
 
@@ -961,7 +965,7 @@ static int find_vbeaf_mode(int w, int h, int v_w, int v_h, int color_depth, AF_M
       }
    }
 
-   ustrcpy(allegro_error, get_config_text("Resolution not supported"));
+   ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Resolution not supported"));
    return 0;
 }
 
@@ -1006,11 +1010,10 @@ static int set_vbeaf_mode(int mode, int w, int h, int v_w, int v_h, int *width,
 
 
 
-/* vbeaf_init:
- *  Tries to enter the specified graphics mode, and makes a screen bitmap
- *  for it.
+/* vbeaf_locate_driver:
+ *  locates and loads a VBE/AF driver.
  */
-static BITMAP *vbeaf_init(int w, int h, int v_w, int v_h, int color_depth)
+static int vbeaf_locate_driver(void)
 {
    static char *possible_filenames[] =
    {
@@ -1025,22 +1028,15 @@ static BITMAP *vbeaf_init(int w, int h, int v_w, int v_h, int color_depth)
       NULL
    };
 
-   BITMAP *b;
-   AF_MODE_INFO mode_info;
-   char filename[256], tmp1[128], tmp2[128];
-   int bpp = BYTES_PER_PIXEL(color_depth);
-   int bytes_per_scanline, width, height;
-   int scrollable, mode, attrib;
-   int vseg, ret, i;
-   int rs, gs, bs;
-   void *vaddr;
-   char *p;
+   char filename[1024], tmp1[1024], tmp2[128];
+   AL_CONST char *p;
+   int ret, i, attrib;
 
    /* look for driver in the config file location */
-   p = get_config_string(NULL, uconvert_ascii("vbeaf_driver", tmp1), NULL);
+   p = get_config_string(uconvert_ascii("graphics", tmp1), uconvert_ascii("vbeaf_driver", tmp2), NULL);
 
    if ((p) && (ugetc(p))) {
-      ustrcpy(filename, p);
+      ustrzcpy(filename, sizeof(filename), p);
 
       if (ugetc(get_filename(filename)) == 0) {
 	 append_filename(filename, filename, uconvert_ascii("vbeaf.drv", tmp1), sizeof(filename));
@@ -1081,13 +1077,109 @@ static BITMAP *vbeaf_init(int w, int h, int v_w, int v_h, int color_depth)
    }
 
    /* oops, no driver */
-   ustrcpy(allegro_error, get_config_text("Can't find VBEAF.DRV"));
-   return NULL;
+   ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Can't find VBEAF.DRV"));
+   ret = FALSE;
 
    /* got it! */ 
    found_it:
 
-   if (ret < 0)
+   return ret;
+}
+
+
+
+/*  vbeaf_lowlevel_init:
+ *   loads the VBE/AF driver and initializes it.
+ */
+static int vbeaf_lowlevel_init(void)
+{
+   int ret;
+
+   /* check the driver ID string */
+   if (strcmp(af_driver->Signature, "VBEAF.DRV") != 0) {
+      vbeaf_exit(NULL);
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Bad VBE/AF driver ID string"));
+      return FALSE;
+   }
+
+   /* check the VBE/AF version number */
+   if (af_driver->Version < 0x200) {
+      vbeaf_exit(NULL);
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Obsolete VBE/AF version (need 2.0 or greater)"));
+      return FALSE;
+   }
+
+   /* to run on Linux, we need to be God */
+   #ifdef ALLEGRO_LINUX
+
+      if (!__al_linux_have_ioperms) {
+	 vbeaf_exit(NULL);
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("This driver needs root privileges"));
+	 return FALSE;
+      }
+
+      _vbeaf_selector = _default_ds();
+
+   #endif
+
+   /* detect and initialise the FreeBE/AF extensions */
+   if (strstr(af_driver->OemVendorName, "FreeBE")) {
+      faf_id = TRUE;
+      initialise_freebeaf_extensions();
+   }
+   else {
+      faf_id = FALSE;
+      faf_ext = 0;
+   }
+
+   /* special setup for Plug and Play hardware */
+   if (call_vbeaf_asm(af_driver->PlugAndPlayInit) != 0) {
+      vbeaf_exit(NULL);
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("VBE/AF Plug and Play initialisation failed"));
+      return FALSE;
+   }
+
+   /* deal with all that DPMI memory mapping crap */
+   ret = initialise_vbeaf_driver();
+   if (ret != 0) {
+      vbeaf_exit(NULL);
+      if (ret == -2)
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("VBE/AF nearptrs not supported on this platform"));
+      else
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Can't map memory for VBE/AF"));
+      return FALSE;
+   }
+
+   /* low level driver initialisation */
+   if (call_vbeaf_asm(af_driver->InitDriver) != 0) {
+      vbeaf_exit(NULL);
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("VBE/AF device not present"));
+      return FALSE;
+   }
+
+   return TRUE;
+}
+
+
+
+/* vbeaf_init:
+ *  Tries to enter the specified graphics mode, and makes a screen bitmap
+ *  for it.
+ */
+static BITMAP *vbeaf_init(int w, int h, int v_w, int v_h, int color_depth)
+{
+   BITMAP *b;
+   AF_MODE_INFO mode_info;
+   int bpp = BYTES_PER_PIXEL(color_depth);
+   int bytes_per_scanline, width, height;
+   int scrollable, mode;
+   int vseg;
+   int rs, gs, bs;
+   char tmp1[512];
+   void *vaddr;
+
+   /* locate and load VBE/AF driver from disk */
+   if (!vbeaf_locate_driver())
       return NULL;
 
    LOCK_VARIABLE(af_driver);
@@ -1106,72 +1198,14 @@ static BITMAP *vbeaf_init(int w, int h, int v_w, int v_h, int color_depth)
    LOCK_FUNCTION(vbeaf_draw_sprite);
    LOCK_FUNCTION(vbeaf_blit_from_memory);
 
-   /* check the driver ID string */
-   if (strcmp(af_driver->Signature, "VBEAF.DRV") != 0) {
-      vbeaf_exit(NULL);
-      ustrcpy(allegro_error, get_config_text("Bad VBE/AF driver ID string"));
+   /* init the currently loaded VBE/AF driver */
+   if (!vbeaf_lowlevel_init())
       return NULL;
-   }
-
-   /* check the VBE/AF version number */
-   if (af_driver->Version < 0x200) {
-      vbeaf_exit(NULL);
-      ustrcpy(allegro_error, get_config_text("Obsolete VBE/AF version (need 2.0 or greater)"));
-      return NULL;
-   }
-
-   /* to run on Linux, we need to be God */
-   #ifdef ALLEGRO_LINUX
-
-      if (!__al_linux_have_ioperms) {
-	 vbeaf_exit(NULL);
-	 ustrcpy(allegro_error, get_config_text("This driver needs root privileges"));
-	 return NULL;
-      }
-
-      _vbeaf_selector = _default_ds();
-
-   #endif
-
-   /* detect and initialise the FreeBE/AF extensions */
-   if (strstr(af_driver->OemVendorName, "FreeBE")) {
-      faf_id = TRUE;
-      initialise_freebeaf_extensions();
-   }
-   else {
-      faf_id = FALSE;
-      faf_ext = 0;
-   }
 
    /* bodge to work around bugs in the present (6.51) version of UniVBE */
    if ((v_w > w) && (!faf_id)) {
       vbeaf_exit(NULL);
-      ustrcpy(allegro_error, get_config_text("SciTech VBE/AF drivers do not support wide virtual screens"));
-      return NULL; 
-   }
-
-   /* special setup for Plug and Play hardware */
-   if (call_vbeaf_asm(af_driver->PlugAndPlayInit) != 0) {
-      vbeaf_exit(NULL);
-      ustrcpy(allegro_error, get_config_text("VBE/AF Plug and Play initialisation failed"));
-      return NULL;
-   }
-
-   /* deal with all that DPMI memory mapping crap */
-   ret = initialise_vbeaf_driver();
-   if (ret != 0) {
-      vbeaf_exit(NULL);
-      if (ret == -2)
-	 ustrcpy(allegro_error, get_config_text("VBE/AF nearptrs not supported on this platform"));
-      else
-	 ustrcpy(allegro_error, get_config_text("Can't map memory for VBE/AF"));
-      return NULL;
-   }
-
-   /* low level driver initialisation */
-   if (call_vbeaf_asm(af_driver->InitDriver) != 0) {
-      vbeaf_exit(NULL);
-      ustrcpy(allegro_error, get_config_text("VBE/AF device not present"));
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("SciTech VBE/AF drivers do not support wide virtual screens"));
       return NULL;
    }
 
@@ -1225,14 +1259,14 @@ static BITMAP *vbeaf_init(int w, int h, int v_w, int v_h, int color_depth)
    _sort_out_virtual_width(&width, &gfx_vbeaf);
 
    if (width * height > gfx_vbeaf.vid_mem) {
-      ustrcpy(allegro_error, get_config_text("Insufficient video memory"));
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Insufficient video memory"));
       vbeaf_exit(NULL);
       return NULL;
    }
 
    /* set the mode */
    if (set_vbeaf_mode(mode, w, h, width/bpp, height, &width, &scrollable) != 0) {
-      ustrcpy(allegro_error, get_config_text("Failed to set VBE/AF mode"));
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Failed to set VBE/AF mode"));
       vbeaf_exit(NULL);
       return NULL;
    }
@@ -1244,7 +1278,7 @@ static BITMAP *vbeaf_init(int w, int h, int v_w, int v_h, int color_depth)
       height = MAX(height, (int)af_driver->OffscreenEndY+1);
 
    if ((width/bpp < v_w) || (width/bpp < w) || (height < v_h) || (height < h)) {
-      ustrcpy(allegro_error, get_config_text("Virtual screen size too large"));
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Virtual screen size too large"));
       vbeaf_exit(NULL);
       return NULL;
    }
@@ -1447,30 +1481,31 @@ static BITMAP *vbeaf_init(int w, int h, int v_w, int v_h, int color_depth)
    }
 
    /* set up the VBE/AF description string */
-   ustrcpy(vbeaf_desc, uconvert_ascii(af_driver->OemVendorName, NULL));
+   ustrzcpy(vbeaf_desc, sizeof(vbeaf_desc), uconvert_ascii(af_driver->OemVendorName, tmp1));
 
    if (gfx_vbeaf.linear)
-      ustrcat(vbeaf_desc, uconvert_ascii(", linear", NULL));
+      ustrzcat(vbeaf_desc, sizeof(vbeaf_desc), uconvert_ascii(", linear", tmp1));
    else
-      ustrcat(vbeaf_desc, uconvert_ascii(", banked", NULL));
+      ustrzcat(vbeaf_desc, sizeof(vbeaf_desc), uconvert_ascii(", banked", tmp1));
 
    if (faf_ext > 0)
-      usprintf(vbeaf_desc+ustrsize(vbeaf_desc), uconvert_ascii(", FreeBE ex%02d", NULL), faf_ext);
+      uszprintf(vbeaf_desc+ustrsize(vbeaf_desc), sizeof(vbeaf_desc) - ustrsize(vbeaf_desc),
+		uconvert_ascii(", FreeBE ex%02d", tmp1), faf_ext);
    else if (faf_id)
-      ustrcat(vbeaf_desc, uconvert_ascii(", FreeBE noex", NULL));
+      ustrzcat(vbeaf_desc, sizeof(vbeaf_desc), uconvert_ascii(", FreeBE noex", tmp1));
 
    if (faf_farptr)
-      ustrcat(vbeaf_desc, uconvert_ascii(", farptr", NULL));
+      ustrzcat(vbeaf_desc, sizeof(vbeaf_desc), uconvert_ascii(", farptr", tmp1));
 
    /* is this an accelerated or dumb framebuffer mode? */
    if (mode_info.Attributes & 16) {
       gfx_vbeaf.drawing_mode = vbeaf_drawing_mode;
       vbeaf_drawing_mode();
-      ustrcat(vbeaf_desc, uconvert_ascii(", accel", NULL));
+      ustrzcat(vbeaf_desc, sizeof(vbeaf_desc), uconvert_ascii(", accel", tmp1));
    }
    else {
       gfx_vbeaf.drawing_mode = NULL;
-      ustrcat(vbeaf_desc, uconvert_ascii(", noaccel", NULL));
+      ustrzcat(vbeaf_desc, sizeof(vbeaf_desc), uconvert_ascii(", noaccel", tmp1));
    }
 
    gfx_vbeaf.desc = vbeaf_desc;
@@ -1590,6 +1625,95 @@ static void vbeaf_restore()
 
 
 
+/* vbeaf_fetch_mode_list:
+ *  Attempts to create a list of valid video modes for the VBE/AF driver.
+ *  Returns the mode list on success or NULL on failure.
+ */
+static GFX_MODE_LIST *vbeaf_fetch_mode_list()
+{
+   GFX_MODE_LIST *gfx_mode_list;
+   AF_MODE_INFO *mode_info;
+   unsigned short *mode;
+   int vbeaf_was_off;
+
+   mode_info = malloc(sizeof(AF_MODE_INFO));
+   if (!mode_info) return NULL;
+
+   /* make sure VBE/AF interface is enabled! */
+   if (!af_driver) {
+      if (!vbeaf_locate_driver()) {
+         if (mode_info) free(mode_info);
+	 
+         return NULL;
+      }
+      if (!vbeaf_lowlevel_init()) {
+         if (mode_info) free(mode_info);
+	 
+         return NULL;
+      }
+      vbeaf_was_off = TRUE;
+   }
+   else
+      vbeaf_was_off = FALSE;
+
+   /* start building mode-list */
+   gfx_mode_list = malloc(sizeof(GFX_MODE_LIST));
+   if (!gfx_mode_list) {
+      if (mode_info) free(mode_info);
+      
+      return NULL;
+   }
+   gfx_mode_list->mode = NULL;
+   gfx_mode_list->num_modes = 0;
+
+   /* create mode list */
+   for (mode = af_driver->AvailableModes; *mode != 0xFFFF; mode++) {
+      gfx_mode_list->mode = _al_sane_realloc(gfx_mode_list->mode, sizeof(GFX_MODE) * (gfx_mode_list->num_modes + 1));
+      if (!gfx_mode_list->mode) {
+         if (mode_info) free(mode_info);
+         if (gfx_mode_list) free(gfx_mode_list);
+         
+         return NULL;
+      }
+      if (af_driver->GetVideoModeInfo(af_driver, *mode, mode_info) != 0) {
+         if (mode_info) free(mode_info);
+         if (gfx_mode_list->mode) free(gfx_mode_list->mode);
+         if (gfx_mode_list) free(gfx_mode_list);
+         
+         return NULL;
+      }
+
+      gfx_mode_list->mode[gfx_mode_list->num_modes].width  = mode_info->XResolution;
+      gfx_mode_list->mode[gfx_mode_list->num_modes].height = mode_info->YResolution;
+      gfx_mode_list->mode[gfx_mode_list->num_modes].bpp    = mode_info->BitsPerPixel;
+
+      gfx_mode_list->num_modes++;
+   }
+
+   /* terminate mode list */
+   gfx_mode_list->mode = _al_sane_realloc(gfx_mode_list->mode, sizeof(GFX_MODE) * (gfx_mode_list->num_modes + 1));
+   if (!gfx_mode_list->mode) {
+      if (mode_info) free(mode_info);
+      if (gfx_mode_list) free(gfx_mode_list);
+      
+      return NULL;
+   }
+
+   gfx_mode_list->mode[gfx_mode_list->num_modes].width  = 0;
+   gfx_mode_list->mode[gfx_mode_list->num_modes].height = 0;
+   gfx_mode_list->mode[gfx_mode_list->num_modes].bpp    = 0;
+
+   /* shut down VBE/AF interface if it wasn't previously loaded */
+   if (vbeaf_was_off)
+      vbeaf_exit(NULL);
+
+   if (mode_info) free(mode_info);
+
+   return gfx_mode_list;
+}
+
+
+
 /* vbeaf_vsync:
  *  VBE/AF vsync routine, needed for cards that don't emulate the VGA 
  *  blanking registers. VBE/AF doesn't provide a vsync function, but we 
@@ -1614,7 +1738,7 @@ static int vbeaf_scroll(int x, int y)
    vbeaf_yscroll = y;
 
    SAFISH_CALL(
-      if (af_driver->WaitTillIdle)
+      if (_wait_for_vsync && af_driver->WaitTillIdle)
 	 af_driver->WaitTillIdle(af_driver);
 
       af_driver->SetDisplayStart(af_driver, x, y, 1);
@@ -1631,7 +1755,7 @@ static int vbeaf_scroll(int x, int y)
 /* vbeaf_set_palette_range:
  *  Uses VBE/AF functions to set the palette.
  */
-static void vbeaf_set_palette_range(PALETTE p, int from, int to, int vsync)
+static void vbeaf_set_palette_range(AL_CONST PALETTE p, int from, int to, int vsync)
 {
    AF_PALETTE tmp[256];
    int c;
@@ -2121,7 +2245,7 @@ static void prepare_mono_pattern(BITMAP *bmp)
 /* go_accel:
  *  Prepares the hardware for an accelerated drawing operation.
  */
-static INLINE void go_accel()
+static INLINE void go_accel(void)
 {
    /* turn on the accelerator */
    if (!_accel_active) {
@@ -2527,17 +2651,17 @@ static int vbeaf_triangle(BITMAP *bmp, int x1, int y1, int x2, int y2, int x3, i
 /* vbeaf_draw_glyph:
  *  Monochrome text plotter.
  */
-static void vbeaf_draw_glyph(BITMAP *bmp, FONT_GLYPH *glyph, int x, int y, int color)
+static void vbeaf_draw_glyph(BITMAP *bmp, AL_CONST FONT_GLYPH *glyph, int x, int y, int color, int bg)
 {
-   unsigned char *data = glyph->dat;
+   AL_CONST unsigned char *data = glyph->dat;
    int w = glyph->w;
    int h = glyph->h;
    int stride = (w+7)/8;
    int d;
 
    /* give up if we can't draw this */
-   if ((x < bmp->cl) || (x+w >= bmp->cr) || ((w&7) && (_textmode >= 0))) {
-      orig_draw_glyph(bmp, glyph, x, y, color);
+   if ((x < bmp->cl) || (x+w >= bmp->cr) || ((w&7) && (bg >= 0))) {
+      orig_draw_glyph(bmp, glyph, x, y, color, bg);
       return;
    }
 
@@ -2562,9 +2686,9 @@ static void vbeaf_draw_glyph(BITMAP *bmp, FONT_GLYPH *glyph, int x, int y, int c
 
    SAFE_CALL(
       /* set the mix mode */
-      if (_textmode >= 0) {
+      if (bg >= 0) {
 	 af_driver->SetMix(af_driver, 0, 0);
-	 d = _textmode;
+	 d = bg;
       }
       else {
 	 af_driver->SetMix(af_driver, 0, 4);
@@ -2576,7 +2700,7 @@ static void vbeaf_draw_glyph(BITMAP *bmp, FONT_GLYPH *glyph, int x, int y, int c
 
       af_driver->PutMonoImage(af_driver, color, d,
 			      x+bmp->x_ofs, y+bmp->y_ofs,
-			      stride, 0, 0, stride*8, h, data);
+			      stride, 0, 0, stride*8, h, (unsigned char *)data);
 
       af_driver->SetMix(af_driver, vbeaf_fg_mix, vbeaf_bg_mix);
    );
