@@ -20,7 +20,7 @@
 
 
 #include "allegro.h"
-#include "allegro/aintern.h"
+#include "allegro/internal/aintern.h"
 
 
 
@@ -41,8 +41,7 @@ MOUSE_DRIVER mousedrv_none =
 
 
 MOUSE_DRIVER *mouse_driver = NULL;     /* the active driver */
-int mouse_type = MOUSEDRV_AUTODETECT;  /* driver ID */
-
+int _mouse_type = MOUSEDRV_AUTODETECT;  /* driver ID */
 int _mouse_installed = FALSE;
 
 volatile int mouse_x = 0;              /* user-visible position */
@@ -65,8 +64,8 @@ volatile int freeze_mouse_flag = FALSE;
 
 void (*mouse_callback)(int flags) = NULL;
 
-int _mouse_x_focus = 1;                /* focus point in mouse sprite */
-int _mouse_y_focus = 1;
+int mouse_x_focus = 1;                 /* focus point in mouse sprite */
+int mouse_y_focus = 1;
 
 
 #define MOUSE_OFFSCREEN    -4096       /* somewhere to put unwanted cursors */
@@ -98,14 +97,14 @@ static char mouse_pointer_data[DEFAULT_SPRITE_H][DEFAULT_SPRITE_W] =
 
 BITMAP *_mouse_pointer = NULL;         /* default mouse pointer */
 
-BITMAP *_mouse_sprite = NULL;          /* current mouse pointer */
+BITMAP *mouse_sprite = NULL;	       /* current mouse pointer */
 
 BITMAP *_mouse_screen = NULL;          /* where to draw the pointer */
 
 static int got_hw_cursor = FALSE;      /* hardware pointer available? */
 static int hw_cursor_dirty = FALSE;    /* need to set a new pointer? */
 
-static int mx, my;                     /* previous mouse position */
+static int mx, my;                     /* previous mouse pointer position */
 static BITMAP *ms = NULL;              /* previous screen data */
 static BITMAP *mtemp = NULL;           /* double-buffer drawing area */
 
@@ -133,27 +132,27 @@ static void draw_mouse_doublebuffer(int newx, int newy)
    int x1, y1, w, h;
 
    /* grab bit of screen containing where we are and where we'll be */
-   x1 = MIN(mx, newx) - _mouse_x_focus;
-   y1 = MIN(my, newy) - _mouse_y_focus;
+   x1 = MIN(mx, newx) - mouse_x_focus;
+   y1 = MIN(my, newy) - mouse_y_focus;
 
    /* get width of area */
-   w = MAX(mx, newx) - x1 + _mouse_sprite->w+1;
-   h = MAX(my, newy) - y1 + _mouse_sprite->h+1;
+   w = MAX(mx, newx) - x1 + mouse_sprite->w+1;
+   h = MAX(my, newy) - y1 + mouse_sprite->h+1;
 
    /* make new co-ords relative to 'mtemp' bitmap co-ords */
-   newx -= _mouse_x_focus+x1;
-   newy -= _mouse_y_focus+y1;
+   newx -= mouse_x_focus+x1;
+   newy -= mouse_y_focus+y1;
 
    /* save screen image in 'mtemp' */
    blit(_mouse_screen, mtemp, x1, y1, 0, 0, w, h);
 
    /* blit saved image in 'ms' to corect place in this buffer */
-   blit(ms, mtemp, 0, 0, mx-_mouse_x_focus-x1, my-_mouse_y_focus-y1,
-			 _mouse_sprite->w, _mouse_sprite->h);
+   blit(ms, mtemp, 0, 0, mx-mouse_x_focus-x1, my-mouse_y_focus-y1,
+			 mouse_sprite->w, mouse_sprite->h);
 
    /* draw mouse at correct place in 'mtemp' */
-   blit(mtemp, ms, newx, newy, 0, 0, _mouse_sprite->w, _mouse_sprite->h);
-   draw_sprite(mtemp, _mouse_sprite, newx, newy);
+   blit(mtemp, ms, newx, newy, 0, 0, mouse_sprite->w, mouse_sprite->h);
+   draw_sprite(mtemp, mouse_sprite, newx, newy);
 
    /* blit 'mtemp' to screen */
    blit(mtemp, _mouse_screen, 0, 0, x1, y1, w, h);
@@ -194,18 +193,18 @@ static void draw_mouse(int remove, int add)
       mon = TRUE;
 
    if (!normal_draw) {
-      if ((newmx <= mx-_mouse_sprite->w) || (newmx >= mx+_mouse_sprite->w) ||
-	  (newmy <= my-_mouse_sprite->h) || (newmy >= my+_mouse_sprite->h))
+      if ((newmx <= mx-mouse_sprite->w) || (newmx >= mx+mouse_sprite->w) ||
+	  (newmy <= my-mouse_sprite->h) || (newmy >= my+mouse_sprite->h))
 	 normal_draw = 1;
    }
 
    if (normal_draw) {
       if (remove)
-	 blit(ms, _mouse_screen, 0, 0, mx-_mouse_x_focus, my-_mouse_y_focus, _mouse_sprite->w, _mouse_sprite->h);
+	 blit(ms, _mouse_screen, 0, 0, mx-mouse_x_focus, my-mouse_y_focus, mouse_sprite->w, mouse_sprite->h);
 
       if (add) {
-	 blit(_mouse_screen, ms, newmx-_mouse_x_focus, newmy-_mouse_y_focus, 0, 0, _mouse_sprite->w, _mouse_sprite->h);
-	 draw_sprite(_mouse_screen, _mouse_sprite, newmx-_mouse_x_focus, newmy-_mouse_y_focus);
+	 blit(_mouse_screen, ms, newmx-mouse_x_focus, newmy-mouse_y_focus, 0, 0, mouse_sprite->w, mouse_sprite->h);
+	 draw_sprite(_mouse_screen, mouse_sprite, newmx-mouse_x_focus, newmy-mouse_y_focus);
       }
    }
    else
@@ -228,7 +227,7 @@ END_OF_STATIC_FUNCTION(draw_mouse);
 /* update_mouse:
  *  Worker function to update the mouse position variables with new values.
  */
-static void update_mouse()
+static void update_mouse(void)
 {
    int x, y, z, b, flags = 0;
 
@@ -330,13 +329,13 @@ static void mouse_move(void)
 	 }
       }
       else {
-	 /* bodge to avoid using MMX code inside a timer handler */
-	 int old_mmx = cpu_mmx;
-	 cpu_mmx = FALSE;
+	 /* bodge to avoid using non legacy 386 asm code inside a timer handler */
+	 int old_capabilities = cpu_capabilities;
+	 cpu_capabilities = 0;
 
 	 draw_mouse(TRUE, TRUE);
 
-	 cpu_mmx = old_mmx;
+	 cpu_capabilities = old_capabilities;
       }
 
       release_bitmap(_mouse_screen);
@@ -353,7 +352,7 @@ END_OF_STATIC_FUNCTION(mouse_move);
  *  Callback for an asynchronous driver to tell us when it has changed the
  *  position.
  */
-void _handle_mouse_input()
+void _handle_mouse_input(void)
 {
    if (!mouse_polled)
       update_mouse();
@@ -363,6 +362,34 @@ END_OF_FUNCTION(_handle_mouse_input);
 
 
 
+/* create_mouse_pointer:
+ *  Creates the default arrow mouse sprite using the current color depth
+ *  and palette.
+ */
+static BITMAP *create_mouse_pointer(void)
+{
+   BITMAP *bmp;
+   int x, y;
+   int col;
+   
+   bmp = create_bitmap(DEFAULT_SPRITE_W, DEFAULT_SPRITE_H);
+
+   for (y=0; y<DEFAULT_SPRITE_H; y++) {
+      for (x=0; x<DEFAULT_SPRITE_W; x++) {
+	 switch (mouse_pointer_data[y][x]) {
+	    case 1:  col = makecol(255, 255, 255);  break;
+	    case 2:  col = makecol(0, 0, 0);        break;
+	    default: col = bmp->vtable->mask_color; break;
+	 }
+	 putpixel(bmp, x, y, col);
+      }
+   }
+
+   return bmp;
+}
+
+
+
 /* set_mouse_sprite:
  *  Sets the sprite to be used for the mouse pointer. If the sprite is
  *  NULL, restores the default arrow.
@@ -378,28 +405,33 @@ void set_mouse_sprite(struct BITMAP *sprite)
       show_mouse(NULL);
 
    if (sprite)
-      _mouse_sprite = sprite;
-   else
-      _mouse_sprite = _mouse_pointer;
+      mouse_sprite = sprite;
+   else {
+      if (_mouse_pointer)
+	 destroy_bitmap(_mouse_pointer);
+      _mouse_pointer = create_mouse_pointer();
+      mouse_sprite = _mouse_pointer;
+   }
 
-   lock_bitmap(_mouse_sprite);
+   lock_bitmap((struct BITMAP*)mouse_sprite);
 
    /* make sure the ms bitmap is big enough */
-   if ((!ms) || (ms->w < _mouse_sprite->w) || (ms->h < _mouse_sprite->h)) {
+   if ((!ms) || (ms->w < mouse_sprite->w) || (ms->h < mouse_sprite->h) ||
+       (bitmap_color_depth(mouse_sprite) != bitmap_color_depth(ms))) {
       if (ms) {
 	 destroy_bitmap(ms);
 	 destroy_bitmap(mtemp);
       }
 
-      ms = create_bitmap(_mouse_sprite->w, _mouse_sprite->h);
+      ms = create_bitmap(mouse_sprite->w, mouse_sprite->h);
       lock_bitmap(ms);
 
-      mtemp = create_bitmap(_mouse_sprite->w*2, _mouse_sprite->h*2);
+      mtemp = create_bitmap(mouse_sprite->w*2, mouse_sprite->h*2);
       lock_bitmap(mtemp);
    }
 
-   _mouse_x_focus = 1;
-   _mouse_y_focus = 1;
+   mouse_x_focus = 1;
+   mouse_y_focus = 1;
 
    hw_cursor_dirty = TRUE;
 
@@ -418,8 +450,8 @@ void set_mouse_sprite_focus(int x, int y)
    if (!mouse_driver)
       return;
 
-   _mouse_x_focus = x;
-   _mouse_y_focus = y;
+   mouse_x_focus = x;
+   mouse_y_focus = y;
 
    hw_cursor_dirty = TRUE;
 }
@@ -464,7 +496,7 @@ void show_mouse(BITMAP *bmp)
 	 got_hw_cursor = FALSE;
 
 	 if ((gfx_driver) && (gfx_driver->set_mouse_sprite) && (!_dispsw_status))
-	    if (gfx_driver->set_mouse_sprite(_mouse_sprite, _mouse_x_focus, _mouse_y_focus) == 0)
+	    if (gfx_driver->set_mouse_sprite(mouse_sprite, mouse_x_focus, mouse_y_focus) == 0)
 	       got_hw_cursor = TRUE;
 
 	 hw_cursor_dirty = FALSE;
@@ -495,7 +527,7 @@ void show_mouse(BITMAP *bmp)
  *  cursor is in use). This operation can later be reversed by calling
  *  unscare_mouse().
  */
-void scare_mouse()
+void scare_mouse(void)
 {
    if (!mouse_driver)
       return;
@@ -537,10 +569,10 @@ void scare_mouse_area(int x, int y, int w, int h)
       was_frozen = freeze_mouse_flag;
       freeze_mouse_flag = TRUE;
 
-      if ((mouse_x - _mouse_x_focus < x + w) &&
-	  (mouse_y - _mouse_y_focus < y + h) &&
-	  (mouse_x - _mouse_x_focus + _mouse_sprite->w >= x) &&
-	  (mouse_y - _mouse_y_focus + _mouse_sprite->h >= y)) {
+      if ((mx - mouse_x_focus < x + w) &&
+	   (my - mouse_y_focus < y + h) &&
+	   (mx - mouse_x_focus + mouse_sprite->w >= x) &&
+	   (my - mouse_y_focus + mouse_sprite->h >= y)) {
 
 	 if (scared_size < SCARED_SIZE) {
 	    scared_screen[scared_size] = _mouse_screen;
@@ -579,7 +611,7 @@ void scare_mouse_area(int x, int y, int w, int h)
  *  Restores the original mouse state, after a call to scare_mouse() or
  *  scare_mouse_area.
  */
-void unscare_mouse()
+void unscare_mouse(void)
 {
    if (!mouse_driver)
       return;
@@ -653,6 +685,10 @@ void position_mouse_z(int z)
 void set_mouse_range(int x1, int y1, int x2, int y2)
 {
    BITMAP *old_mouse_screen = _mouse_screen;
+   ASSERT(x1 >= 0);
+   ASSERT(y1 >= 0);
+   ASSERT(x2 >= x1);
+   ASSERT(y2 >= y2);
 
    if (!mouse_driver)
       return;
@@ -710,7 +746,7 @@ void get_mouse_mickeys(int *mickeyx, int *mickeyy)
  *  and will no longer operate asynchronously even if the driver actually
  *  does support that.
  */
-int poll_mouse()
+int poll_mouse(void)
 {
    if (!mouse_driver)
       return -1;
@@ -732,7 +768,7 @@ END_OF_FUNCTION(poll_mouse);
 /* mouse_needs_poll:
  *  Checks whether the current driver uses polling.
  */
-int mouse_needs_poll()
+int mouse_needs_poll(void)
 {
    return mouse_polled;
 }
@@ -741,59 +777,19 @@ END_OF_FUNCTION(mouse_needs_poll);
 
 
 
-/* _set_mouse_range:
+/* set_mouse_etc:
  *  Hook for setting up the motion range, cursor graphic, etc, called by
  *  the mouse init and whenever we change the graphics mode.
  */
-void _set_mouse_range()
+static void set_mouse_etc(void)
 {
-   int x, y;
-   int col;
-
    if ((!mouse_driver) || (!gfx_driver))
       return;
 
    if ((!_mouse_pointer) || 
        ((screen) && (_mouse_pointer) &&
-	(bitmap_color_depth(_mouse_pointer) != bitmap_color_depth(screen)))) {
-
-      if (_mouse_pointer) {
-	 destroy_bitmap(_mouse_pointer);
-	 _mouse_pointer = NULL;
-      }
-
-      if (ms) {
-	 destroy_bitmap(ms);
-	 ms = NULL;
-
-	 destroy_bitmap(mtemp);
-	 mtemp = NULL;
-      }
-
-      _mouse_pointer = create_bitmap(DEFAULT_SPRITE_W, DEFAULT_SPRITE_H);
-
-      for (y=0; y<DEFAULT_SPRITE_H; y++) {
-	 for (x=0; x<DEFAULT_SPRITE_W; x++) {
-	    if (bitmap_color_depth(_mouse_pointer) == 8) {
-	       switch (mouse_pointer_data[y][x]) {
-		  case 1:  col = 16;  break;
-		  case 2:  col = 255; break;
-		  default: col = 0;   break;
-	       }
-	    }
-	    else {
-	       switch (mouse_pointer_data[y][x]) {
-		  case 1:  col = makecol(255, 255, 255);             break;
-		  case 2:  col = makecol(0, 0, 0);                   break;
-		  default: col = _mouse_pointer->vtable->mask_color; break;
-	       }
-	    } 
-	    putpixel(_mouse_pointer, x, y, col);
-	 }
-      }
-
-      set_mouse_sprite(_mouse_pointer);
-   }
+	(bitmap_color_depth(_mouse_pointer) != bitmap_color_depth(screen))))
+      set_mouse_sprite(NULL);
    else
       hw_cursor_dirty = TRUE;
 
@@ -809,12 +805,13 @@ void _set_mouse_range()
  *  other mouse functions. Return -1 if it can't find a mouse driver,
  *  otherwise the number of buttons on the mouse.
  */
-int install_mouse()
+int install_mouse(void)
 {
    _DRIVER_INFO *driver_list;
    int num_buttons = -1;
-   char *emulate;
-   char tmp[64];
+   int config_num_buttons;
+   AL_CONST char *emulate;
+   char tmp1[64], tmp2[64];
    int i;
 
    if (mouse_driver)
@@ -836,9 +833,9 @@ int install_mouse()
    LOCK_VARIABLE(emulate_three);
    LOCK_VARIABLE(freeze_mouse_flag);
    LOCK_VARIABLE(mouse_callback);
-   LOCK_VARIABLE(_mouse_x_focus);
-   LOCK_VARIABLE(_mouse_y_focus);
-   LOCK_VARIABLE(_mouse_sprite);
+   LOCK_VARIABLE(mouse_x_focus);
+   LOCK_VARIABLE(mouse_y_focus);
+   LOCK_VARIABLE(mouse_sprite);
    LOCK_VARIABLE(mouse_pointer_data);
    LOCK_VARIABLE(_mouse_pointer);
    LOCK_VARIABLE(_mouse_screen);
@@ -861,12 +858,12 @@ int install_mouse()
    else
       driver_list = _mouse_driver_list;
 
-   if (mouse_type == MOUSEDRV_AUTODETECT)
-      mouse_type = get_config_id(NULL, uconvert_ascii("mouse", tmp), MOUSEDRV_AUTODETECT);
+   if (_mouse_type == MOUSEDRV_AUTODETECT)
+      _mouse_type = get_config_id(uconvert_ascii("mouse", tmp1), uconvert_ascii("mouse", tmp2), MOUSEDRV_AUTODETECT);
 
-   if (mouse_type != MOUSEDRV_AUTODETECT) {
+   if (_mouse_type != MOUSEDRV_AUTODETECT) {
       for (i=0; driver_list[i].driver; i++) {
-	 if (driver_list[i].id == mouse_type) {
+	 if (driver_list[i].id == _mouse_type) {
 	    mouse_driver = driver_list[i].driver;
 	    break;
 	 }
@@ -893,8 +890,12 @@ int install_mouse()
       return -1;
    }
 
-   num_buttons = get_config_int(NULL, uconvert_ascii("num_buttons", tmp), num_buttons);
-   emulate = get_config_string(NULL, uconvert_ascii("emulate_three", tmp), NULL);
+   config_num_buttons = get_config_int(uconvert_ascii("mouse", tmp1), uconvert_ascii("num_buttons", tmp2), -1);
+   emulate = get_config_string(uconvert_ascii("mouse", tmp1), uconvert_ascii("emulate_three", tmp2), NULL);
+
+   /* clamp config_num_buttons to zero/positive values */
+   if (config_num_buttons >= 0)
+      num_buttons = config_num_buttons;
 
    if ((emulate) && ((i = ugetc(emulate)) != 0)) {
       if ((i == 'y') || (i == 'Y') || (i == '1'))
@@ -903,17 +904,14 @@ int install_mouse()
 	 emulate_three = FALSE;
    }
    else {
-      if (num_buttons < 3)
-	 emulate_three = TRUE;
-      else
-	 emulate_three = FALSE;
+      emulate_three = FALSE;
    }
 
    mouse_polled = (mouse_driver->poll) ? TRUE : FALSE;
 
    _mouse_installed = TRUE;
 
-   _set_mouse_range();
+   set_mouse_etc();
    _add_exit_func(remove_mouse);
 
    if (mouse_driver->timer_poll)
@@ -963,3 +961,23 @@ void remove_mouse(void)
 
    _remove_exit_func(remove_mouse);
 }
+
+
+
+/* _mouse_constructor:
+ *  Register mouse functions if this object file is linked in.
+ */
+#ifdef ALLEGRO_USE_CONSTRUCTOR
+   CONSTRUCTOR_FUNCTION(void _mouse_constructor(void));
+#endif
+
+static struct _AL_LINKER_MOUSE mouse_linker = {
+   set_mouse_etc,
+   show_mouse,
+   &_mouse_screen
+};
+
+void _mouse_constructor(void)
+{
+   _al_linker_mouse = &mouse_linker;
+}
