@@ -21,24 +21,28 @@
 #include <string.h>
 
 #include "allegro.h"
-#include "allegro/aintern.h"
+#include "allegro/internal/aintern.h"
+
+
+
+extern DIGI_DRIVER digi_none;
 
 
 
 /* dummy functions for the nosound drivers */
 int  _dummy_detect(int input) { return TRUE; }
-int  _dummy_init(int input, int voices) { digi_none.desc = midi_none.desc = get_config_text("The sound of silence"); return 0; }
+int  _dummy_init(int input, int voices) { digi_none.desc = _midi_none.desc = get_config_text("The sound of silence"); return 0; }
 void _dummy_exit(int input) { }
 int  _dummy_mixer_volume(int volume) { return 0; }
-void _dummy_init_voice(int voice, SAMPLE *sample) { }
+void _dummy_init_voice(int voice, AL_CONST SAMPLE *sample) { }
 void _dummy_noop1(int p) { }
 void _dummy_noop2(int p1, int p2) { }
 void _dummy_noop3(int p1, int p2, int p3) { }
 int  _dummy_get_position(int voice) { return -1; }
 int  _dummy_get(int voice) { return 0; }
 void _dummy_raw_midi(int data) { }
-int  _dummy_load_patches(char *patches, char *drums) { return 0; }
-void _dummy_adjust_patches(char *patches, char *drums) { }
+int  _dummy_load_patches(AL_CONST char *patches, AL_CONST char *drums) { return 0; }
+void _dummy_adjust_patches(AL_CONST char *patches, AL_CONST char *drums) { }
 void _dummy_key_on(int inst, int note, int bend, int vol, int pan) { }
 
 /* put this after all the dummy functions, so they will all get locked */
@@ -46,7 +50,7 @@ END_OF_FUNCTION(_dummy_detect);
 
 
 
-DIGI_DRIVER digi_none =
+static DIGI_DRIVER digi_none =
 {
    DIGI_NONE,
    empty_string,
@@ -87,7 +91,7 @@ DIGI_DRIVER digi_none =
 };
 
 
-MIDI_DRIVER midi_none =
+MIDI_DRIVER _midi_none =
 {
    MIDI_NONE,
    empty_string,
@@ -113,14 +117,14 @@ MIDI_DRIVER midi_none =
 int digi_card = DIGI_AUTODETECT;          /* current driver ID numbers */
 int midi_card = MIDI_AUTODETECT;
 
-int digi_input_card = MIDI_AUTODETECT;
+int digi_input_card = DIGI_AUTODETECT;
 int midi_input_card = MIDI_AUTODETECT;
 
 DIGI_DRIVER *digi_driver = &digi_none;    /* these things do all the work */
-MIDI_DRIVER *midi_driver = &midi_none;
+MIDI_DRIVER *midi_driver = &_midi_none;
 
 DIGI_DRIVER *digi_input_driver = &digi_none;
-MIDI_DRIVER *midi_input_driver = &midi_none;
+MIDI_DRIVER *midi_input_driver = &_midi_none;
 
 void (*digi_recorder)(void) = NULL;
 void (*midi_recorder)(unsigned char data) = NULL;
@@ -139,7 +143,7 @@ int _digi_volume = -1;                    /* current volume settings */
 int _midi_volume = -1;
 
 int _sound_flip_pan = FALSE;              /* reverse l/r sample panning? */
-int _sound_hq = 1;                        /* mixer speed vs. quality */
+int _sound_hq = 2;                        /* mixer speed vs. quality */
 
 int _sound_freq = -1;                     /* common hardware parameters */
 int _sound_stereo = -1;
@@ -153,15 +157,12 @@ int _sound_irq = -1;
 static void update_sweeps(void);
 static void sound_lock_mem(void);
 
-int (*_midi_init)(void) = NULL;
-void (*_midi_exit)(void) = NULL;
-
 
 
 /* read_sound_config:
  *  Helper for reading the sound hardware configuration data.
  */
-static void read_sound_config()
+static void read_sound_config(void)
 {
    char tmp1[64], tmp2[64];
    char *sound = uconvert_ascii("sound", tmp1);
@@ -260,12 +261,12 @@ int detect_midi_driver(int driver_id)
 	 else
 	    ret = 0;
 
-	 midi_driver = &midi_none;
+	 midi_driver = &_midi_none;
 	 return ret;
       }
    }
 
-   return midi_none.max_voices;
+   return _midi_none.max_voices;
 }
 
 
@@ -293,7 +294,7 @@ void reserve_voices(int digi_voices, int midi_voices)
  *  allegro.h. Pass DIGI_AUTODETECT and MIDI_AUTODETECT if you don't know 
  *  what the soundcard is.
  */
-int install_sound(int digi, int midi, char *cfg_path)
+int install_sound(int digi, int midi, AL_CONST char *cfg_path)
 {
    char tmp1[64], tmp2[64];
    char *sound = uconvert_ascii("sound", tmp1);
@@ -313,8 +314,8 @@ int install_sound(int digi, int midi, char *cfg_path)
       _phys_voice[c].num = -1;
 
    /* initialise the MIDI file player */
-   if (_midi_init)
-      if (_midi_init() != 0)
+   if (_al_linker_midi)
+      if (_al_linker_midi->init() != 0)
 	 return -1;
 
    usetc(allegro_error, 0);
@@ -360,16 +361,19 @@ int install_sound(int digi, int midi, char *cfg_path)
 	 digi_driver = digi_drivers[c].driver;
 	 if (!digi_driver->detect(FALSE)) {
 	    digi_driver = &digi_none; 
-	    if (_midi_exit)
-	       _midi_exit();
+	    if (_al_linker_midi)
+	       _al_linker_midi->exit();
 	    if (!ugetc(allegro_error))
-	       ustrcpy(allegro_error, get_config_text("Digital sound driver not found"));
+	       ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Digital sound driver not found"));
 	    return -1;
 	 }
 	 break;
       }
    }
 
+   if (digi_card == DIGI_NONE)
+      digi_driver = &digi_none;
+
    /* autodetect digital driver */
    if (!digi_driver) {
       for (c=0; digi_drivers[c].driver; c++) {
@@ -378,8 +382,14 @@ int install_sound(int digi, int midi, char *cfg_path)
 	    digi_driver = digi_drivers[c].driver;
 	    if (digi_driver->detect(FALSE))
 	       break;
+	    digi_driver = NULL;
 	 }
       }
+
+      if (!digi_driver) {
+	 digi_card = DIGI_NONE;
+	 digi_driver = &digi_none;
+      }
    }
 
    /* set up midi sound driver */
@@ -401,17 +411,20 @@ int install_sound(int digi, int midi, char *cfg_path)
 	 midi_driver = midi_drivers[c].driver;
 	 if (!midi_driver->detect(FALSE)) {
 	    digi_driver = &digi_none; 
-	    midi_driver = &midi_none; 
-	    if (_midi_exit)
-	       _midi_exit();
+	    midi_driver = &_midi_none; 
+	    if (_al_linker_midi)
+	       _al_linker_midi->exit();
 	    if (!ugetc(allegro_error))
-	       ustrcpy(allegro_error, get_config_text("MIDI music driver not found"));
+	       ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("MIDI music driver not found"));
 	    return -1;
 	 }
 	 break;
       }
    }
 
+   if (midi_card == MIDI_NONE)
+      midi_driver = &_midi_none;
+
    /* autodetect MIDI driver */
    if (!midi_driver) {
       for (c=0; midi_drivers[c].driver; c++) {
@@ -420,8 +433,14 @@ int install_sound(int digi, int midi, char *cfg_path)
 	    midi_driver = midi_drivers[c].driver;
 	    if (midi_driver->detect(FALSE))
 	       break;
+	    midi_driver = NULL;
 	 }
       }
+
+      if (!midi_driver) {
+	 midi_card = MIDI_NONE;
+	 midi_driver = &_midi_none;
+      }
    }
 
    /* work out how many voices to allocate for each driver */
@@ -449,22 +468,23 @@ int install_sound(int digi, int midi, char *cfg_path)
 
    /* make sure this is a reasonable number of voices to use */
    if ((digi_voices > DIGI_VOICES) || (midi_voices > MIDI_VOICES)) {
-      usprintf(allegro_error, get_config_text("Insufficient %s voices available"), (digi_voices > DIGI_VOICES) ? get_config_text("digital") : get_config_text("MIDI"));
+      uszprintf(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Insufficient %s voices available"),
+		(digi_voices > DIGI_VOICES) ? get_config_text("digital") : get_config_text("MIDI"));
       digi_driver = &digi_none; 
-      midi_driver = &midi_none; 
-      if (_midi_exit)
-	 _midi_exit();
+      midi_driver = &_midi_none; 
+      if (_al_linker_midi)
+	 _al_linker_midi->exit();
       return -1;
    }
 
    /* initialise the digital sound driver */
    if (digi_driver->init(FALSE, digi_voices) != 0) {
       digi_driver = &digi_none; 
-      midi_driver = &midi_none; 
-      if (_midi_exit)
-	 _midi_exit();
+      midi_driver = &_midi_none; 
+      if (_al_linker_midi)
+	 _al_linker_midi->exit();
       if (!ugetc(allegro_error))
-	 ustrcpy(allegro_error, get_config_text("Failed to init digital sound driver"));
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Failed to init digital sound driver"));
       return -1;
    }
 
@@ -472,11 +492,11 @@ int install_sound(int digi, int midi, char *cfg_path)
    if (midi_driver->init(FALSE, midi_voices) != 0) {
       digi_driver->exit(FALSE);
       digi_driver = &digi_none; 
-      midi_driver = &midi_none; 
-      if (_midi_exit)
-	 _midi_exit();
+      midi_driver = &_midi_none; 
+      if (_al_linker_midi)
+	 _al_linker_midi->exit();
       if (!ugetc(allegro_error))
-	 ustrcpy(allegro_error, get_config_text("Failed to init MIDI music driver"));
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Failed to init MIDI music driver"));
       return -1;
    }
 
@@ -486,13 +506,14 @@ int install_sound(int digi, int midi, char *cfg_path)
    /* check that we actually got enough voices */
    if ((digi_driver->voices < digi_voices) || 
        ((midi_driver->voices < midi_voices) && (!midi_driver->raw_midi))) {
-      usprintf(allegro_error, get_config_text("Insufficient %s voices available"), (digi_driver->voices < digi_voices) ? get_config_text("digital") : get_config_text("MIDI"));
+      uszprintf(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Insufficient %s voices available"),
+                (digi_driver->voices < digi_voices) ? get_config_text("digital") : get_config_text("MIDI"));
       midi_driver->exit(FALSE);
       digi_driver->exit(FALSE);
       digi_driver = &digi_none; 
-      midi_driver = &midi_none; 
-      if (_midi_exit)
-	 _midi_exit();
+      midi_driver = &_midi_none; 
+      if (_al_linker_midi)
+	 _al_linker_midi->exit();
       return -1;
    }
 
@@ -543,7 +564,7 @@ int install_sound_input(int digi, int midi)
       return 0;
 
    if (!_sound_installed) {
-      ustrcpy(allegro_error, get_config_text("Sound system not installed"));
+      ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Sound system not installed"));
       return -1;
    }
 
@@ -568,7 +589,7 @@ int install_sound_input(int digi, int midi)
    else
       digi_drivers = _digi_driver_list;
 
-   for (c=0; (digi_drivers[c].driver) && (digi_drivers[c].driver != &digi_none); c++) { 
+   for (c=0; digi_drivers[c].driver; c++) {
       if ((digi_drivers[c].id == digi_input_card) || (digi_input_card == DIGI_AUTODETECT)) {
 	 digi_input_driver = digi_drivers[c].driver;
 	 if (digi_input_driver->detect(TRUE)) {
@@ -579,7 +600,8 @@ int install_sound_input(int digi, int midi)
 	    digi_input_driver = &digi_none;
 	    if (digi_input_card != DIGI_AUTODETECT) {
 	       if (!ugetc(allegro_error))
-		  usprintf(allegro_error, get_config_text("%s does not support audio input"), ((DIGI_DRIVER *)digi_drivers[c].driver)->name);
+		  uszprintf(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("%s does not support audio input"),
+			    ((DIGI_DRIVER *)digi_drivers[c].driver)->name);
 	       break;
 	    }
 	 }
@@ -589,7 +611,7 @@ int install_sound_input(int digi, int midi)
    /* did we find one? */
    if ((digi_input_driver == &digi_none) && (digi_input_card != DIGI_NONE)) {
       if (!ugetc(allegro_error))
-	 ustrcpy(allegro_error, get_config_text("Digital input driver not found"));
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Digital input driver not found"));
       return -1;
    }
 
@@ -601,7 +623,7 @@ int install_sound_input(int digi, int midi)
    else
       midi_drivers = _midi_driver_list;
 
-   for (c=0; (midi_drivers[c].driver) && (midi_drivers[c].driver != &midi_none); c++) { 
+   for (c=0; midi_drivers[c].driver; c++) { 
       if ((midi_drivers[c].id == midi_input_card) || (midi_input_card == MIDI_AUTODETECT)) {
 	 midi_input_driver = midi_drivers[c].driver;
 	 if (midi_input_driver->detect(TRUE)) {
@@ -609,10 +631,11 @@ int install_sound_input(int digi, int midi)
 	    break;
 	 }
 	 else {
-	    midi_input_driver = &midi_none;
+	    midi_input_driver = &_midi_none;
 	    if (midi_input_card != MIDI_AUTODETECT) {
 	       if (!ugetc(allegro_error))
-		  usprintf(allegro_error, get_config_text("%s does not support MIDI input"), ((MIDI_DRIVER *)midi_drivers[c].driver)->name);
+		  uszprintf(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("%s does not support MIDI input"),
+			    ((MIDI_DRIVER *)midi_drivers[c].driver)->name);
 	       break;
 	    }
 	 }
@@ -620,19 +643,19 @@ int install_sound_input(int digi, int midi)
    }
 
    /* did we find one? */
-   if ((midi_input_driver == &midi_none) && (midi_input_card != MIDI_NONE)) {
+   if ((midi_input_driver == &_midi_none) && (midi_input_card != MIDI_NONE)) {
       digi_input_driver = &digi_none;
       if (!ugetc(allegro_error))
-	 ustrcpy(allegro_error, get_config_text("MIDI input driver not found"));
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("MIDI input driver not found"));
       return -1;
    }
 
    /* initialise the digital input driver */
    if (digi_input_driver->init(TRUE, 0) != 0) {
       digi_input_driver = &digi_none;
-      midi_input_driver = &midi_none;
+      midi_input_driver = &_midi_none;
       if (!ugetc(allegro_error))
-	 ustrcpy(allegro_error, get_config_text("Failed to init digital input driver"));
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Failed to init digital input driver"));
       return -1;
    }
 
@@ -640,9 +663,9 @@ int install_sound_input(int digi, int midi)
    if (midi_input_driver->init(TRUE, 0) != 0) {
       digi_input_driver->exit(TRUE);
       digi_input_driver = &digi_none;
-      midi_input_driver = &midi_none;
+      midi_input_driver = &_midi_none;
       if (!ugetc(allegro_error))
-	 ustrcpy(allegro_error, get_config_text("Failed to init MIDI input driver"));
+	 ustrzcpy(allegro_error, ALLEGRO_ERROR_SIZE, get_config_text("Failed to init MIDI input driver"));
       return -1;
    }
 
@@ -668,11 +691,11 @@ void remove_sound(void)
 	 if (_voice[c].sample)
 	    deallocate_voice(c);
 
-      if (_midi_exit)
-	 _midi_exit();
+      if (_al_linker_midi)
+	 _al_linker_midi->exit();
 
       midi_driver->exit(FALSE);
-      midi_driver = &midi_none; 
+      midi_driver = &_midi_none; 
 
       digi_driver->exit(FALSE);
       digi_driver = &digi_none; 
@@ -694,7 +717,7 @@ void remove_sound_input()
       digi_input_driver = &digi_none;
 
       midi_input_driver->exit(TRUE);
-      midi_input_driver = &midi_none;
+      midi_input_driver = &_midi_none;
 
       digi_recorder = NULL;
       midi_recorder = NULL;
@@ -707,31 +730,55 @@ void remove_sound_input()
 
 /* set_volume:
  *  Alters the global sound output volume. Specify volumes for both digital
- *  samples and MIDI playback, as integers from 0 to 255. If possible this
- *  routine will use a hardware mixer to control the volume, otherwise it
- *  will tell the sample mixer and MIDI player to simulate a mixer in
- *  software.
+ *  samples and MIDI playback, as integers from 0 to 255. This routine will
+ *  not alter the volume of the hardware mixer if it exists.
  */
 void set_volume(int digi_volume, int midi_volume)
+{
+   int *voice_vol;
+   int i;
+
+   if (digi_volume >= 0) {
+      voice_vol = malloc(sizeof(int)*VIRTUAL_VOICES);
+
+      /* Retrieve the (relative) volume of each voice. */
+      for (i=0; i<VIRTUAL_VOICES; i++)
+	 voice_vol[i] = voice_get_volume(i);
+
+      _digi_volume = MID(0, digi_volume, 255);
+
+      /* Set the new (relative) volume for each voice. */
+      for (i=0; i<VIRTUAL_VOICES; i++)
+	 voice_set_volume(i, voice_vol[i]);
+
+      free(voice_vol);
+   }
+
+   if (midi_volume >= 0)
+      _midi_volume = MID(0, midi_volume, 255);
+}
+
+
+
+/* set_hardware_volume:
+ *  Alters the hardware sound output volume. Specify volumes for both digital
+ *  samples and MIDI playback, as integers from 0 to 255. This routine will
+ *  use the hardware mixer to control the volume if it exists, or do nothing.
+ */
+void set_hardware_volume(int digi_volume, int midi_volume)
 {
    if (digi_volume >= 0) {
       digi_volume = MID(0, digi_volume, 255);
 
-      if ((digi_driver->mixer_volume) && 
-	  (digi_driver->mixer_volume(digi_volume) == 0))
-	 _digi_volume = -1;
-      else
-	 _digi_volume = digi_volume;
+      if (digi_driver->mixer_volume)
+	 digi_driver->mixer_volume(digi_volume);
    }
 
    if (midi_volume >= 0) {
       midi_volume = MID(0, midi_volume, 255);
 
-      if ((midi_driver->mixer_volume) && 
-	  (midi_driver->mixer_volume(midi_volume) == 0))
-	 _midi_volume = -1;
-      else
-	 _midi_volume = midi_volume;
+      if (midi_driver->mixer_volume)
+	 midi_driver->mixer_volume(midi_volume);
    }
 }
 
@@ -743,32 +790,18 @@ void set_volume(int digi_volume, int midi_volume)
  */
 void lock_sample(SAMPLE *spl)
 {
+   ASSERT(spl);
    LOCK_DATA(spl, sizeof(SAMPLE));
    LOCK_DATA(spl->data, spl->len * ((spl->bits==8) ? 1 : sizeof(short)) * ((spl->stereo) ? 2 : 1));
 }
 
 
 
-/* load_sample:
- *  Loads a sample from disk.
- */
-SAMPLE *load_sample(char *filename)
-{
-   if (ustricmp(get_extension(filename), uconvert_ascii("wav", NULL)) == 0)
-      return load_wav(filename);
-   else if (ustricmp(get_extension(filename), uconvert_ascii("voc", NULL)) == 0)
-      return load_voc(filename);
-   else
-      return NULL;
-}
-
-
-
 /* load_voc:
  *  Reads a mono VOC format sample file, returning a SAMPLE structure, 
  *  or NULL on error.
  */
-SAMPLE *load_voc(char *filename)
+SAMPLE *load_voc(AL_CONST char *filename)
 {
    PACKFILE *f;
    char buffer[30];
@@ -777,7 +810,8 @@ SAMPLE *load_voc(char *filename)
    SAMPLE *spl = NULL;
    int len;
    int x, ver;
-   signed short s;
+   int s;
+   ASSERT(filename);
 
    f = pack_fopen(filename, F_READ);
    if (!f) 
@@ -815,8 +849,7 @@ SAMPLE *load_voc(char *filename)
       spl = create_sample(8, FALSE, freq, len);
 
       if (spl) {
-	 pack_fread(spl->data, len, f);
-	 if (*allegro_errno) {
+	 if (pack_fread(spl->data, len, f) < len) {
 	    destroy_sample(spl);
 	    spl = NULL;
 	 }
@@ -842,19 +875,22 @@ SAMPLE *load_voc(char *filename)
 
       if (spl) {
 	 if (bits == 8) {
-	    pack_fread(spl->data, len, f);
+	    if (pack_fread(spl->data, len, f) < len) {
+	       destroy_sample(spl);
+	       spl = NULL;
+	    }
 	 }
 	 else {
 	    len /= 2;
 	    for (x=0; x<len; x++) {
-	       s = pack_igetw(f);
-	       ((signed short *)spl->data)[x] = s^0x8000;
+	       if ((s = pack_igetw(f)) == EOF) {
+		  destroy_sample(spl);
+		  spl = NULL;
+		  break;
+	       }
+	       ((signed short *)spl->data)[x] = (signed short)s^0x8000;
 	    }
 	 }
-	 if (*allegro_errno) {
-	    destroy_sample(spl);
-	    spl = NULL;
-	 }
       }
    }
 
@@ -869,7 +905,7 @@ SAMPLE *load_voc(char *filename)
  *  Reads a RIFF WAV format sample file, returning a SAMPLE structure, 
  *  or NULL on error.
  */
-SAMPLE *load_wav(char *filename)
+SAMPLE *load_wav(AL_CONST char *filename)
 {
    PACKFILE *f;
    char buffer[25];
@@ -878,8 +914,9 @@ SAMPLE *load_wav(char *filename)
    int freq = 22050;
    int bits = 8;
    int channels = 1;
-   signed short s;
+   int s;
    SAMPLE *spl = NULL;
+   ASSERT(filename);
 
    f = pack_fopen(filename, F_READ);
    if (!f)
@@ -889,7 +926,7 @@ SAMPLE *load_wav(char *filename)
    if (memcmp(buffer, "RIFF", 4) || memcmp(buffer+8, "WAVE", 4))
       goto getout;
 
-   while (!pack_feof(f)) {
+   while (TRUE) {
       if (pack_fread(buffer, 4, f) != 4)
 	 break;
 
@@ -926,23 +963,25 @@ SAMPLE *load_wav(char *filename)
 
 	 spl = create_sample(bits, ((channels == 2) ? TRUE : FALSE), freq, len);
 
-	 if (spl) { 
+	 if (spl) {
 	    if (bits == 8) {
-	       pack_fread(spl->data, length, f);
+	       if (pack_fread(spl->data, length, f) < length) {
+		  destroy_sample(spl);
+		  spl = NULL;
+	       }
 	    }
 	    else {
 	       for (i=0; i<len*channels; i++) {
-		  s = pack_igetw(f);
-		  ((signed short *)spl->data)[i] = s^0x8000;
+		  if ((s = pack_igetw(f)) == EOF) {
+		     destroy_sample(spl);
+		     spl = NULL;
+		     break;
+		  }
+		  ((signed short *)spl->data)[i] = (signed short)s^0x8000;
 	       }
 	    }
 
 	    length = 0;
-
-	    if (*allegro_errno) {
-	       destroy_sample(spl);
-	       spl = NULL;
-	    }
 	 }
       }
 
@@ -1020,8 +1059,9 @@ void destroy_sample(SAMPLE *spl)
  *  Converts a pitch from the relative 1000 = original format to an absolute
  *  value in hz.
  */
-static INLINE int absolute_freq(int freq, SAMPLE *spl)
+static INLINE int absolute_freq(int freq, AL_CONST SAMPLE *spl)
 {
+   ASSERT(spl);
    if (freq == 1000)
       return spl->freq;
    else
@@ -1040,10 +1080,12 @@ static INLINE int absolute_freq(int freq, SAMPLE *spl)
  *  stop_sample(), and can be manipulated while it is playing by calling
  *  adjust_sample().
  */
-int play_sample(SAMPLE *spl, int vol, int pan, int freq, int loop)
+int play_sample(AL_CONST SAMPLE *spl, int vol, int pan, int freq, int loop)
 {
-   int voice = allocate_voice(spl);
+   int voice;
+   ASSERT(spl);
 
+   voice = allocate_voice(spl);
    if (voice >= 0) {
       voice_set_volume(voice, vol);
       voice_set_pan(voice, pan);
@@ -1069,9 +1111,10 @@ END_OF_FUNCTION(play_sample);
  *  first one it comes across. If the sample is not playing it has no
  *  effect.
  */
-void adjust_sample(SAMPLE *spl, int vol, int pan, int freq, int loop)
+void adjust_sample(AL_CONST SAMPLE *spl, int vol, int pan, int freq, int loop)
 {
    int c;
+   ASSERT(spl);
 
    for (c=0; c<VIRTUAL_VOICES; c++) { 
       if (_voice[c].sample == spl) {
@@ -1093,9 +1136,10 @@ END_OF_FUNCTION(adjust_sample);
  *  in looped mode. If there are several copies of the sample playing,
  *  it will stop them all.
  */
-void stop_sample(SAMPLE *spl)
+void stop_sample(AL_CONST SAMPLE *spl)
 {
    int c;
+   ASSERT(spl);
 
    for (c=0; c<VIRTUAL_VOICES; c++)
       if (_voice[c].sample == spl)
@@ -1176,7 +1220,7 @@ static INLINE int allocate_physical_voice(int priority)
  *  others to make room, as we allow up to 256 virtual voices to be used
  *  simultaneously.
  */
-static INLINE int allocate_virtual_voice()
+static INLINE int allocate_virtual_voice(void)
 {
    int virt_voices, c;
 
@@ -1222,10 +1266,13 @@ static INLINE int allocate_virtual_voice()
  *  since there are 256 virtual voices and anyone who needs more than that
  *  needs some urgent repairs to their brain :-)
  */
-int allocate_voice(SAMPLE *spl)
+int allocate_voice(AL_CONST SAMPLE *spl)
 {
-   int phys = allocate_physical_voice(spl->priority);
-   int virt = allocate_virtual_voice();
+   int phys, virt;
+   ASSERT(spl);
+   
+   phys = allocate_physical_voice(spl->priority);
+   virt = allocate_virtual_voice();
 
    if (virt >= 0) {
       _voice[virt].sample = spl;
@@ -1277,9 +1324,10 @@ END_OF_FUNCTION(deallocate_voice);
 /* reallocate_voice:
  *  Switches an already-allocated voice to use a different sample.
  */
-void reallocate_voice(int voice, SAMPLE *spl)
+void reallocate_voice(int voice, AL_CONST SAMPLE *spl)
 {
    int phys = _voice[voice].num;
+   ASSERT(spl);
 
    if (phys >= 0) {
       digi_driver->stop_voice(phys);
@@ -1375,7 +1423,7 @@ SAMPLE *voice_check(int voice)
 	 if (voice_get_position(voice) < 0)
 	    return NULL;
 
-      return _voice[voice].sample;
+      return (SAMPLE*)_voice[voice].sample;
    }
    else
       return NULL;
@@ -1885,7 +1933,7 @@ END_OF_FUNCTION(read_sound_input);
 static void sound_lock_mem()
 {
    LOCK_VARIABLE(digi_none);
-   LOCK_VARIABLE(midi_none);
+   LOCK_VARIABLE(_midi_none);
    LOCK_VARIABLE(digi_card);
    LOCK_VARIABLE(midi_card);
    LOCK_VARIABLE(digi_driver);
