@@ -21,10 +21,19 @@
 
 
 BEGIN_DIGI_DRIVER_LIST
-#ifdef DIGI_ESD
+#if (defined DIGI_JACK) && (!defined ALLEGRO_WITH_MODULES)
+   DIGI_DRIVER_JACK
+#endif
+#if (defined DIGI_SGIAL) && (!defined ALLEGRO_WITH_MODULES)
+   DIGI_DRIVER_SGIAL
+#endif
+#if (defined DIGI_ARTS) && (!defined ALLEGRO_WITH_MODULES)
+   DIGI_DRIVER_ARTS
+#endif
+#if (defined DIGI_ESD) && (!defined ALLEGRO_WITH_MODULES)
    DIGI_DRIVER_ESD
 #endif
-#ifdef DIGI_ALSA
+#if (defined DIGI_ALSA) && (!defined ALLEGRO_WITH_MODULES)
    DIGI_DRIVER_ALSA
 #endif
 #ifdef DIGI_OSS
@@ -35,6 +44,9 @@ END_DIGI_DRIVER_LIST
 
 BEGIN_MIDI_DRIVER_LIST
    MIDI_DRIVER_DIGMID
+#if (defined MIDI_ALSA) && (!defined ALLEGRO_WITH_MODULES)
+   MIDI_DRIVER_ALSA
+#endif
 #ifdef MIDI_OSS
    MIDI_DRIVER_OSS
 #endif
