@@ -16,8 +16,8 @@
  */
 
 #include "allegro.h"
-#include "allegro/aintern.h"
-#include "allegro/aintunix.h"
+#include "allegro/internal/aintern.h"
+#include "allegro/platform/aintunix.h"
 #include "xwin.h"
 
 
@@ -25,7 +25,6 @@
 static GFX_VTABLE _xwin_vtable;
 
 
-
 static void _xwin_putpixel(BITMAP *dst, int dx, int dy, int color);
 static void _xwin_vline(BITMAP *dst, int dx, int dy1, int dy2, int color);
 static void _xwin_hline(BITMAP *dst, int dx1, int dy, int dx2, int color);
@@ -38,12 +37,12 @@ static void _xwin_draw_sprite_vh_flip(BITMAP *dst, BITMAP *src, int dx, int dy);
 static void _xwin_draw_trans_sprite(BITMAP *dst, BITMAP *src, int dx, int dy);
 static void _xwin_draw_trans_rgba_sprite(BITMAP *dst, BITMAP *src, int dx, int dy);
 static void _xwin_draw_lit_sprite(BITMAP *dst, BITMAP *src, int dx, int dy, int color);
-static void _xwin_draw_rle_sprite(BITMAP *dst, RLE_SPRITE *src, int dx, int dy);
-static void _xwin_draw_trans_rle_sprite(BITMAP *dst, RLE_SPRITE *src, int dx, int dy);
-static void _xwin_draw_trans_rgba_rle_sprite(BITMAP *dst, RLE_SPRITE *src, int dx, int dy);
-static void _xwin_draw_lit_rle_sprite(BITMAP *dst, RLE_SPRITE *src, int dx, int dy, int color);
-static void _xwin_draw_character(BITMAP *dst, BITMAP *src, int dx, int dy, int color);
-static void _xwin_draw_glyph(BITMAP *dst, FONT_GLYPH *src, int dx, int dy, int color);
+static void _xwin_draw_rle_sprite(BITMAP *dst, AL_CONST RLE_SPRITE *src, int dx, int dy);
+static void _xwin_draw_trans_rle_sprite(BITMAP *dst, AL_CONST RLE_SPRITE *src, int dx, int dy);
+static void _xwin_draw_trans_rgba_rle_sprite(BITMAP *dst, AL_CONST RLE_SPRITE *src, int dx, int dy);
+static void _xwin_draw_lit_rle_sprite(BITMAP *dst, AL_CONST RLE_SPRITE *src, int dx, int dy, int color);
+static void _xwin_draw_character(BITMAP *dst, BITMAP *src, int dx, int dy, int color, int bg);
+static void _xwin_draw_glyph(BITMAP *dst, AL_CONST FONT_GLYPH *src, int dx, int dy, int color, int bg);
 static void _xwin_blit_anywhere(BITMAP *src, BITMAP *dst, int sx, int sy,
 				int dx, int dy, int w, int h);
 static void _xwin_blit_backward(BITMAP *src, BITMAP *dst, int sx, int sy,
@@ -59,13 +58,12 @@ static void _xwin_clear_to_color(BITMAP *dst, int color);
  */
 void _xwin_replace_vtable(struct GFX_VTABLE *vtable)
 {
-   DISABLE();
-
    memcpy(&_xwin_vtable, vtable, sizeof(GFX_VTABLE));
 
    vtable->putpixel = _xwin_putpixel;
    vtable->vline = _xwin_vline;
    vtable->hline = _xwin_hline;
+   vtable->hfill = _xwin_hline;
    vtable->rectfill = _xwin_rectfill;
    vtable->draw_sprite = _xwin_draw_sprite;
    vtable->draw_256_sprite = _xwin_draw_256_sprite;
@@ -90,14 +88,12 @@ void _xwin_replace_vtable(struct GFX_VTABLE *vtable)
    vtable->blit_to_self_backward = _xwin_blit_backward;
    vtable->masked_blit = _xwin_masked_blit;
    vtable->clear_to_color = _xwin_clear_to_color;
-
-   ENABLE();
 }
 
 
 
 /* _xwin_update_video_bitmap:
- *  Update screen with correct offset for sub bitmap.
+ *  Update screen with correct offset for sub-bitmap.
  */
 static void _xwin_update_video_bitmap(BITMAP *dst, int x, int y, int w, int h)
 {
@@ -142,6 +138,7 @@ static void _xwin_hline(BITMAP *dst, int dx1, int dy, int dx2, int color)
       dx1 = dx2;
       dx2 = tmp;
    }
+
    if (dst->clip) {
       if (dx1 < dst->cl)
 	 dx1 = dst->cl;
@@ -174,6 +171,7 @@ static void _xwin_vline(BITMAP *dst, int dx, int dy1, int dy2, int color)
       dy1 = dy2;
       dy2 = tmp;
    }
+
    if (dst->clip) {
       if (dy1 < dst->ct)
 	 dy1 = dst->ct;
@@ -312,49 +310,52 @@ static void _xwin_masked_blit(BITMAP *src, BITMAP *dst, int sx, int sy,
 
 
 
+#define CLIP_BOX(dst, x, y, w, h, x_orig, y_orig, w_orig, h_orig)  \
+{                                                                  \
+   if (dst->clip) {                                                \
+      int tmp, x_delta, y_delta;                                   \
+                                                                   \
+      tmp = dst->cl - x_orig;                                      \
+      x_delta = ((tmp < 0) ? 0 : tmp);                             \
+      x = x_orig + x_delta;                                        \
+                                                                   \
+      tmp = dst->cr - x_orig;                                      \
+      w = ((tmp > w_orig) ? w_orig : tmp) - x_delta;               \
+      if (w <= 0)                                                  \
+	 return;                                                   \
+                                                                   \
+      tmp = dst->ct - y_orig;                                      \
+      y_delta = ((tmp < 0) ? 0 : tmp);                             \
+      y = y_orig + y_delta;                                        \
+                                                                   \
+      tmp = dst->cb - y_orig;                                      \
+      h = ((tmp > h_orig) ? h_orig : tmp) - y_delta;               \
+      if (h <= 0)                                                  \
+	 return;                                                   \
+   }                                                               \
+   else {                                                          \
+      x = x_orig;                                                  \
+      y = y_orig;                                                  \
+      w = w_orig;                                                  \
+      h = h_orig;                                                  \
+   }                                                               \
+}
+
+
+
 /* _xwin_draw_sprite:
  *  Wrapper for draw_sprite.
  */
 static void _xwin_draw_sprite(BITMAP *dst, BITMAP *src, int dx, int dy)
 {
-   int w, h;
-   int dxbeg, dybeg;
-   int sxbeg, sybeg;
+   int dxbeg, dybeg, w, h;
 
    if (_xwin_in_gfx_call) {
       _xwin_vtable.draw_sprite(dst, src, dx, dy);
       return;
    }
 
-   if (dst->clip) {
-      int tmp;
-
-      tmp = dst->cl - dx;
-      sxbeg = ((tmp < 0) ? 0 : tmp);
-      dxbeg = sxbeg + dx;
-
-      tmp = dst->cr - dx;
-      w = ((tmp > src->w) ? src->w : tmp) - sxbeg;
-      if (w <= 0)
-	 return;
-
-      tmp = dst->ct - dy;
-      sybeg = ((tmp < 0) ? 0 : tmp);
-      dybeg = sybeg + dy;
-
-      tmp = dst->cb - dy;
-      h = ((tmp > src->h) ? src->h : tmp) - sybeg;
-      if (h <= 0)
-	 return;
-   }
-   else {
-      w = src->w;
-      h = src->h;
-      sxbeg = 0;
-      sybeg = 0;
-      dxbeg = dx;
-      dybeg = dy;
-   }
+   CLIP_BOX(dst, dxbeg, dybeg, w, h, dx, dy, src->w, src->h)
 
    _xwin_in_gfx_call = 1;
    _xwin_vtable.draw_sprite(dst, src, dx, dy);
@@ -369,44 +370,14 @@ static void _xwin_draw_sprite(BITMAP *dst, BITMAP *src, int dx, int dy)
  */
 static void _xwin_draw_256_sprite(BITMAP *dst, BITMAP *src, int dx, int dy)
 {
-   int w, h;
-   int dxbeg, dybeg;
-   int sxbeg, sybeg;
+   int dxbeg, dybeg, w, h;
 
    if (_xwin_in_gfx_call) {
       _xwin_vtable.draw_256_sprite(dst, src, dx, dy);
       return;
    }
 
-   if (dst->clip) {
-      int tmp;
-
-      tmp = dst->cl - dx;
-      sxbeg = ((tmp < 0) ? 0 : tmp);
-      dxbeg = sxbeg + dx;
-
-      tmp = dst->cr - dx;
-      w = ((tmp > src->w) ? src->w : tmp) - sxbeg;
-      if (w <= 0)
-	 return;
-
-      tmp = dst->ct - dy;
-      sybeg = ((tmp < 0) ? 0 : tmp);
-      dybeg = sybeg + dy;
-
-      tmp = dst->cb - dy;
-      h = ((tmp > src->h) ? src->h : tmp) - sybeg;
-      if (h <= 0)
-	 return;
-   }
-   else {
-      w = src->w;
-      h = src->h;
-      sxbeg = 0;
-      sybeg = 0;
-      dxbeg = dx;
-      dybeg = dy;
-   }
+   CLIP_BOX(dst, dxbeg, dybeg, w, h, dx, dy, src->w, src->h)
 
    _xwin_in_gfx_call = 1;
    _xwin_vtable.draw_256_sprite(dst, src, dx, dy);
@@ -421,44 +392,14 @@ static void _xwin_draw_256_sprite(BITMAP *dst, BITMAP *src, int dx, int dy)
  */
 static void _xwin_draw_sprite_v_flip(BITMAP *dst, BITMAP *src, int dx, int dy)
 {
-   int w, h;
-   int dxbeg, dybeg;
-   int sxbeg, sybeg;
+   int dxbeg, dybeg, w, h;
 
    if (_xwin_in_gfx_call) {
       _xwin_vtable.draw_sprite_v_flip(dst, src, dx, dy);
       return;
    }
 
-   if (dst->clip) {
-      int tmp;
-
-      tmp = dst->cl - dx;
-      sxbeg = ((tmp < 0) ? 0 : tmp);
-      dxbeg = sxbeg + dx;
-
-      tmp = dst->cr - dx;
-      w = ((tmp > src->w) ? src->w : tmp) - sxbeg;
-      if (w <= 0)
-	 return;
-
-      tmp = dst->ct - dy;
-      sybeg = ((tmp < 0) ? 0 : tmp);
-      dybeg = sybeg + dy;
-
-      tmp = dst->cb - dy;
-      h = ((tmp > src->h) ? src->h : tmp) - sybeg;
-      if (h <= 0)
-	 return;
-   }
-   else {
-      w = src->w;
-      h = src->h;
-      sxbeg = 0;
-      sybeg = 0;
-      dxbeg = dx;
-      dybeg = dy;
-   }
+   CLIP_BOX(dst, dxbeg, dybeg, w, h, dx, dy, src->w, src->h)
 
    _xwin_in_gfx_call = 1;
    _xwin_vtable.draw_sprite_v_flip(dst, src, dx, dy);
@@ -473,44 +414,14 @@ static void _xwin_draw_sprite_v_flip(BITMAP *dst, BITMAP *src, int dx, int dy)
  */
 static void _xwin_draw_sprite_h_flip(BITMAP *dst, BITMAP *src, int dx, int dy)
 {
-   int w, h;
-   int dxbeg, dybeg;
-   int sxbeg, sybeg;
+   int dxbeg, dybeg, w, h;
 
    if (_xwin_in_gfx_call) {
       _xwin_vtable.draw_sprite_h_flip(dst, src, dx, dy);
       return;
    }
 
-   if (dst->clip) {
-      int tmp;
-
-      tmp = dst->cl - dx;
-      sxbeg = ((tmp < 0) ? 0 : tmp);
-      dxbeg = sxbeg + dx;
-
-      tmp = dst->cr - dx;
-      w = ((tmp > src->w) ? src->w : tmp) - sxbeg;
-      if (w <= 0)
-	 return;
-
-      tmp = dst->ct - dy;
-      sybeg = ((tmp < 0) ? 0 : tmp);
-      dybeg = sybeg + dy;
-
-      tmp = dst->cb - dy;
-      h = ((tmp > src->h) ? src->h : tmp) - sybeg;
-      if (h <= 0)
-	 return;
-   }
-   else {
-      w = src->w;
-      h = src->h;
-      sxbeg = 0;
-      sybeg = 0;
-      dxbeg = dx;
-      dybeg = dy;
-   }
+   CLIP_BOX(dst, dxbeg, dybeg, w, h, dx, dy, src->w, src->h)
 
    _xwin_in_gfx_call = 1;
    _xwin_vtable.draw_sprite_h_flip(dst, src, dx, dy);
@@ -525,44 +436,14 @@ static void _xwin_draw_sprite_h_flip(BITMAP *dst, BITMAP *src, int dx, int dy)
  */
 static void _xwin_draw_sprite_vh_flip(BITMAP *dst, BITMAP *src, int dx, int dy)
 {
-   int w, h;
-   int dxbeg, dybeg;
-   int sxbeg, sybeg;
+   int dxbeg, dybeg, w, h;
 
    if (_xwin_in_gfx_call) {
       _xwin_vtable.draw_sprite_vh_flip(dst, src, dx, dy);
       return;
    }
 
-   if (dst->clip) {
-      int tmp;
-
-      tmp = dst->cl - dx;
-      sxbeg = ((tmp < 0) ? 0 : tmp);
-      dxbeg = sxbeg + dx;
-
-      tmp = dst->cr - dx;
-      w = ((tmp > src->w) ? src->w : tmp) - sxbeg;
-      if (w <= 0)
-	 return;
-
-      tmp = dst->ct - dy;
-      sybeg = ((tmp < 0) ? 0 : tmp);
-      dybeg = sybeg + dy;
-
-      tmp = dst->cb - dy;
-      h = ((tmp > src->h) ? src->h : tmp) - sybeg;
-      if (h <= 0)
-	 return;
-   }
-   else {
-      w = src->w;
-      h = src->h;
-      sxbeg = 0;
-      sybeg = 0;
-      dxbeg = dx;
-      dybeg = dy;
-   }
+   CLIP_BOX(dst, dxbeg, dybeg, w, h, dx, dy, src->w, src->h)
 
    _xwin_in_gfx_call = 1;
    _xwin_vtable.draw_sprite_vh_flip(dst, src, dx, dy);
@@ -577,44 +458,14 @@ static void _xwin_draw_sprite_vh_flip(BITMAP *dst, BITMAP *src, int dx, int dy)
  */
 static void _xwin_draw_trans_sprite(BITMAP *dst, BITMAP *src, int dx, int dy)
 {
-   int w, h;
-   int dxbeg, dybeg;
-   int sxbeg, sybeg;
+   int dxbeg, dybeg, w, h;
 
    if (_xwin_in_gfx_call) {
       _xwin_vtable.draw_trans_sprite(dst, src, dx, dy);
       return;
    }
 
-   if (dst->clip) {
-      int tmp;
-
-      tmp = dst->cl - dx;
-      sxbeg = ((tmp < 0) ? 0 : tmp);
-      dxbeg = sxbeg + dx;
-
-      tmp = dst->cr - dx;
-      w = ((tmp > src->w) ? src->w : tmp) - sxbeg;
-      if (w <= 0)
-	 return;
-
-      tmp = dst->ct - dy;
-      sybeg = ((tmp < 0) ? 0 : tmp);
-      dybeg = sybeg + dy;
-
-      tmp = dst->cb - dy;
-      h = ((tmp > src->h) ? src->h : tmp) - sybeg;
-      if (h <= 0)
-	 return;
-   }
-   else {
-      w = src->w;
-      h = src->h;
-      sxbeg = 0;
-      sybeg = 0;
-      dxbeg = dx;
-      dybeg = dy;
-   }
+   CLIP_BOX(dst, dxbeg, dybeg, w, h, dx, dy, src->w, src->h)
 
    _xwin_in_gfx_call = 1;
    _xwin_vtable.draw_trans_sprite(dst, src, dx, dy);
@@ -629,44 +480,14 @@ static void _xwin_draw_trans_sprite(BITMAP *dst, BITMAP *src, int dx, int dy)
  */
 static void _xwin_draw_trans_rgba_sprite(BITMAP *dst, BITMAP *src, int dx, int dy)
 {
-   int w, h;
-   int dxbeg, dybeg;
-   int sxbeg, sybeg;
+   int dxbeg, dybeg, w, h;
 
    if (_xwin_in_gfx_call) {
       _xwin_vtable.draw_trans_rgba_sprite(dst, src, dx, dy);
       return;
    }
 
-   if (dst->clip) {
-      int tmp;
-
-      tmp = dst->cl - dx;
-      sxbeg = ((tmp < 0) ? 0 : tmp);
-      dxbeg = sxbeg + dx;
-
-      tmp = dst->cr - dx;
-      w = ((tmp > src->w) ? src->w : tmp) - sxbeg;
-      if (w <= 0)
-	 return;
-
-      tmp = dst->ct - dy;
-      sybeg = ((tmp < 0) ? 0 : tmp);
-      dybeg = sybeg + dy;
-
-      tmp = dst->cb - dy;
-      h = ((tmp > src->h) ? src->h : tmp) - sybeg;
-      if (h <= 0)
-	 return;
-   }
-   else {
-      w = src->w;
-      h = src->h;
-      sxbeg = 0;
-      sybeg = 0;
-      dxbeg = dx;
-      dybeg = dy;
-   }
+   CLIP_BOX(dst, dxbeg, dybeg, w, h, dx, dy, src->w, src->h)
 
    _xwin_in_gfx_call = 1;
    _xwin_vtable.draw_trans_rgba_sprite(dst, src, dx, dy);
@@ -681,44 +502,14 @@ static void _xwin_draw_trans_rgba_sprite(BITMAP *dst, BITMAP *src, int dx, int d
  */
 static void _xwin_draw_lit_sprite(BITMAP *dst, BITMAP *src, int dx, int dy, int color)
 {
-   int w, h;
-   int dxbeg, dybeg;
-   int sxbeg, sybeg;
+   int dxbeg, dybeg, w, h;
 
    if (_xwin_in_gfx_call) {
       _xwin_vtable.draw_lit_sprite(dst, src, dx, dy, color);
       return;
    }
 
-   if (dst->clip) {
-      int tmp;
-
-      tmp = dst->cl - dx;
-      sxbeg = ((tmp < 0) ? 0 : tmp);
-      dxbeg = sxbeg + dx;
-
-      tmp = dst->cr - dx;
-      w = ((tmp > src->w) ? src->w : tmp) - sxbeg;
-      if (w <= 0)
-	 return;
-
-      tmp = dst->ct - dy;
-      sybeg = ((tmp < 0) ? 0 : tmp);
-      dybeg = sybeg + dy;
-
-      tmp = dst->cb - dy;
-      h = ((tmp > src->h) ? src->h : tmp) - sybeg;
-      if (h <= 0)
-	 return;
-   }
-   else {
-      w = src->w;
-      h = src->h;
-      sxbeg = 0;
-      sybeg = 0;
-      dxbeg = dx;
-      dybeg = dy;
-   }
+   CLIP_BOX(dst, dxbeg, dybeg, w, h, dx, dy, src->w, src->h)
 
    _xwin_in_gfx_call = 1;
    _xwin_vtable.draw_lit_sprite(dst, src, dx, dy, color);
@@ -731,49 +522,19 @@ static void _xwin_draw_lit_sprite(BITMAP *dst, BITMAP *src, int dx, int dy, int
 /* _xwin_draw_character:
  *  Wrapper for draw_character.
  */
-static void _xwin_draw_character(BITMAP *dst, BITMAP *src, int dx, int dy, int color)
+static void _xwin_draw_character(BITMAP *dst, BITMAP *src, int dx, int dy, int color, int bg)
 {
-   int w, h;
-   int dxbeg, dybeg;
-   int sxbeg, sybeg;
+   int dxbeg, dybeg, w, h;
 
    if (_xwin_in_gfx_call) {
-      _xwin_vtable.draw_character(dst, src, dx, dy, color);
+      _xwin_vtable.draw_character(dst, src, dx, dy, color, bg);
       return;
    }
 
-   if (dst->clip) {
-      int tmp;
-
-      tmp = dst->cl - dx;
-      sxbeg = ((tmp < 0) ? 0 : tmp);
-      dxbeg = sxbeg + dx;
-
-      tmp = dst->cr - dx;
-      w = ((tmp > src->w) ? src->w : tmp) - sxbeg;
-      if (w <= 0)
-	 return;
-
-      tmp = dst->ct - dy;
-      sybeg = ((tmp < 0) ? 0 : tmp);
-      dybeg = sybeg + dy;
-
-      tmp = dst->cb - dy;
-      h = ((tmp > src->h) ? src->h : tmp) - sybeg;
-      if (h <= 0)
-	 return;
-   }
-   else {
-      w = src->w;
-      h = src->h;
-      sxbeg = 0;
-      sybeg = 0;
-      dxbeg = dx;
-      dybeg = dy;
-   }
+   CLIP_BOX(dst, dxbeg, dybeg, w, h, dx, dy, src->w, src->h)
 
    _xwin_in_gfx_call = 1;
-   _xwin_vtable.draw_character(dst, src, dx, dy, color);
+   _xwin_vtable.draw_character(dst, src, dx, dy, color, bg);
    _xwin_in_gfx_call = 0;
    _xwin_update_video_bitmap(dst, dxbeg, dybeg, w, h);
 }
@@ -783,49 +544,19 @@ static void _xwin_draw_character(BITMAP *dst, BITMAP *src, int dx, int dy, int c
 /* _xwin_draw_glyph:
  *  Wrapper for draw_glyph.
  */
-static void _xwin_draw_glyph(BITMAP *dst, FONT_GLYPH *src, int dx, int dy, int color)
+static void _xwin_draw_glyph(BITMAP *dst, AL_CONST FONT_GLYPH *src, int dx, int dy, int color, int bg)
 {
-   int w, h;
-   int dxbeg, dybeg;
-   int sxbeg, sybeg;
+   int dxbeg, dybeg, w, h;
 
    if (_xwin_in_gfx_call) {
-      _xwin_vtable.draw_glyph(dst, src, dx, dy, color);
+      _xwin_vtable.draw_glyph(dst, src, dx, dy, color, bg);
       return;
    }
 
-   if (dst->clip) {
-      int tmp;
-
-      tmp = dst->cl - dx;
-      sxbeg = ((tmp < 0) ? 0 : tmp);
-      dxbeg = sxbeg + dx;
-
-      tmp = dst->cr - dx;
-      w = ((tmp > src->w) ? src->w : tmp) - sxbeg;
-      if (w <= 0)
-	 return;
-
-      tmp = dst->ct - dy;
-      sybeg = ((tmp < 0) ? 0 : tmp);
-      dybeg = sybeg + dy;
-
-      tmp = dst->cb - dy;
-      h = ((tmp > src->h) ? src->h : tmp) - sybeg;
-      if (h <= 0)
-	 return;
-   }
-   else {
-      w = src->w;
-      h = src->h;
-      sxbeg = 0;
-      sybeg = 0;
-      dxbeg = dx;
-      dybeg = dy;
-   }
+   CLIP_BOX(dst, dxbeg, dybeg, w, h, dx, dy, src->w, src->h)
 
    _xwin_in_gfx_call = 1;
-   _xwin_vtable.draw_glyph(dst, src, dx, dy, color);
+   _xwin_vtable.draw_glyph(dst, src, dx, dy, color, bg);
    _xwin_in_gfx_call = 0;
    _xwin_update_video_bitmap(dst, dxbeg, dybeg, w, h);
 }
@@ -835,46 +566,16 @@ static void _xwin_draw_glyph(BITMAP *dst, FONT_GLYPH *src, int dx, int dy, int c
 /* _xwin_draw_rle_sprite:
  *  Wrapper for draw_rle_sprite.
  */
-static void _xwin_draw_rle_sprite(BITMAP *dst, RLE_SPRITE *src, int dx, int dy)
+static void _xwin_draw_rle_sprite(BITMAP *dst, AL_CONST RLE_SPRITE *src, int dx, int dy)
 {
-   int w, h;
-   int dxbeg, dybeg;
-   int sxbeg, sybeg;
+   int dxbeg, dybeg, w, h;
 
    if (_xwin_in_gfx_call) {
       _xwin_vtable.draw_rle_sprite(dst, src, dx, dy);
       return;
    }
 
-   if (dst->clip) {
-      int tmp;
-
-      tmp = dst->cl - dx;
-      sxbeg = ((tmp < 0) ? 0 : tmp);
-      dxbeg = sxbeg + dx;
-
-      tmp = dst->cr - dx;
-      w = ((tmp > src->w) ? src->w : tmp) - sxbeg;
-      if (w <= 0)
-	 return;
-
-      tmp = dst->ct - dy;
-      sybeg = ((tmp < 0) ? 0 : tmp);
-      dybeg = sybeg + dy;
-
-      tmp = dst->cb - dy;
-      h = ((tmp > src->h) ? src->h : tmp) - sybeg;
-      if (h <= 0)
-	 return;
-   }
-   else {
-      w = src->w;
-      h = src->h;
-      sxbeg = 0;
-      sybeg = 0;
-      dxbeg = dx;
-      dybeg = dy;
-   }
+   CLIP_BOX(dst, dxbeg, dybeg, w, h, dx, dy, src->w, src->h)
 
    _xwin_in_gfx_call = 1;
    _xwin_vtable.draw_rle_sprite(dst, src, dx, dy);
@@ -887,46 +588,16 @@ static void _xwin_draw_rle_sprite(BITMAP *dst, RLE_SPRITE *src, int dx, int dy)
 /* _xwin_draw_trans_rle_sprite:
  *  Wrapper for draw_trans_rle_sprite.
  */
-static void _xwin_draw_trans_rle_sprite(BITMAP *dst, RLE_SPRITE *src, int dx, int dy)
+static void _xwin_draw_trans_rle_sprite(BITMAP *dst, AL_CONST RLE_SPRITE *src, int dx, int dy)
 {
-   int w, h;
-   int dxbeg, dybeg;
-   int sxbeg, sybeg;
+   int dxbeg, dybeg, w, h;
 
    if (_xwin_in_gfx_call) {
       _xwin_vtable.draw_trans_rle_sprite(dst, src, dx, dy);
       return;
    }
 
-   if (dst->clip) {
-      int tmp;
-
-      tmp = dst->cl - dx;
-      sxbeg = ((tmp < 0) ? 0 : tmp);
-      dxbeg = sxbeg + dx;
-
-      tmp = dst->cr - dx;
-      w = ((tmp > src->w) ? src->w : tmp) - sxbeg;
-      if (w <= 0)
-	 return;
-
-      tmp = dst->ct - dy;
-      sybeg = ((tmp < 0) ? 0 : tmp);
-      dybeg = sybeg + dy;
-
-      tmp = dst->cb - dy;
-      h = ((tmp > src->h) ? src->h : tmp) - sybeg;
-      if (h <= 0)
-	 return;
-   }
-   else {
-      w = src->w;
-      h = src->h;
-      sxbeg = 0;
-      sybeg = 0;
-      dxbeg = dx;
-      dybeg = dy;
-   }
+   CLIP_BOX(dst, dxbeg, dybeg, w, h, dx, dy, src->w, src->h)
 
    _xwin_in_gfx_call = 1;
    _xwin_vtable.draw_trans_rle_sprite(dst, src, dx, dy);
@@ -939,46 +610,16 @@ static void _xwin_draw_trans_rle_sprite(BITMAP *dst, RLE_SPRITE *src, int dx, in
 /* _xwin_draw_trans_rgba_rle_sprite:
  *  Wrapper for draw_trans_rgba_rle_sprite.
  */
-static void _xwin_draw_trans_rgba_rle_sprite(BITMAP *dst, RLE_SPRITE *src, int dx, int dy)
+static void _xwin_draw_trans_rgba_rle_sprite(BITMAP *dst, AL_CONST RLE_SPRITE *src, int dx, int dy)
 {
-   int w, h;
-   int dxbeg, dybeg;
-   int sxbeg, sybeg;
+   int dxbeg, dybeg, w, h;
 
    if (_xwin_in_gfx_call) {
       _xwin_vtable.draw_trans_rgba_rle_sprite(dst, src, dx, dy);
       return;
    }
 
-   if (dst->clip) {
-      int tmp;
-
-      tmp = dst->cl - dx;
-      sxbeg = ((tmp < 0) ? 0 : tmp);
-      dxbeg = sxbeg + dx;
-
-      tmp = dst->cr - dx;
-      w = ((tmp > src->w) ? src->w : tmp) - sxbeg;
-      if (w <= 0)
-	 return;
-
-      tmp = dst->ct - dy;
-      sybeg = ((tmp < 0) ? 0 : tmp);
-      dybeg = sybeg + dy;
-
-      tmp = dst->cb - dy;
-      h = ((tmp > src->h) ? src->h : tmp) - sybeg;
-      if (h <= 0)
-	 return;
-   }
-   else {
-      w = src->w;
-      h = src->h;
-      sxbeg = 0;
-      sybeg = 0;
-      dxbeg = dx;
-      dybeg = dy;
-   }
+   CLIP_BOX(dst, dxbeg, dybeg, w, h, dx, dy, src->w, src->h)
 
    _xwin_in_gfx_call = 1;
    _xwin_vtable.draw_trans_rgba_rle_sprite(dst, src, dx, dy);
@@ -991,46 +632,16 @@ static void _xwin_draw_trans_rgba_rle_sprite(BITMAP *dst, RLE_SPRITE *src, int d
 /* _xwin_draw_lit_rle_sprite:
  *  Wrapper for draw_lit_rle_sprite.
  */
-static void _xwin_draw_lit_rle_sprite(BITMAP *dst, RLE_SPRITE *src, int dx, int dy, int color)
+static void _xwin_draw_lit_rle_sprite(BITMAP *dst, AL_CONST RLE_SPRITE *src, int dx, int dy, int color)
 {
-   int w, h;
-   int dxbeg, dybeg;
-   int sxbeg, sybeg;
+   int dxbeg, dybeg, w, h;
 
    if (_xwin_in_gfx_call) {
       _xwin_vtable.draw_lit_rle_sprite(dst, src, dx, dy, color);
       return;
    }
 
-   if (dst->clip) {
-      int tmp;
-
-      tmp = dst->cl - dx;
-      sxbeg = ((tmp < 0) ? 0 : tmp);
-      dxbeg = sxbeg + dx;
-
-      tmp = dst->cr - dx;
-      w = ((tmp > src->w) ? src->w : tmp) - sxbeg;
-      if (w <= 0)
-	 return;
-
-      tmp = dst->ct - dy;
-      sybeg = ((tmp < 0) ? 0 : tmp);
-      dybeg = sybeg + dy;
-
-      tmp = dst->cb - dy;
-      h = ((tmp > src->h) ? src->h : tmp) - sybeg;
-      if (h <= 0)
-	 return;
-   }
-   else {
-      w = src->w;
-      h = src->h;
-      sxbeg = 0;
-      sybeg = 0;
-      dxbeg = dx;
-      dybeg = dy;
-   }
+   CLIP_BOX(dst, dxbeg, dybeg, w, h, dx, dy, src->w, src->h)
 
    _xwin_in_gfx_call = 1;
    _xwin_vtable.draw_lit_rle_sprite(dst, src, dx, dy, color);
