@@ -25,13 +25,13 @@
 /* this program is not portable! */
 #if (defined ALLEGRO_DOS) || (defined ALLEGRO_LINUX_VBEAF)
 
-#include "allegro/aintern.h"
+#include "allegro/internal/aintern.h"
 
 #ifdef ALLEGRO_INTERNAL_HEADER
    #include ALLEGRO_INTERNAL_HEADER
 #endif
 
-#ifdef ALLEGRO_DJGPP
+#if (defined ALLEGRO_DJGPP) && (!defined SCAN_DEPEND)
    #include <sys/nearptr.h>
 #endif
 
@@ -217,18 +217,18 @@ MMAP af_linear_mem = NOMM;
 #ifdef ALLEGRO_DOS
 
 /* hooks to let the driver call BIOS routines */
-extern void _af_int86(), _af_call_rm();
+extern void _af_int86(void), _af_call_rm(void);
 
 #else
 
 /* no BIOS on this platform, my friend! */
-void nobios() { }
+void nobios(void) { }
 
 #endif
 
 
 
-void af_exit()
+void af_exit(void)
 {
    int c;
 
@@ -322,7 +322,7 @@ int load_vbeaf_driver(char *filename)
 
 
 
-void initialise_freebeaf_extensions()
+void initialise_freebeaf_extensions(void)
 {
    typedef unsigned long (*EXT_INIT_FUNC)(AF_DRIVER *af, unsigned long id);
    EXT_INIT_FUNC ext_init;
@@ -377,7 +377,7 @@ void initialise_freebeaf_extensions()
 
 
 
-static int initialise_vbeaf()
+static int initialise_vbeaf(void)
 {
    int c;
 
@@ -469,7 +469,7 @@ void print_af_attributes(unsigned long attrib)
 
 
 
-int get_af_info()
+int get_af_info(void)
 {
    static char *possible_filenames[] =
    {
@@ -485,7 +485,7 @@ int get_af_info()
    };
 
    char filename[256];
-   char *p;
+   AL_CONST char *p;
    int c, attrib;
 
    p = get_config_string(NULL, "vbeaf_driver", NULL);
@@ -735,7 +735,15 @@ int main(int argc, char *argv[])
 {
    int c;
 
-   allegro_init();
+   if (allegro_init() != 0)
+      return 1;
+
+   #ifdef SYSTEM_XWINDOWS
+      if (system_driver->id == SYSTEM_XWINDOWS) {
+	 allegro_message("Sorry, VBE/AF is not available under X, run AFINFO from Linux console\n");
+	 return 1;
+      }
+   #endif
 
    #ifdef ALLEGRO_LINUX
       if (!__al_linux_have_ioperms) {
@@ -771,7 +779,7 @@ int main(int argc, char *argv[])
 #else       /* ifdef VBE/AF cool on this platform */
 
 
-int main()
+int main(void)
 {
    allegro_init();
    allegro_message("Sorry, the AFINFO program only works on DOS or Linux\n");
