@@ -12,11 +12,13 @@
  *
  *      By Shawn Hargreaves.
  *
+ *      Nathan Smith added the recursive handling of directories.
+ *
  *      See readme.txt for copyright information.
  */
 
 
-#define USE_CONSOLE
+#define ALLEGRO_USE_CONSOLE
 
 #include <stdio.h>
 #include <string.h>
@@ -25,21 +27,21 @@
 #include "allegro.h"
 #include "datedit.h"
 
-#if (defined ALLEGRO_DOS) || (defined ALLEGRO_MSVC)
+#if ((defined ALLEGRO_DOS) || (defined ALLEGRO_WINDOWS)) && (!defined SCAN_DEPEND)
    #define HAVE_CONIO_H
    #include <conio.h>
 #endif
 
 
 
-static DATAFILE *datafile = NULL;
-
 static int err = 0;
 static int changed = 0;
 
 static int opt_command = 0;
 static int opt_compression = -1;
 static int opt_strip = -1;
+static int opt_sort = -1;
+static int opt_relf = FALSE;
 static int opt_verbose = FALSE;
 static int opt_keepnames = FALSE;
 static int opt_colordepth = -1;
@@ -47,7 +49,7 @@ static int opt_gridx = -1;
 static int opt_gridy = -1;
 static int opt_gridw = -1;
 static int opt_gridh = -1;
-static char *opt_datafile = NULL;
+static char *opt_datafilename = NULL;
 static char *opt_outputname = NULL;
 static char *opt_headername = NULL;
 static char *opt_dependencyfile = NULL;
@@ -56,19 +58,27 @@ static char *opt_prefixstring = NULL;
 static char *opt_password = NULL;
 static char *opt_palette = NULL;
 
+static int attrib_ok = FA_ARCH|FA_RDONLY;
+
+static char canonical_datafilename[1024];
+
 #define MAX_FILES    256
 
 static char *opt_proplist[MAX_FILES];
 static int opt_numprops = 0;
 
+static char *opt_fixedproplist[MAX_FILES];
+static int opt_numfixedprops = 0;
+
 static char *opt_namelist[MAX_FILES];
-static int opt_usedname[MAX_FILES];
 static int opt_numnames = 0;
 
+static int opt_usedname[MAX_FILES];
+
 
 
 /* display help on the command syntax */
-static void usage()
+static void usage(void)
 {
    printf("\nDatafile archiving utility for Allegro " ALLEGRO_VERSION_STR ", " ALLEGRO_PLATFORM_STR "\n");
    printf("By Shawn Hargreaves, " ALLEGRO_DATE_STR "\n\n");
@@ -82,21 +92,28 @@ static void usage()
    printf("\t'-d' deletes the named objects from the datafile\n");
    printf("\t'-dither' dithers when reducing color depths\n");
    printf("\t'-e' extracts the named objects from the datafile\n");
+   printf("\t'-f' store references to original files as relative filenames\n");
    printf("\t'-g x y w h' grabs bitmap data from a specific grid location\n");
    printf("\t'-h outputfile.h' sets the output header file\n");
-   printf("\t'-k' keeps the original file names when grabbing objects\n");
+   printf("\t'-k' keeps the original filenames when grabbing objects\n");
    printf("\t'-l' lists the contents of the datafile\n");
    printf("\t'-m dependencyfile' outputs makefile dependencies\n");
+   printf("\t'-n0' no sort: list the objects in the order they were added\n");
+   printf("\t'-n1' sort the objects of the datafile alphabetically by name\n");
    printf("\t'-o output' sets the output file or directory when extracting data\n");
    printf("\t'-p prefixstring' sets the prefix for the output header file\n");
    printf("\t'-pal objectname' specifies which palette to use\n");
+   printf("\t'-r' recursively adds directories as nested datafiles\n");
    printf("\t'-s0' no strip: save everything\n");
    printf("\t'-s1' strip grabber specific information from the file\n");
    printf("\t'-s2' strip all object properties and names from the file\n");
+   printf("\t'-s-PROP' do not strip object property PROP from the file\n");
    printf("\t'-t type' sets the object type when adding files\n");
+   printf("\t'-transparency' preserves transparency through color conversion\n");
    printf("\t'-u' updates the contents of the datafile\n");
    printf("\t'-v' selects verbose mode\n");
    printf("\t'-w' always updates the entire contents of the datafile\n");
+   printf("\t'-x' alias for -e\n");
    printf("\t'-007 password' sets the file encryption key\n");
    printf("\t'PROP=value' sets object properties\n");
 }
@@ -104,7 +121,7 @@ static void usage()
 
 
 /* callback for outputting messages */
-void datedit_msg(char *fmt, ...)
+void datedit_msg(AL_CONST char *fmt, ...)
 {
    va_list args;
    char buf[1024];
@@ -119,7 +136,7 @@ void datedit_msg(char *fmt, ...)
 
 
 /* callback for starting a 2-part message output */
-void datedit_startmsg(char *fmt, ...)
+void datedit_startmsg(AL_CONST char *fmt, ...)
 {
    va_list args;
    char buf[1024];
@@ -135,7 +152,7 @@ void datedit_startmsg(char *fmt, ...)
 
 
 /* callback for ending a 2-part message output */
-void datedit_endmsg(char *fmt, ...)
+void datedit_endmsg(AL_CONST char *fmt, ...)
 {
    va_list args;
    char buf[1024];
@@ -150,7 +167,7 @@ void datedit_endmsg(char *fmt, ...)
 
 
 /* callback for printing errors */
-void datedit_error(char *fmt, ...)
+void datedit_error(AL_CONST char *fmt, ...)
 {
    va_list args;
    char buf[1024];
@@ -167,7 +184,7 @@ void datedit_error(char *fmt, ...)
 
 
 /* callback for asking questions */
-int datedit_ask(char *fmt, ...)
+int datedit_ask(AL_CONST char *fmt, ...)
 {
    va_list args;
    char buf[1024];
@@ -244,7 +261,7 @@ int datedit_ask(char *fmt, ...)
 
 
 /* checks if a string is one of the names specified on the command line */
-static int is_name(char *name)
+static int is_name(AL_CONST char *name)
 {
    char str1[256], str2[256];
    int i, e;
@@ -277,10 +294,10 @@ static int is_name(char *name)
 
 
 /* does a view operation */
-static void do_view(DATAFILE *dat, char *parentname)
+static void do_view(AL_CONST DATAFILE *dat, AL_CONST char *parentname)
 {
    int i, j;
-   char *name;
+   AL_CONST char *name;
    DATAFILE_PROPERTY *prop;
    char tmp[256];
 
@@ -325,7 +342,7 @@ static void do_view(DATAFILE *dat, char *parentname)
 static void do_export(DATAFILE *dat, char *parentname)
 {
    int i;
-   char *name;
+   AL_CONST char *name;
    char tmp[256];
 
    for (i=0; dat[i].type != DAT_END; i++) {
@@ -356,7 +373,7 @@ static void do_export(DATAFILE *dat, char *parentname)
 static void do_delete(DATAFILE **dat, char *parentname)
 {
    int i;
-   char *name;
+   AL_CONST char *name;
    char tmp[256];
 
    for (i=0; (*dat)[i].type != DAT_END; i++) {
@@ -381,19 +398,68 @@ static void do_delete(DATAFILE **dat, char *parentname)
 
 
 
-/* adds a file to the archive */
-static void do_add_file(char *filename, int attrib, int param)
+/* open a sub-datafile in the parent datafile */
+static DATAFILE *open_sub_datafile(DATAFILE *parent, AL_CONST char *name)
 {
+   DATAFILE *dat;
+   int c;
+
+   /* Return the datafile if it already exists. */
+   for (c=0; parent[c].type != DAT_END; c++) {
+      if ((parent[c].type == DAT_FILE) && (stricmp(name, get_datafile_property(parent+c, DAT_NAME)) == 0))
+	 return parent[c].dat;
+   }
+
+   /* Otherwise create a new datafile. */
+   printf("Creating sub-datafile %s\n", name);
+
+   dat = malloc(sizeof(DATAFILE));
+   dat->dat = NULL;
+   dat->type = DAT_END;
+   dat->size = 0;
+   dat->prop = NULL;
+
+   return dat;
+}
+
+
+
+/* close a sub-datafile from the parent datafile */
+static DATAFILE *close_sub_datafile(DATAFILE *parent, AL_CONST char *name, DATAFILE *dat)
+{
+   int c;
+
+   /* Adjust the datafile if it already exists. */
+   for (c=0; parent[c].type != DAT_END; c++) {
+      if ((parent[c].type == DAT_FILE) && (stricmp(name, get_datafile_property(parent+c, DAT_NAME)) == 0)) {
+	 parent[c].dat = dat;
+	 return parent;
+      }
+   }
+
+   /* Otherwise insert the datafile. */
+   return datedit_insert(parent, NULL, name, DAT_FILE, dat, 0);
+}
+
+
+
+/* adds a file to the archive */
+static int do_add_file(AL_CONST char *filename, int attrib, void *param)
+{
+   DATEDIT_GRAB_PARAMETERS params;
+   DATAFILE **dat = param;
+   DATAFILE *d;
    char fname[256];
    char name[256];
    int c;
-   DATAFILE *d;
 
-   strcpy(fname, filename);
-   fix_filename_case(fname);
+   canonicalize_filename(fname, filename, sizeof(fname));   
 
    strcpy(name, get_filename(fname));
 
+   if ((strcmp(name, ".") == 0) || (strcmp(name, "..") == 0))
+      return 0;
+
    if (!opt_keepnames) {
       strupr(name);
 
@@ -402,34 +468,73 @@ static void do_add_file(char *filename, int attrib, int param)
 	    name[c] = '_';
    }
 
-   for (c=0; datafile[c].type != DAT_END; c++) {
-      if (stricmp(name, get_datafile_property(datafile+c, DAT_NAME)) == 0) {
+   if (attrib & FA_DIREC) {
+      DATAFILE *sub_dat = open_sub_datafile(*dat, name);
+
+      if (!sub_dat) {
+	 fprintf(stderr, "Error: failed to create sub-datafile %s\n", name);
+	 err = 1;
+	 return -1;
+      }
+
+      append_filename(fname, fname, "*", sizeof(fname));
+
+      if (for_each_file_ex(fname, 0, ~attrib_ok, do_add_file, &sub_dat) <= 0) {
+	 fprintf(stderr, "Error: failed to add objects to %s\n", name);
+	 err = 1;
+	 return -1;
+      }
+
+      *dat = close_sub_datafile(*dat, name, sub_dat);
+      changed = TRUE;
+      return 0;
+   }
+
+   params.datafile = canonical_datafilename;
+   params.filename = fname;
+   params.name = name;
+   params.type = datedit_clean_typename(opt_objecttype);
+   params.x = opt_gridx;
+   params.y = opt_gridy;
+   params.w = opt_gridw;
+   params.h = opt_gridh;
+   params.colordepth = opt_colordepth;
+   params.relative = opt_relf;
+
+   for (c=0; (*dat)[c].type != DAT_END; c++) {
+      if (stricmp(name, get_datafile_property(*dat+c, DAT_NAME)) == 0) {
 	 printf("Replacing %s -> %s\n", fname, name);
-	 if (!datedit_grabreplace(datafile+c, fname, name, opt_objecttype, opt_colordepth, opt_gridx, opt_gridy, opt_gridw, opt_gridh))
-	    errno = err = 1;
-	 else
+	 if (!datedit_grabreplace(*dat+c, &params)) {
+	    err = 1;
+	    return -1;
+	 }
+	 else {
 	    changed = TRUE;
-	 return;
+	    return 0;
+	 }
       }
    }
 
    printf("Inserting %s -> %s\n", fname, name);
-   d = datedit_grabnew(datafile, fname, name, opt_objecttype, opt_colordepth, opt_gridx, opt_gridy, opt_gridw, opt_gridh);
-   if (!d)
-      errno = err = 1;
+   d = datedit_grabnew(*dat, &params);
+   if (!d) {
+      err = 1;
+      return -1;
+   }
    else {
-      datafile = d;
+      *dat = d;
       changed = TRUE;
+      return 0;
    }
 }
 
 
 
 /* does an update operation */
-static void do_update(DATAFILE *dat, char *parentname)
+static void do_update(DATAFILE *dat, int force, char *parentname)
 {
    int i;
-   char *name;
+   AL_CONST char *name;
    char tmp[256];
 
    for (i=0; dat[i].type != DAT_END; i++) {
@@ -439,41 +544,12 @@ static void do_update(DATAFILE *dat, char *parentname)
 
       if (dat[i].type == DAT_FILE) {
 	 strcat(tmp, "/");
-	 do_update((DATAFILE *)dat[i].dat, tmp);
+	 do_update((DATAFILE *)dat[i].dat, force, tmp);
 	 if (err)
 	    return;
       }
       else if ((opt_numnames <= 0) || (is_name(tmp))) {
-	 if (!datedit_update(dat+i, opt_verbose, &changed)) {
-	    err = 1;
-	    return;
-	 }
-      }
-   }
-}
-
-
-
-/* does a forced update operation */
-static void do_force_update(DATAFILE *dat, char *parentname)
-{
-   int i;
-   char *name;
-   char tmp[256];
-
-   for (i=0; dat[i].type != DAT_END; i++) {
-      name = get_datafile_property(dat+i, DAT_NAME);
-      strcpy(tmp, parentname);
-      strcat(tmp, name);
-
-      if (dat[i].type == DAT_FILE) {
-	 strcat(tmp, "/");
-	 do_force_update((DATAFILE *)dat[i].dat, tmp);
-	 if (err)
-	    return;
-      }
-      else if ((opt_numnames <= 0) || (is_name(tmp))) {
-	 if (!datedit_force_update(dat+i, opt_verbose, &changed)) {
+	 if (!datedit_update(dat+i, canonical_datafilename, force, opt_verbose, &changed)) {
 	    err = 1;
 	    return;
 	 }
@@ -487,7 +563,7 @@ static void do_force_update(DATAFILE *dat, char *parentname)
 static void do_set_props(DATAFILE *dat, char *parentname)
 {
    int i, j;
-   char *name;
+   AL_CONST char *name;
    char tmp[256];
    char propname[256], *propvalue;
    int type;
@@ -541,7 +617,7 @@ static void do_set_props(DATAFILE *dat, char *parentname)
 static void do_setpal(DATAFILE **dat, char *parentname)
 {
    int i;
-   char *name;
+   AL_CONST char *name;
    char tmp[256];
 
    for (i=0; (*dat)[i].type != DAT_END; i++) {
@@ -577,7 +653,7 @@ static void do_setpal(DATAFILE **dat, char *parentname)
 /* recursive helper for writing out datafile dependencies */
 static void save_dependencies(DATAFILE *dat, FILE *f, int *depth)
 {
-   char *orig;
+   AL_CONST char *orig;
    int c, i;
    int hasspace;
 
@@ -677,10 +753,12 @@ static int do_save_dependencies(DATAFILE *dat, char *srcname, char *depname)
 
 int main(int argc, char *argv[])
 {
-   int c;
+   int c, colorconv_mode = 0;
+   int *opt_fixed_prop = NULL;
+   DATAFILE *datafile = NULL;
 
-   install_allegro(SYSTEM_NONE, &errno, atexit);
-   set_color_conversion(COLORCONV_NONE);
+   if (install_allegro(SYSTEM_NONE, &errno, atexit) != 0)
+      return 1;
    datedit_init();
 
    for (c=0; c<PAL_SIZE; c++)
@@ -693,7 +771,7 @@ int main(int argc, char *argv[])
 
 	    case 'd':
 	       if (stricmp(argv[c]+2, "ither") == 0) {
-		  set_color_conversion(COLORCONV_DITHER);
+		  colorconv_mode |= COLORCONV_DITHER;
 		  break;
 	       }
 	       /* fall through */
@@ -734,6 +812,10 @@ int main(int argc, char *argv[])
 	       opt_compression = argv[c][2] - '0'; 
 	       break;
 
+	    case 'f':
+	       opt_relf = TRUE;
+	       break;
+
 	    case 'g':
 	       if ((opt_gridx > 0) || (c >= argc-4)) {
 		  usage();
@@ -743,7 +825,7 @@ int main(int argc, char *argv[])
 	       opt_gridy = atoi(argv[++c]);
 	       opt_gridw = atoi(argv[++c]);
 	       opt_gridh = atoi(argv[++c]);
-	       if ((opt_gridx <= 0) || (opt_gridy <= 0) ||
+	       if ((opt_gridx <  0) || (opt_gridy <  0) ||
 		   (opt_gridw <= 0) || (opt_gridh <= 0)) {
 		  usage();
 		  return 1;
@@ -770,6 +852,15 @@ int main(int argc, char *argv[])
 	       opt_dependencyfile = argv[++c];
 	       break;
 
+	    case 'n':
+	       if ((opt_sort >= 0) ||
+		   (argv[c][2] < '0') || (argv[c][2] > '1')) {
+		  usage();
+		  return 1;
+	       }
+	       opt_sort = argv[c][2] - '0';
+	       break;
+
 	    case 'o':
 	       if ((opt_outputname) || (c >= argc-1)) {
 		  usage();
@@ -796,17 +887,32 @@ int main(int argc, char *argv[])
 		  opt_prefixstring = argv[++c];
 	       }
 	       break;
+               
+	    case 'r':
+	       attrib_ok |= FA_DIREC;
+	       break;
 
 	    case 's':
-	       if ((opt_strip >= 0) || 
-		   (argv[c][2] < '0') || (argv[c][2] > '2')) {
-		  usage();
-		  return 1;
+	       if (argv[c][2] == '-') {
+		  if (opt_numfixedprops < MAX_FILES)
+		     opt_fixedproplist[opt_numfixedprops++] = argv[c]+3;
+	       }
+	       else {
+		  if ((opt_strip >= 0) || 
+		      (argv[c][2] < '0') || (argv[c][2] > '2')) {
+		     usage();
+		     return 1;
+	          }
+	          opt_strip = argv[c][2] - '0';
 	       }
-	       opt_strip = argv[c][2] - '0'; 
 	       break;
 
 	    case 't':
+	       if (stricmp(argv[c]+2, "ransparency") == 0) {
+		  colorconv_mode |= COLORCONV_KEEP_TRANS;
+		  break;
+	       }
+
 	       if ((opt_objecttype) || (c >= argc-1)) {
 		  usage();
 		  return 1;
@@ -826,6 +932,17 @@ int main(int argc, char *argv[])
 	       opt_password = argv[++c];
 	       break;
 
+	    case '-':
+	       if (stricmp(argv[c]+2, "help") == 0) {
+	          usage();
+	          return 1;
+	       }
+	       break;
+
+	    case '?':
+	       usage();
+	       return 1;
+
 	    default:
 	       printf("Unknown option '%s'\n", argv[c]);
 	       return 1;
@@ -837,8 +954,8 @@ int main(int argc, char *argv[])
 	       opt_proplist[opt_numprops++] = argv[c];
 	 }
 	 else {
-	    if (!opt_datafile)
-	       opt_datafile = argv[c];
+	    if (!opt_datafilename)
+	       opt_datafilename = argv[c];
 	    else {
 	       if (opt_numnames < MAX_FILES) {
 		  opt_namelist[opt_numnames] = argv[c];
@@ -850,10 +967,11 @@ int main(int argc, char *argv[])
       }
    }
 
-   if ((!opt_datafile) || 
+   if ((!opt_datafilename) || 
        ((!opt_command) && 
 	(opt_compression < 0) && 
 	(opt_strip < 0) && 
+	(opt_sort < 0) &&
 	(!opt_numprops) &&
 	(!opt_headername) &&
 	(!opt_dependencyfile))) {
@@ -861,7 +979,14 @@ int main(int argc, char *argv[])
       return 1;
    }
 
-   datafile = datedit_load_datafile(opt_datafile, FALSE, opt_password);
+   canonicalize_filename(canonical_datafilename, opt_datafilename, sizeof(canonical_datafilename));
+
+   if (colorconv_mode)
+      set_color_conversion(colorconv_mode);
+   else
+      set_color_conversion(COLORCONV_NONE);
+
+   datafile = datedit_load_datafile(opt_datafilename, FALSE, opt_password);
 
    if (datafile) {
 
@@ -878,9 +1003,8 @@ int main(int argc, char *argv[])
 	       }
 	       else {
 		  for (c=0; c<opt_numnames; c++) {
-		     if (for_each_file(opt_namelist[c], FA_ARCH | FA_RDONLY, do_add_file, 0) <= 0) {
-			if (errno)
-			   fprintf(stderr, "Error: %s not found\n", opt_namelist[c]);
+		     if (for_each_file_ex(opt_namelist[c], 0, ~attrib_ok, do_add_file, &datafile) <= 0) {
+			fprintf(stderr, "Error: %s not found\n", opt_namelist[c]);
 			err = 1;
 			break;
 		     }
@@ -914,11 +1038,11 @@ int main(int argc, char *argv[])
 	       break;
 
 	    case 'u':
-	       do_update(datafile, "");
+	       do_update(datafile, FALSE, "");
 	       break;
 
 	    case 'w':
-	       do_force_update(datafile, "");
+	       do_update(datafile, TRUE, "");
 	       break;
 	 }
       }
@@ -952,20 +1076,48 @@ int main(int argc, char *argv[])
 	       }
 	    }
 	 }
+
+	 if (opt_numfixedprops > 0) {
+	    if (opt_strip < 0) {
+	       printf("Error: no strip mode\n");
+	       err = 1;
+	    }
+	    else {
+	       opt_fixed_prop = malloc((opt_numfixedprops+1)*sizeof(int));
+
+	       for (c=0; c<opt_numfixedprops; c++)
+		  opt_fixed_prop[c] = datedit_clean_typename(opt_fixedproplist[c]);
+
+	       opt_fixed_prop[c] = 0;  /* end of array delimiter */
+	    }
+	 }
       }
 
-      if ((!err) && ((changed) || (opt_compression >= 0) || (opt_strip >= 0)))
-	 if (!datedit_save_datafile(datafile, opt_datafile, opt_strip, opt_compression, opt_verbose, TRUE, FALSE, opt_password))
+      if ((!err) && ((changed) || (opt_compression >= 0) || (opt_strip >= 0) || (opt_sort >= 0))) {
+	 DATEDIT_SAVE_DATAFILE_OPTIONS options;
+
+	 options.pack = opt_compression;
+	 options.strip = opt_strip;
+	 options.sort = opt_sort;
+	 options.verbose = opt_verbose;
+	 options.write_msg = TRUE;
+	 options.backup = FALSE;
+
+	 if (!datedit_save_datafile(datafile, opt_datafilename, opt_fixed_prop, &options, opt_password))
 	    err = 1;
+      }
 
       if ((!err) && (opt_headername))
-	 if (!datedit_save_header(datafile, opt_datafile, opt_headername, "dat", opt_prefixstring, opt_verbose))
+	 if (!datedit_save_header(datafile, opt_datafilename, opt_headername, "dat", opt_prefixstring, opt_verbose))
 	    err = 1;
 
       if ((!err) && (opt_dependencyfile))
-	 if (!do_save_dependencies(datafile, opt_datafile, opt_dependencyfile))
+	 if (!do_save_dependencies(datafile, opt_datafilename, opt_dependencyfile))
 	    err = 1;
 
+      if (opt_fixed_prop)
+	 free(opt_fixed_prop);
+
       unload_datafile(datafile);
    }
 
