@@ -16,17 +16,29 @@
  */
 
 
-#define USE_CONSOLE
+#define ALLEGRO_USE_CONSOLE
 
 #include <stdio.h>
 #include <string.h>
 #include <time.h>
 
 #include "allegro.h"
-#include "allegro/aintern.h"
+#include "allegro/internal/aintern.h"
 #include "datedit.h"
 
 
+/* unused callbacks for datedit.c */
+void datedit_msg(AL_CONST char *fmt, ...) { }
+void datedit_startmsg(AL_CONST char *fmt, ...) { }
+void datedit_endmsg(AL_CONST char *fmt, ...) { }
+void datedit_error(AL_CONST char *fmt, ...) { }
+int datedit_ask(AL_CONST char *fmt, ...) { return 0; }
+
+
+/* this program is not portable! */
+#ifdef ALLEGRO_I386
+
+
 #ifndef ALLEGRO_ASM_PREFIX
    #define ALLEGRO_ASM_PREFIX    ""
 #endif
@@ -34,12 +46,15 @@
 
 static int err = 0;
 static int truecolor = FALSE;
+static int convert_compiled_sprites = FALSE;
 
 static char prefix[80] = "";
 
 static char *infilename = NULL;
 static char *outfilename = NULL;
 static char *outfilenameheader = NULL;
+static char *dataname = NULL;
+static char default_dataname[]="data";
 static char *password = NULL;
 
 static DATAFILE *data = NULL;
@@ -50,16 +65,8 @@ static FILE *outfileheader = NULL;
 static void output_object(DATAFILE *object, char *name);
 
 
-/* unused callbacks for datedit.c */
-void datedit_msg(char *fmt, ...) { }
-void datedit_startmsg(char *fmt, ...) { }
-void datedit_endmsg(char *fmt, ...) { }
-void datedit_error(char *fmt, ...) { }
-int datedit_ask(char *fmt, ...) { return 0; }
 
-
-
-static void usage()
+static void usage(void)
 {
    printf("\nDatafile -> asm conversion utility for Allegro " ALLEGRO_VERSION_STR ", " ALLEGRO_PLATFORM_STR "\n");
    printf("By Shawn Hargreaves, " ALLEGRO_DATE_STR "\n\n");
@@ -67,7 +74,9 @@ static void usage()
    printf("Options:\n");
    printf("\t'-o outputfile.s' sets the output file (default stdout)\n");
    printf("\t'-h outputfile.h' sets the output header file (default none)\n");
-   printf("\t'-p prefix' sets the object name prefix string\n");
+   printf("\t'-p prefix' sets the object name prefix string (default none)\n");
+   printf("\t'-S' converts COMPILED_SPRITEs to BITMAPs (default abort)\n");
+   printf("\t'-n name' gives a name for the datafile (default 'data')\n");
    printf("\t'-007 password' sets the datafile password\n");
 }
 
@@ -137,9 +146,9 @@ static void output_bitmap(BITMAP *bmp, char *name, int global)
    fprintf(outfile, "\t.long %-16d# cr\n", bmp->w);
    fprintf(outfile, "\t.long %-16d# ct\n", 0);
    fprintf(outfile, "\t.long %-16d# cb\n", bmp->h);
-   fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "__linear_vtable%d\n", bpp);
-   fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "_stub_bank_switch\n");
-   fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "_stub_bank_switch\n");
+   fprintf(outfile, "\t.long %-16d# bpp\n", bpp);
+   fprintf(outfile, "\t.long %-16d# write_bank\n", 0);
+   fprintf(outfile, "\t.long %-16d# read_bank\n", 0);
    fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "%s%s_data\n", prefix, name);
    fprintf(outfile, "\t.long %-16d# bitmap_id\n", 0);
    fprintf(outfile, "\t.long %-16d# extra\n", 0);
@@ -206,66 +215,103 @@ static void output_midi(MIDI *midi, char *name)
 
 
 
-static void output_font(FONT *f, char *name, int depth)
+static void output_font_color(FONT_COLOR_DATA *cf, char *name, int depth)
 {
-   FONT_GLYPH *g;
-   char buf[160], goodname[160];
-   int c;
+   char buf[1000], goodname[1000];
+   int ch;
 
-   if (f->next)
-      output_font(f->next, name, depth+1);
+   if (cf->next) output_font_color(cf->next, name, depth + 1);
 
-   if (depth > 0)
-      sprintf(goodname, "%s_r%d", name, depth+1);
-   else
+   if (depth > 0) 
+      sprintf(goodname, "%s_r%d", name, depth + 1);
+   else 
       strcpy(goodname, name);
 
-   for (c=f->start; c<=f->end; c++) {
-      sprintf(buf, "%s_char_%04X", goodname, c);
-
-      if (f->mono) {
-	 g = f->glyphs[c-f->start];
-
-	 fprintf(outfile, "# glyph\n");
-	 fprintf(outfile, ".balign 4\n" ALLEGRO_ASM_PREFIX "%s%s:\n", prefix, buf);
-	 fprintf(outfile, "\t.short %-15d# w\n", g->w);
-	 fprintf(outfile, "\t.short %-15d# h\n", g->h);
-
-	 write_data(g->dat, ((g->w+7)/8) * g->h);
-
-	 fprintf(outfile, "\n");
-      }
-      else
-	 output_bitmap(f->glyphs[c-f->start], buf, FALSE);
+   for (ch = cf->begin; ch < cf->end; ch++) {
+      sprintf(buf, "%s_char_%04X", goodname, ch);
+      output_bitmap(cf->bitmaps[ch - cf->begin], buf, FALSE);
    }
 
    fprintf(outfile, "# glyph list\n");
    fprintf(outfile, ".balign 4\n" ALLEGRO_ASM_PREFIX "%s%s_glyphs:\n", prefix, goodname);
 
-   for (c=f->start; c<=f->end; c++)
-      fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "%s%s_char_%04X\n", prefix, goodname, c);
+   for (ch = cf->begin; ch < cf->end; ch++)
+      fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "%s%s_char_%04X\n", prefix, goodname, ch);
+   fprintf(outfile, "\n# FONT_COLOR_DATA\n");
 
+   fprintf(outfile, ".balign 4\n" ALLEGRO_ASM_PREFIX "%s%s_data:\n", prefix, goodname);
+   fprintf(outfile, "\t.long 0x%04X%10c# begin\n", cf->begin, ' ');
+   fprintf(outfile, "\t.long 0x%04X%10c# end\n", cf->end, ' ');
+   fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "%s%s_glyphs\n", prefix, goodname);
+   if (cf->next) 
+      fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "%s%s_r%d_data\n", prefix, name, depth + 2);
+   else
+      fprintf(outfile, "\t.long %-16d#next\n", 0);
    fprintf(outfile, "\n");
+}
+
+
+
+static void output_font_mono(FONT_MONO_DATA *mf, char *name, int depth)
+{
+   char buf[1000], goodname[1000];
+   int ch;
+
+   if (mf->next) output_font_mono(mf->next, name, depth + 1);
+
+   if (depth > 0) 
+      sprintf(goodname, "%s_r%d", name, depth + 1);
+   else
+      strcpy(goodname, name);
+
+   for (ch = mf->begin; ch < mf->end; ch++) {
+      FONT_GLYPH *g = mf->glyphs[ch - mf->begin];
+
+      sprintf(buf, "%s_char_%04X", goodname, ch);
+      fprintf(outfile, "# glyph\n");
+      fprintf(outfile, ".balign 4\n" ALLEGRO_ASM_PREFIX "%s%s:\n", prefix, buf);
+      fprintf(outfile, "\t.short %-15d# w\n", g->w);
+      fprintf(outfile, "\t.short %-15d# h\n", g->h);
+      
+      write_data(g->dat, ((g->w + 7) / 8) * g->h);
+      fprintf(outfile, "\n");
+   }
+
+   fprintf(outfile, "# glyph list\n");
+   fprintf(outfile, ".balign 4\n" ALLEGRO_ASM_PREFIX "%s%s_glyphs:\n", prefix, goodname);
+
+   for (ch = mf->begin; ch < mf->end; ch++)
+      fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "%s%s_char_%04X\n", prefix, goodname, ch);
+   fprintf(outfile, "\n# FONT_COLOR_DATA\n");
+
+   fprintf(outfile, ".balign 4\n" ALLEGRO_ASM_PREFIX "%s%s_data:\n", prefix, goodname);
+   fprintf(outfile, "\t.long 0x%04X%10c# begin\n", mf->begin, ' ');
+   fprintf(outfile, "\t.long 0x%04X%10c# end\n", mf->end, ' ');
+   fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "%s%s_glyphs\n", prefix, goodname);
+   if (mf->next) 
+      fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "%s%s_r%d_data\n", prefix, name, depth + 2);
+   else
+      fprintf(outfile, "\t.long %-16d#next\n", 0);
+   fprintf(outfile, "\n");
+}
+
+
+
+static void output_font(FONT *f, char *name, int depth)
+{
+   int color_flag = (f->vtable == font_vtable_color ? 1 : 0);
+
+   if (color_flag)
+      output_font_color(f->data, name, 0);
+   else
+      output_font_mono(f->data, name, 0);
 
    fprintf(outfile, "# font\n");
-
-   if (depth == 0)
-      fprintf(outfile, ".globl " ALLEGRO_ASM_PREFIX "%s%s\n", prefix, goodname);
-
-   fprintf(outfile, ".balign 4\n" ALLEGRO_ASM_PREFIX "%s%s:\n", prefix, goodname);
-   fprintf(outfile, "\t.long %-16d# mono\n", f->mono);
-   fprintf(outfile, "\t.long 0x%04X%10c# start\n", f->start, ' ');
-   fprintf(outfile, "\t.long 0x%04X%10c# end\n", f->end, ' ');
-   fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "%s%s_glyphs\n", prefix, goodname);
-
-   if (f->next)
-      fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "%s%s_r%d\n", prefix, name, depth+2);
-   else
-      fprintf(outfile, "\t.long %-16d# next\n", 0);
-
-   fprintf(outfile, "\t.long %-16d# renderhook\n", 0);
-   fprintf(outfile, "\t.long %-16d# widthhook\n", 0);
-   fprintf(outfile, "\t.long %-16d# heighthook\n", 0);
+   fprintf(outfile, ".globl " ALLEGRO_ASM_PREFIX "%s%s\n", prefix, name);
+   fprintf(outfile, ".balign 4\n" ALLEGRO_ASM_PREFIX "%s%s:\n", prefix, name);
+   fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "%s%s_data\n", prefix, name);
+   fprintf(outfile, "\t.long %-16d# height\n", f->height);
+   fprintf(outfile, "\t.long %-16d# color flag\n", color_flag);
    fprintf(outfile, "\n");
 }
 
@@ -285,58 +331,13 @@ static void output_rle_sprite(RLE_SPRITE *sprite, char *name)
    fprintf(outfile, "\t.long %-16d# color depth\n", bpp);
    fprintf(outfile, "\t.long %-16d# size\n", sprite->size);
 
-   write_data(sprite->dat, sprite->size);
+   write_data((unsigned char *)sprite->dat, sprite->size);
 
    fprintf(outfile, "\n");
 }
 
 
 
-#ifndef ALLEGRO_USE_C
-
-static void output_compiled_sprite(COMPILED_SPRITE *sprite, char *name)
-{
-   char buf[160];
-   int c;
-
-   if (sprite->color_depth != 8) {
-      fprintf(stderr, "Error: truecolor compiled sprites not supported (%s)\n", name);
-      err = 1;
-      return;
-   }
-
-   for (c=0; c<4; c++) {
-      if (sprite->proc[c].draw) {
-	 sprintf(buf, "%s_plane_%d", name, c);
-	 output_data(sprite->proc[c].draw, sprite->proc[c].len, buf, "compiled sprite code", FALSE);
-      }
-   } 
-
-   fprintf(outfile, "# compiled sprite\n.globl " ALLEGRO_ASM_PREFIX "%s%s\n", prefix, name);
-   fprintf(outfile, ".balign 4\n" ALLEGRO_ASM_PREFIX "%s%s:\n", prefix, name);
-   fprintf(outfile, "\t.short %-15d# planar\n", sprite->planar);
-   fprintf(outfile, "\t.short %-15d# color depth\n", sprite->color_depth);
-   fprintf(outfile, "\t.short %-15d# w\n", sprite->w);
-   fprintf(outfile, "\t.short %-15d# h\n", sprite->h);
-
-   for (c=0; c<4; c++) {
-      if (sprite->proc[c].draw) {
-	 fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "%s%s_plane_%d\n", prefix, name, c);
-	 fprintf(outfile, "\t.long %-16d# len\n", sprite->proc[c].len);
-      }
-      else {
-	 fprintf(outfile, "\t.long 0\n");
-	 fprintf(outfile, "\t.long %-16d# len\n", 0);
-      }
-   }
-
-   fprintf(outfile, "\n");
-}
-
-#endif      /* ifndef ALLEGRO_USE_C */
-
-
-
 static void get_object_name(char *buf, char *name, DATAFILE *dat, int root)
 {
    if (!root) {
@@ -443,14 +444,15 @@ static void output_object(DATAFILE *object, char *name)
 
       case DAT_C_SPRITE:
       case DAT_XC_SPRITE:
-	 if (outfileheader)
-	    fprintf(outfileheader, "extern COMPILED_SPRITE %s%s;\n", prefix, name);
-
-	 #ifdef ALLEGRO_USE_C
-	    output_rle_sprite((RLE_SPRITE *)object->dat, name);
-	 #else
-	    output_compiled_sprite((COMPILED_SPRITE *)object->dat, name);
-	 #endif
+	 if (convert_compiled_sprites) {
+	    object->type = DAT_BITMAP;
+	    output_object(object, name);
+	 }
+	 else {
+	    fprintf(stderr, "Error: encountered a compiled sprite (%s). Please\n"
+	                    "see documentation for more information.\n", name);
+	    err = 1;
+	 }
 	 break;
 
       case DAT_FILE:
@@ -483,7 +485,8 @@ int main(int argc, char *argv[])
    char tm[80];
    time_t now;
 
-   install_allegro(SYSTEM_NONE, &errno, atexit);
+   if (install_allegro(SYSTEM_NONE, &errno, atexit) != 0)
+      return 1;
    datedit_init();
 
    time(&now);
@@ -514,6 +517,13 @@ int main(int argc, char *argv[])
 	 }
 	 strcpy(prefix, argv[++c]);
       }
+      else if (stricmp(argv[c], "-n") == 0) {
+	 if ((dataname) || (c >= argc-1)) {
+	    usage();
+	    return 1;
+	 }
+	 dataname = argv[++c];
+      }
       else if (stricmp(argv[c], "-007") == 0) {
 	 if ((password) || (c >= argc-1)) {
 	    usage();
@@ -521,6 +531,13 @@ int main(int argc, char *argv[])
 	 }
 	 password = argv[++c];
       }
+      else if (stricmp(argv[c], "-S") == 0) {
+	 if (convert_compiled_sprites) {
+	    usage();
+	    return 1;
+	 }
+	 convert_compiled_sprites = TRUE;
+      }
       else {
 	 if ((argv[c][0] == '-') || (infilename)) {
 	    usage();
@@ -538,9 +555,12 @@ int main(int argc, char *argv[])
    if ((prefix[0]) && (prefix[strlen(prefix)-1] != '_'))
       strcat(prefix, "_");
 
+   if (!dataname)
+      dataname = default_dataname;
+
    set_color_conversion(COLORCONV_NONE);
 
-   data = datedit_load_datafile(infilename, TRUE, password);
+   data = datedit_load_datafile(infilename, FALSE, password);
    if (!data) {
       fprintf(stderr, "Error reading %s\n", infilename);
       err = 1; 
@@ -579,22 +599,26 @@ int main(int argc, char *argv[])
    if (outfilename)
       printf("Converting %s to %s...\n", infilename, outfilename);
 
-   output_datafile(data, "data", TRUE);
+   output_datafile(data, dataname, TRUE);
 
-   #ifdef ALLEGRO_DJGPP
+   #ifdef ALLEGRO_USE_CONSTRUCTOR
 
       fprintf(outfile, ".text\n");
       fprintf(outfile, ".balign 4\n");
-      fprintf(outfile, ALLEGRO_ASM_PREFIX "_construct_me:\n");
+      fprintf(outfile, "construct_me:\n");
       fprintf(outfile, "\tpushl %%ebp\n");
       fprintf(outfile, "\tmovl %%esp, %%ebp\n");
-      fprintf(outfile, "\tpushl $" ALLEGRO_ASM_PREFIX "%sdata\n", prefix);
+      fprintf(outfile, "\tpushl $" ALLEGRO_ASM_PREFIX "%s%s\n", prefix, dataname);
       fprintf(outfile, "\tcall " ALLEGRO_ASM_PREFIX "_construct_datafile\n");
       fprintf(outfile, "\taddl $4, %%esp\n");
       fprintf(outfile, "\tleave\n");
       fprintf(outfile, "\tret\n\n");
+      #ifdef ALLEGRO_DJGPP      
       fprintf(outfile, ".section .ctor\n");
-      fprintf(outfile, "\t.long " ALLEGRO_ASM_PREFIX "_construct_me\n");
+      #else
+      fprintf(outfile, ".section .ctors\n");
+      #endif
+      fprintf(outfile, "\t.long construct_me\n");
 
    #endif
 
@@ -620,18 +644,36 @@ int main(int argc, char *argv[])
 	 delete_file(outfilenameheader);
    }
    else {
-      #ifdef ALLEGRO_DJGPP
+      #ifdef ALLEGRO_USE_CONSTRUCTOR
+
 	 if (truecolor) {
 	    printf("\nI noticed some truecolor images, so you must call fixup_datafile()\n");
 	    printf("before using this data! (after setting a video mode).\n");
 	 }
+
       #else
+
 	 printf("\nI don't know how to do constructor functions on this platform, so you must\n");
 	 printf("call fixup_datafile() before using this data! (after setting a video mode).\n");
+
       #endif
    }
 
    return err;
 }
 
+
+#else       /* ifdef ALLEGRO_I386 */
+
+
+int main(void)
+{
+   allegro_init();
+   allegro_message("Sorry, the DAT2S program only works on x86 processors.\n");
+   return 1;
+}
+
+
+#endif
+
 END_OF_MAIN();
