@@ -18,15 +18,15 @@
  */
 
 
-#define USE_CONSOLE
+#define ALLEGRO_USE_CONSOLE
 
 #include <stdio.h>
 #include <string.h>
-#include <math.h>
 #include <time.h>
+#include <math.h>
 
 #include "allegro.h"
-#include "allegro/aintern.h"
+#include "allegro/internal/aintern.h"
 #include "datedit.h"
 
 
@@ -212,7 +212,7 @@ static char *short_drum_names[] =
 
 
 /* display help on the command syntax */
-static void usage()
+static void usage(void)
 {
    printf("\nPatch grabber for the Allegro DIGMID driver " ALLEGRO_VERSION_STR ", " ALLEGRO_PLATFORM_STR "\n");
    printf("By Shawn Hargreaves, " ALLEGRO_DATE_STR "\n\n");
@@ -233,12 +233,12 @@ static void usage()
 
 
 /* unused callback for datedit.c */
-int datedit_ask(char *fmt, ...) { return 0; }
+int datedit_ask(AL_CONST char *fmt, ...) { return 0; }
 
 
 
 /* callback for outputting messages */
-void datedit_msg(char *fmt, ...)
+void datedit_msg(AL_CONST char *fmt, ...)
 {
    va_list args;
    char buf[1024];
@@ -253,7 +253,7 @@ void datedit_msg(char *fmt, ...)
 
 
 /* callback for starting a 2-part message output */
-void datedit_startmsg(char *fmt, ...)
+void datedit_startmsg(AL_CONST char *fmt, ...)
 {
    va_list args;
    char buf[1024];
@@ -269,7 +269,7 @@ void datedit_startmsg(char *fmt, ...)
 
 
 /* callback for ending a 2-part message output */
-void datedit_endmsg(char *fmt, ...)
+void datedit_endmsg(AL_CONST char *fmt, ...)
 {
    va_list args;
    char buf[1024];
@@ -284,7 +284,7 @@ void datedit_endmsg(char *fmt, ...)
 
 
 /* callback for printing errors */
-void datedit_error(char *fmt, ...)
+void datedit_error(AL_CONST char *fmt, ...)
 {
    va_list args;
    char buf[1024];
@@ -410,7 +410,7 @@ static void scan_patches(MIDI *midi, char *patches, char *drums)
 
 
 /* reads a MIDI file to see what patches it requires */
-static void read_midi(char *filename, int attrib, int param)
+static int read_midi(AL_CONST char *filename, int attrib, void *param)
 {
    char patches[128], drums[128];
    char fname[256];
@@ -425,8 +425,8 @@ static void read_midi(char *filename, int attrib, int param)
    midi = load_midi(fname);
    if (!midi) {
       fprintf(stderr, "Error reading %s\n", fname);
-      errno = err = 1;
-      return;
+      err = 1;
+      return -1;
    }
 
    scan_patches(midi, patches, drums);
@@ -454,6 +454,7 @@ static void read_midi(char *filename, int attrib, int param)
    }
 
    destroy_midi(midi);
+   return 0;
 }
 
 
@@ -539,7 +540,7 @@ static void mem_write_file(PACKFILE *f, int size)
 
 
 /* imports data from a GUS patch file */
-static void *grab_patch(char *filename, long *size, int x, int y, int w, int h, int depth)
+static DATAFILE *grab_patch(int type, AL_CONST char *filename, DATAFILE_PROPERTY **prop, int depth)
 {
    PACKFILE *f;
    long sz = file_size(filename);
@@ -585,8 +586,7 @@ static void *grab_patch(char *filename, long *size, int x, int y, int w, int h,
 	 }
       }
 
-      *size = sz;
-      return mem;
+      return datedit_construct(type, mem, sz, prop);
    }
 
    /* have to parse the file to adjust the sample format */
@@ -736,8 +736,7 @@ static void *grab_patch(char *filename, long *size, int x, int y, int w, int h,
       return NULL;
    }
 
-   *size = mem_size;
-   return mem;
+   return datedit_construct(type, mem, mem_size, prop);
 }
 
 
@@ -772,9 +771,10 @@ static int parse_string(char *buf, char *argv[])
 
 
 /* inserts all the required patches into the datafile */
-static void add_gus_patches()
+static void add_gus_patches(void)
 {
    char dir[256], file[256], buf[256], filename[256];
+   DATEDIT_GRAB_PARAMETERS params;
    PACKFILE *f;
    DATAFILE *d;
    char *argv[16];
@@ -797,7 +797,7 @@ static void add_gus_patches()
       strcpy(file, get_filename(opt_configfile));
    }
    else {
-      if (!_digmid_find_patches(dir, file)) {
+      if (!_digmid_find_patches(dir, sizeof dir, file, sizeof file)) {
 	 fprintf(stderr, "\nError: MIDI patch set not found!\n\n");
 	 fprintf(stderr, "You should have a directory containing a collection of GUS format .pat\n");
 	 fprintf(stderr, "files and a default.cfg index file, or a SoundFont 2 .sf2 sample bank.\n");
@@ -814,8 +814,19 @@ static void add_gus_patches()
       if (buf[i] == '\\')
 	 buf[i] = '/';
 
+   params.datafile = NULL;  /* only with absolute filenames */
+   params.filename = buf;
+   params.name = "default_cfg";
+   params.type = DAT_DATA;
+   /* params.x = */
+   /* params.y = */
+   /* params.w = */
+   /* params.h = */
+   /* params.colordepth = */
+   params.relative = FALSE;  /* required (see above) */
+
    printf("Grabbing %s\n", buf);
-   d = datedit_grabnew(datafile, buf, "default_cfg", "DATA", -1, -1, -1, -1, -1);
+   d = datedit_grabnew(datafile, &params);
    if (!d) {
       errno = err = 1;
       return;
@@ -887,8 +898,11 @@ static void add_gus_patches()
 			break;
 
 		  if (datafile[j].type == DAT_END) {
+		     params.filename = filename;
+		     params.name = argv[1];
+		     params.type = DAT_PATCH;
 		     printf("Grabbing %s\n", filename);
-		     d = datedit_grabnew(datafile, filename, argv[1], "PAT", -1, -1, -1, -1, -1);
+		     d = datedit_grabnew(datafile, &params);
 		  }
 		  else
 		     d = datafile;
@@ -1869,19 +1883,32 @@ static void print_sf_string(FILE *f, char *title)
 
 
 /* inserts all the required patches into the datafile */
-static void add_soundfont_patches()
+static void add_soundfont_patches(void)
 {
    RIFF_CHUNK file, chunk, subchunk;
+   DATEDIT_GRAB_PARAMETERS params;
    DATAFILE *d;
    FILE *f;
    time_t now;
-   char buf[256];
+   char buf[256] = "XXXXXX";
    char tm[80];
    int i;
 
+   #ifdef HAVE_MKSTEMP
+
+      int tmp_fd;
+
+      tmp_fd = mkstemp(buf);
+      close(tmp_fd);
+
+   #else
+
+      tmpnam(buf);
+
+   #endif
+   
    printf("Generating index file\n");
 
-   tmpnam(buf);
    f = fopen(buf, F_WRITE);
    if (!f) {
       fprintf(stderr, "Error writing temporary file\n");
@@ -1913,7 +1940,18 @@ static void add_soundfont_patches()
 
    fclose(f);
 
-   d = datedit_grabnew(datafile, buf, "default_cfg", "DATA", -1, -1, -1, -1, -1);
+   params.datafile = NULL;  /* only with absolute filenames */
+   params.filename = buf;
+   params.name = "default_cfg";
+   params.type = DAT_DATA;
+   /* params.x = */
+   /* params.y = */
+   /* params.w = */
+   /* params.h = */
+   /* params.colordepth = */
+   params.relative = FALSE;  /* required (see above) */
+
+   d = datedit_grabnew(datafile, &params);
 
    delete_file(buf);
 
@@ -2269,7 +2307,8 @@ int main(int argc, char *argv[])
 {
    int c;
 
-   install_allegro(SYSTEM_NONE, &errno, atexit);
+   if (install_allegro(SYSTEM_NONE, &errno, atexit) != 0)
+       return 1;
    datedit_init();
 
    for (c=0; c<128; c++)
@@ -2388,12 +2427,9 @@ int main(int argc, char *argv[])
       need_patches[0] = TRUE;    /* always load the piano */
 
       for (c=0; c<opt_numnames; c++) {
-	 if (for_each_file(opt_namelist[c], FA_ARCH | FA_RDONLY, read_midi, 0) <= 0) {
+	 if (for_each_file_ex(opt_namelist[c], 0, ~(FA_ARCH | FA_RDONLY), read_midi, NULL) <= 0) {
 	    fprintf(stderr, "Error: %s not found\n", opt_namelist[c]);
 	    err = 1;
-	 }
-	 if (errno) {
-	    err = 1;
 	    break;
 	 }
       }
@@ -2429,7 +2465,16 @@ int main(int argc, char *argv[])
 	 add_gus_patches();
 
       if (!err) {
-	 if (!datedit_save_datafile(datafile, opt_datafile, 1, opt_compression, (opt_veryverbose || (opt_verbose && opt_compression)), TRUE, FALSE, NULL))
+	 DATEDIT_SAVE_DATAFILE_OPTIONS options;
+
+	 options.pack = opt_compression;
+	 options.strip = 1;
+	 options.sort = 1;
+	 options.verbose = (opt_veryverbose || (opt_verbose && opt_compression));
+	 options.write_msg = TRUE;
+	 options.backup = FALSE;
+
+	 if (!datedit_save_datafile(datafile, opt_datafile, NULL, &options, NULL))
 	    err = 1;
       }
 
