@@ -22,6 +22,7 @@
 #include <string.h>
 
 #include "allegro.h"
+#include "allegro/internal/aintern.h"
 #include "../datedit.h"
 
 
@@ -48,7 +49,7 @@ static int bitmap_has_alpha(BITMAP *bmp)
 
 
 /* checks whether this RLE sprite has an alpha channel */
-static int rle_has_alpha(RLE_SPRITE *spr)
+static int rle_has_alpha(AL_CONST RLE_SPRITE *spr)
 {
    signed long *p32;
    int x, y, c;
@@ -89,9 +90,8 @@ static void *makenew_bitmap(long *size)
 {
    BITMAP *bmp = create_bitmap_ex(8, 32, 32);
 
-   clear(bmp);
-   text_mode(-1);
-   textout_centre(bmp, font, "Hi!", 16, 12, 1);
+   clear_bitmap(bmp);
+   textout_centre_ex(bmp, font, "Hi!", 16, 12, 1, -1);
 
    return bmp;
 }
@@ -99,7 +99,7 @@ static void *makenew_bitmap(long *size)
 
 
 /* returns a bitmap description string */
-static void get_bitmap_desc(DATAFILE *dat, char *s)
+static void get_bitmap_desc(AL_CONST DATAFILE *dat, char *s)
 {
    BITMAP *bmp = (BITMAP *)dat->dat;
 
@@ -112,7 +112,7 @@ static void get_bitmap_desc(DATAFILE *dat, char *s)
 
 
 /* exports a bitmap into an external file */
-static int export_bitmap(DATAFILE *dat, char *filename)
+static int export_bitmap(AL_CONST DATAFILE *dat, AL_CONST char *filename)
 {
    return (save_bitmap(filename, (BITMAP *)dat->dat, datedit_current_palette) == 0);
 }
@@ -120,7 +120,7 @@ static int export_bitmap(DATAFILE *dat, char *filename)
 
 
 /* draws a bitmap onto the grabber object view window */
-static void plot_bitmap(DATAFILE *dat, int x, int y)
+static void plot_bitmap(AL_CONST DATAFILE *dat, int x, int y)
 {
    BITMAP *b = dat->dat;
    int w = b->w;
@@ -143,12 +143,12 @@ static void plot_bitmap(DATAFILE *dat, int x, int y)
 
    if ((bitmap_color_depth(screen) == 8) && (bitmap_color_depth(b) > 8)) {
       if ((w != b->w) || (h != b->h))
-	 textout(screen, font, "<reduced color scaled preview>", x, y+18, gui_fg_color);
+	 textout_ex(screen, font, "<reduced color scaled preview>", x, y+18, gui_fg_color, gui_bg_color);
       else
-	 textout(screen, font, "<reduced color preview>", x, y+18, gui_fg_color);
+	 textout_ex(screen, font, "<reduced color preview>", x, y+18, gui_fg_color, gui_bg_color);
    }
    else if ((w != b->w) || (h != b->h))
-      textout(screen, font, "<scaled preview>", x, y+18, gui_fg_color);
+      textout_ex(screen, font, "<scaled preview>", x, y+18, gui_fg_color, gui_bg_color);
 
    if (bitmap_color_depth(b) == bitmap_color_depth(screen)) {
       if (dat->type != DAT_BITMAP)
@@ -322,7 +322,7 @@ static int view_bitmap(DATAFILE *dat)
       poll_mouse();
    } while (mouse_b);
 
-   clear(screen);
+   clear_bitmap(screen);
    show_mouse(screen);
 
    return D_REDRAW;
@@ -331,9 +331,10 @@ static int view_bitmap(DATAFILE *dat)
 
 
 /* reads a bitmap from an external file */
-static void *grab_bitmap(char *filename, long *size, int x, int y, int w, int h, int depth)
+static DATAFILE *grab_bitmap(int type, AL_CONST char *filename, DATAFILE_PROPERTY **prop, int depth)
 {
    BITMAP *bmp;
+   int x, y, w, h;
 
    if (depth > 0) {
       int oldcolordepth = _color_depth;
@@ -352,25 +353,31 @@ static void *grab_bitmap(char *filename, long *size, int x, int y, int w, int h,
    if (!bmp)
       return NULL;
 
+   x = datedit_numprop(prop, DAT_XPOS);
+   y = datedit_numprop(prop, DAT_YPOS);
+   w = datedit_numprop(prop, DAT_XSIZ);
+   h = datedit_numprop(prop, DAT_YSIZ);
+
    if ((x >= 0) && (y >= 0) && (w >= 0) && (h >= 0)) {
       BITMAP *b2 = create_bitmap_ex(bitmap_color_depth(bmp), w, h);
       clear_to_color(b2, bitmap_mask_color(b2));
       blit(bmp, b2, x, y, 0, 0, w, h);
       destroy_bitmap(bmp);
-      return b2;
+      bmp = b2;
    }
-   else 
-      return bmp;
+   
+   return datedit_construct(type, bmp, 0, prop);
 }
 
 
 
 /* saves a bitmap into the datafile format */
-static void save_datafile_bitmap(DATAFILE *dat, int packed, int packkids, int strip, int verbose, int extra, PACKFILE *f)
+static int save_datafile_bitmap(DATAFILE *dat, AL_CONST int *fixed_prop, int pack, int pack_kids, int strip, int sort, int verbose, int extra, PACKFILE *f)
 {
    BITMAP *bmp = (BITMAP *)dat->dat;
    int x, y, c, r, g, b, a;
    unsigned short *p16;
+   unsigned char *p24;
    unsigned long *p32;
    int depth;
 
@@ -379,6 +386,8 @@ static void save_datafile_bitmap(DATAFILE *dat, int packed, int packkids, int st
    else
       depth = bitmap_color_depth(bmp);
 
+   *allegro_errno = 0;
+
    pack_mputw(depth, f);
    pack_mputw(bmp->w, f);
    pack_mputw(bmp->h, f);
@@ -412,13 +421,17 @@ static void save_datafile_bitmap(DATAFILE *dat, int packed, int packkids, int st
       case 24:
 	 /* 24 bit truecolor */
 	 for (y=0; y<bmp->h; y++) {
+	    p24 = (unsigned char *)bmp->line[y];
+
 	    for (x=0; x<bmp->w; x++) {
-	       r = bmp->line[y][x*3+_rgb_r_shift_24/8];
-	       g = bmp->line[y][x*3+_rgb_g_shift_24/8];
-	       b = bmp->line[y][x*3+_rgb_b_shift_24/8];
+	       c = READ3BYTES(p24);
+	       r = getr24(c);
+	       g = getg24(c);
+	       b = getb24(c);
 	       pack_putc(r, f);
 	       pack_putc(g, f);
 	       pack_putc(b, f);
+	       p24 += 3;
 	    }
 	 }
 	 break;
@@ -459,6 +472,11 @@ static void save_datafile_bitmap(DATAFILE *dat, int packed, int packkids, int st
 	 }
 	 break;
    }
+
+   if (*allegro_errno)
+      return FALSE;
+   else
+      return TRUE;
 }
 
 
@@ -477,7 +495,7 @@ static void *makenew_rle_sprite(long *size)
 
 
 /* draws an RLE sprite onto the grabber object view window */
-static void plot_rle_sprite(DATAFILE *dat, int x, int y)
+static void plot_rle_sprite(AL_CONST DATAFILE *dat, int x, int y)
 {
    RLE_SPRITE *s = dat->dat;
    BITMAP *b = create_bitmap_ex(s->color_depth, s->w, s->h);
@@ -518,7 +536,7 @@ static int view_rle_sprite(DATAFILE *dat)
 
 
 /* returns a description string for an RLE sprite */
-static void get_rle_sprite_desc(DATAFILE *dat, char *s)
+static void get_rle_sprite_desc(AL_CONST DATAFILE *dat, char *s)
 {
    RLE_SPRITE *spr = (RLE_SPRITE *)dat->dat;
 
@@ -531,7 +549,7 @@ static void get_rle_sprite_desc(DATAFILE *dat, char *s)
 
 
 /* exports an RLE sprite into an external file */
-static int export_rle_sprite(DATAFILE *dat, char *filename)
+static int export_rle_sprite(AL_CONST DATAFILE *dat, AL_CONST char *filename)
 {
    BITMAP *bmp;
    RLE_SPRITE *rle = (RLE_SPRITE *)dat->dat;
@@ -550,31 +568,33 @@ static int export_rle_sprite(DATAFILE *dat, char *filename)
 
 
 /* reads an RLE sprite from an external file */
-static void *grab_rle_sprite(char *filename, long *size, int x, int y, int w, int h, int depth)
+static DATAFILE *grab_rle_sprite(int type, AL_CONST char *filename, DATAFILE_PROPERTY **prop, int depth)
 {
+   DATAFILE *dat;
    BITMAP *bmp;
    RLE_SPRITE *spr;
 
-   bmp = (BITMAP *)grab_bitmap(filename, size, x, y, w, h, depth);
-   if (!bmp)
+   dat = grab_bitmap(type, filename, prop, depth);
+   if (!dat)
       return NULL;
 
-   spr = get_rle_sprite(bmp);
+   bmp = dat->dat;
+   dat->dat = get_rle_sprite(bmp);
    destroy_bitmap(bmp);
 
-   return spr;
+   return dat;
 }
 
 
 
 /* saves an RLE sprite into the datafile format */
-static void save_rle_sprite(DATAFILE *dat, int packed, int packkids, int strip, int verbose, int extra, PACKFILE *f)
+static int save_rle_sprite(DATAFILE *dat, AL_CONST int *fixed_prop, int pack, int pack_kids, int strip, int sort, int verbose, int extra, PACKFILE *f)
 {
    RLE_SPRITE *spr = (RLE_SPRITE *)dat->dat;
    int x, y, c, r, g, b, a;
    signed short *p16;
    signed long *p32;
-   unsigned int eol_marker;
+   unsigned long eol_marker;
    int depth;
 
    if (rle_has_alpha(spr))
@@ -582,6 +602,8 @@ static void save_rle_sprite(DATAFILE *dat, int packed, int packkids, int strip,
    else
       depth = spr->color_depth;
 
+   *allegro_errno = 0;
+
    pack_mputw(depth, f);
    pack_mputw(spr->w, f);
    pack_mputw(spr->h, f);
@@ -601,7 +623,7 @@ static void save_rle_sprite(DATAFILE *dat, int packed, int packkids, int strip,
 	 eol_marker = (depth == 15) ? MASK_COLOR_15 : MASK_COLOR_16;
 
 	 for (y=0; y<spr->h; y++) {
-	    while ((unsigned short)*p16 != eol_marker) {
+	    while ((unsigned short)*p16 != (unsigned short)eol_marker) {
 	       if (*p16 < 0) {
 		  /* skip count */
 		  pack_iputw(*p16, f);
@@ -710,12 +732,17 @@ static void save_rle_sprite(DATAFILE *dat, int packed, int packkids, int strip,
 	 }
 	 break;
    }
+
+   if (*allegro_errno)
+      return FALSE;
+   else
+      return TRUE;
 }
 
 
 
 /* returns a description string for a compiled sprite object */
-static void get_c_sprite_desc(DATAFILE *dat, char *s)
+static void get_c_sprite_desc(AL_CONST DATAFILE *dat, char *s)
 {
    BITMAP *bmp = (BITMAP *)dat->dat;
 
@@ -725,7 +752,7 @@ static void get_c_sprite_desc(DATAFILE *dat, char *s)
 
 
 /* returns a description string for a mode-X compiled sprite object */
-static void get_xc_sprite_desc(DATAFILE *dat, char *s)
+static void get_xc_sprite_desc(AL_CONST DATAFILE *dat, char *s)
 {
    BITMAP *bmp = (BITMAP *)dat->dat;
 
@@ -758,7 +785,8 @@ DATEDIT_GRABBER_INFO datbitmap_grabber =
    "bmp;lbm;pcx;tga",
    "bmp;pcx;tga",
    grab_bitmap,
-   export_bitmap
+   export_bitmap,
+   "XPOS;YPOS;XSIZ;YSIZ"
 };
 
 
@@ -784,7 +812,8 @@ DATEDIT_GRABBER_INFO datrlesprite_grabber =
    "bmp;lbm;pcx;tga",
    "bmp;pcx;tga",
    grab_rle_sprite,
-   export_rle_sprite
+   export_rle_sprite,
+   "XPOS;YPOS;XSIZ;YSIZ"
 };
 
 
@@ -810,7 +839,8 @@ DATEDIT_GRABBER_INFO datcsprite_grabber =
    "bmp;lbm;pcx;tga",
    "bmp;pcx;tga",
    grab_bitmap,
-   export_bitmap
+   export_bitmap,
+   "XPOS;YPOS;XSIZ;YSIZ"
 };
 
 
@@ -836,6 +866,7 @@ DATEDIT_GRABBER_INFO datxcsprite_grabber =
    "bmp;lbm;pcx;tga",
    "bmp;pcx;tga",
    grab_bitmap,
-   export_bitmap
+   export_bitmap,
+   "XPOS;YPOS;XSIZ;YSIZ"
 };
 
